/*****************************************************************************/
/* mat_cplex.c                                                               */
/* Konversion von Netzmatrizen in das CPLEX-Format.                          */
/*****************************************************************************/
/*                                            Do., 03.03.1994, Stefan Roemer */
/*****************************************************************************/

#define _TEST_CPLEX__

#include <stdlib.h>
#include "mat_cplex.h"
#ifdef _TEST_CPLEX_
#include "convertPEP.h"
#endif
#include "cpxdefs.inc"

/*****************************************************************************/

char     probname[NAMELENGTH];
int      mac;                             /* Anzahl Zeilen/Gleichungen */
int      mar;                             /* Anzahl Spalten */
int      objsen;                          /* -1 => Maximierungsproblem */
double   * objx;                          /* Koeffizienten der Zielfunktion */
double   * rhsx;                          /* Koeffizienten der rechten Seite */
char     * senx;                          /* Typ Un-/Gleichung: L/E */
                                   /* Beschreibung der Matrix: */
                                   /* Nur Eintraege != 0 werden gespeichert! */
int      * matbeg;                        /* "Startindizes" */
int      * matcnt;                        /* "Anzahlen" pro Spalte */
int      * matind;                        /* Zeilennummern */
double   * matval;                        /* entspr. Eintraege */

double   * bdl;                               /* untere Schranken */
double   * bdu;                               /* obere Schranken */
char     * dataname = NULL;
char     * objname = NULL;
char     * rhsname = NULL;
char     * rngname = NULL;
char     * bndname = NULL;
char     * cstore = NULL;                     /* Namen der Variablen */
char     ** cname = NULL;
char     * rstore = NULL;                     /* Namen der "Zeilen"*/
char     ** rname = NULL;
int      macsz;                            /* Wert >= mac: Groesse von objx,
                                             matbeg, matcnt, bdl, bdu, cname */
int      marsz;                          /* Wert >= mar: Groesse von rhsx,senx,
                                                               rname, rngval */
int      matsz;                            /* Groesse von matind, matval */
unsigned cstorsz;                          /* Laenge von cstore */
unsigned rstorsz;                          /* Laenge von rstore */
int      lpstat;
double   obj;
double   * x;
double   * pi;
double   * slack;
double   * dj;

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

int calc_simplex( int * sol_status, double * sol_solution )

#else

int calc_simplex( sol_status, sol_solution )

  int    * sol_status;
  double * sol_solution;

#endif

  {

    struct cpxlp  * lp = NULL;
    FILE          * logfile = NULL,
                  * changefile = NULL;
    int           status = 1,
                  toosmall,
                  toobig;

    *sol_status = 255;
    *sol_solution = (double) INFBOUND;

    /* Set up CPLEX message output.  setlogfile sends messages
       on all four CPLEX-defined channels to the file specified.
       setscr_ind sends messages on all four channels to stdout.
       The CPLEX Message Handler must be initialized for CPLEX to
       generate any messages, even error messages. */

    if (    ! ( logfile = fopen( "cplex_logfile.log", "w" ) )
         || setlogfile( logfile )
         || setscr_ind( 1 ) )
      {
        printf ("Failure to connect logging channel.\n");
        goto TERMINATE;
      }

    /* Loading, optimizing and obtaining a solution to the problem. */
    lp = loadprob( probname, mac, mar, 0, objsen, objx, rhsx,
                   senx, matbeg, matcnt, matind, matval,
                   bdl, bdu, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                   dataname, objname, rhsname, rngname, bndname,
                   cname, cstore, rname, rstore, NULL, NULL,
                   macsz, marsz, matsz, 0, 0, cstorsz, rstorsz, 0 );

    /* A returned pointer of NULL may mean that not enough memory
       was available or there was some other problem.  For most 
       other CPLEX functions, a return status of 0 means success,
       and 1 means failure.  In the case of failure, an error message
       will have been written to the error channel, which for this
       example is both stdout and the logfile, from inside CPLEX. */
    if ( ! lp )
      {
        goto TERMINATE;
      } /* if */

    /* Set iteration logging to log every factorization */
    if ( status = setitfoind( 1, &toosmall, &toobig ) )
      {
        goto TERMINATE; 
      } /* if */

    /* Set iteration limit to a large value; if failure,
       set limit to largest allowable value which is returned
       by first call to setitlim. */
    if ( status = setitlim( 10000, &toosmall, &toobig ) )
      {
        setitlim( toobig, &toosmall, &toobig );
      } /* if */

    /* Optimize the problem and obtain solution. */
    if ( status = optimize( lp ) )
      {
        goto TERMINATE;
      } /* if */

    if ( status = solution( lp, &lpstat, &obj, x, pi, slack, dj ) )
      {
        goto TERMINATE;
      } /* if */

    /* Write solution output using CPLEX Message Handler, using
       the results channel. */
    /* BG
    cpxmsg( cpxresults, "\n\n" );
    cpxmsg( cpxresults, "\n" );
    cpxmsg( cpxresults, "Status          = %d\n", status );
    cpxmsg( cpxresults, "Solution status = %d\n", lpstat );
    cpxmsg( cpxresults, "Solution value  = %f\n\n", obj ); */
    *sol_status = lpstat;
    *sol_solution = obj;

/*
#define _O_U_T_P_U_T_ 
*/
#ifdef _O_U_T_P_U_T_
printf("status=%d.\n",status);
printf("sol_status=%d.\n",*sol_status);
printf("sol_solution=%f.\n",*sol_solution);
    for ( i = 0 ; i < getmar( lp ) ; i++ )
      {
        cpxmsg( cpxresults, "Row %d:  Slack = %10f  Pi = %10f\n",
                            i, slack[i], pi[i] );
      } /* for */

    for ( j = 0 ; j < getmac( lp ) ; j++ )
      {
        cpxmsg( cpxresults, 
                "Column %d:  Value = %10f  Reduced cost = %10f\n",
                            j, x[j], dj[j] );
      } /* for */
#endif
   
TERMINATE:
    /* Flush all channels before deleting files as destinations. */
    flushchannel( cpxresults );
    flushchannel( cpxlog );
    flushchannel( cpxerror );
    flushchannel( cpxwarning );

    /* Delete log file from all four channels and close. */
    setlogfile( NULL );     
    if ( logfile )
      {
        fclose( logfile );	   
      } /* if */
 
    /* Delete changefile from results channel, then close. */
    delfpdest( cpxresults, changefile );
    if ( changefile )
      {
        fclose( changefile );
      } /* if */
    
    return ( status );

  } /* calc_simplex() */


/*****************************************************************************/
/* Konversion einer Netzmatrix in das CPLEX-Matrixformat.                    */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void convert_matrix( LMPNet net )

#else

void convert_matrix( net )

  LMPNet net;

#endif

  {

    Place      pl;
    Transition tr;
    STEdge     st;
    TSEdge     ts;
    int        i,
               count = 0,
               * ptr_matbeg,
               * ptr_matcnt,
               * ptr_matind;
    double     * ptr_matval;

    /* show_matrix( net ); */
    strcpy( probname, "net_matrix" );

    mac = (int) net->NumPlaces;        /* Anzahl Zeilen/Gleichungen */
    mar = (int) net->NumTransitions;   /* Anzahl Spalten */
    macsz = mac + 1;                   /* Wert >= mac: Groesse von objx,
                                            matbeg, matcnt, bdl, bdu, cname */
    marsz = mar + 2;                   /* Wert >= mar: Groesse von rhsx,senx,
                                                              rname, rngval */
    objsen = -1;                       /* -1 => Maximierungsproblem */
    dataname = objname = rhsname = rngname = bndname = NULL;
    cstore = rstore = NULL;
    cname = rname = NULL;
    cstorsz = rstorsz = 0;

    /* Zielfunktion und Schranken festlegen */
    objx = (double *) Malloc( macsz * sizeof( double ) );
    bdl = (double *) Malloc( macsz * sizeof( double ) );
    bdu = (double *) Malloc( macsz * sizeof( double ) );
    x = (double *) Malloc( macsz * sizeof( double ) );
    dj = (double *) Malloc( macsz * sizeof( double ) );
    i = mac;
    while ( i-- )
      {
        objx[i] = (double) 1.0;
        bdl[i] = (double) 0.0;
        bdu[i] = (double) INFBOUND;
      } /* while */
  
    /* Rechte Seiten und Art der (Un-)Gleichungen bestimmen */
    rhsx = (double *) Malloc( marsz * sizeof( double ) );
    senx = (char *) Malloc( marsz * sizeof( char ) );
    pi = (double *) Malloc( marsz * sizeof( double ) );
    slack = (double *) Malloc( marsz * sizeof( double ) );
    i = mar;
    while ( i-- )
      {
        rhsx[i] = (double) 0.0;       /* Koeffizienten der rechten Seite */
        senx[i] = 'E';                /* Typ: Gleichung (E) */
      } /* while */

    /* Anzahl der Knoten mit Kantengewicht > 0 ermitteln */
    pl = net->Places;
    while ( pl )
      {
        st = pl->postset;
        while ( st )
          {
            count++;
            st = st->nexttrans;
          } /* while */
        ts = pl->preset;
        while ( ts )
          {
            count++;
            ts = ts->nexttrans;
          } /* while */
        pl = pl->nextplace;
      } /* while */
    matsz = count;                /* Groesse von matind, matval */

    num_net_nodes(net);

    /* Konstruktion der Matrix und der notwendigen Hilfsfelder */
    *( ptr_matbeg = matbeg
                  = (int *) Malloc( ( macsz + 1 ) * sizeof( int ) ) ) = 0;
    ptr_matcnt = matcnt = (int *) Malloc( ( macsz + 1 ) * sizeof( int ) );
    ptr_matind = matind = (int *) Malloc( ( matsz + 1 ) * sizeof( int ) );
    ptr_matval = matval
               = (double *) Malloc( ( matsz + 1 ) * sizeof( double ) );
    pl = net->Places;
    while ( pl )
      {
        st = pl->postset;
        ts = pl->preset;
        i = 0;
        while ( st || ts )
          {
            i++;
            if ( ts && st && ( st->trans->numname == ts->trans->numname ) )
              {
                fprintf( stderr, "ERROR: (Matrixkonversion) " );
                fprintf( stderr, "Kein schlingenfreies Netz!\n");
                exit( 1 );
              } /* if */
            if ( !ts || ( st && ( st->trans->numname < ts->trans->numname ) ) )
              {
                *( ptr_matval++ ) = (double) st->weight;
                *( ptr_matind++ ) = (int) st->trans->numname;
                st = st->nexttrans;
              } /* if */
            else
              {
                *( ptr_matval++ ) = - ( (double) ts->weight );
                *( ptr_matind++ ) = (int) ts->trans->numname;
                ts = ts->nexttrans;
              } /* else */
          } /* while ( st || ts ) */
        *( ptr_matcnt++ ) = i;
        i += *ptr_matbeg;
        *++ptr_matbeg = i;
        pl = pl->nextplace;
      } /* while ( pl ) */

#ifdef _O_U_T_P_U_T_
    ptr_matbeg = matbeg;
    ptr_matcnt = matcnt;
    ptr_matind = matind;
    ptr_matval = matval;
    i = mac;
    printf( "matbeg = (%d", *ptr_matbeg );
    while ( --i )
      {
        printf( ", %d", *++ptr_matbeg );
      } /* while */
    printf( ")\n" );
    i = mac;
    printf( "matcnt = (%d", *ptr_matcnt );
    while ( --i )
      {
        printf( ", %d", *++ptr_matcnt );
      } /* while */
    printf( ")\n" );
    i = count;
    printf( "matind = (%d", *ptr_matind );
    while ( --i )
      {
        printf( ", %d", *++ptr_matind );
      } /* while */
    printf( ")\n" );
    i = count;
    printf( "matval = (%f", *ptr_matval );
    while ( --i )
      {
        printf( ", %f", *++ptr_matval );
      } /* while */
    printf( ")\n" );
#endif

  } /* convert_matrix() */


/*****************************************************************************/
/* Freigabe des fuer die Felder angeforderten Speicherplatzes.               */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void free_matrix( void )

#else

void free_matrix()

#endif

  {

    free( (char *) objx );
    free( (char *) bdl );
    free( (char *) bdu );
    free( (char *) x );
    free( (char *) dj );
    free( (char *) rhsx );
    free( (char *) senx );
    free( (char *) pi );
    free( (char *) slack );
    free( (char *) matbeg );
    free( (char *) matcnt );
    free( (char *) matind );
    free( (char *) matval );

  } /* free_matrix() */


/*****************************************************************************/
/* Testroutine.                                                              */
/*****************************************************************************/

#ifdef _TEST_CPLEX_

#if defined(__STDC__) || defined(__cplusplus)

void main( int argc, char ** argv )

#else

void main( argc, argv )

  int argc;
  char ** argv;

#endif

  {

    /* Deklarationen */ 

    LMPNet net; 
    char   in_filename[NAMELENGTH];

    if ( argc == 2 )
      {
        strcpy( in_filename, argv[1] );
      } /* if */
    else
      {
        printf("TESTPROGRAMM fuer Matrixkonversion\n");
        printf("Name der PEP-Netzdatei: ");
        gets(in_filename);
      } /* else */

    net = read_PEP( in_filename );
    convert_matrix( net );
    calc_simplex();
    free_matrix();
    /*
    write_matrix(net, out_filename); 
    write_netmarking(net, out_filename); 
    */
    show_matrix(net); 

    remove_net(&net);

  } /* main() */

#endif

/*****************************************************************************/
/* END OF FILE mat.c                                                         */
/*****************************************************************************/
