/******************************************************************************/
/***       Praktikum Werkzeuge zur Analyse von Petri-Netzen, SS 1992        ***/
/***       C-Quelldatei fuer Programmteil 'Test auf Beschraenktheit'        ***/
/***                                        Juli 1992, Stefan Roemer        ***/
/******************************************************************************/
/***    Version 2.0 ( neue Datenstruktur )    Stefan Roemer, 10.05.1993     ***/
/***    Version 2.5 ( Test unabhaengig von FC-Test,                         ***/
/***                  n-Beschraenktheit  )    Stefan Roemer, 17.02.1994     ***/
/******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include "nbound.h"

/******************************************************************************/
/*          Z U G R U N D E L I E G E N D E R   A L G O R I T H M U S         */
/*----------------------------------------------------------------------------*/
/*  "Berechne Ueberdeckungsgraphen"                                           */
/*  Eingabe: Netz N = ( S, T, F ) ;  Anfangsmarkierung Mo .                   */
/*  Ausgabe: zugehoeriger Ueberdeckungsgraph UDG = ( V, E ) .                 */
/*                                                                            */
/*  BEGIN                                                                     */
/*    E := {};                                      "Einrichten der Wurzel."  */
/*    V := {Mo};                                                              */
/*    WHILE                                                                   */
/*      ex. v in V : ex. t in T : ( t ist v-aktiviert AND                     */
/*                                  fuer alle (v,x,r) aus E : r <> t )        */
/*         "fuer einen Knoten v ist eine Transition v-aktiviert, mit der      */
/*          noch keine von v ausgehende Kante beschriftet ist."               */
/*      DO                                                                    */
/*        v' := v + N(t);                           "Bilde Folgemarkierung."  */
/*        NV := (0,0,...,0);                                                  */
/*        FOR ALL v^ aus V                                                    */
/*          DO                                                                */
/*            IF ( ex. Pfad von v^ nach v ) AND ( v^ < v' )                   */
/*              THEN NV := NV + ( v' - v^ )                                   */
/*          OD;                                                               */
/*             "Teste, ob und an welchen Stellen v'  v oder einen Vorgaenger  */
/*                                                   von v echt ueberdeckt."  */
/*        v' := v' + w * NV;                                     "w = omega"  */
/*                                           "Diese werden durch w ersetzt."  */
/*        IF v' nicht in V             "v' kommt noch nicht als Knoten vor."  */
/*          THEN V := V + {v'};                             "neu aufnehmen."  */
/*        E := E + { (v,v',t) }                                               */
/*      OD                                                                    */
/*  END.                                                                      */
/*----------------------------------------------------------------------------*/
/* Test auf Beschraenktheit: Das getestete System ist unbeschraenkt, falls    */
/* beim Aufbau des Ueberdeckungsgraphen eine Markierung erzeugt wird, die ein */
/* Omega als Markenanzahl enthaelt; andernfalls ist es beschraenkt.           */
/******************************************************************************/

/******************************************************************************/
/*                     R E A L I S I E R U N G   I N   C                      */
/*----------------------------------------------------------------------------*/
/*  Vorbemerkungen:                                                           */
/*    Um eine effiziente Speicherausnutzung zu erreichen, sind im folgenden   */
/*    Programm die Stellen des zu untersuchenden Systems umbenannt, und zwar  */
/*    derart, dass sie in der Reihenfolge, in der sie im Originalnetz stehen, */
/*    eine 'Platznummer', beginnend mit der Nummer 1, zugewiesen bekommen.    */
/*    Um nicht mit den Grundroutinen in Konflikt zu geraten, beginnen sie     */
/*    jedoch auch weiterhin mit einem 'P', d.h. die Stellennamen lauten im    */
/*    folgenden Programm P1,P2,P3,...,Pn, wobei n=|S| die Anzahl der Stellen  */
/*    im Netz sei. Vor Verlassen der Funktion 'beschraenkt()' werden den Stel-*/
/*    len wieder die urspruenglichen Namen zugewiesen, so dass der Benutzer   */
/*    der Funktion von dieser Manipulation nichts mitbekommt. Die Transitio-  */
/*    nennamen bleiben unveraendert; auch sonst wird am zu testenden System   */
/*    keine Manipulation vorgenommen.                                         */
/*    Fuer die Stellennummern und Markenzahlen ist ein Datentyp 'VCardinal'   */
/*    eingerichtet worden, der in dieser Programmversion ( fuer die SUN ) so  */
/*    gewaehlt ist, dass bis zu 65535 Stellen verwendet werden koennen. Diese */
/*    Zahl ist auch gleichzeitig mit der Konstanten OMEGA besetzt worden, die */
/*    anzeigt, dass eine Stelle unbeschraenkt viele Marken aufnimmt. Hier     */
/*    muss der Benutzer AUFPASSEN : Falls ein System getestet werden soll,    */
/*    bei dem die Stellen schon zu Beginn sehr viele Marken tragen, koennen   */
/*    sich hier durch Addition von Markierungsvektoren sehr schnell grosse    */
/*    Zahlen ergeben, die den Wert OMEGA ueberschreiten; das Programm selbst  */
/*    faengt diese Situation NICHT ab. Ein eventuelles Ergebnis ist in einem  */
/*    solchen Fall mit Vorsicht zu geniessen!                                 */
/*                                                                            */
/*  Markierungsvektoren:                                                      */
/*    Ein Markierungsvektor gibt an, wieviele Marken eine jede Stelle zu      */
/*    einem Zeitpunkt besitzt. Da die Stellen nach den Vorbemerkungen durch-  */
/*    numeriert worden sind, laesst sich ein solcher Vektor angeben als       */
/*    M = ( #P1, #P2, ..., #Pn ), wobei #Pi die Anzahl der Marken von Stelle  */
/*    Pi sei. Da jedoch sehr haeufig nicht alle Stellen markiert sind, ent-   */
/*    haelt ein solcher Vektor viele Nulleintraege; ausserdem sind Operatio-  */
/*    nen wie Addition, Test zweier Vektoren auf Gleichheit, u.s.w. bei einer */
/*    solchen Realisierung fuer eine groessere Anzahl von Stellen sehr inef-  */
/*    fizient, da immer der ganze Vektor ( der Laenge n ) durchlaufen werden  */
/*    muss. C unterstuetzt jedoch die Benutzung dynamischer Felder, d.h.      */
/*    Felder, deren Laenge erst zur Laufzeit bekannt werden. Deshalb werden   */
/*    im folgenden nur die Eintraege gespeichert, die nicht null sind; dafuer */
/*    muss allerdings auch die Stellennummer gespeichert werden, so dass      */
/*    Markierungsvektoren die folgende Form besitzen:                         */
/*                                                                            */
/*       +-----+-----+-----+-----+-----+-----+ ... +-----+-----+-----+        */
/*    -->|  1  | #P1 |  4  | #P4 |  5  | #P5 |     |  n  | #Pn |  0  |        */
/*       +-----+-----+-----+-----+-----+-----+ ... +-----+-----+-----+        */
/*                                                                            */
/*    (unter der beispielhaften Annahme, dass n > 5, #P2 = #P3 = 0 und alle   */
/*    anderen Eintraege ungleich null sind). Die '0' als Stellennummer kenn-  */
/*    zeichnet das Ende des Markierungsvektors. Wenn man erst einmal die An-  */
/*    zahl der Eintraege eines solchen Vektors kennt, kann man sich ueber     */
/*    'malloc()' Speicher zur Ablage geben lassen ( dynamisch ). Fuer einige  */
/*    Operationen ( etwa Addition ) kann man die Laenge jedoch nicht vorher   */
/*    festlegen, hierfuer sind im Programm 4 Hilfsvektoren deklariert, die    */
/*    die Maximalzahl an Stellen ( n ) aufnehmen koennen.                     */
/*                                                                            */
/*  Ueberdeckungsgraph: Knotenmenge V                                         */
/*    Die Menge der Knoten, d.h. der Markierungsvektoren, ist als lineare     */
/*    Liste realisiert; jeder Knoten enthaelt daneben noch einige Zusatz-     */
/*    informationen. Die Kantenmenge ist NICHT explizit in einer eigenen      */
/*    Datenstruktur abgelegt, sondern wird direkt in eine Datei geschrieben.  */
/*    Die Knotenmenge besitzt die folgende Form:                              */
/*                                                                            */
/*                 V E                                                        */
/*                  |                                                         */
/*                  V                                                         */
/*            +------------+                                                  */
/*            | Markierung-------> Zeiger auf Markierungsvektor.              */
/*    Knoten  |  PfadVorg---->                                                */
/*       1    |   besucht  |    -- Flag, das beim Durchgehen der Vorgaenger-  */
/*            |    next    |       markierungen anzeigt, ob dieser Knoten     */
/*            +---- | -----+       bereits untersucht worden ist.             */
/*                  |                                                         */
/*                  V                                                         */
/*            +------------+                                                  */
/*            | Markierung--->                                                */
/*    Knoten  |  PfadVorg---->                                                */
/*       2    |   besucht  |    -- Flag, das beim Durchgehen der Vorgaenger-  */
/*            |    next    |       markierungen anzeigt, ob dieser Knoten     */
/*            +---- | -----+       bereits untersucht worden ist.             */
/*                  |                                                         */
/*                  V                                                         */
/*                  .                                                         */
/*                  .                                                         */
/*                  .                                                         */
/*                  V                                                         */
/*            +------------+                                                  */
/*            | Markierung--->                                                */
/*    Knoten  |  PfadVorg--------> Zeiger auf eine Liste, in der Zeiger auf   */
/*       m    |   besucht  |       diejenigen Knoten enthalten sind, die      */
/*            |    next    |       eine Kante auf diesen Knoten (m) besitzen, */
/*            +---- | -----+       also direkte Vorgaengermarkierungen sind.  */
/*                  |                                                         */
/*                  V                                                         */
/*                 NULL                                                       */
/*                                                                            */
/*                                                                            */
/*  Spezialitaeten:                                                           */
/*    Da der vorliegende Algorithmus exponentielle Komplexitaet besitzt, ist  */
/*    besonderer Augenmerk auf eine effiziente Implementierung gelegt worden  */
/*    (wie ein schneller Blick ueber das Programmlisting beweist). Im ein-    */
/*    zelnen bedeutet dies:                                                   */
/*    - Der erste Existenzquantor ( ex. v in V ) wurde derart realisiert,     */
/*      dass eine Schleife GENAU EINMAL ueber alle ( im Verlauf des Algo-     */
/*      rithmus entestehenden ) Markierungs-'Knoten' v laeuft.                */
/*    - Die Formel ( ex. t in T : ( t ist v-aktiviert AND fuer alle (v,x,r)   */
/*      aus E : r<>t ) ) ist folgendermassen aufgeloest worden:               */
/*      In JEDEM Schleifendurchlauf fuer v ( siehe oben ) wird EINMAL durch   */
/*      die gesamte Transitionenmenge T gelaufen und jeweils getestet, ob     */
/*      die Transition t durch die Markierung v aktiviert wird.               */
/*    - Die FOR-Schleife des Algorithmus wurde 'umgedreht':                   */
/*      Anstatt fuer alle Markierungsvektoren aus VE nachzupruefen, ob der    */
/*      gerade untersuchte Vektor v ueber eine Kantenfolge erreicht werden    */
/*      kann, werden zu jedem Knoten die direkten Pfadvorgaenger der jeweili- */
/*      gen Markierung gespeichert und die Liste aller Vorgaengermarkierungen */
/*      von v untersucht ( mittels eines rekursiven Standardalgorithmus geht  */
/*      das in linearer Zeit, d.h. linear zu der aktuellen Anzahl der Knoten  */
/*      in VE ).                                                              */
/*    Der Ueberdeckungsgraph wird ausserdem in eine Datei 'TEST.UDG' ge-      */
/*    schrieben ( Default-Name, falls die Funktion 'beschraenkt()' aufgerufen */
/*    wird; bei Benutzung der Funktion 'Ueberdeckungsgraph()' kann ein belie- */
/*    biger Dateiname ( ohne Endung ) uebergeben werden, in die der Graph     */
/*    geschrieben werden soll. ). Aus diesem Grunde berechnet der Algorithmus */
/*    den vollstaendigen Ueberdeckungsgraphen; fuer die vorliegende Aufgabe,  */
/*    den Test auf Beschraenktheit, haette es gereicht, mit dem Wert FALSE    */
/*    abzubrechen, sobald ein OMEGA im Ueberdeckungsgraphen auftaucht und     */
/*    somit eine Stelle unbeschraenkt viele Marken aufnimmt.                  */
/*    Die Orthogonalitaet der Sprache C wurde extensiv genutzt, bei Aenderun- */
/*    gen und sonstigen Eingriffen in das Programm bitte aufpassen; an den    */
/*    benutzten Zeigerstrukturen und damit verbundenen Anweisungen sollte nur */
/*    bei genauer Kenntnis ihrer Arbeitsweise etwas geaendert werden, diese   */
/*    sind mit Bedacht gewaehlt.                                              */
/*                                                                            */
/******************************************************************************/


/******************************************************************************/
/*                             Globale Variablen                              */
/******************************************************************************/

bool BOUND_statistics = False;
    /* BG Der Einfachheit halber hier: */
    bool BOUND_progress = False;
    Cardinal32           AnzahlKanten = 1;
    Cardinal32           AnzahlMarkierungen = 1;


/******************************************************************************/
/*       Freigabe des vom Ueberdeckungsgraphen besetzten Speicherplatzes      */
/******************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void RemoveUDG ( KnotenMenge KnotenPtr )

#else

void RemoveUDG ( KnotenPtr )

  KnotenMenge KnotenPtr;

#endif 

  {
    
    KnotenMenge     KnotenDavor;
    VorgaengerListe VorgPtr,
        	    VorgDavor;

    while ( KnotenPtr )
      {
        VorgPtr = KnotenPtr->PfadVorg;
        while ( VorgPtr )
          {
            VorgPtr = ( VorgDavor = VorgPtr ) -> next;
            Free( VorgDavor );
          } /* while */
        Free( KnotenPtr->Markierung );
        KnotenPtr = ( KnotenDavor = KnotenPtr ) -> next;
        Free( KnotenDavor );
      } /* while */

  } /* RemoveUDG () */


/******************************************************************************/
/*        Die im Netz gegebenen Stellennamen speichern und durch neue,        */
/*                       numerisch geordnete ersetzen                         */
/******************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void StellenBenennen( LMPNet System, Place * Stellenfeld )

#else

void StellenBenennen( System, Stellenfeld )

  LMPNet System;                         /* Zeiger auf markiertes Netz.       */
  Place  * Stellenfeld;                  /* Feld von Zeigern auf die Stellen. */

#endif
 

  {

    register VCardinal Nummer = 1;
    register Place     StellenPtr = System->Places,
                       * FeldPtr = Stellenfeld - 1;

    while ( StellenPtr )
      {
        ( *( ++FeldPtr ) = StellenPtr ) -> numname = Nummer++;
        StellenPtr = StellenPtr->nextplace;
      } /* while */

  } /* StellenBenennen() */


/******************************************************************************/
/*        Prozedur, die einen Markierungsvektor in eine Datei schreibt        */
/******************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static void SchreibeMarkierung( FILE * Datei, Vektor vek, Place * Stellenfeld,
                                bool Newline )

#else

static void SchreibeMarkierung( Datei, vek, Stellenfeld, Newline )

  FILE    * Datei;                    /* Zeiger auf die Zieldatei.            */
  Vektor  vek;                        /* Auszugebender Markierungsvektor.     */
  Place   * Stellenfeld;              /* Liste der Stellen des Netzes.        */
  bool    Newline;                    /* Flag: Falls True, wird nach dem      */
        			      /*       Vektor ein '\n' geschrieben.   */

#endif

  {

    register Vektor vekPtr = vek;   /* Zeiger auf aktuelle Position im Vektor */

    putc( '(', Datei );

    while ( *(vekPtr++) )
      {
        if ( *vekPtr > 1 )
          {
            fprintf( Datei, *vekPtr == OMEGA ? " w" : " %hu", *vekPtr );
          } /* if */
        fprintf( Datei, " %s", Stellenfeld[ (*(vekPtr-1))-1 ] -> name );
        if ( *(++vekPtr) )
          {
            putc ( ',', Datei );
          } /* if */
      } /* while */

    fprintf( Datei, Newline ? " )\n" : " )" ); 

  } /* SchreibeMarkierung() */


/******************************************************************************/
/*     Funktion zur Untersuchung der Vorgaengermarkierungen des  bereits      */
/*     erstellten Ueberdeckungsgraphen, ob diese die aktuelle Markierung      */
/*        jeweils echt ueberdecken; dies ist eine REKURSIVE Funktion.         */
/******************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void UntersuchePfad( KnotenMenge VE, KnotenMenge vQuer, Vektor vStrich, Vektor NV, Vektor HilfsVektor )

#else

void UntersuchePfad( VE, vQuer, vStrich, NV, HilfsVektor )

  KnotenMenge VE,                   /* Die Menge der Markierungsvektoren.     */
              vQuer;                /* Aktuell zu bearbeitende Markierung.    */
  Vektor      vStrich,              /* Zu vergleichender Vektor.              */
              NV,                   /* Vektor der unbeschraenkten Stellen.    */
              HilfsVektor;          /* Hilfsvektor zur Aufnahme des Zwischen- */
        		            /* ergebnisses bei der Addition.          */

#endif

  {

    register VorgaengerListe vVorg;

    vQuer->besucht = True;
    if ( VektorKleiner( vQuer->Markierung, vStrich ) )
      {
        /* NV := NV + ( v' - v^ ) */
        Addiere( NV, vStrich, HilfsVektor );
        Subtrahiere( NV, vQuer->Markierung );
      } /* if */
    vVorg = vQuer->PfadVorg;
    while ( vVorg )
      {
        if ( ! vVorg->KnotenPtr->besucht )
          {
            UntersuchePfad( VE, vVorg->KnotenPtr, vStrich, NV, HilfsVektor );
          } /* if */
        vVorg = vVorg->next;
      } /* while */

  } /* UntersuchePfad() */


/******************************************************************************/
/*       Funktion zur Berechnung des Ueberdeckungsgraphen eines Systems       */
/*  Liefert 1, falls das System beschraenkt ist, und 0, falls unbeschraenkt.  */
/*    Falls der uebergebene Dateiname leer ( "" ) ist, wird keine Datei be-   */
/*    schrieben, ansonsten wird eine Datei mit dem Dateinamen und der Endung  */
/*  '.UDG' angelegt, in die der erzeugte Ueberdeckungsgraph geschrieben wird. */
/******************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

VCardinal Ueberdeckungsgraph_n( LMPNet System, char * DateiName, VCardinal Schranke )
  
#else

VCardinal Ueberdeckungsgraph_n( System, DateiName, Schranke )
  
  LMPNet    System;
  char      * DateiName;
  VCardinal Schranke;

#endif

  {

    register KnotenMenge VE,
                         v,
                         vQuer,
        	         vStrichPtr;
    register Place       Stelle,
                         * Stellenfeld;
    register Transition  t;
    char                 UDGDateiName[200];
    register VCardinal   Stellenzahl,
                         MaxMarken = 0;
    register Vektor      NV,
        	         vStrich,
                         vtest,
                         HilfsVektor,
        	         HilfsPtr,
                         SchaltVektor;
    VorgaengerListe      VorgPtr;
    bool	         UDGschreiben = False,
                         interruption = False;
    FILE                 * Datei=stdout;
    /* BG war vorher hier:
    Cardinal32           AnzahlKanten = 1;
    Cardinal32           AnzahlMarkierungen = 1; */

    /* Ermittle Anzahl der Stellen im Netz */
    Stellenzahl = System->NumPlaces;

    /* Zaehle die Stellen im Netz durch */
    Stellenfeld = (Place *) Malloc( Stellenzahl * sizeof( Place ) );
    StellenBenennen( System, Stellenfeld );

    /* Datei fuer den Markierungsgraphen oeffnen */
    if ( *DateiName )
      {
        strcat( strcpy( UDGDateiName, DateiName ) , ".UDG" );
                                                 /* UDG = UeberDeckungsGraph */
        if (( Datei = fopen( UDGDateiName, "w" ) ))
          {
            fprintf( Datei, "* Cover graph for system '%s':\n\n",
          		    DateiName );
            UDGschreiben = True;
          } /* if */
        else
          {
            fprintf( stderr,
                     "ERROR: could not write file '%s'!\n",
                     UDGDateiName );
            exit ( 1 );
          } /* else */
      } /* if */

    /* Initialisiere einige Vektoren */
    NV           = NeuerVektor( Stellenzahl );
    vStrich      = NeuerVektor( Stellenzahl );
    HilfsVektor  = NeuerVektor( Stellenzahl );
    SchaltVektor = NeuerVektor( Stellenzahl );

    /* Kopiere die Anfangsmarkierung */
    Stelle = System->Places;
    HilfsPtr = HilfsVektor;
    while ( Stelle )
      {
        if ( Stelle->marks )
          {
            *(HilfsPtr++) = Stelle->numname;
            *(HilfsPtr++) = Stelle->marks;
          } /* if */
        Stelle = Stelle->nextplace;
      } /* while */
    *HilfsPtr = 0;

    /* Initialisiere Knotenmenge: Einrichten der Wurzel */
    VE = (KnotenMenge) Malloc( sizeof( Knoten ) );
    VE->Markierung = ErzeugeVektor( HilfsVektor );
    VE->next = NULL;
    VE->PfadVorg = NULL; 
    if ( UDGschreiben )
      {
        SchreibeMarkierung( Datei, VE->Markierung, Stellenfeld, False );
        fputs( " ist die Startmarkierung des Systems\n\n", Datei );
      } /* if */

    t = System->Transitions;
    v = VE;

    /* Hauptschleife des Algorithmus */
    while ( v )
      {

        VektorAusSTListe( SchaltVektor, t->preset );

        if ( VektorKleiner( SchaltVektor, v->Markierung ) 
             || VektorGleich( SchaltVektor, v->Markierung ) )
          {
            /* ==> v aktiviert t */

            /* Bilde Folgemarkierung: v' := v + N(t) */
            Kopiere( vStrich, v->Markierung );
            Subtrahiere( vStrich, SchaltVektor );
            VektorAusTSListe( SchaltVektor, t->postset );
            Addiere( vStrich, SchaltVektor, HilfsVektor );

            /* NV := (0,0,...,0) */
            *NV = 0;

            /* Teste, ob und an welchen Stellen v'  v oder einen */
            /* Vorgaenger von v echt ueberdeckt. */
            vQuer = VE;
            while ( vQuer )
              {
        	vQuer->besucht = False;
        	vQuer = vQuer->next;
              } /* while */
            UntersuchePfad( VE, v, vStrich, NV, HilfsVektor );

            /* v' := v' + omega * NV */
            if ( * MultipliziereMitOmega( NV ) )
              {
        	MaxMarken = OMEGA;
                Addiere( vStrich, NV, HilfsVektor ); 
        	if ( ! UDGschreiben )
        	  {
        	    /* Falls kein Ueberdeckungsgraph generiert werden soll, */
        	    /* wird an dieser Stelle bereits das System als         */
        	    /* unbeschraenkt erkannt und die weitere Berechnung     */
        	    /* abgebrochen. */
                    interruption = True;
        	    break; /* while ( v ) */
        	  }
              } /* if */

            /* n-Beschraenktheit testen */
            if ( MaxMarken != OMEGA )
              {
                vtest = vStrich;
                while ( *( vtest++ ) )
                  {
                    if ( *vtest > MaxMarken )
                      {
                        MaxMarken = *vtest;
                      } /* if */
                    vtest++;
                  } /* while */
                if ( ( MaxMarken > Schranke ) && ! UDGschreiben )
                  {
                    break; /* while */
                  } /* if */
              } /* if */

            /* v' in VE suchen; falls nicht da: neu einfuegen */
            vStrichPtr = VE;
            if ( ! VektorGleich( vStrichPtr->Markierung, vStrich ) )
              {
        	while ( vStrichPtr->next && 
        		! VektorGleich( vStrichPtr->next->Markierung, vStrich ) )
        	  {
        	    vStrichPtr = vStrichPtr->next;
        	  } /* while */
        	if ( vStrichPtr->next )
        	  {
        	    vStrichPtr = vStrichPtr->next;
        	  } /* if */
        	else
        	  {
        	    /* neuen Knoten fuer v' einrichten */
        	    vStrichPtr = vStrichPtr->next 
        		       = (KnotenMenge) Malloc( sizeof( Knoten ) );
        	    vStrichPtr->Markierung = ErzeugeVektor( vStrich );

/* BG */
AnzahlMarkierungen++;
if ( BOUND_progress && ( ( AnzahlMarkierungen % 1000 ) == 0 ) )
	printf ("Markings: %u\n", AnzahlMarkierungen);
/* BG */

        	    vStrichPtr->PfadVorg = NULL;
        	    vStrichPtr->next = NULL;
        	  } /* else */
              } /* if */

            /* neue Kante v --t--> v' einfuegen, d.h. in Datei schreiben */
            AnzahlKanten++;

/* BG */
if (BOUND_progress && ( ( AnzahlKanten % 1000 ) == 0 ) )
	printf ("Edges: %u\n", AnzahlKanten);
/* BG */

            if ( UDGschreiben )
              {
                SchreibeMarkierung( Datei, v->Markierung, Stellenfeld, False );
                fprintf( Datei, " --%s--> ", t->name );
                SchreibeMarkierung( Datei, vStrichPtr->Markierung, Stellenfeld,
                                    True );
              }

            /* Pfad-Vorgaengerliste von v' aktualisieren */
            if (( VorgPtr = vStrichPtr->PfadVorg ))
              {
        	if ( ! VektorGleich( VorgPtr->KnotenPtr->Markierung, v->Markierung ) )
        	  {
        	    while ( VorgPtr->next && 
        		    ! VektorGleich( VorgPtr->next->KnotenPtr->Markierung,
        			      v->Markierung )
        		  )
        	      {
        		VorgPtr = VorgPtr->next;
        	      } /* while */
        	  } /* if */
        	if ( ! VorgPtr->next )
        	  {
        	    VorgPtr = VorgPtr->next
        		    = (VorgaengerListe) Malloc( sizeof(Vorgaenger) );
        	    VorgPtr->next = NULL;
        	    VorgPtr->KnotenPtr = v;
        	  } /* if */
              } /* if */
            else
              {
        	VorgPtr = vStrichPtr->PfadVorg
        	        = (VorgaengerListe) Malloc( sizeof( Vorgaenger ) );
        	VorgPtr->next = NULL;
        	VorgPtr->KnotenPtr = v;
              } /* else */
          } /* if */

        /* Naechste Transition bzw. naechste Markierung untersuchen */
        if ( ! ( t = t->nexttrans ) )
          {
            t = System->Transitions;
            v = v->next;
            if ( UDGschreiben )
              {
        	putc( '\n', Datei );
              } /* if */
          } /* if */

      } /* while */

    /* ... und jetzt wird aufgeraeumt: */
    if ( UDGschreiben )
      {
        fclose( Datei );
      } /* if */

    if ( BOUND_statistics )
      {
        register Cardinal32 count = 0,
                            size,
                            sum   = 0;

        v = VE;
        while ( v )
          {
            size = 1;
            HilfsPtr = v->Markierung;
            while ( *( HilfsPtr++ ) )
              {
                size++;
                size++;
                HilfsPtr++;
              } /* while */
            (void) ( ( ( size <<= 1 ) & 7 )
                 && ( ( size &= (Cardinal32) -8 ), size += (Cardinal32) 8 ) );
            size += (Cardinal32) 8;
            sum += size;
            count++;
            v = v->next;
          } /* while */
        printf( ">>> %u different reachable markings.\n", count );
        printf( ">>> %u arcs in the reachability graph.\n", AnzahlKanten );
        printf( ">>> Memory used for marking vectors : %8u Bytes.\n", sum );
        (void) ( ( ( size = sizeof( Knoten ) ) & 7 )
                 && ( ( size &= (Cardinal32) -8 ), size += (Cardinal32) 8 ) );
        size += (Cardinal32) 8;
        printf( ">>> Memory used for list of markings: %8u Bytes.\n",
                                                               count * size );
      } /* if */

    Free( NV );
    Free( vStrich );
    Free( HilfsVektor );
    Free( SchaltVektor );
    Free( Stellenfeld );

    RemoveUDG( VE );

    return ( MaxMarken );

  } /* Ueberdeckungsgraph_n() */


/******************************************************************************/
/*         Hauptprogramm fuer Programmteil 'Test auf Beschraenktheit'         */
/******************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool n_bounded( LMPNet System, Cardinal32 Schranke, char * UDG_Dateiname )

#else

bool n_bounded( System, Schranke, UDG_Dateiname )

  LMPNet     System;
  Cardinal32 Schranke;
  char       * UDG_Dateiname;

#endif

  {

    if ( ! System || ! System->Places || ( System->NumPlaces > 0xFFFF ) )
      {
        fprintf( stderr, "ERROR: n_bounded(): No System!\n" );
        return ( False );
      } /* if */

    return ( (bool) ( Ueberdeckungsgraph_n( System, UDG_Dateiname, Schranke )
                                                                <= Schranke ) );

  } /* n_bounded() */


/******************************************************************************/
/* END OF FILE 'bound.c'                                                      */
/******************************************************************************/

