/*****************************************************************************/
/* FILE bpn.y                                                                */
/* BISON source file: parser for b(pn)^2                                     */
/* 1993 by Burkhard Graves                                                   */
/* 1994 by Lars Jenner                                                       */
/* completely reviewed February 1995 by Stephan Merkel                       */
/* winter 1996/97 by Martin Ackermann                                        */
/*****************************************************************************/
/* Note: Rules containing square brackets ('[' and ']') are no longer        */
/* actively supported and provided only for historical continuity with older */
/* versions.                                                                 */
/* The default action '{ $$ = $1 }' is not explicitly given in rules.        */
/*****************************************************************************/

%{
#include <string.h>
#include "tree.h"
#include "set.h"

SymTabEntry *ProcId;
TreeNode *ExprTree;

#define YYERROR_VERBOSE 
#undef  YYERROR_VERBOSE 
  /* this provides for verbose error messages with bison as well as with */
  /* yacc when using the proper driver file (yacc_dri.own)               */

#define MAX_EXPECT 20
  /* this sets a limit to the number of expected tokens printed into */
  /* error messages                                                  */

int exprerror(char *msg)
{
	fprintf(stderr,"error while parsing: %s\n", msg);
	return 1;
}

ShortEntryPtr ref;

#define _create_action_(erg, p1, p2, p3)	\
  {						\
    ref=MYmalloc(sizeof(ShortEntry));		\
    ref->s_ref.start_l = p1.start_l;		\
    ref->s_ref.start_c = p1.start_c;		\
    ref->s_ref.end_l   = p3.end_l;		\
    ref->s_ref.end_c   = p3.end_c;		\
    ref->s_type        = _ACTION;		\
    erg=GenNode(_EXPR,p2,(TreeNode*)ref);	\
  }

extern int   exprlex(void);
extern char *exprtext;

%}

%union {
  SymTabEntry   *sym;    /* defined in file symtab.h */
  IdentStruct    ident;  /* defined in file ident.h  */
  TreeNode      *node;   /* defined in file tree.h   */
  ArgListEntry  *arg;    /* defined in file tree.h   */
  SetRange      *range;  /* defined in file set.h    */
  RefStruct      ref;    /* defined in file symtab.h */
  int            wert;
}

%token  KONST
%token  VAR
%token  NOT
%token  AND
%token  OR
%token  TRUE
%token  FALSE
%token  MAX
%token  <ref> TILDE
%token  <ref> AT
%token  <wert>  NUMBER
%token  <wert>  CONST_IDENT
%token  <ident> IDENT
%token  EMPTY_
%token  FULL
%token  GE
%token  LE
%token  MOD

%left           ';' ','
%left           OR
%right          AND
%right          NOT
%nonassoc       '=' '#' GE LE '<' '>'
%left           '+' '-'
%left           '*' '/' MOD
%right          Unaryminus

%type <wert>  pos_expr
%type <node>  com expr val_expr simple_expr

%start  com

%%

com
	: TILDE expr TILDE AT
		{
		  _create_action_($$, $1, $2, $3);
		  ExprTree = $$;
		}
	;

expr
	: '(' expr ')'
		{
		  $$=GenNode(N_KLAMM,$2,NULL);
		}
	| IDENT
		{
		  chk_ident($1.sym,_VAR);
		  $1.sym->s_special|=_LG_UNPRIMED;
		  $$=GenNode(N_VAR,(TreeNode *)$1.sym,(TreeNode *)_UNPRIMED);
		}
	| val_expr
	;

val_expr
	: simple_expr
	| expr OR expr
		{
		  $$=GenNode(N_OR,$1,$3);
		}
	| expr AND expr
		{
		  $$=GenNode(N_AND,$1,$3);
		}
	| '!' expr %prec NOT
		{
		  $$=GenNode(N_NOT,$2,NULL);		  
		}
	| NOT expr
		{
		  $$=GenNode(N_NOT,$2,NULL);		  
		}
	| expr '=' expr
		{
		  $$=GenNode(N_EQ,$1,$3);
		}
	| expr '#' expr
		{
		  $$=GenNode(N_UNEQ,$1,$3);
		}
	| expr '<' expr
		{
		  $$=GenNode(N_LT,$1,$3);
		}
	| expr '>' expr
		{
		  $$=GenNode(N_GT,$1,$3);
		}
	| expr GE expr
		{
		  $$=GenNode(N_GE,$1,$3);
		}
	| expr LE expr
		{
		  $$=GenNode(N_LE,$1,$3);
		}
	| expr '*' expr
		{
		  $$=GenNode(N_MULT,$1,$3);
		}
	| expr '/' expr
		{
		  $$=GenNode(N_DIV,$1,$3);
		}
	| expr MOD expr
		{
		  $$=GenNode(N_MOD,$1,$3);
		}
	| expr '+' expr
		{
		  $$=GenNode(N_PLUS,$1,$3);
		}
	| expr '-' expr
		{
		  $$=GenNode(N_MIN,$1,$3);
		}
	| '-' expr %prec Unaryminus
		{
		  $$=GenNode(N_SIGN,$2,NULL);
		}
	;

simple_expr
	: '\'' IDENT
		{
		  chk_ident($2.sym,_VAR);
		  $2.sym->s_special|=_LG_PRIMED;
		  $$=GenNode(N_VAR,(TreeNode *)$2.sym,(TreeNode *)_PREPRIMED);
		}
	| IDENT '\''
		{
		  chk_ident($1.sym,_VAR);
		  $1.sym->s_special|=_LG_PRIMED;
		  $$=GenNode(N_VAR,(TreeNode *)$1.sym,(TreeNode *)_POSTPRIMED);
		}
	| IDENT '?'
		{
		  chk2_ident($1.sym,_CHANNEL,_STACK);
		  $1.sym->s_special|=_LG_READING;
		  $$=GenNode(N_VAR,(TreeNode *)$1.sym,(TreeNode *)_READ);
		}
	| IDENT '!'
		{
		  chk2_ident($1.sym,_CHANNEL,_STACK);
		  $1.sym->s_special|=_LG_WRITING;
		  $$=GenNode(N_VAR,(TreeNode *)$1.sym,(TreeNode *)_WRITE);
		}
	| IDENT EMPTY_
		{
		  chk2_ident($1.sym,_CHANNEL,_STACK);
		  if ($1.sym->s_init_or_cap == 0)
		    Msg(ERR,"illegal use of chan-0-identifier %s",
			$1.sym->s_name);
		  $1.sym->s_special|=_LG_EMPTY;
		  $$=GenNode(N_VAR,(TreeNode *)$1.sym,(TreeNode *)_EMPTY);
		}
	| IDENT FULL
		{
		  chk2_ident($1.sym,_CHANNEL,_STACK);
		  if ($1.sym->s_init_or_cap == 0)
		    Msg(ERR,"illegal use of chan-0-identifier %s",
			$1.sym->s_name);
		  $1.sym->s_special|=_LG_FULL;
		  $$=GenNode(N_VAR,(TreeNode *)$1.sym,(TreeNode *)_FULL);
		}
	| pos_expr
		{
		  $$=GenNode(N_CONST,(TreeNode *)$1,NULL);
		}
	;

pos_expr
	: NUMBER
	| TRUE
		{
		  $$=TT;
		}
	| FALSE
		{
		  $$=FF;
		}
	| CONST_IDENT
	;


/******************************************************************************/
/* END OF FILE bpn.y                                                          */
/******************************************************************************/
