// Edge.h
// Tobias Himstedt 1995
// Burkhard Bieber 1995/96
// last-update: 05.04.96

#ifndef __EDGE_H__
#define __EDGE_H__

  //**********
  // includes
  //**********

#include "NetObject.h"
#include "../include/VHMap.h"

  //********************
  // Forwards fr Casts
  //********************

class Vertex;
class Block;
class Edge;
class Net;
class HLEdge;

  //-------------------------------------------------------------------------
  // Klasse      : Edge
  // Beschreibung: Defintion der grundlegenden Funktionen fr Kanten
  //-------------------------------------------------------------------------
class Edge : public NetObject {
  protected:
    static int edge_counter;
    static int active_objects;
    int id;             // eindeutige Id im Netz
    Vertex *from;       // von Knoten from
    Vertex *to;         // nach Knoten to
    int weight;         // Kantengewicht
    int rx_weight,      // rel. Pos. des Kantengewichtes
      ry_weight;
    int splinex,        // abs. Pos. des Splinemittelpunktes
      spliney;
    Net *net;           // zugehriges Netz
    int phantom_flag;   // Phantomkante oder nicht
    VHMap<String, Block *> blocks_from;  // falls Blockkante, dann Quellblock
    VHMap<String, Block *> blocks_to;    // falls Blockkante, dann Zielblock
      // I/O-Behandlung
    virtual void Write(ostream &f) const;
    virtual void ReadStandard(istream &f);
    virtual void ReadOptional(istream &f);
  public:
      //--------------------
      // Konstante Methoden
      //--------------------
      // liefert die Id zurck
    virtual int Id() const;
      // liefert Anfangsknoten der Kante
    const Vertex &Src(void) const;
      // liefert Endknoten der Kante
    const Vertex &Dest(void) const;
      // liefert Anfangsblock der Kante
    VHMap<String, Block *> &BlockSrc(void);
      // liefert Endknoten der Kante
    VHMap<String, Block *> &BlockDest(void);
      // liefert das Kantengewicht
    unsigned int Weight(void) const;
      // Vergleichsoperator
    int operator ==(const Edge &e) const;
      //-------------------------
      //  Konvertierungsmethoden
      //-------------------------
    virtual Edge *asEdge() const;
    virtual HLEdge *asHLEdge() const;
      //-------------------------
      // Modifizierende Methoden
      //-------------------------
      // Kapazitt setzen
    void SetWeight(int c);
      // als normale Kante setzen
    void SetNormalEdge(void) {phantom_flag = 0;}
      // als Phantomkante setzen
    void SetPhantomEdge(void) {phantom_flag = 1;}
      // neuen Quellblock eintragen
    int getPhantom(void) {return phantom_flag;}
      // ist PhantomEdge oder nicht
    void setPhantom(int w) {phantom_flag = w;}
      // setzen des Phantomflags
    int getSplinex(void) {return splinex;}
      // liefert die x-Koordinate des Splinemittelpunktes
    int getSpliney(void) {return spliney;}
      // liefert die y-Koordinate des Splinemittelpunktes
    void setSplinex(int a) {if(a>=0) splinex = a;}
      // setzt die x-Koordinate des Splinemittelpunktes
    void setSpliney(int a) {if(a>=0) spliney = a;}
      // setzt die y-Koordinate des Splinemittelpunktes
    void AddBlockSrc(Block &b);
      // neuen Zielblock eintragen
    void AddBlockDest(Block &b);
      // Quellblockverweis lschen
    void ResetBlockSrc(void);
      // Zielblockverweis lschen
    void ResetBlockDest(void);
      // Zuweisungsoperator
    const Edge &operator =(const Edge &e);
      // Konstruktoren
    Edge(Vertex &f, Vertex &t, const Net &n, const String &Type = "Edge");
    Edge(const Net &n, const String &Type = "Edge");
    Edge(const Edge &e);
      // Destruktoren
    virtual ~Edge();
      // Tcl-Schnittstelle
    int ObjectGet(Tcl_Interp *ip, int argc, CONST char *argv[]);
    int ObjectSet(Tcl_Interp *ip, int argc, CONST char *argv[]);
};  // class Edge

  //******************
  // Inlines fr Edge
  //******************

inline
Edge::Edge(const Edge &e) : NetObject(e), id(edge_counter++),
  from(e.from), to(e.to),
  weight(e.weight), rx_weight(e.rx_weight), ry_weight(e.ry_weight),
  net(e.net), phantom_flag(e.phantom_flag),
  blocks_from(e.blocks_from), blocks_to(e.blocks_to)
{
  active_objects++;
  base_classes.add("NetObject");
}  // Edge::Edge

inline const Edge &
Edge::operator =(const Edge &e) 
{
  NetObject::operator =(e);
  rx_weight = e.rx_weight;
  ry_weight = e.ry_weight;
  //from = e.from;
  //to = e.to;
  visibility = e.visibility;
  splinex = e.splinex;
  spliney = e.spliney;
  net = e.net;
  phantom_flag = e.phantom_flag;
  blocks_from = e.blocks_from;
  blocks_to = e.blocks_to;
  return *this;
}  // Edge::operator =

inline void
Edge::SetWeight(int w)
{
  if (w < 0)
    cerr << "error: edge weight < 0" << endl;
  else
    weight = w;
}  // Edge::SetWeight

inline unsigned int 
Edge::Weight(void) const 
{
  return weight;
}  // Edge::Weight

inline const Vertex &
Edge::Src(void) const
{
  return *from;
}  // Edge::Src

inline const Vertex &
Edge::Dest(void) const
{
  return *to;
}  // Edge::Dest

inline VHMap<String, Block *> &
Edge::BlockSrc(void)
{
  return blocks_from;
}  // Edge::BlockSrc

inline VHMap<String, Block *> &
Edge::BlockDest(void)
{
  return blocks_to;
}  // Edge::BlockDest

inline int
Edge::operator ==(const Edge &e) const
{
    // die Blockbezge blocks_from und blocks_to sind hier unwichtig
  return (this == &e) || (from == e.from) && (to == e.to);
}  // Edge::operator ==

#endif  // #ifndef __EDGE_H__

  //*** Dateiende Edge.h ***

