// Link.h
// Asynchronous links
//-----------------------------------------
// class Link: Link symbols
// class LinkTerm: Valued link term
//-----------------------------------------
// Cecile Bui Thanh 2001
// last-update: Dec. 2001

#ifndef __LINK_H__
#define __LINK_H__


// includes
#include <iostream>
#include "../include/Utils.h"
#include "../include/VHSet.h"
#include "Place.h"
#include "../High/Val.h"

/*****************************************************************
*                          CLASS LINK
******************************************************************/

class Link {
  private:
	String my_name;
	String my_string_type;
	VHSet<Val> my_type;
    VHMap<String, Place *> places;	//place(s) associated to this link
	int my_arity;

  public:
    /*----------- Constructors -----------*/
	Link();
	Link(const String &name, const String string_type, const VHSet<Val> &type);
	~Link();
    /*----------- Operations -----------*/
	String get_name() const;
	String get_string_type() const;
	VHSet<Val> get_type() const;
	VHMap<String, Place *> get_places() const;
	int get_arity() const;
	void add_place(Place *p);
    void clear_places();
	void set_type(String new_string_type, VHSet<Val> new_type);

    /*----------- Operators -----------*/
    int operator ==(const Link l) const;
	Link& operator =(const Link l);

    /*----------- Hashing, for using sets of Link -----------*/
	unsigned int Hash() const;

    /*----------- Write -----------*/
    friend ostream &operator <<(ostream &f, const Link &n);
    void Write(ostream &f) const;
}; // class Link


          /************ Inlines for Link ************/
inline
Link::Link():
  my_name("#"), my_string_type(""), my_type(),
  places((Place *) NULL,1), my_arity(0)
{
}

// premise: type is a valable type, i.e. every Vals have the same arity
inline
Link::Link(const String &name, const String string_type, const VHSet<Val> &type):
  my_name(name), my_string_type(string_type), my_type(type),
  places((Place *) NULL,5)
{
  if (type.empty())
	{
	  my_arity = 0;
	  Val *d = new Val(dot);
	  my_type.add(*d);
	}
  else
	{
	  Val v = my_type (my_type.first());
	  if (v.Type()!=TUPEL)
		  my_arity = 1;
	  else
		  my_arity = v.getTupel()->getDimension();
	}
}

inline
Link::~Link()
{
	my_type.clear();
	places.clear();
}

inline String
Link::get_name() const
{
  return my_name;
}

inline String
Link::get_string_type() const
{
  if (my_string_type == "")
    return "{dot}";
  else
    return my_string_type;
}

inline VHSet<Val>
Link::get_type() const
{
  return my_type;
}

inline VHMap<String, Place *>
Link::get_places() const
{
  return places;
}

inline int
Link::get_arity() const
{
  return my_arity;
}

inline void
Link::add_place(Place *p)
{
  if (p!=NULL && !places.contains(p->Name()))
	places[p->Name()] = p;
}

inline void
Link::clear_places()
{
  places.clear();
}

inline void
Link::set_type(String new_string_type, VHSet<Val> new_type)
{ // we assume the set new_type is correct, wrt the new_string_type
  my_type = new_type;
  my_string_type = new_string_type;

  if (my_string_type == "")
     my_arity = 0;
  else
	{
	  Val v = my_type (my_type.first());
	  if (v.Type()!=TUPEL)
		  my_arity = 1;
	  else
		  my_arity = v.getTupel()->getDimension();
	};
}

inline Link&
Link::operator =(const Link l)
{
   my_name = l.get_name();
   my_type = l.get_type();
   my_string_type = l.get_string_type();
   places = l.get_places();				//place(s) associated to this link
   my_arity = l.get_arity();
   return *this;
}

inline unsigned int
Link::Hash() const
{
	return hash_string(my_name);
}


/*-------- EMPTY VHSET ---------*/
#ifndef EMPTY_VHSET
#define EMPTY_VHSET
const VHSet<Val> EmptySet(1);
#endif

/*-------- EMPTY LINK ----------*/
#ifndef EMPTY_LINK
#define EMPTY_LINK
const Link EmptyLink("#","",EmptySet);
#endif


/*****************************************************************
*                          CLASS LINKTERM
******************************************************************/

class LinkTerm {
  private:
	String my_name;
	String args_string;
	Val my_args;
  public:
    /*----------- Constructors -----------*/
	LinkTerm();
	LinkTerm(const String &name, const String &args);
	~LinkTerm(){};

    /*----------- Operations -----------*/
	String get_name() const;
	String get_args_string() const;
	Val get_args() const;
	int get_arity();

    /*----------- Operators -----------*/
    int operator ==(const LinkTerm &l) const;
	LinkTerm& operator =(const LinkTerm l);

    /*----------- Hashing, for using sets of LinkTerm -----------*/
	unsigned int Hash() const;

}; // class LinkTerm

          /************ Inlines for LinkTerm ************/
inline
LinkTerm::LinkTerm():
my_name(""), args_string(), my_args()
{
}

inline String
LinkTerm::get_name() const
{
  return my_name;
}

inline String
LinkTerm::get_args_string() const
{
  return args_string;
}

inline Val
LinkTerm::get_args() const
{
  return my_args;
}

inline int
LinkTerm::get_arity()
{
  if (args_string == "")
	return 0;
  else if ( my_args.Type()!=TUPEL)
       return 1;
  else return (my_args.getTupel()->getDimension());
}

inline int
LinkTerm::operator ==(const LinkTerm &l) const
{
  return (l.get_name()==my_name
           && l.get_args() == my_args);
}

inline LinkTerm&
LinkTerm::operator =(const LinkTerm l)
{
  my_name = l.get_name();
  args_string = l.get_args_string();
  my_args = l.get_args();
  return *this;
}

inline unsigned int
LinkTerm::Hash() const
{
  return hash_string(my_name+args_string);
}

#endif	// #ifndef __LINK_H__

  //*** Dateiende Link.h ***
