// Vertex.h
// Tobias Himstedt 1995
// Burkhard Bieber 1995/96
//   - Block/HLBlock eingefhrt
//   - Blockverweismenge
//   - Phantomkanten
// Cecile Bui Thanh: added a default constructor (2001)
// last-update: Nov. 2001

#ifndef __VERTEX_H__
#define __VERTEX_H__

const String VertexPrefix = "Vertex";

  //**********
  // includes
  //**********
  
#include "NetObject.h"
#include "../include/VHMap.h"

  //********************
  // Forwards fr Casts
  //********************

class Block;
class Place;
class Transition;
class PTransition;
class Net;
class HLBlock;
class HLPlace;
class HLTransition;
class HLPTransition;
class Box;
class HLBox;

  //-------------------------------------------------------------------------
  // Klasse      : Vertex
  //-------------------------------------------------------------------------
  // Beschreibung: Definition der grundlegenden Funktionen fr Knoten
  //-------------------------------------------------------------------------
class Vertex : public NetObject {
  protected:
      // ---------------------------------------------------------------
      //  Konstante Funktionen
      // ---------------------------------------------------------------
    virtual void Write(ostream &f) const;
    virtual void ReadStandard(istream &f);
    virtual void ReadOptional(istream &f);
    virtual const String &PreFix() const {return VertexPrefix;};
      // ---------------------------------------------------------------
      //  Datenelemente
      // ---------------------------------------------------------------
    int id;
    String name;
    String meaning;
    String reference;
    int px, py;
    int rx_name, ry_name;
    int rx_mean, ry_mean;
    Net *net;
    String box;
    Box *boxPointer;
      // Einganskanten
    VHMap<String, Edge *> incoming_edges;
      // Ausgangskanten;
    VHMap<String, Edge *> outgoing_edges;
      // Phantomeinganskanten
    VHMap<String, Edge *> incoming_pedges;
      // Phantomausgangskanten;
    VHMap<String, Edge *> outgoing_pedges;
      // Menge der Blcke zu der dieser Knoten gehrt
    VHMap<String, Block *> blocks;
  public:
      // ---------------------------------------------------------------
      //  Konstante Funktionen
      // ---------------------------------------------------------------
      // Menge der eingehenden Kanten zurckgeben
    const Map<String, Edge *> &OutputEdges(void) const;
      // Dito andersherum
    const Map<String, Edge *> &InputEdges(void) const;
      // Menge der eingehenden Phantomkanten zurckgeben
    const Map<String, Edge *> &OutputPEdges(void) const;
      // Dito andersherum
    const Map<String, Edge *> &InputPEdges(void) const;
      // Ist e Eingangsknoten?
    int IsOutput(const String &name) const;
      // Ist e Ausgangsknoten?
    int IsInput(const String &name) const;
      // Ist e Eingangsknoten?
    int PIsOutput(const String &name) const;
      // Ist e Ausgangsknoten?
    int PIsInput(const String &name) const;
      // Kante zurck die von this e zeigt
    Edge *EdgeTo(const String &name) const;
      // Kante von e nach this
    Edge *EdgeFrom(const String &name) const;
      // Phantomkante zurck die von this e zeigt
    Edge *PEdgeTo(const String &name) const;
      // Phantomkante von e nach this
    Edge *PEdgeFrom(const String &name) const;
      // Menge der Blcke zurckgeben
    VHMap<String, Block *> &Blocks(void);
      // Menge der Blcke als Set zurckgeben, nur Keys
    VHSet<String> BlockKeySet(void);
    virtual void ChangeMyName(const String &new_name);
    const String &Name() const;
    const String &Meaning() const;
    const String &BoxName() const;
    const Box *BoxPointer() const;
    int Id() const;
      // ---------------------------------------------------------------
      //  Konvertierungsfunktionen
      // ---------------------------------------------------------------
    virtual Transition *asTransition() const;
    virtual PTransition *asPTransition() const;
    virtual HLTransition *asHLTransition() const;
    virtual HLPTransition *asHLPTransition() const;
    virtual Place *asPlace() const;
    virtual HLPlace *asHLPlace() const;
    virtual Block *asBlock() const;
    virtual HLBlock *asHLBlock() const;
    virtual Box *asBox() const;
    virtual HLBox *asHLBox() const;
      // ---------------------------------------------------------------
      //  Modifizierende Funktionen
      // ---------------------------------------------------------------
    static void ResetMaxIdTo(int nr);
    virtual void SetMaxId(int id) = 0;
    void ChangeOutName(const String &old_name, const String &new_name);
    void ChangeInName(const String &old_name, const String &new_name);
    void ChangeOutPName(const String &old_name, const String &new_name);
    void ChangeInPName(const String &old_name, const String &new_name);
    void SetX(int i) {px = i;};
    void SetY(int i) {py = i;};
    int GetX() {return px;};
    int GetY() {return py;};
    virtual void SetPosition(int x, int y);
      // Edge e ist Ausgangskante von *this
    void SetOutEdge(const Edge &e);
      // Edge e ist Eingangskante von *this
    void SetInEdge(const Edge &i);
      // Edge e ist Phantomausgangskante von *this
    void SetOutPEdge(const Edge &e);
      // Edge e ist Phantomeingangskante von *this
    void SetInPEdge(const Edge &i);
      // Edge e ist nicht mehr Ausgangskante von *this
    void RemoveOutEdge(const Edge &e);
      // Edge e ist nicht mehr Eingangskante von *this
    void RemoveInEdge(const Edge &x);
      // Edge e ist nicht mehr Phantomausgangskante von *this
    void RemoveOutPEdge(const Edge &e);
      // Edge e ist nicht mehr Phantomeingangskante von *this
    void RemoveInPEdge(const Edge &x);
      // Zuweisungsoperator
    void RemoveInEdges(void);
    void RemoveOutEdges(void);
    void RemoveInPEdges(void);
    void RemoveOutPEdges(void);
    void RemoveBlock(String& name);
    void AddBlock(String name, Block *ptr);
    void setBox(String name, Box *b) {box = name; boxPointer = b;};
    virtual void DeleteBox(String &name);
    const Vertex &operator =(const Vertex &v);
      // Konstruktoren
    Vertex();
    Vertex(const Vertex &);
    Vertex(const Net &n, const String &type = "Vertex");
      // Destruktor
    virtual ~Vertex();
      // Tcl-Interface
    int ObjectGet(Tcl_Interp *ip, int argc, CONST char *argv[]);
    int ObjectSet(Tcl_Interp *ip, int argc, CONST char *argv[]);
};  // class Vertex

  //********************
  // Inlines fr Vertex
  //********************

inline
Vertex::Vertex(const Vertex &v) : NetObject(v), 
  px(v.px), py(v.py), rx_name(v.rx_name), ry_name(v.ry_name),
  rx_mean(v.rx_mean), ry_mean(v.ry_mean), net(v.net), box(v.box), boxPointer(v.boxPointer),
  incoming_edges(v.incoming_edges), outgoing_edges(v.outgoing_edges),
  incoming_pedges(v.incoming_pedges), outgoing_pedges(v.outgoing_pedges),
  blocks(v.blocks)
{
  // Nix;
}  // Vertex:: Vertex

inline void
Vertex::SetPosition(int x, int y)
{
  px = x;
  py = y;
}  // Vertex::SetPosition

inline const String &
Vertex::Name() const
{
  return name;
}  // Vertex::Name

inline const String &
Vertex::Meaning() const
{
  return meaning;
}  // Vertex::Meaning

inline const String &
Vertex::BoxName() const
{
  return box;
}  // Vertex::box

inline const Box *
Vertex::BoxPointer() const
{
  return boxPointer;
}  // Vertex::BoxPointer

inline int 
Vertex::Id() const
{
  return id;
}  // Vertex::Id

inline void
Vertex::ChangeMyName(const String &new_name) 
{
  name = new_name;
}  // Vertex::ChangeMyName

inline const Map<String, Edge *> & 
Vertex::InputEdges(void) const
  // Dito andersrum
{
  return incoming_edges;
}  // Vertex::InputEdges

inline const Map<String, Edge *> & 
Vertex::OutputEdges(void) const 
  // Menge der eingehenden Kanten zurck
{
  return outgoing_edges;           
}  // Vertex::OutputEdges

inline const Map<String, Edge *> & 
Vertex::InputPEdges(void) const
  // Menge der eingehenden Phantomkanten zurck
{
  return incoming_pedges;
}  // Vertex::InputPEdges

inline const Map<String, Edge *> & 
Vertex::OutputPEdges(void) const 
  // Dito andersrum
{
  return outgoing_pedges;           
}  // Vertex::OutputPEdges

inline int
Vertex::IsOutput(const String &n) const
  // Ist e Eingangskonten?
{
  return EdgeFrom(n) != NULL;
}  // Vertex::IsOutput

inline int
Vertex::IsInput(const String &n) const
  // Ist e Ausgangsknoten?
{
  return EdgeTo(n) != NULL;
}  // Vertex::IsInput

inline int
Vertex::PIsOutput(const String &n) const
  // Ist e Eingangskonten?
{
  return PEdgeFrom(n) != NULL;
}  // Vertex::PIsOutput

inline int
Vertex::PIsInput(const String &n) const
  // Ist e Ausgangsknoten?
{
  return PEdgeTo(n) != NULL;
}  // Vertex::PIsInput

inline void
Vertex::SetOutEdge(const Edge &e)
{
  outgoing_edges[e.Dest().Name()] = (Edge *) &e;
}  // Vertex::SetOutEdge

inline void
Vertex::SetInEdge(const Edge &e)
{
  incoming_edges[e.Src().Name()] = (Edge *) &e;
}  // Vertex::SetInEdge

inline void
Vertex::SetOutPEdge(const Edge &e)
{
  outgoing_pedges[e.Dest().Name()] = (Edge *) &e;
}  // Vertex::SetOutPEdge

inline void
Vertex::SetInPEdge(const Edge &e)
{
  incoming_pedges[e.Src().Name()] = (Edge *) &e;
}  // Vertex::SetInPEdge

inline void
Vertex::RemoveOutEdge(const Edge &e)
{
  outgoing_edges.del(e.Dest().Name());
}  // Vertex::RemoveOutEdge

inline void
Vertex::RemoveInEdge(const Edge &e)
{
  incoming_edges.del(e.Src().Name());
}  // Vertex::RemoveIdEdge

inline void
Vertex::RemoveInEdges(void)
{
  incoming_edges.clear();
}  // Vertex::RemoveInEdges

inline void
Vertex::RemoveOutEdges(void)
{
  outgoing_edges.clear();
}  // Vertex::RemoveOutEdges

inline void
Vertex::RemoveInPEdges(void)
{
  incoming_pedges.clear();
}  // Vertex::RemoveInPEdges

inline void
Vertex::RemoveOutPEdges(void)
{
  outgoing_pedges.clear();
}  // Vertex::RemoveOutPEdges

inline void
Vertex::RemoveOutPEdge(const Edge &e)
{
  outgoing_pedges.del(e.Dest().Name());
}  // Vertex::RemoveOutPEdge

inline void
Vertex::RemoveInPEdge(const Edge &e)
{
  incoming_pedges.del(e.Src().Name());
}  // Vertex::RemoveInPEdge

inline void
Vertex::RemoveBlock(String& name)
{
  blocks.del(name);
}  // Vertex::RemoveBlock

inline void
Vertex::AddBlock(String name, Block *ptr)
{
  if(!blocks.contains(name))
    blocks[name] = ptr;
}  // Vertex::AddBlock

inline VHMap<String, Block *> &
Vertex::Blocks(void)
  // Menge der Blcke zurck
{
  return blocks;
}  // Vertex::Blocks

#endif  // #ifndef __VERTEX_H__

  //*** Dateiende Vertex.h ***

