// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// Modified for Templates by Tobias Himstedt
// Universitaet Hildesheim
// 2/95

#ifndef _OSLSet_h
#define _OSLSet_h 1

#include "Set.h"
#include "SLList.h"

template<class Type> inline int
LessEq(const Type& a, const Type& b)
{
    return a<=b;
}


template<class Type> inline int
Compare(const Type& a, const Type& b)
{
    return LessEq(a,b) ? ( Equal(a,b) ? 0 : -1 ) : 1;
}

template<class Type>
class OSLSet : public SET<Type> {
protected:
    SLList<Type>     p;
    
public:
    OSLSet();
    OSLSet(const OSLSet<Type>&);
    
    Pix           add(const Type& item);
    void          del(const Type& item);
    int           contains(const Type& item) const;
    
    void          clear();
    
    Pix           first() const;
    void          next(Pix& i) const;
    Type&         operator () (Pix i) const;
    int           owns(Pix i) const;
    Pix           seek(const Type& item) const;


    const OSLSet<Type>& operator =(const OSLSet<Type>& b);
    
    void          operator |= (const OSLSet<Type>& b);
    void          operator -= (const OSLSet<Type>& b);
    void          operator &= (const OSLSet<Type>& b);
    
    int           operator == (const OSLSet<Type>& b) const;
    int           operator != (const OSLSet<Type>& b) const;
    int           operator <= (const OSLSet<Type>& b) const; 
    
    int           OK() const;
};


template<class Type> inline 
OSLSet<Type>::OSLSet() : p() 
{ 
    count = 0; 
}

template<class Type> inline 
OSLSet<Type>::OSLSet(const OSLSet<Type>& s) : p(s.p) 
{ 
    count = s.count; 
}

template<class Type> inline const OSLSet<Type>&
OSLSet<Type>::operator =(const OSLSet<Type>& s)
{ 
    if (this == &s)
	return *this;
    p = s.p;
    count = s.count; 
    return *this;
}


template<class Type> inline Pix 
OSLSet<Type>::first() const
{
    return p.first();
}

template<class Type> inline void 
OSLSet<Type>::next(Pix& idx) const
{
    p.next(idx);
}

template<class Type> inline Type& 
OSLSet<Type>::operator ()(Pix  idx) const
{
    return p(idx);
}

template<class Type> inline void 
OSLSet<Type>::clear()
{
    count = 0;  
    p.clear();
}

template<class Type> inline int 
OSLSet<Type>::contains (const Type& item) const
{
    return seek(item) != 0;
}

template<class Type> inline int 
OSLSet<Type>::owns (Pix   idx) const
{
    return p.owns(idx);
}

template<class Type> inline int 
OSLSet<Type>::operator != (const OSLSet<Type>& b) const
{
    return !(*this == b);
}


template<class Type> Pix 
OSLSet<Type>::seek(const Type& item) const
{
    for (Pix i = p.first(); i != 0; p.next(i))
	{
	    int cmp = Compare(item, p(i));
	    if (cmp == 0)
		return i;
	    else if (cmp < 0)
		return 0;
	}
    return 0;
}

template<class Type> Pix 
OSLSet<Type>::add(const Type& item)
{
    Pix i = p.first();
    if (i == 0) 
	{
	    ++count;
	    return p.prepend(item);
	}
    int cmp = Compare(item, p(i));
    if (cmp == 0)
	return i;
    else if (cmp < 0)
	{
	    ++count;
	    return p.prepend(item);
	}
    else
	{
	    Pix trail = i;
	    p.next(i);
	    for (;;)
		{
		    if (i == 0)
			{
			    ++count;
			    return p.append(item);
			}
		    cmp = Compare(item, p(i));
		    if (cmp == 0)
			return i;
		    else if (cmp < 0)
			{
			    ++count;
			    return p.ins_after(trail, item);
			}
		    else
			{
			    trail = i;
			    p.next(i);
			}
		}
	}
}

template<class Type> void 
OSLSet<Type>::del(const Type& item)
{
    Pix i = p.first();
    if (i == 0)
	return;
    int cmp = Compare(item, p(i));
    if (cmp < 0)
	return;
    else if (cmp == 0)
	{
	    --count;
	    p.del_front();
	}
    else
	{
	    Pix trail = i;
	    p.next(i);
	    while (i != 0)
		{
		    cmp = Compare(item, p(i));
		    if (cmp < 0)
			return;
		    else if (cmp == 0)
			{
			    --count;
			    p.del_after(trail);
			    return;
			}
		    else
			{
			    trail = i;
			    p.next(i);
			}
		}
	}
}


template<class Type> int 
OSLSet<Type>::operator <= (const OSLSet<Type>& b) const
{
    if (count > b.count) return 0;
    Pix i = first();
    Pix j = b.first();
    for (;;)
	{
	    if (i == 0)
		return 1;
	    else if (j == 0)
		return 0;
	    int cmp = Compare(p(i), b.p(j));
	    if (cmp == 0)
		{
		    next(i); b.next(j);
		}
	    else if (cmp < 0)
		return 0;
	    else
		b.next(j);
	}
}

template<class Type> int 
OSLSet<Type>::operator == (const OSLSet<Type>& b) const
{
    if (count != b.count) return 0;
    if (count == 0) return 1;
    Pix i = p.first();
    Pix j = b.p.first();
    while (i != 0)
	{
	    if (!Equal(p(i),b.p(j))) return 0;
	    next(i);
	    b.next(j);
	}
    return 1;
}


template<class Type> void 
OSLSet<Type>::operator |= (const OSLSet<Type>& b)
{
    if (&b == this || b.count == 0)
	return;
    else
	{
	    Pix j = b.p.first();
	    Pix i = p.first();
	    Pix trail = 0;
	    for (;;)
		{
		    if (j == 0)
			return;
		    else if (i == 0)
			{
			    for (; j != 0; b.next(j))
				{
				    ++count;
				    p.append(b.p(j));
				}
			    return;
			}
		    int cmp = Compare(p(i), b.p(j));
		    if (cmp <= 0)
			{
			    if (cmp == 0) b.next(j);
			    trail = i;
			    next(i);
			}
		    else
			{
			    ++count;
			    if (trail == 0)
				trail = p.prepend(b.p(j));
			    else
				trail = p.ins_after(trail, b.p(j));
			    b.next(j);
			}
		}
	}
}


template<class Type> void 
OSLSet<Type>::operator -= (const OSLSet<Type>& b)
{
    if (&b == this)
	clear();
    else if (count != 0 && b.count != 0)
	{
	    Pix i = p.first();
	    Pix j = b.p.first();
	    Pix trail = 0;
	    for (;;)
		{
		    if (j == 0 || i == 0)
			return;
		    int cmp = Compare(p(i), b.p(j));
		    if (cmp == 0)
			{
			    --count;
			    b.next(j);
			    if (trail == 0)
				{
				    p.del_front();
				    i = p.first();
				}
			    else
				{
				    next(i);
				    p.del_after(trail);
				}
			}
		    else if (cmp < 0)
			{
			    trail = i;
			    next(i);
			}
		    else
			b.next(j);
		}
	}
}

template<class Type> void 
OSLSet<Type>::operator &= (const OSLSet<Type>& b)
{
    if (b.count == 0)
	clear();
    else if (&b != this && count != 0)
	{
	    Pix i = p.first();
	    Pix j = b.p.first();
	    Pix trail = 0;
	    for (;;)
		{
		    if (i == 0)
			return;
		    else if (j == 0)
			{
			    if (trail == 0)
				{
				    p.clear();
				    count = 0;
				}
			    else
				{
				    while (i != 0)
					{
					    --count;
					    next(i);
					    p.del_after(trail);
					}
				}
			    return;
			}
		    int cmp = Compare(p(i), b.p(j));
		    
		    if (cmp == 0)
			{
			    trail = i;
			    next(i);
			    b.next(j);
			}
		    else if (cmp < 0)
			{
			    --count;
			    if (trail == 0)
				{
				    p.del_front();
				    i = p.first();
				}
			    else
				{
				    next(i);
				    p.del_after(trail);
				}
			}
		    else
			b.next(j);
		}
	}
}


template<class Type> int 
OSLSet<Type>::OK() const
{
    int v = p.OK();
    v &= count == p.length();
    Pix trail = p.first();
    if (trail == 0)
	v &= count == 0;
    else
	{
	    Pix i = trail; next(i);
	    while (i != 0)
		{
		    v &= Compare(p(trail), p(i)) < 0;
		    trail = i;
		    next(i);
		}
	}
    if (!v) error("invariant failure");
    return v;
}

#endif
