// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


// Modified for Templates by Tobias Himstedt
// Universitaet Hildesheim
// 1/95


#ifndef _VHBag_h
#define _VHBag_h 1

#include "Bag.h"

#ifndef DEFAULT_INITIAL_CAPACITY
#define DEFAULT_INITIAL_CAPACITY 20
#endif

// HASHTABLE_TOO_CROWDED(COUNT, SIZE) is true iff a hash table with COUNT
// elements and SIZE slots is too full, and should be resized.
// This is so if available space is less than 1/8.

#define HASHTABLE_TOO_CROWDED(COUNT, SIZE) ((SIZE) - ((SIZE) >> 3) <= (COUNT))



template<class Type>
class VHBag : public Bag<Type> {
  protected:
    Type*          tab;
    char*         status;
    unsigned int  size;
    
  public:
    VHBag(unsigned int sz = DEFAULT_INITIAL_CAPACITY);
    VHBag(const VHBag<Type>& a);
    ~VHBag();
    
    Pix           add(const Type& item);
    void          del(const Type& item);
    
    void          clear();
    
    Pix           first() const;
    void          next(Pix& i) const;
    Type&         operator () (Pix i) const;
    Pix           seek(const Type& item) const;
    virtual void  nof(const Type& i) const;

    const VHBag<Type>& operator =(const VHBag<Type>& b);

    int           operator == (const VHBag<Type>& b) const;
    int           operator != (const VHBag<Type>& b) const;
    
    int           capacity() const;
    void          resize(unsigned int newsize = 0);
    
    int           OK() const;
};


template<class Type> inline 
VHBag<Type>::~VHBag()
{
    delete [] tab;
    delete [] status;
}


template<class Type> inline int 
VHBag<Type>::capacity() const
{
    return size;
}


template<class Type> inline Type& 
VHBag<Type>::operator () (Pix i) const
{
    if (i == 0) error("null Pix");
    return *((Type*)i);
}

template<class Type> inline int 
VHBag<Type>::operator != (const VHBag<Type>& b) const
{
    return ! ((*this) == b);
}

#define EMPTYCELL   0
#define VALIDCELL   1
#define DELETEDCELL 2


template<class Type>
VHBag<Type>::VHBag(unsigned int sz)
{
    tab = new Type[size = sz];
    status = new char[size];
    for (unsigned int i = 0; i < size; ++i) status[i] = EMPTYCELL;
    count = 0;
}

template<class Type>
VHBag<Type>::VHBag(const VHBag<Type>& a)
{
    tab = new Type[size = a.size];
    status = new char[size];
    for (unsigned int i = 0; i < size; ++i) status[i] = EMPTYCELL;
    count = 0;
    for (Pix p = a.first(); p; a.next(p)) add(a(p));
}

template<class Type> const VHBag<Type>&
VHBag<Type>::operator =(const VHBag<Type>& a)
{
    if (this == &a)
	return *this;

    delete [] tab;
    delete [] status;

    tab = new Type[size = a.size];
    status = new char[size];
    for (unsigned int i = 0; i < size; ++i) status[i] = EMPTYCELL;
    count = 0;
    for (Pix p = a.first(); p; a.next(p)) add(a(p));
    return *this;
}    
    

/* 
 * hashing method: double hash based on high bits of hash fct,
 * followed by linear probe. Can't do too much better if table
 * sizes not constrained to be prime.
*/

#ifndef DOUBLEHASH
#define DOUBLEHASH
static inline unsigned int 
doublehashinc(unsigned int h, unsigned int s)
{
    unsigned int dh =  ((h / s) % s);
    return (dh > 1)? dh : 1;
}
#endif

template<class Type> Pix 
VHBag<Type>::seek(const Type& key) const
{
    unsigned int hashval = Hash(key);
    unsigned int h = hashval % size;
    for (unsigned int i = 0; i <= size; ++i)
	{
	    if (status[h] == EMPTYCELL)
		return 0;
	    else if (status[h] == VALIDCELL && Equal(key, tab[h]))
		return Pix(&tab[h]);
	    if (i == 0)
		h = (h + doublehashinc(hashval, size)) % size;
	    else if (++h >= size)
		h -= size;
	}
    return 0;
}


template<class Type> Pix 
VHBag<Type>::add(const Type& item)
{
  if (HASHTABLE_TOO_CROWDED(count, size))
    resize();

  unsigned int bestspot = size;
  unsigned int hashval = Hash(item);
  unsigned int h = hashval % size;
  for (unsigned int i = 0; i <= size; ++i)
      {
	  if (status[h] == EMPTYCELL)
	      {
		  if (bestspot >= size) bestspot = h;
		  tab[bestspot] = item;
		  status[bestspot] = VALIDCELL;
		  ++count;
		  return Pix(&tab[bestspot]);
	      }
	  else if (status[h] == DELETEDCELL)
	      {
		  if (bestspot >= size) bestspot = h;
	      }
	  if (i == 0)
	      h = (h + doublehashinc(hashval, size)) % size;
	  else if (++h >= size)
	      h -= size;
      }
  tab[bestspot] = item;
  status[bestspot] = VALIDCELL;
  ++count;
  return Pix(&tab[bestspot]);
  
}


template<class Type> void 
VHBag<Type>::del(const Type& key)
{
    unsigned int hashval = Hash(key);
    unsigned int h = hashval % size;
    for (unsigned int i = 0; i <= size; ++i)
	{
	    if (status[h] == EMPTYCELL)
		return;
	    else if (status[h] == VALIDCELL && Equal(key, tab[h]))
		{
		    status[h] = DELETEDCELL;
		    --count;
		    return;
		}
	    if (i == 0)
		h = (h + doublehashinc(hashval, size)) % size;
	    else if (++h >= size)
		h -= size;
	}
}


template<class Type> void 
VHBag<Type>::clear() 
{
  for (unsigned int i = 0; i < size; ++i) status[i] = EMPTYCELL;
  count = 0;
}

template<class Type> void 
VHBag<Type>::resize(unsigned int newsize)
{
    if (newsize <= count)
	{
	    newsize = DEFAULT_INITIAL_CAPACITY;
	    while (HASHTABLE_TOO_CROWDED(count, newsize))  newsize <<= 1;
	}
    Type* oldtab = tab;
    char* oldstatus = status;
    unsigned int oldsize = size;
    tab = new Type[size = newsize];
    status = new char[size];

    for (unsigned int i = 0; i < size; ++i) status[i] = EMPTYCELL;
    count = 0;
    for (i = 0; i < oldsize; ++i) 
	if (oldstatus[i] == VALIDCELL) add(oldtab[i]);
    delete [] oldtab;
    
    delete [] oldstatus;
}

template<class Type> Pix 
VHBag<Type>::first() const
{
    for (unsigned int pos = 0; pos < size; ++pos)
	if (status[pos] == VALIDCELL) return Pix(&tab[pos]);
    return 0;
}

template<class Type> void 
VHBag<Type>::next(Pix& i) const
{
    if (i == 0) return;
    unsigned int pos = ((unsigned)i - (unsigned)tab) / sizeof(Type) + 1;
    for (; pos < size; ++pos)
	if (status[pos] == VALIDCELL)
	    {
		i = Pix(&tab[pos]);
		return;
	    }
    i = 0;
}
  
template<class Type> int 
VHBag<Type>:: operator == (const VHBag<Type>& b) const
{
    if (count != b.count)
	return 0;
    else
	{
	    for (unsigned int i = 0; i < size; ++i)
		if (status[i] == VALIDCELL && b.seek(tab[i]) == 0)
		    return 0;
	    for (i = 0; i < b.size; ++i)
		if (b.status[i] == VALIDCELL && seek(b.tab[i]) == 0)
		    return 0;
	    return 1;
	}
}

template<class Type> int 
VHBag<Type>::nof(const Type& item) const
{
    int n = 0;
    unsigned int hashval = Hash(item);
    unsigned int h = hashval % size;
    unsigned int firsth = size;
    for (unsigned int i = 0; i <= size; ++i)
	{
	    if (status[h] == EMPTYCELL)
		return n;
	    else if (h != firsth && status[h] == VALIDCELL && 
		     Equal(item, tab[h]))
		{
		    ++n;
		    if (firsth >= size)
			firsth = h;
		}
	    if (i == 0)
		h = (h + doublehashinc(hashval, size)) % size;
	    else if (++h >= size)
		h -= size;
	}
    return n;
}

template<class Type> int 
VHBag<Type>::OK() const
{
    int v = tab != 0;
    v &= status != 0;
    int n = 0;
    for (unsigned int i = 0; i < size; ++i) 
	{
	    if (status[i] == VALIDCELL) ++n;
	    else if (status[i] != DELETEDCELL && status[i] != EMPTYCELL)
		v = 0;
	}
    v &= n == count;
    if (!v) error("invariant failure");
    return v;
}

#endif
