// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


// Modified for Templates by Tobias Himstedt
// Universitaet Hildesheim
// 1/95

// Modified for gcc 2.7.2 by Burkhard Bieber
// last-update: 10.07.96

#ifndef _VHMap_h
#define _VHMap_h 1

#include "Map.h"

#ifndef DEFAULT_INITIAL_CAPACITY
#define DEFAULT_INITIAL_CAPACITY 100
#endif

// HASHTABLE_TOO_CROWDED(COUNT, SIZE) is true iff a hash table with COUNT
// elements and SIZE slots is too full, and should be resized.
// This is so if available space is less than 1/8.

#define HASHTABLE_TOO_CROWDED(COUNT, SIZE) ((SIZE) - ((SIZE) >> 3) <= (COUNT))

template<class Key, class Value>
class VHMap : public Map<Key,Value> {
  protected:
    Key*           tab;
    Value*         cont;
    char*          status;
    unsigned int   size;

  public:
    VHMap(const Value& dflt,unsigned int sz=DEFAULT_INITIAL_CAPACITY);
    VHMap(const VHMap<Key,Value>& a);
    ~VHMap();
    
    Value&        operator [] (const Key& key);
    
    void          del(const Key& key);
    
    Pix           first() const;
    void          next(Pix& i) const ;
    Key&          key(Pix i) const;
    Value&        contents(Pix i) const;

    Pix           seek(const Key& key) const;
    int           contains(const Key& key) const;
    
    void          clear(); 
    void          resize(unsigned int newsize = 0);

    const VHMap<Key, Value>& operator =(const VHMap<Key,Value>& b);
    int           OK() const;
};

template<class Key, class Value> inline 
VHMap<Key,Value>::~VHMap()
{
    // ein Test
    delete [] tab;
    delete [] cont;
    delete [] status;
}

template<class Key, class Value> inline int 
VHMap<Key,Value>::contains(const Key& key) const
{
    return seek(key) != 0;
}

template<class Key, class Value> inline Key& 
VHMap<Key,Value>::key(Pix i) const
{
    if (i == 0) error("null Pix");
    return *((Key*)i);
}

template<class Key, class Value> inline Value& 
VHMap<Key,Value>::contents(Pix i) const
{
    if (i == 0) error("null Pix");
    return cont[((unsigned)(i) - (unsigned)(tab)) / sizeof(Key)];
}

#define EMPTYCELL   0
#define VALIDCELL   1
#define DELETEDCELL 2

template<class Key, class Value>
VHMap<Key,Value>::VHMap(const Value& dflt, unsigned int sz) : 
    Map<Key,Value>(dflt)
{
    tab = new Key[size = sz];
    cont = new Value[size];
    status = new char[size];
    for (unsigned int i = 0; i < size; ++i) status[i] = EMPTYCELL;
}

template<class Key, class Value>
VHMap<Key,Value>::VHMap(const VHMap<Key,Value>& a) : Map<Key,Value>(a.def)
{
    tab = new Key[size = a.size];
    cont = new Value[size];
    status = new char[size];
    for (unsigned int i = 0; i < size; ++i) status[i] = EMPTYCELL;
    count = 0;
    for (Pix p = a.first(); p; a.next(p)) (*this)[a.key(p)] = a.contents(p);
}

template<class Key, class Value> const VHMap<Key, Value>&
VHMap<Key,Value>::operator =(const VHMap<Key,Value>& a) 
{
    if (this == &a)
	return *this;

    Map<Key,Value>::operator =(a);

    // ein Test
    delete [] tab;
    delete [] cont;
    delete [] status;

    tab = new Key[size = a.size];
    cont = new Value[size];
    status = new char[size];
    for (unsigned int i = 0; i < size; ++i) status[i] = EMPTYCELL;
    count = 0;
    for (Pix p = a.first(); p; a.next(p)) (*this)[a.key(p)] = a.contents(p);
    return *this;
}


/* 
 * hashing method: double hash based on high bits of hash fct,
 * followed by linear probe. Can't do too much better if table
 * sizes not constrained to be prime.
*/


#ifndef DOUBLEHASH
#define DOUBLEHASH
static inline unsigned int 
doublehashinc(unsigned int h, unsigned int s)
{
    unsigned int dh =  ((h / s) % s);
    return (dh > 1)? dh : 1;
}
#endif


template<class Key, class Value> Pix 
VHMap<Key,Value>::seek(const Key& key) const
{
    unsigned int hashval = Hash(key);
    unsigned int h = hashval % size;
    for (unsigned int i = 0; i <= size; ++i)
	{
	    if (status[h] == EMPTYCELL)
		return 0;
	    else if (status[h] == VALIDCELL && Equal(key, tab[h]))
		return Pix(&tab[h]);
	    if (i == 0)
		h = (h + doublehashinc(hashval, size)) % size;
	    else if (++h >= size)
		h -= size;
	}
    return 0;
}

template<class Key, class Value> Value& 
VHMap<Key,Value>::operator [](const Key& item)
{
    if (HASHTABLE_TOO_CROWDED(count, size))
	resize(2*size);
    
    unsigned int bestspot = size;
    unsigned int hashval = Hash(item);
    unsigned int h = hashval % size;
    for (unsigned int i = 0; i <= size; ++i)
	{
	    if (status[h] == EMPTYCELL)
		{
		    ++count;
		    if (bestspot >= size) bestspot = h;
		    tab[bestspot] = item;
		    status[bestspot] = VALIDCELL;
		    cont[bestspot] = def;
		    return cont[bestspot];
		}
	    else if (status[h] == DELETEDCELL)
		{
		    if (bestspot >= size) bestspot = h;
		}
	    else if (Equal(tab[h],item) && status[h] == VALIDCELL) {
		return cont[h];
            }
	    
	    if (i == 0)
		h = (h + doublehashinc(hashval, size)) % size;
	    else if (++h >= size)
		h -= size;
	}
    
    ++count;
    status[bestspot] = VALIDCELL;
    tab[bestspot] = item;
    cont[bestspot] = def;
    return cont[bestspot];
}


template<class Key, class Value> void 
VHMap<Key,Value>::del(const Key& key)
{
    unsigned int hashval = Hash(key);
    unsigned int h = hashval % size;
    for (unsigned int i = 0; i <= size; ++i)
	{
	    if (status[h] == EMPTYCELL)
		return;
	    else if (status[h] == VALIDCELL && Equal(key, tab[h]))
		{
		    status[h] = DELETEDCELL;
		    --count;
		    return;
		}
	    if (i == 0)
		h = (h + doublehashinc(hashval, size)) % size;
	    else if (++h >= size)
		h -= size;
	}
}


template<class Key, class Value> void 
VHMap<Key,Value>::clear()
{
    for (unsigned int i = 0; i < size; ++i) status[i] = EMPTYCELL;
    count = 0;
}

template<class Key, class Value> void 
VHMap<Key,Value>::resize(unsigned int newsize)
{
    if (newsize <= count)
	{
	    newsize = DEFAULT_INITIAL_CAPACITY;
	    while (HASHTABLE_TOO_CROWDED(count, newsize)) newsize <<= 1;
	}
    // cout << "Hash Tabelle: Neue Groesse : " << newsize << endl;
    Key* oldtab = tab;
    Value* oldcont = cont;
    char* oldstatus = status;
    unsigned int oldsize = size;
    tab = new Key[size = newsize];
    cont = new Value[size];
    status = new char[size];
    unsigned int i;
    for (i = 0; i < size; ++i) status[i] = EMPTYCELL;
    count = 0;
    for (i = 0; i < oldsize; ++i) 
	if (oldstatus[i] == VALIDCELL) 
	    (*this)[oldtab[i]] = oldcont[i];
    // ein Test
    delete [] oldtab;
    delete [] oldcont;
    delete [] oldstatus;
}

template<class Key, class Value> Pix 
VHMap<Key,Value>::first() const
{
    for (unsigned int pos = 0; pos < size; ++pos)
	if (status[pos] == VALIDCELL) return Pix(&tab[pos]);
    return 0;
}

template<class Key, class Value> void 
VHMap<Key,Value>::next(Pix& i) const
{
    if (i == 0) return;
    unsigned int pos = ((unsigned)i - (unsigned)tab) / sizeof(Key) + 1;
    for (; pos < size; ++pos)
	if (status[pos] == VALIDCELL)
	    {
		i = Pix(&tab[pos]);
		return;
	    }
    i = 0;
}


template<class Key, class Value> int 
VHMap<Key,Value>::OK() const
{
    int v = tab != 0;
    v &= status != 0;
    unsigned int n = 0;
    for (unsigned int i = 0; i < size; ++i) 
	{
	    if (status[i] == VALIDCELL) ++n;
	    else if (status[i] != DELETEDCELL && status[i] != EMPTYCELL)
		v = 0;
	}
    v &= n == count;
    if (!v) error("invariant failure");
    return v;
}

#endif
