/*****************************************************************************/
/* parser.c -- Functions used while parsing formulae                         */
/*****************************************************************************/ 

#include <string.h>
#include "parser.h"
#include "transform.h"
#include "sdlform.h"
#include "formula.h"
#include "scanner.h"
#include "output.h"
#include "convertPEP.h"

/*****************************************************************************/
/* Global variables.                                                         */
/*****************************************************************************/

MacroTable      FT_MacroTable = NULL;

#define PlaceNode(pl) NewNode(PLACE,NULL,(BTree) pl)

/*****************************************************************************/
/* Check whether the given net is a Petri Box; abort if not.                 */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
 
void TestPetriBox( void )

#else

void TestPetriBox()

#endif

  {
    if ( strcmp( CM_nettype, "PetriBox" ) )
      yyerror( "wrong net type; 'Petri Box' expected" );
  } /* TestPetriBox() */


/*****************************************************************************/
/* Generate a new node for the parse tree.                                   */
/*****************************************************************************/ 
#if defined(__STDC__) || defined(__cplusplus)
BTree NewNode(int op, BTree left, BTree right)
#else
BTree NewNode(op, left, right)
  int   op;
  BTree left;
  BTree right;
#endif

{
	register BTree node;

	(node = (BTree) Malloc(sizeof(BTreeElem)))->op = op;
	node->left = left;
	node->right = right;

	return node;
} /* NewNode() */


/*****************************************************************************/
/* Return a AND/OR list of n places.                                         */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
BTree form_ext_pl_label(SET plset, int op)
#else
BTree form_ext_pl_label(plset, op)
  SET plset;
  int op;
#endif
{
	BTree node = NULL;

	while(plset)
	{
		node = node? NewNode(op, node, PlaceNode(plset->item))
			   : PlaceNode(plset->item);
	  plset = plset->next;
	}

	return node? node : NewNode(op == AND? TRUE : FALSE, NULL, NULL);
} /* form_ext_pl_label() */


/*****************************************************************************/
/* Mutual exclusion of n > 1 places.                                         */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
static BTree form_exclusion(ASET aset)
#else
static BTree form_exclusion(aset)
  ASET aset;
#endif
{
	register SET	list,list2;
	register BTree	node,pnode;

	node = NULL;
	for (list = aset->set; list; list = list->next)
	    for (list2 = list->next; list2; list2 = list2->next)
	    {
		pnode = NewNode('(', NULL, NewNode(AND,
				PlaceNode(list->item),PlaceNode(list2->item)));
		node = node? NewNode(OR, node, pnode) : pnode;
	    }
	return node? NewNode(NOT, NULL,
			NewNode(DIAMOND, NULL,
				NewNode('(', NULL, node)))
		   : NewNode(TRUE,NULL,NULL);
} /* form_exclusion() */


/*****************************************************************************/
/* Concurrency of n > 1 transitions.                                         */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
static BTree form_concurrent(ASET aset)
#else
static BTree form_concurrent(aset)
  ASET aset;
#endif
{
	register Place      pl;
	register STEdge     st;
	register SET        list;
	register BTree	    node;

	for (pl = FT_net->Places; pl; pl = pl->nextplace)
		pl->inscript = NULL;

	for (list = aset->set; list; list = list->next)
	{
	    for (st = ((Transition) list->item)->preset; st; st = st->nextplace)
	    {
		    if (st->place->inscript)
		    {
			fprintf(stderr,
				"WARNING: the given transitions have common "
				"place(s) in their pre-sets, so the generated "
				"formula is %s (false).\n", sym_false);
			return NewNode(FALSE,NULL,NULL);
		    }
		    st->place->inscript = (char*) 1;
	    }
	}

	for (pl = FT_net->Places; pl && !pl->inscript; pl = pl->nextplace);
	if (!pl) return NewNode(TRUE,NULL,NULL);
	 
	node = PlaceNode(pl);
	while ((pl = pl->nextplace))
		if (pl->inscript) node = NewNode(AND, node, PlaceNode(pl));

	return NewNode(DIAMOND,NULL,NewNode('(',NULL,node));

} /* form_concurrent() */


/*****************************************************************************/
/* Reachability of a marking.                                                */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
static BTree form_reachable(ASET aset)
#else
static BTree form_reachable(aset)
  ASET aset;
#endif
{
	register Place	pl;
	register SET	list;
	register BTree	node,pnode;

	for (pl = FT_net->Places; pl; pl = pl->nextplace)
		pl->inscript = NULL;

	for (list = aset->set; list; list = list->next)
		list->item->inscript = (char*) 1;

	node = NULL;
        for (pl = FT_net->Places; pl; pl = pl->nextplace)
	{
		pnode = PlaceNode(pl);
		if (!pl->inscript) pnode = NewNode(NOT, NULL, pnode);
		node = node? NewNode(AND, node, pnode) : pnode;
	}

	return node? NewNode(DIAMOND, NULL, NewNode('(',NULL,node))
		   : NewNode(TRUE,NULL,NULL);

} /* form_reachable() */


/*****************************************************************************/
/* Partial reachability of a marking.                                        */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
static BTree form_partially_reachable(ASET aset)
#else
static BTree form_partially_reachable(aset)
  ASET aset;
#endif
{
	return aset->set? NewNode(DIAMOND, NULL,
			    NewNode('(',NULL, form_ext_pl_label(aset->set,AND)))
			: NewNode(TRUE, NULL, NULL);

} /* form_partially_reachable() */


/*****************************************************************************/
/* Reversibility of the start marking ( "cyclicity" ).                       */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
static BTree form_cyclicity(ASET aset)
#else
static BTree form_cyclicity(aset)
  ASET aset;
#endif
{
	register Place pl;

	aset = allocASET();
	for (pl = FT_net->Places; pl; pl = pl->nextplace)
		if (pl->marks) set_insert(&(aset->set),pl);

	return NewNode(BOX, NULL, form_reachable(aset));
} /* form_cyclicity() */


/*****************************************************************************/
/* Termination ( Reachability of the end marking of a Box ).                 */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
static BTree form_box_termination(ASET aset)
#else
static BTree form_box_termination(aset)
  ASET aset;
#endif
{
	register Place pl;
 
	aset = allocASET();
	TestPetriBox();
	for (pl = FT_net->Places; pl; pl = pl->nextplace)
		if (pl->label == EXIT) set_insert(&(aset->set),pl);

	return aset->set? form_partially_reachable(aset)
			: NewNode(FALSE,NULL,NULL);
} /* form_box_termination() */
 

#if defined(__STDC__) || defined(__cplusplus)
static BTree form_spin_deadlock(ASET aset)
#else
static BTree form_spin_deadlock(aset)
  ASET aset;
#endif
{
  Place pl;

  pl= (Place) Malloc( sizeof( PlElem ) );
  pl->name=MYstrdup("DEADLOCK");
  return NewNode(PLACE,NULL,(BTree) pl);
}

/*****************************************************************************/
/* Check whether the net is deadlock-free.                                   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
static BTree form_deadlockfree(ASET aset)
#else
static BTree form_deadlockfree(aset)
  ASET aset;
#endif
{
	register Place      pl;
	register Transition tr,tr2;
	register STEdge     st;
	register Cardinal32 trcount;
	register bool       equalPresets;
	register BTree	    node,pnode;

	for (tr = FT_net->Transitions; tr; tr = tr->nexttrans)
	{
		for (trcount = 0, st = tr->preset; st;)
			trcount++, st = st->nextplace;
		tr->inscript = (char*) trcount;
	}
 
	for (pl = FT_net->Places; pl; pl = pl->nextplace)
		pl->inscript = NULL;

	node = NULL;
	trcount = 0;
	for (tr = FT_net->Transitions; tr; tr = tr->nexttrans)
	{
		trcount++;
		for (st = tr->preset; st; st = st->nextplace)
			st->place->inscript = (char*) trcount;

		equalPresets = False;
		for (tr2 = tr->prevtrans; tr2; tr2 = tr2->prevtrans)
		    if (tr2->inscript == tr->inscript)
		    {
			for (st = tr2->preset;
			     st && (Cardinal32) st->place->inscript == trcount;
			     st = st->nextplace);
			if ((equalPresets = !st)) break;
		    } /* if */

		if (!equalPresets)
		{
		    pnode = NULL;
		    for (st = tr->preset; st; st = st->nextplace)
			pnode = pnode? NewNode(AND,pnode,PlaceNode(st->place))
				     : PlaceNode(st->place);
		    if (pnode)
			node = node? NewNode(OR,node,NewNode('(',NULL,pnode))
				   : NewNode('(',NULL,pnode);
		}
	}

	return node? NewNode(BOX,NULL,NewNode('(',NULL,node))
		   : NewNode(FALSE,NULL,NULL);
} /* form_deadlockfree() */


/*****************************************************************************/
/* Liveness test.                                                            */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
static BTree form_live(ASET aset)
#else
static BTree form_live(aset)
  ASET aset;
#endif
{
	register Place      pl;
	register Transition tr,tr2;
	register STEdge     st;
	register Cardinal32 trcount;
	register bool       equalPresets;
	register BTree	    node,pnode;
	register SET	    list;

	for (tr = FT_net->Transitions; tr; tr = tr->nexttrans)
	{
		for (trcount = 0, st = tr->preset; st;)
			trcount++, st = st->nextplace;
		tr->inscript = (char*) trcount;
	}
 
	for (pl = FT_net->Places; pl; pl = pl->nextplace)
		pl->inscript = NULL;

	node = NULL;
	trcount = 0;
	for (list = aset->set; list; list = list->next)
	{
		trcount++;
		tr = (Transition) list->item;
		for (st = tr->preset; st; st = st->nextplace)
			st->place->inscript = (char*) trcount;

		equalPresets = False;
		for (tr2 = tr->prevtrans; tr2; tr2 = tr2->prevtrans)
		    if (tr2->inscript == tr->inscript)
		    {
			for (st = tr2->preset;
			     st && (Cardinal32) st->place->inscript == trcount;
			     st = st->nextplace);
			if ((equalPresets = !st)) break;
		    } /* if */

		if (!equalPresets)
		{
		    pnode = NULL;
		    for (st = tr->preset; st; st = st->nextplace)
			pnode = pnode? NewNode(AND,pnode,PlaceNode(st->place))
				     : PlaceNode(st->place);
		    if (pnode)
		    {
			pnode = NewNode(BOX,NULL,
				NewNode(DIAMOND,NULL,
				NewNode('(',NULL,pnode)));
			node = node? NewNode(aset->or? OR:AND, node, pnode)
				   : pnode;
		    }
		}
	}

	return node? NewNode('(', NULL, node)
		   : NewNode(aset->or? FALSE:TRUE, NULL, NULL);

} /* form_live() */


/*****************************************************************************/
/* Initialize the macro table.                                               */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void initMacroTable( bool use_spin)

#else

void initMacroTable( use_spin )

  bool use_spin;

#endif

  {

#define MTABLE_SIZE 15

#define MEntry( NAME, CLASS, FUNCTION ) \
    ( mtab->name       = (NAME), \
      mtab->mclass     = (CLASS), \
      mtab++->function = (FUNCTION) )

    register MacroTable mtab;

    mtab = FT_MacroTable = (MacroTable) Malloc( ( MTABLE_SIZE + 1 )
                                                 * sizeof( MacroTableEntry ) );

    MEntry( "COncurrent",          F_MACRO_CONC,    form_concurrent );
    if (use_spin)
    {
      MEntry( "NODEADlock",        F_MACRO_DEAD,    form_deadlockfree );
      MEntry( "DEADLOCK",        F_MACRO_DEAD,    form_spin_deadlock );
    }
    else
      MEntry( "DEADlockfree",        F_MACRO_DEAD,    form_deadlockfree );
    MEntry( "EXclusion",           F_MACRO_EXCL,    form_exclusion );
    MEntry( "Live",                F_MACRO_LIVE,    form_live );
    MEntry( "PARTially_reachable", F_MACRO_PREACH,  form_partially_reachable );
    MEntry( "POSTset",             POSTSET,         NULL );
    MEntry( "PREset",              PRESET,          NULL );
    MEntry( "REACHable",           F_MACRO_REACH,   form_reachable );
    MEntry( "REVersible",          F_MACRO_REV,     form_cyclicity );
    MEntry( "TERMination",         F_MACRO_TERM,    form_box_termination );
    /* close the table with a null entry */
    MEntry( NULL,                  0,               NULL );

  } /* initMacroTable() */
