/* block.c                                                                  */
/* Behandlung von Bloecken im Uebersetzer von B(PN)2 nach High-level Netzen */
/* Burkhard Bieber 1995/96                                                  */
/* last-update: 08.04.96                                                    */

  /************/
  /* includes */
  /************/

#include <ctype.h>
#include <string.h>
#include "block.h"

  /***********/
  /* defines */
  /***********/

#define BLOCKTABLE_BUFFER_SIZE 10

  /*********************/
  /* Globale Variablen */
  /*********************/

Cardinal32 MaxBlocks  = BLOCKTABLE_BUFFER_SIZE;
Cardinal32 BlockCounter = 0;

static BlockTablePtr BlockTable = NULL;
static Cardinal unnamed_block_counter = 0;  /* Zaehler unnamed-blocks */

  /***********/
  /* defines */
  /***********/

  /* Die folgende Definition zeigt, dass die Dinger nur hier, */
  /* eben in der Datei block.c alloziert werden !             */
#define M_allocBlock() ((M_Block) MYmalloc(sizeof(M_BlockElem)))

  /**************/
  /* Prototypen */
  /**************/

  /* Einen Block kopieren. */
static void M_CopyBlock(BlockPtr b2, const BlockPtr b1);

  /* Stellt fest, ob eine BlockNr im irgendeinem Knoten verwendet wird. */
static int BlockNr_is_ever_used(M_Net net, BlockNr bnr);

  /**************/
  /* Funktionen */
  /**************/

  /* Den BlockTable initialisieren. */
void InitBlockTable(void)
{
  if (BlockTable != NULL) {
    RemoveBlockTable();
  } /* if bereits was drinne */
  BlockTable = (BlockTablePtr) MYmalloc(MaxBlocks * sizeof(BlockPtr));
  *BlockTable = NULL;
  MaxBlocks = BLOCKTABLE_BUFFER_SIZE;
  BlockCounter = 0;
}  /* InitBlockTable */

  /* Den BlockTable wegschmeissen. */
void RemoveBlockTable(void)
{
  register BlockTablePtr bt_ptr;
  register BlockPtr b_ptr_rem;

  if (BlockTable == NULL)  /* Ist ja garnichts da! */
    return;

  bt_ptr = BlockTable;
  while (*bt_ptr != NULL) {
    b_ptr_rem = *bt_ptr;
    bt_ptr++;
    MYfree(b_ptr_rem->blockname);  /* Blockname */
    MYfree(b_ptr_rem->children);   /* Kinderchen */
    MYfree(b_ptr_rem);             /* BlockTable-Eintrag selber */
  }  /* while noch Bloecke vorhanden */
  MYfree(BlockTable);
  BlockTable = NULL;
  BlockCounter = 0;
  MaxBlocks = 0;
}  /* RemoveBlockTable */

  /* Kopiert eine kompletten Block (b1 wird nach b2 kopiert). */
static void M_CopyBlock(BlockPtr b2, const BlockPtr b1)
{
  unsigned int si;

  b2->blockname = MYstrdup(b1->blockname);
  b2->blockdeep = b1->blockdeep;
  b2->blocktags = b1->blocktags;
  b2->childcount = b1->childcount;
  b2->childmax = b1->childmax;
  si = sizeof(BlockNr) * b2->childmax;
  b2->children = (BlockNr *) MYmalloc(si);
  memcpy(b2->children, b1->children, si);
  b2->ref = b1->ref;
}  /* M_CopyBlock */

  /* Der Zeiger auf die BlockNr enthaelt nach erfolgreichem Aufruf die Nummer */
  /* des Blocks im BlockTable. In C++ sollte es eine Referenz sein (grrr..).  */
  /* Diese Funktion sieht also aehnlich aus, wie die StructToRef aus          */
  /* parser.c, funktioniert aber etwas anders.                                */
static void StructsToBlockTable(SymTabEntry *entry, BlockNr *bl, RefNr ref,
                                Cardinal blocktags)
{
  register BlockTablePtr bt_ptr;
  BlockPtr b_ptr;

  b_ptr = (BlockPtr) MYmalloc(sizeof(BlockStruct));
  b_ptr->blockname = MYstrdup(entry->s_name);
  b_ptr->blockdeep = entry->s_blknum;
  b_ptr->blocktags = blocktags;
  b_ptr->childcount = 0;  /* keine Kinderchen zur Zeit */
  b_ptr->childmax = BLOCK_CHUNK_CHILD_MAX;
  b_ptr->children = (BlockNr *) MYmalloc(BLOCK_CHUNK_CHILD_MAX*sizeof(BlockNr));
  *b_ptr->children = 0;  /* Listenende markieren */
  b_ptr->ref = ref;

  if (++BlockCounter >= MaxBlocks) {
    BlockTable = (BlockTablePtr) MYrealloc((char *) BlockTable,
      (MaxBlocks += BLOCKTABLE_BUFFER_SIZE) * sizeof(BlockPtr));
  } /* if Tabelle vergroessern */
  bt_ptr = BlockTable + BlockCounter - 1;
  *bt_ptr = MYmalloc(sizeof(BlockStruct));
  M_CopyBlock(*bt_ptr, b_ptr);
  *(bt_ptr + 1) = NULL;

    /* Blocknummer per Pointer auf BlockNr zurueckliefern */
  *bl = (Cardinal) (bt_ptr - BlockTable + 1);
    /* Speicher fuer unsere Hilfsstruktur freigeben */
  (void) MYfree(b_ptr->blockname);
  (void) MYfree(b_ptr->children);
  (void) MYfree(b_ptr); 
}  /* StructsToBlockTable */

  /* Funktion zum Umwandeln des Blockkodes `blk` in den entsprechenden. */
  /* Zeiger auf die Blockstruktur.                                      */
BlockPtr BlockToStruct(BlockNr blk)
{
  return ((BlockPtr) (BlockTable[blk - 1]));
}  /* BlockToStruct */

  /* Den Block in den BlockTable aufnehmen.                  */
  /* Vorher pruefen, ob Namen gleich und noch nicht bekannt. */
void add_block(SymTabEntry *entry, BlockNr *bl, RefNr ref, Cardinal tags)
{
  StructsToBlockTable(entry, bl, ref, tags);
}  /* add_block */

  /* Einen Block ohne eigenen Namen in den BlockTable aufnehmen. */
void add_unnamed_block(Integer blknum, BlockNr *bl, RefNr ref, Cardinal tags)
{
  SymTabEntry entry;  /* Hilfsstruktur */

  unnamed_block_counter++;
  entry.s_name = MYmalloc(20);
  sprintf(entry.s_name, "unnamed_block_%u", unnamed_block_counter); 
  entry.s_blknum = blknum;
  StructsToBlockTable(&entry, bl, ref, tags);
  (void) MYfree(entry.s_name);
}  /* add_unnamed_block */

  /* Fuegt zu allen Stellen den Verweis auf den Block bl hinzu. */
  /* Die Entry- und Exit-Stellen werden NICHT markiert!         */
void M_AddAllPlacesBlock(M_Net net, BlockNr bl)
{
  M_Place pl = net->Places;

  while (pl) {
    if (((pl->label && ENTRY) == 0) && ((pl->label && EXIT) == 0)) {
      if (pl->BlockEntries++)
        pl->blocks = MYrealloc(pl->blocks, pl->BlockEntries * sizeof(BlockNr));
      else
        pl->blocks = MYmalloc(sizeof(BlockNr));
      pl->blocks[pl->BlockEntries - 1] = bl;
    }  /* if Entry- und Exit-Stellen nicht markieren */
    pl = pl->nextplace;
  }  /* while noch Stellen zu behandeln */
}  /* M_AddAllPlacesBlock */

  /* Fuegt zu allen Transitionen den Verweis auf den Block bl hinzu. */
void M_AddAllTransBlock(M_Net net, BlockNr bl)
{
  M_Transition tr = net->Transitions;

  while (tr) {
    if (tr->BlockEntries++)
      tr->blocks = MYrealloc(tr->blocks, tr->BlockEntries * sizeof(BlockNr));
    else
      tr->blocks = MYmalloc(sizeof(BlockNr));
    tr->blocks[tr->BlockEntries - 1] = bl;
    tr = tr->nexttrans;
  }  /* while noch Transitionen zu behandeln */
}  /* M_AddAllTransBlock */

  /* Fuegt zu allen Phantomtransitionen den Verweis auf den Block bl hinzu. */
void M_AddAllPTransBlock(M_Net net, BlockNr bl)
{
  M_Transition ptr = net->PTransitions;

  while (ptr) {
    if (ptr->BlockEntries++)
      ptr->blocks = MYrealloc(ptr->blocks, ptr->BlockEntries * sizeof(BlockNr));
    else
      ptr->blocks = MYmalloc(sizeof(BlockNr));
    ptr->blocks[ptr->BlockEntries - 1] = bl;
    ptr = ptr->nexttrans;
  }  /* while noch Phantomtransitionen zu behandeln */
}  /* M_AddAllPTransBlock */

  /* Eine Funktion zum Kopieren der Blockverweise aus der Stelle src    */
  /* in die Stelle dest. Dabei wird der Speicher fuer die Blockverweise */
  /* in der Zielstelle alloziert.                                       */
void M_CopyPlaceBlockEntries(M_Place dest, M_Place src)
{
  int blocks_size;

  dest->BlockEntries = src->BlockEntries;
  if (src->BlockEntries) {
    blocks_size = src->BlockEntries * sizeof(BlockNr);
    dest->blocks = MYmalloc(blocks_size);
    memcpy(dest->blocks, src->blocks, blocks_size);
  }  /* if es gibt etwas zu kopieren */
  else {
    dest->blocks = NULL;
  }  /* else, keine Verweise */
}  /* M_CopyPlaceBlockEntries */

  /* Eine Funktion zum Kopieren der Blockverweise aus der Transition src    */
  /* in die Transition dest. Dabei wird der Speicher fuer die Blockverweise */
  /* in der Zieltransition alloziert.                                       */
void M_CopyTransBlockEntries(M_Transition dest, M_Transition src)
{
  int blocks_size;

  dest->BlockEntries = src->BlockEntries;
  if (src->BlockEntries) {
    blocks_size = src->BlockEntries * sizeof(BlockNr);
    dest->blocks = MYmalloc(blocks_size);
    memcpy(dest->blocks, src->blocks, blocks_size);
  }  /* if es gibt etwas zu kopieren */
  else {
    dest->blocks = NULL;
  }  /* else, keine Verweise */
}  /* M_CopyTransBlockEntries */

#ifdef DEBUG
  /* Debug-Ausgabe fuer Blockverweise einer Stelle/Transition */
static void debug_blocks(Cardinal blockentries, BlockNr *blocks)
{
  Cardinal i;
  BlockNr *bnr;

  printf(" #be=%u (", blockentries);
  i = 1;
  bnr = blocks;
  while (i <= blockentries) {
    printf("'%s',", (BlockToStruct(*bnr))->blockname);
    i++;
    bnr++;
  }  /* while immer noch etwas auzugeben */
  printf(")\n");
}  /* debug_blocks */
#endif

  /*******************************************/
  /* Funktionen fuer M_BlockElem's und Netze */
  /*******************************************/

  /* Den Inhalt von Block src nach Block dest kopieren.                */
  /* Die Zeiger und der BlockTable-Eintrag werden dabei NICHT kopiert. */
static void M_CopyBlockElem(M_Block dest, M_Block src)
{
  dest->blocknr = src->blocknr;
  dest->numname = src->numname;
  dest->ever_used = src->ever_used;
  dest->x_min = src->x_min;
  dest->y_min = src->y_min;
  dest->x_max = src->x_max;
  dest->y_max = src->y_max;
  dest->x_pos = src->x_pos;
  dest->y_pos = src->y_pos;
}  /* M_CopyBlockElem */

  /* In dem Netz werden die Zeiger auf die Blockelementeliste und */
  /* der Zaehler auf passende Startwerte (naemlich 0) gesetzt.    */
void M_InitBlockInNet(M_Net net)
{
  net->NumBlocks = 0;
  net->Blocks = NULL;
  net->LastBlock = NULL;
}  /* M_InitBlockInNet */

  /* Die Liste der Bloecke aus dem Netz wegschmeissen, NICHT BlockTable. */
void M_RemoveBlocksFromNet(M_Net net)
{
  M_Block mb, mbrem;

  mb = net->Blocks;
  while (mb != NULL) {
    mbrem = mb;          /* merken fuer das Loeschen */
    mb = mb->nextblock;  /* zum naechsten Block hopsen */
    MYfree(mbrem);       /* Speicher freigeben */
  }  /* while noch Bloecke vorhanden */
  M_InitBlockInNet(net);  /* Netzdaten zuruecksetzen */
}  /* M_RemoveBlocksFromNet */

void M_CopyBlockInNets(M_Net net_dest, M_Net net_src)
  /* Es werden nur die M_Block's kopiert, die Verweise in den BlockTable */
  /* bleiben wie sie sind.                                               */
{
  M_Block mb_src, mb_dest=NULL, mb_dest_prev;

  mb_src = net_src->Blocks;
  mb_dest_prev = NULL;
  net_dest->Blocks = NULL;
  while (mb_src != NULL) {
    mb_dest = M_allocBlock();  /* Speicher allozieren */
      /* neuen Block verzeigern */
    if (net_dest->Blocks == NULL)  /* Anfang eintragen */
      net_dest->Blocks = mb_dest;
    mb_dest->prevblock = mb_dest_prev;
    mb_dest->nextblock = NULL;
    if (mb_dest_prev != NULL)
      mb_dest_prev->nextblock = mb_dest;
    M_CopyBlockElem(mb_dest, mb_src);  /* Daten uebernehmen */
      /* und den naechsten */
    mb_dest_prev = mb_dest;
    mb_src = mb_src->nextblock;
  }  /* while noch Bloecke zu kopieren */
    /* Restliches in das Netz eintragen */
  net_dest->NumBlocks = net_src->NumBlocks;
  if (net_dest->NumBlocks != 0)
    net_dest->LastBlock = mb_dest;
  else {
    net_dest->Blocks = NULL;
    net_dest->LastBlock = NULL;
  }  /* else, keine Bloecke */
}  /* M_CopyBlockInNets */

  /* Blockinformationen vereinigen; net1 = net1 vereinigt net2 */
void M_JointBlockUnion(M_Net net1, M_Net net2)
{
  M_Block mb1, mb2;

  if (net2->NumBlocks == 0)  /* dann ist wohl nichts zu tun */
    return;
  mb2 = net2->Blocks;
  while (mb2 != NULL) {
      /* die aktuelle Blocknummer im ersten Netz suchen */
    mb1 = net1->Blocks;
    while ((mb1 != NULL) && (mb1->blocknr != mb2->blocknr))
      mb1 = mb1->nextblock;
    if (mb1 == NULL) {  /* die Blocknummer ist bisher unbekannt in net1 */
      mb1 = M_allocBlock();
      M_CopyBlockElem(mb1, mb2);  /* Inhalt kopieren, ohne Zeiger */
      mb1->prevblock = net1->LastBlock;
      if (net1->LastBlock != NULL)
        net1->LastBlock->nextblock = mb1;
      mb1->nextblock = NULL;
      if (net1->Blocks == NULL)
        net1->Blocks = mb1;
      net1->LastBlock = mb1;
      net1->NumBlocks++;
    }  /* if unbekannt */
    mb2 = mb2->nextblock;  /* auf zum naechsten Block */
  }  /* while noch Bloecke im zweiten Netz zu bearbeiten */
}  /* M_JointBlockUnion */

  /* Die angebene BlockNr wird innerhalb der M_BlockElem-Liste        */
  /* des Netzes gesucht und dort wird geprueft, ob das ever_used-Flag */
  /* gesetzt worden ist oder nicht.                                   */
static int BlockNr_is_ever_used(M_Net net, BlockNr bnr)
{
  M_Block mb;

  if (net->NumBlocks == 0)  /* dann ist wohl nichts zu tun */
    return False;
  mb = net->Blocks;
  while ((mb != NULL) && (mb->blocknr != bnr))
    mb = mb->nextblock;
  if (mb == NULL)  /* die Blocknummer ist unbekannt in net */
    return False;
  else  /* Blocknummer ist bekannt */
    return (mb->ever_used);
}  /* BlockNr_is_ever_used */

  /* Einen Block in die Liste der Bloecke des Netzes aufnehmen. */
void M_AppendBlock(M_Net net, BlockNr bnr)
{
  M_Block mb, mb2;
  BlockPtr b_ptr, b2_ptr;
  BlockNr *bnr_ptr;
  unsigned int offset;

    /* 1. das neue M_BlockElem erzeugen und grundsaetzlich initialisieren */
    /*    x_min, y_min, x_max, y_max, x_pos, y_pos */
    /*    werden spaeter initialisiert             */
  mb = M_allocBlock();
  mb->blocknr = bnr;  /* Index im BlockTable eintragen */
  mb->numname = 0;  /* Anzahl der untersuchten Stellen/Transitionen */
  mb->ever_used = False;

    /* 2. im Eintrag des BlockTable's (neuer Block) die Baumstruktur merken */
  b_ptr = BlockToStruct(mb->blocknr);  /* BlockTable-Eintrag holen */
  mb2 = net->Blocks;
  while (mb2 != NULL) {
    b2_ptr = BlockToStruct(mb2->blocknr);
    if (b_ptr->blockdeep+1 == b2_ptr->blockdeep) {
      bnr_ptr = b_ptr->children;
      while ((*bnr_ptr != 0) && (*bnr_ptr != mb2->blocknr))
        bnr_ptr++;
      if (*bnr_ptr == 0) {  /* noch nicht als Kind vermerkt */
        b_ptr->childcount++;
        if (b_ptr->childcount+1 > b_ptr->childmax) {  /* if kein Platz mehr */
            /* dann eben neuen Platz schaffen */
          offset = bnr_ptr - b_ptr->children;  /* rel. Position bestimmen */
          b_ptr->childmax += BLOCK_CHUNK_CHILD_MAX;
          b_ptr->children = (BlockNr *) MYrealloc(b_ptr->children,
            BLOCK_CHUNK_CHILD_MAX * b_ptr->childmax);
          bnr_ptr = b_ptr->children + offset;  /* neuer Eintrag */
        }  /* if kein Platz mehr */
        *bnr_ptr = mb2->blocknr;
        *(bnr_ptr + 1) = 0;  /* Listenende */
      }  /* if noch nicht als Kind vermerkt */
    }  /* if neuer Block ist Papi geworden */
    mb2 = mb2->nextblock;
  }  /* while noch Bloecke in dem Netz zu bearbeiten */

    /* 3. den neuen Block an das Netz haengen */
  if (net->NumBlocks == 0) {
    mb->prevblock = NULL;
    mb->nextblock = NULL;
    net->Blocks = mb;
    net->LastBlock = mb;
    net->NumBlocks++;
  }  /* if es ist der erste Block im Netz */
  else {  /* irgendwelche Bloecke sind vorhanden */
    mb->prevblock = net->LastBlock;
    mb->nextblock = NULL;
    mb->prevblock->nextblock = mb;
    net->LastBlock = mb;
    net->NumBlocks++;
  }  /* else, Bloecke bereits vorhanden */
}  /* M_AppendBlock */

  /* Ausgabe der Blockverweisliste eines Blocks einer Stelle */
  /* oder einer Transition in die bereits offene Datei ofp.  */
  /* Der Kode fuer den Blockverweis ist 'u' - used in block. */
void PrintBlocks(FILE *ofp, M_Net net, Cardinal numblocks, BlockNr *blocks)
{
  BlockNr *blocknr_ptr;
  Cardinal c, blocks_written;

  if (!(flag & BLOCK_OUTPUT) || (numblocks == 0))
    return;  /* keine Ausgabe gewuenscht oder keine Blockverweise vorhanden */
  blocknr_ptr = blocks;
  c = 0;
  blocks_written = 0;
  while (c < numblocks) {
    if (BlockNr_is_ever_used(net, *blocknr_ptr)) {
      if (blocks_written == 0)  /* erster zu schreibende Block */
        fprintf(ofp, "u\"(" );
      else  /* weitere zu schreibende Bloecke ( > 1) */
        fprintf(ofp, "," );
      fprintf(ofp, "%u", *blocknr_ptr);
      blocks_written++;
    }  /* if Block wurde verwendet */
    blocknr_ptr++;
    c++;
  }  /* while noch Bloecke auszugeben */
  if (blocks_written != 0)
    fprintf(ofp, ")\"");
}  /* PrintBlocks */

  /* Falls gewuenscht (BLOCK_OUTPUT-Flag auf True), dann werden die     */
  /* im Netz verwendeten Bloecke in die Datei im PEP-Format ausgegeben. */
  /* Die Datei ofp muss bereits zum Schreiben offen sein.               */
void PrintNet_Blocks_PEP(M_Net net, FILE *ofp, bool in_ref_file )
{
  M_Block bl;
  BlockPtr bl_ptr;
  Cardinal written_blocks = 0;  /* Hilfszaehler fuer die Ausgabe */

  if (flag & BLOCK_OUTPUT) {
    bl = net->Blocks;
    while (bl != NULL) {
      bl_ptr = BlockToStruct(bl->blocknr);
      if (bl->ever_used) {
        if (written_blocks == 0) {
          fprintf(ofp, "BL\n");
        }  /* if bisher noch kein Block geschrieben */
        written_blocks++;
	if (!in_ref_file)
	  fprintf(ofp, "%u \"",             /* BlockId aus BlockTable-Index */
		  bl->blocknr);
        fprintf(ofp, "B%u",                 /* Name */
          bl->blocknr);
	if (in_ref_file)
	  {
	    fprintf(ofp," T\"%d:0\"", bl_ptr->ref);
	  }
	else
	  {
	    fprintf(ofp, "\"%d@%d",         /* Mittelpunkt */
		    scale * bl->x_pos, scale * bl->y_pos);
            /* fprintf(ofp, "P%d", bl_ptr->blocktags); */
	    fprintf(ofp, "b\"%s\"",         /* Meaning begin-end-Name */
		    bl_ptr->blockname);
	    PrintBlocks(ofp, net, bl_ptr->childcount, bl_ptr->children);
	  }
        fprintf(ofp, "\n");
      }  /* if Block definiert und im zu schreibenden Netz verwendet worden */
      bl = bl->nextblock;
    }  /* while noch Bloecke vorhanden */
  }  /* if Bloecke ausgeben */
}  /* PrintNet_Blocks_PEP */

  /* Aus den bereits fertig positionierten Stellen und Transitionen des      */
  /* Netzes die Position der Bloecke bestimmen. Sie werden zentriert in      */
  /* Bounding-Box der zugehoerigen Elemente gesetzt. Synchronisations-       */
  /* transitionen werden dabei nicht beruecksichtigt. Auch Zugriffe auf      */
  /* globale Variablen werden nicht beruecksichtigt. Die Liste der Bloecke   */
  /* fuer das Netz wird erst jetzt initialisiert. Vorher werden die Daten im */
  /* BlockTable gesammelt (diese werden natuerlich uebernommen).             */
  /* Es werden nur die Bloecke erstellt, die auch tatsaechlich im            */
  /* fertigen Netz benutzt werden. Der BlockTable wird beim Parsen aufge-    */
  /* baut, er koennte somit auch ueberfluessige Bloecke enthalten.           */
void M_CreateBlockElements(M_Net net)
{
  M_Block mb;
  M_Place pl_ptr;
  M_Transition tr_ptr;
  Cardinal act_blocknr_count;
  BlockNr *act_blocknr_ptr;

    /* 1. Alle Stellen des Netzes durchlaufen und pruefen, ob der Block */
    /*    vorhanden ist (das sollte wohl der Fall sein). Anschliessend  */
    /*    die Position der Stelle mit der derzeitigen Bounding-Box des  */
    /*    Blocks vergleichen und die BBox evtenuell angleichen.         */
    /*    Den betreffenden Block als ever_used markieren.               */
  pl_ptr = net->Places;
  while (pl_ptr != NULL) {
    act_blocknr_count = 0;
    act_blocknr_ptr = pl_ptr->blocks;
    while (act_blocknr_count < pl_ptr->BlockEntries) {
      mb = net->Blocks;
      while ((mb != NULL) && (mb->blocknr != *act_blocknr_ptr))
        mb = mb->nextblock;
      if (mb != NULL) {
        if (mb->ever_used) {
          mb->x_min = Min(mb->x_min, pl_ptr->x_pos);
          mb->y_min = Min(mb->y_min, pl_ptr->y_pos);
          mb->x_max = Max(mb->x_max, pl_ptr->x_pos);
          mb->y_max = Max(mb->y_max, pl_ptr->y_pos);
        }  /* if bereits Netzknoten untersucht */
        else {
          mb->x_min = pl_ptr->x_pos;  mb->y_min = pl_ptr->y_pos;
          mb->x_max = pl_ptr->x_pos;  mb->y_max = pl_ptr->y_pos;
        }  /* else, noch keine Netzknoten untersucht */
        mb->ever_used = True;
      }  /* if BlockNr gefunden */
      else
        fprintf(stderr, "Internal error: places: blocknr (%u) not found\n",
          *act_blocknr_ptr);
      act_blocknr_count++;
      act_blocknr_ptr++;
    }  /* while noch Bloecke in der Stelle vermerkt */
    pl_ptr = pl_ptr->nextplace;
  }  /* while noch Stellen zu bearbeiten */

    /* 2. Alle Transitionen nach dem gleichen Schema wie unter 1. behandeln. */
  tr_ptr = net->Transitions;
  while (tr_ptr != NULL) {
    act_blocknr_count = 0;
    act_blocknr_ptr = tr_ptr->blocks;
    while (act_blocknr_count < tr_ptr->BlockEntries) {
      mb = net->Blocks;
      while ((mb != NULL) && (mb->blocknr != *act_blocknr_ptr))
        mb = mb->nextblock;
      if (mb != NULL) {
        if (mb->ever_used) {
          mb->x_min = Min(mb->x_min, tr_ptr->x_pos);
          mb->y_min = Min(mb->y_min, tr_ptr->y_pos);
          mb->x_max = Max(mb->x_max, tr_ptr->x_pos);
          mb->y_max = Max(mb->y_max, tr_ptr->y_pos);
        }  /* if bereits Netzknoten untersucht */
        else {
          mb->x_min = tr_ptr->x_pos;  mb->y_min = tr_ptr->y_pos;
          mb->x_max = tr_ptr->x_pos;  mb->y_max = tr_ptr->y_pos;
        }  /* else, noch keine Netzknoten untersucht */
        mb->ever_used = True;
      }  /* if BlockNr gefunden */
      else
        fprintf(stderr,
          "Internal error: blocknr (%u) not found\n",
          *act_blocknr_ptr);
      act_blocknr_count++;
      act_blocknr_ptr++;
    }  /* while noch Bloecke in der Stelle vermerkt */
    tr_ptr = tr_ptr->nexttrans;
  }  /* while noch Transitionen zu bearbeiten */

    /* 3. Die Position der Bloecke bestimmen. Mittelpunkt der Bounding Box. */
  mb = net->Blocks;
  while (mb != NULL) {
    if (mb->ever_used) {
      mb->x_pos = (mb->x_min + mb->x_max) / 2;
      mb->y_pos = (mb->y_min + mb->y_max) / 2;
    }  /* if Block ist verwendet worden */
    else {  /* Standardwerte eintragen, rein willkuerlich */
      mb->x_min = 10;  mb->x_max = 10;  mb->x_pos = 10;
      mb->y_min = 10;  mb->y_max = 10;  mb->y_pos = 10;
    }  /* else, Standardwerte eintragen */
    mb = mb->nextblock;
  }  /* while nicht am Ende der Blockliste */

}  /* M_CreateBlockElements */

  /*** Dateiende block.c ***/

