/****************************************************************************/
/* box_sub.c                                                                */
/*                                                                          */
/* aus Elementaroperatoren zusammengesetzte Routinen zur kompositionellen   */
/* Verknuepfung von M-Netz                                                  */
/*                                                                          */
/* Burkhard Graves, 26.6.'93                                                */
/* Martin Ackermann, Sommer 95                                              */
/* Burkhard Bieber 1995:                                                    */
/*   - Blocktypen ergaenzt                                                  */
/*   - Sync-/Phantom-Transitionen ergaenzt                                  */
/* Martin Ackermann, Winter 1996/97                                         */
/*   - Prozeduren, andere Referenzen, mehr Effizienz, Fehler beseitigt      */
/****************************************************************************/

  /************/
  /* includes */
  /************/

#include <string.h>
#include "block.h"
#include "ptrans.h"
#include "box_sub.h"
#include "move.h"
#include "multiset.h"
#include "var_mset.h"
#include "renaming.h"
#include "iter.h"
#include "tables.h"
#include "annotat.h"
#include "f_parse.h"
#include "macros.h"

  /********************************/
  /* Inits fuer globale Variablen */
  /********************************/

Cardinal dead_eliminated = 0;  /* Anzahl der entfernten toten Transitionen */

char     actname[MAX_IDENT_LENGTH] = ""; /* Name der aktuellen Sync-Aktion */

  /************************/
  /* typedefs und defines */
  /************************/

/*****************************************************************************/
/* Strukturen fuer die Guard-Evaluation                                      */
/*****************************************************************************/

typedef struct bind_struct { 
                             Integer* range;       /* Wertebereich           */
                             Cardinal index;       /* aktueller range-Index  */
                           } binding, *bind_ptr;

/* realisiere den binaeren logischen Operator op */
#define log( op )                                               \
{                                                               \
  if ( EvaluateTree( tree, n->op1, bindings, used_vars )==TT    \
         op                                                     \
       EvaluateTree( tree, n->op2, bindings, used_vars )==TT )  \
    return TT;                                                  \
  else                                                          \
    return FF;                                                  \
}

/* realisiere den Werte-Vergleich op */
#define cmp( op )                                           \
{                                                           \
  if ( EvaluateTree( tree, n->op1, bindings, used_vars )    \
         op                                                 \
       EvaluateTree( tree, n->op2, bindings, used_vars ) )  \
    return TT;                                              \
  else                                                      \
    return FF;                                              \
}

/* realisiere den binaeren Rechenoperator op */
#define ret( op )                                               \
{                                                               \
  return ( EvaluateTree( tree, n->op1, bindings, used_vars )    \
             op                                                 \
           EvaluateTree( tree, n->op2, bindings, used_vars ) ); \
}


/*****************************************************************************/
/* Rechne rekursiv den (Teil-)baum aus, der in tree mit Knoten nr beginnt.   */
/* Falls dabei eine Variable verwendet wird, lies den aktuellen Wert aus den */
/* bindings ab. Trage die Variable in used_vars ein.                         */
/*                                                                           */
/* Rueckgabewert: das Ergebnis des (Teil-)Baums                              */
/*****************************************************************************/

static Integer EvaluateTree( node_ptr tree, Cardinal nr, bind_ptr bindings, 
                             Integer *used_vars )
{
  node_ptr n = tree + nr;
  
/*  printf( "Node %d: %d, %d, %d \n", nr, n->typ, n->op1, n->op2 ); */

  switch ( n->typ )
    {
      case N_HEAD   : break;
      case N_CONST  : return n->op1;
      case N_VAR    : {
                        bind_ptr b = bindings;
                        Cardinal i = 0;
                        Cardinal b_index = n->op1 * MAX_DIM;

                        if ( n->op2 != -1 )
                          b_index += n->op2;

                        while( used_vars[i] != -1 && 
                               used_vars[i] != (Integer) b_index )
                          i++;
                        if( used_vars[i] == -1 )
                          {
                            used_vars[i]   = b_index;
                            used_vars[i+1] = -1;
                          }

                        b += b_index;
                        return b->range[ b->index ];
                      }
      case N_KLAMM  : return EvaluateTree( tree, n->op1, bindings, used_vars );
      case N_OR     : log( || );
      case N_AND    : log( && );
      case N_NOT    : if( EvaluateTree(tree, n->op1, bindings, used_vars)==TT )
                        return FF;
                      else
                        return TT;
      case N_EQ     : cmp( == );
      case N_UNEQ   : cmp( != );
      case N_LT     : cmp( <  );
      case N_GT     : cmp( >  );
      case N_LE     : cmp( <= );
      case N_GE     : cmp( >= );
      case N_PLUS   : ret( +  );
      case N_MIN    : ret( -  );
      case N_MULT   : ret( *  );
      case N_DIV    : ret( /  );
      case N_MOD    : ret( %  );
      case N_REM    : ret( %  );
      case N_SIGN   : return -EvaluateTree( tree, n->op1, bindings, used_vars );
      default: Msg( FAT, "Internal error: Unknown node type!" );
    }   
  return TT;
}


/*****************************************************************************/
/* Liefert False, falls der guard unter den gegebenen bindings unerfuellbar  */
/* ist. Fuer used_vars muss ein gueltiger Speicherbereich uebergeben werden. */
/*****************************************************************************/

static bool CheckVT( node_ptr guard, bind_ptr bindings, Integer *used_vars,
                        Cardinal max_bindings_index )
{
  node_ptr sub_tree, n;
  bool     forward; /* False, falls die bindings geaendert worden sind */
  Integer  ergebnis;
  Cardinal i,j,k;
  Integer  swap;

  sub_tree = guard;
  forward  = True;

  while ( sub_tree->typ != N_HEAD || sub_tree->op1 != 0 )
    {
      n = sub_tree + sub_tree->op1 - 1;
      if( forward )
        *used_vars = -1;
      ergebnis = EvaluateTree(sub_tree, sub_tree->op1-1, bindings, used_vars);

#ifdef DEBUG_MSG
          printf( "index of partial tree %d, result %d: ",(int)(sub_tree-guard),
                  (int) (ergebnis == TT) );
          i = 0;
          while( used_vars[i] != -1 )
            printf( "%d[%d] ", (int) (used_vars[i++]/MAX_DIM), (int) 
              bindings[used_vars[i]].range[bindings[used_vars[i]].index ] );
          printf( "\n" );
#endif

      if( ergebnis != TT )
        {
          forward = False;

          /* sortiere used_vars absteigend mit selection sort */
          if( used_vars[0] != -1 )
            {
              i = 0;
              while( used_vars[i+1] != -1 )
                {
                  j = i;
                  k = i+1;
                  while( used_vars[k] != -1 )
                    {
                      if( used_vars[k] > used_vars[j] )
                        j = k;
                      k++;
                    }
                  if( i != j )
                    {
                      swap = used_vars[i];
                      used_vars[i] = used_vars[j];
                      used_vars[j] = swap;
                    }
                  i++;
                }
            }

          /* suche die naechste Variable, deren Wert noch erhoeht werden kann */
          i = 0;
          while( used_vars[i] != -1 && 
                 bindings[ used_vars[i] ].index >=
                 bindings[ used_vars[i] ].range[0] - 1 )
            {
              i++;
            }

          /* Kann kein Variablenwert mehr erhoeht werden ? */
          if( used_vars[i] == -1 )
            return False; /* keine Loesung moeglich */
          else
            {
              /* alle Variablen mit hoeheren Nummern zuruecksetzen */
              for( j=max_bindings_index; j>(Cardinal)used_vars[i]; j-- )
                bindings[ j ].index = 1; 
              bindings[ used_vars[i] ].index++; /* Wert erhoehen */
            }
        }
      else
        {
          /* War alles OK bzw. wurde sowieso der erste Teilbaum betrachtet ? */
          if ( forward || (sub_tree == guard) )
            sub_tree += sub_tree->op1; /* dann nimm den naechsten Teilbaum */
          else
            sub_tree = guard; /* sonst fang von vorne an */
          forward  = True;
        }
    }

  return True;
}

/*****************************************************************************/
/* Von zwei Wertebereichen wird die Schnittmenge gebildet. Dazu wird ein     */
/* neuer Speicherbereich angefordert, dessen Adresse zurueckgeliefert wird.  */
/*****************************************************************************/

Integer* *allocated_range;     /* hier werden alle Speicherbereiche eingetragen,
                                  die fuer Schnittmengen angelegt worden sind */
Cardinal anz_allocated_ranges; /* Anzahl der Eintraege in allocated_range     */


static Integer* IntersectRanges( Integer *r1, Integer *r2 )
{
  Cardinal max, new, i1, i2, end1, end2;
  Integer *new_r;
  
  end1 = r1[0]; end2 = r2[0];

  if ( end1 > end2 )
    max = end1;
  else
    max = end2;
  new_r = MYmalloc( max * sizeof( Integer ) );

  new = 1; i1 = 1; i2 = 1;
  while( i1 < end1 && i2 < end2 )
    {
      if ( r1[i1] < r2[i2] )
        i1++;
      else if ( r1[i1] > r2[i2] )
        i2++;
      else
        {
          new_r[new] = r1[i1];
          new++; i1++; i2++;
        }
    }
  new_r[0] = new;
  
  new_r = MYrealloc( new_r, new * sizeof( Integer ) );
  allocated_range[ anz_allocated_ranges++ ] = new_r;
  return new_r;
}

/*****************************************************************************/
/* Erzeugt in bindings fuer jede Variable aus VarSet die Bindung an den      */
/* Wertebereich von pl. Wenn pl einen Kreuz-Wertebereich hat, entstehen in   */
/* bindings mehrere Eintraege.                                               */
/*                                                                           */
/* Falls schon vorher fuer eine Variable eine Bindung eingetragen war,       */
/* werden bei einer der alte und der neue Wertebereich geschnitten.          */
/* Die Speicherverwaltung fuer diese Schnittmengen muss gesondert behandelt  */
/* werden und wird deshalb in allocated_range protokolliert.                 */
/*****************************************************************************/

static void Create_Bindings( VarMSet VarSet, bind_ptr bindings, 
                             M_Place pl, M_Transition tr )
{
  Cardinal i;
  bind_ptr b;
  
  while( VarSet->anzahl ) {
    if ( VarSet->nr < V_CONST_OFFSET )
      {
        b = bindings + VarSet->nr * MAX_DIM ;
        
        i=0;
        /* Fuer jeden Teilbereich */
        while( pl->range[i] && i < MAX_DIM )
          {
            if( b[i].range )
              { 
                if ( !M_SingleRangesAreEqual( b[i].range, pl->range[i] ) )
                  {
                    if ( flag & INTERSECT_WARNING )
                      Msg( INF, "[%s] 1-dim variable %s rebound !", actname, 
                           VariableToStr( tr->vars.ComArray[VarSet->nr].var, 
                                          NULL ) );
                    b[i].range = IntersectRanges( b[i].range, pl->range[i] );
                  }
              }
            else
              b[i].range = pl->range[i];
            b[i].index = 1;
            i++;
          }
      }
    else if ( VarSet->nr < V_TUPEL_OFFSET )
      {
        Integer first = 1;

        if ( pl->range[1] )
          Msg( FAT, 
               "[%s] Single constant at an arc from/to a place with tupels !", 
               actname );
        else
          {
            Integer *range = *pl->range;

            while (first <= *range-1 && range[first]< VarSet->nr-V_CONST_NULL)
              first++;

            if ( range[first] != VarSet->nr-V_CONST_NULL )
              Msg( FAT, 
                   "[%s] Constant at an arc is not in the corresponding set !", 
                   actname );
          }
      }
    else
      {
        Var *v = TplToArray( VarSet->nr - V_TUPEL_OFFSET );
        Cardinal i;
        
        i=0;
        /* Fuer jeden Teilbereich */
        while( pl->range[i] && i < MAX_DIM )
          {
            if ( i+1 >= *v )
              Msg( FAT, 
                   "[%s] Tupel has smaller dim than the corresponding place !",
                   actname );

            if ( v[i+1] < V_CONST_OFFSET )
              {
                b = bindings + v[i+1] * MAX_DIM ;
                
                if( b[1].range )
                  Msg( FAT,
                       "[%s] Variable bound to ranges of different dimensions !"
                       , actname);
                
                if( b->range )
                  { 
                    if ( !M_SingleRangesAreEqual( b->range, pl->range[i] ) )
                      {
                        if ( flag & INTERSECT_WARNING )
                          Msg( INF, "[%s] Tupel element %s rebound !", actname,
                               VariableToStr( tr->vars.ComArray[ v[i+1] ].var, 
                                              NULL) );
                        b->range = IntersectRanges( b->range, pl->range[i] );
                      }
                  }
                else
                  b->range = pl->range[i];
                b->index = 1;
              }
            i++;
          }

        if ( i+1<*v )
          Msg( FAT,
               "[%s] Tupel has bigger dimension than the corresponding place !",
               actname );
      }                      
    VarSet++;
  }
}


/*****************************************************************************/
/* Testet, ob der Guard einer Transition nie schalten kann.                  */
/*                                                                           */
/* Dazu werden alle Bindungen, die durch Kanten oder Definitionen gegeben    */
/* sind, herausgesucht. Wenn dadurch alle Variablen, die im Guard vorkommen, */
/* gebunden sind, kann der Guard getestet werden.                            */
/*                                                                           */
/* Falls der Guard unerfuellbar ist, wird 0 geliefert.                       */
/*****************************************************************************/

static int Check_Transition( M_Transition tr )
{
    M_STEdge st;
    M_TSEdge ts;
    bool     is_not_dead = True;
    Cardinal i;
    bind_ptr  b, bindings;
    Integer *range;
    node_ptr n;
    Integer *used_vars;

    /* falls die Option -e fehlt, keinen Test machen */
    if ( !(flag&ELIMINATE_DEAD) ) 
        return 1;

    bindings = MYmalloc( (tr->vars.ComEntries * MAX_DIM) * sizeof( binding ) );

    b = bindings;
    for( i = 0; i < MAX_DIM * tr->vars.ComEntries; i++ )
      b[i].range = NULL;

    /* Bindungen durch  Preset-Kanten heraussuchen */
    st = tr->preset;
    while( st ) {
        Create_Bindings( st->label, bindings, st->place, tr ) ;
        st = st->nextplace;
    }

    /* Bindungen durch Postset-Kanten heraussuchen */
    ts = tr->postset;
    while( ts ) {
        Create_Bindings( ts->label, bindings, ts->place, tr ) ;
        ts = ts->nextplace;
    }

    /* alle Variablen, die mittels Definition gebunden sind, heraussuchen     */
    range = tr->def + 1;
    while ( *range != -1 )
      {
        b = bindings + *range * MAX_DIM;
        if ( range[1] != -1 )
          b += range[1];
          
        if ( b->range ) /* Schon Bindung vorhanden ? */
          {
            if ( !M_SingleRangesAreEqual( b->range, range + 2 ) )
              {
                if ( flag & INTERSECT_WARNING )
                  Msg( INF, "[%s] %s rebound by definition !", actname,
                       VariableToStr( tr->vars.ComArray[*range].var, NULL ) );
                b->range = IntersectRanges( b->range, range + 2 );
              }
          }
        else
          {
            b->range = range + 2;
            b->index = 1;
          }
        range += range[2] + 2;
      }

    /* pruefen, ob alle Variablen gebunden sind */
    n = tr->guard;
    while( n->typ != N_HEAD || n->op1 != 0 )
      {
        if ( n->typ == N_VAR )
          {
            b = bindings + n->op1 * MAX_DIM;

            if ( n->op2 == -1 )
              {
                if ( b[1].range )
                  Msg( FAT, 
                       "[%s] Identifier of vector used without projection !",
                       actname );
              }
            else
              {
                if ( b->range && !b[1].range )
                  Msg( FAT, 
                       "[%s] Projection applied to a non-vector identifier !",
                       actname );
                if ( b->range && !b[n->op2].range )
                  Msg( FAT, 
                       "[%s] Index of projection too high for specific vector !"
                       , actname );
                b += n->op2;
              }
            if ( ! b->range )
              {
                /* Es gibt ungebundene Variablen -> CheckVT nicht moeglich   */
                MYfree( bindings );
                return 1;
              }
          }
        n++;
      }

    used_vars = MYmalloc( (tr->vars.ComEntries+1) * sizeof( Integer ) );
    is_not_dead = CheckVT( tr->guard, bindings, used_vars, 
                           (MAX_DIM * tr->vars.ComEntries) - 1 );
    MYfree( used_vars );

    if ( !is_not_dead )/* Zaehler fuer geloeschte tote Transitionen erhoehen */
        dead_eliminated++; 

    MYfree( bindings );

    return is_not_dead;
}


/*****************************************************************************/
/* Kopiert die Eintraege im ComArray von 'v2' nach 'new'. Die Eintraege im   */
/* ComArray von 'v1' werden durch 'sub1' substituiert und an 'new'           */
/* angehaengt.                                                               */
/* 'subst' gibt an, wie viele Eintraege durch die Substitution wegfallen.    */
/*                                                                           */
/* Falls durch die Substitution verschiedene Variablen zu einer verschmelzen,*/
/* werden ihre real_names ueberprueft. Wenn sie gleich sind, ist alles ok.   */
/* Bei Ungleichheit gibt es zwei Faelle:                                     */
/* - BPN-Namen werden nicht verwendet -> alles in Ordnung !                  */
/* - BPN-Namen werden       verwendet -> Warnung ausgeben !                  */
/*****************************************************************************/

static void ConcatComVars( Variables new, Variables v1, Variables v2, 
                           ComVar *sub1, Cardinal subst )
{
  Cardinal        i, ArrayLength;

  new->ComEntries = v2->ComEntries + v1->ComEntries - subst;
  
  if ( new->ComEntries > MaxAnzComVars )
    Msg( FAT, 
         "[%s] Synchronisation produced a transition with too many ComVars !",
         actname );

  if (new->ComEntries)
  {
    ArrayLength = new->ComEntries * sizeof(ComArrayEntry);
    new->ComArray = MYmalloc( ArrayLength );
    memcpy( new->ComArray, v2->ComArray, v2->ComEntries * sizeof(ComArrayEntry) );
  }
  else
    new->ComArray=NULL;
  for( i = 0; i < v1->ComEntries; i++ )
  {
    if (sub1[i] >= V_CONST_OFFSET) continue;
    if ( sub1[i] >= v2->ComEntries )
      {
        new->ComArray[ sub1[i] ] = v1->ComArray[i];
      }
    else 
      /* es ist schon ein Eintrag vorhanden */
      if ( new->ComArray[ sub1[i] ].var != v1->ComArray[i].var )
        {
          /* die Eintraege und somit die Namen sind verschieden */

          if ( new->ComArray[ sub1[i] ].flag & VAR_MINOR )
            {
              new->ComArray[ sub1[i] ] = v1->ComArray[i];
            }
          else if ( v1->ComArray[i].flag & VAR_MINOR )
            {
            }
          else if ( var_style==VAR_REAL )
            Msg( WAR, 
                 "[%s] Variables %s and %s were united by synchronisation !",
                 actname, VariableToStr( new->ComArray[ sub1[i] ].var, NULL ), 
                 VariableToStr( v1->ComArray[i].var, NULL ) );
        }
    }
    /* todo: unbenutzt gewordene eintraege entfernen */
}

/*****************************************************************************/
/* Unterroutine von 'Synchronisation()' und 'Scoping()'.                     */
/* Durch die Synchronisation der Transitionen 'tr1' und 'tr2' bezueglich der */
/* Aktion 'action' entsteht im Netz 'net' eine neue Transition mit           */
/* dazugehoerigen neuen Kanten. Ueberschreitet das Kantengewicht einer dieser*/
/* neuen Kanten den Parameter 'max', so wird die neue Transition sofort      */
/* wieder geloescht; dadurch wird das Entstehen unendlich grosser Netze      */
/* (bzw. - realistisch gesehen - ein Programmabbruch wegen Speichermangels)  */
/* verhindert.                                                               */
/*                                                                           */
/* 'a1' zeigt auf die Aktion von 'tr1', die sich mit 'a2' von 'tr2'          */
/* synchronisiert.                                                           */
/* 'a1' und 'a2' haben also gleichen Namen und gleiche Aritaet.              */
/*****************************************************************************/

/* die ComVars von t1 werden durch sub1 umbenannt */
/* die ComVars von t2 bleiben erhalten, einige wenige werden jedoch durch sub2 
   umbenannt */
/* Anmerkung: Dadurch werden ueberfluessige Eintraege mitgeschleppt. */
M_Transition M_sub_Sync_Scope(M_Net net, M_Transition tr1, M_Transition tr2,
                                         Action *a1, Action *a2, Cardinal max )
{
    M_STEdge            st1,st2,newst;
    M_TSEdge            ts1,ts2,newts;
    M_Transition        new_tr;
    ComVar              sub1[ MaxAnzComVars ], sub2[ MaxAnzComVars ];
    Cardinal            i, counter, subst = 0;
    Integer             anz_nodes;
    node_ptr            n;
    Integer             *d;
    bool                is_not_dead;
    M_Transition        phtr;
    int			done = 0;

    if ( (flag & CHECK_NAMES) || (flag & INTERSECT_WARNING) )
      ActionToStr( a2->nr, actname );
    
    /* Substitutions-Arrays initialisieren */
    for (i = 0; i < tr1->vars.ComEntries; i++) sub1[i] = UNUSED;
    for (i = 0; i < tr2->vars.ComEntries; i++) sub2[i] = UNUSED;

    /* Umbenennung von Konstanten verhindern und, wo noetig,
	Variablen auf Konstanten mappen */

    /* todo: das hier sieht aus, als koenne man es auch einfacher formulieren */
    while (!done++)
      for (i = 0; i < a1->arity; i++)
	if (a1->com[i] >= V_CONST_OFFSET)
	    if (a2->com[i] >= V_CONST_OFFSET)
	    {
		if (a1->com[i] != a2->com[i]) return NULL;
	    }
	    else
	    {
		if (sub2[a2->com[i]] != UNUSED)
		{
		    if (sub2[a2->com[i]] != a1->com[i]) return NULL;
		}
		else
		{
		    sub2[a2->com[i]] = a1->com[i];
		    done = 0;
		}
	    }
	else
	    if (a2->com[i] >= V_CONST_OFFSET)
	    {
		if (sub1[a1->com[i]] != UNUSED)
		{
		    if (sub1[a1->com[i]] != a2->com[i]) return NULL;
		}
		else
		{
		    sub1[a1->com[i]] = a2->com[i];
		    done = 0;
		}
	    }
	    else
		if (sub1[a1->com[i]] != UNUSED)
		    if (sub2[a2->com[i]] != UNUSED)
		    {
			if (sub1[a1->com[i]] != sub2[a2->com[i]]) return NULL;
		    }
		    else
		    {
			sub2[a2->com[i]] = sub1[a1->com[i]];
			done = 0;
		    }
		else
		    if (sub2[a2->com[i]] != UNUSED)
		    {
			sub1[a1->com[i]] = sub2[a2->com[i]];
			done = 0;
		    }
		    else
			/* kostet 'ne Menge Ueberzeugungsarbeit, dass dies
			   tatsaechlich keine Konstante ist, nicht wahr? */
			;
       /* end of for (i = 0; i < a1->arity; i++) */	
    /* end of while (!done++) */	
	
    /* vorlaeufige Umbenennungen eintragen, d.h. a1 auf a2 abbilden */
    for( i = 0; i < a1->arity; i++ )              /* a1->arity == a2->arity */
    {
    	if (a1->com[i] >= V_CONST_OFFSET) continue;
    	if (sub1[a1->com[i]] >= V_CONST_OFFSET
		&& sub1[a1->com[i]] != UNUSED) continue;
        if ( sub1[ a1->com[i] ] == UNUSED )
            {
                sub1[ a1->com[i] ] = a2->com[i];
                subst++;
            }
        else if ( sub1[ a1->com[i] ] != a2->com[i] && 
                  sub2[ a2->com[i] ] != sub1[ a1->com[i] ] )
            {
                ComVar sub = a2->com[i];
                Cardinal depth = 0; /* Zaehler, um Kreise zu erkennen */
                while( sub2[ sub ] != UNUSED && depth < tr2->vars.ComEntries )
                    {   /* Substitutionskette bis zum Ende verfolgen */
                        sub = sub2[ sub ];
                        depth++;
                    }
                sub2[ sub ] = sub1[ a1->com[i] ];
                /* falls ein Kreis vorlag, wird er irgendwo aufgebrochen */
            }
    }

    /* Substitutionsarrays fuer Direktzugriff modifizieren */
    for( i = 0; i < tr2->vars.ComEntries; i++ )
        if ( sub2[i] == UNUSED ) 
            sub2[i] = i;
        else if (sub2[i] < V_CONST_OFFSET)
            {
                ComVar sub = sub2[i];
                Cardinal depth = 0; /* Zaehler, um Kreise zu erkennen */
                while( sub2[ sub ] != UNUSED && sub2[ sub ] != sub &&
                       depth < tr2->vars.ComEntries )
                    {   /* Substitutionskette bis zum Ende verfolgen */
                        sub = sub2[ sub ];
                        depth++;
                    }
                if( depth == tr2->vars.ComEntries )
                    sub2[sub] = sub; /* Kreis aufbrechen */
                sub2[i] = sub;
            }
    counter = 0;
    for( i = 0; i < tr1->vars.ComEntries; i++ )
        if ( sub1[i] == UNUSED )
          {
            sub1[i] = tr2->vars.ComEntries + counter++;

            if ( flag & CHECK_NAMES )
              {
                Cardinal j=0;

                /* Namen von ComVar i im ComArray von tr2 suchen */
                while ( (j < tr2->vars.ComEntries) && 
                        (tr2->vars.ComArray[j].var!=tr1->vars.ComArray[i].var) )
                  j++;

                /* Kommt der Name von ComVar i auch in tr2 vor ? */
                if ( j != tr2->vars.ComEntries )
                  {
                    if ( tr2->vars.ComArray[j].flag & VAR_RENAME ) 
                      tr1->vars.ComArray[i].flag |= VAR_RENAME;
                    else if ( tr1->vars.ComArray[i].flag & VAR_RENAME )
                      tr2->vars.ComArray[j].flag |= VAR_RENAME;
                    else
                      {
                        Msg( WAR, "[%s] Different renamings of '%s' united !",
                             actname, 
                             VariableToStr( tr2->vars.ComArray[j].var, NULL ) );

                        if ( var_style == VAR_REAL )
                          {
                            /* Auch intern wird das Weglassen der Umbenennung
                               simuliert, d.h. aus den eigentlich disjunkten
                               Vorkommen einer ComVar wird EINE Variable.
                           
                               Dadurch wird die Suche nach mehrfachen Bindungen
                               ermoeglicht. */
                            sub1[i] = j;
                            subst++;
                            counter--;
                          }
                      }
                  }
              }
          }
        else
            if (sub1[i] < V_CONST_OFFSET) sub1[i] = sub2[ sub1[i] ];

    /* neue Transition mit korrektem Label bilden */

    MScopy(mmLeaf,tr1->label,sub1);
    MScopy(mmSub, tr2->label,sub2);
    MSsum(mmLeaf,mmSub);

    MSremove_action_pairs(mmLeaf,a2,sub2,1);
    new_tr=M_DuplicateTrans(net,tr2,&newst,&newts,sub2);
    new_tr->label=MScreate_copy(mmLeaf);
    new_tr->numname=1;
    new_tr->flags |= tr1->flags;

    /* Variablen der neuen Transition setzen */

    ConcatComVars( &(new_tr->vars), &(tr1->vars), &(tr2->vars), 
                   sub1, subst );

    /* links label der neue Transition setzen */
    if (tr1 -> links)
    {
      if (tr2 -> links)
        { /* concat the links labels, with a coma between */
          new_tr -> links =
              (char *) MYmalloc ( (strlen(tr1->links) + strlen(tr2->links) + 1) * sizeof(char) );
          strcpy ((new_tr->links), (tr1->links));
          strcat ((new_tr->links), ",");
          strcat ((new_tr->links), (tr2->links));
        }

      else
        {
          new_tr -> links =
              (char *) MYmalloc ( (strlen(tr1->links) + 1) * sizeof(char) );
          strcpy ((new_tr->links), (tr1->links));
        }
    }
    else if (tr2 -> links)
    {
          new_tr -> links =
              (char *) MYmalloc ( (strlen(tr2->links) + 1) * sizeof(char) );
          strcpy ((new_tr->links), (tr2->links));
    }

    if (tr1->time_label||tr2->time_label)
    {
      new_tr->time_label=(Time_stamp*)MYmalloc(sizeof(Time_stamp));
      if (!tr2->time_label)
      {
	new_tr->time_label->low=tr1->time_label->low;
	new_tr->time_label->high=tr1->time_label->high;
      }
      else if (!tr1->time_label)
      {
	new_tr->time_label->low=tr2->time_label->low;
	new_tr->time_label->high=tr2->time_label->high;
      }
      else
      {
	new_tr->time_label->low=(tr1->time_label->low>tr2->time_label->low)?
	tr1->time_label->low:tr2->time_label->low;
	new_tr->time_label->high=(tr1->time_label->high<tr2->time_label->high)?
	tr1->time_label->high:tr2->time_label->high;
      }
    }
    /* Definitionen der neuen Transition setzen */

    /* Die Wertebereiche werden einzeln eingefuegt, weil es durch die
       Umbennenungen passieren kann, dass mehrere Definitionen 
       zusammenschrumpfen */

    new_tr->def = MYmalloc( ( *(tr1->def) + *(tr2->def) ) * sizeof( Integer ) );
    new_tr->def[0] = 2;  /* 1 fuer Gesamtanzahl + 1 fuer Schlusselement */
    new_tr->def[1] = -1; /* Schluesselement */

    d = tr1->def + 1;
    /* alle Definitionen von tr1->def nacheinander einfuegen */
    while( *d != -1 )
      { if (sub1[*d] >= V_CONST_OFFSET)
	{    /* falls Variable auf Konstante abgebildet wurde, pruefen,
		ob die Konstante innerhalb des erlaubten Bereiches liegt */
	    for (i = *(d+2)-1; i > 0; i--)
		if (d[i+2] == sub1[*d]) break;
	    if (!i)
	    {
                M_RemoveTransition(net,new_tr,False);
                return NULL;
	    }
	}
	else
            AddSingleRangeToDef( &(new_tr->def), sub1[ *d ], d[1], d+2, False );
        d += d[2] + 2;
      }

    d = tr2->def + 1;
    /* alle Definitionen von tr2->def nacheinander einfuegen */
    while( *d != -1 )
      { if (sub2[*d] >= V_CONST_OFFSET)
	{   /* dito Konstante->Variable */
	    for (i = *(d+2)-1; i > 0; i--)
		if (d[i+2] == sub2[*d]) break;
	    if (!i)
	    {
                M_RemoveTransition(net,new_tr,False);
                return NULL;
	    }
	}
	else
            AddSingleRangeToDef( &(new_tr->def), sub2[ *d ], d[1], d+2, False );
        d += d[2] + 2;
      }

    new_tr->def = MYrealloc( new_tr->def, *(new_tr->def) * sizeof( Integer ) );
    

    /* Guard der neuen Transition setzen */

    anz_nodes = tr1->guard->op2 + tr2->guard->op2 - 1;
    new_tr->guard = MYmalloc( anz_nodes * sizeof( node ) );
    memcpy( new_tr->guard, tr2->guard, tr2->guard->op2 * sizeof( node ) );
    memcpy( new_tr->guard + tr2->guard->op2 - 1, tr1->guard, 
            tr1->guard->op2 * sizeof( node ) );
    new_tr->guard->op2 = anz_nodes;

    n = new_tr->guard;
    while( n->typ != N_HEAD || n->op1 != 0 )
      {
        if ( n->typ == N_VAR )
	{
	  int i; /* gnaa... *musste* n->op1 shortint sein? */
          if ( n - new_tr->guard < tr2->guard->op2 - 1 )
            i = sub2[ n->op1 ];
          else
            i = sub1[ n->op1 ];
	  if (i >= V_CONST_OFFSET)
	  {
	    n->typ = N_CONST;
	    i -= V_CONST_NULL;
	    n->op2 = 0;
	  }
	  n->op1 = i;
	}
        n++;
      } 

    /* Die durch Duplikation von Transition 'tr2' neu entstandene Transition */
    /* um den Preset von Transition 'tr1' ergaenzen.                         */

    st1=tr1->preset; st2=new_tr->preset;
    while (st1) {
        if (st1->place->postset->trans!=new_tr) { /*Weitere Kante hinzufuegen,*/
            (newst=M_allocSTEdge())->prevtrans=NULL;
            newst->trans=new_tr;
            newst->place=st1->place;
            newst->label=V_MScreate_copy(st1->label, sub1);
            newst->flags=st1->flags;
            newst->nexttrans=st1->place->postset;
            st1->place->postset=newst->nexttrans->prevtrans=newst;
            if (st2) {
              while (st2->nextplace &&
                     st2->place->numname<st1->place->numname)   /* aber wo ? */
                st2=st2->nextplace;
              if (st2->place->numname>st1->place->numname) {    /* Vor st2 ! */
                if ((newst->prevplace=st2->prevplace)!=NULL)
                  newst->prevplace->nextplace=newst;
                else
                  new_tr->preset=newst;
                (newst->nextplace=st2)->prevplace=newst;
              } else {                                       /* Hinter st2 ! */
                (st2->nextplace=newst)->prevplace=st2;
                newst->nextplace=NULL;
              }
            } else {
              newst->prevplace=newst->nextplace=NULL;
              new_tr->preset=newst;
            }
            st2=newst;
        } else {                                         /* Label dazufuegen */
            V_MScopy( V_mmLeaf, st1->label, sub1 );
            V_MSsum( V_mmLeaf, st1->place->postset->label );
            VSremove( st1->place->postset->label );
            st1->place->postset->label = V_MScreate_copy( V_mmLeaf, NULL );
            if ( V_MScardinality( st1->place->postset->label ) > max ) {
                M_RemoveTransition(net,new_tr,False);
                return NULL;
            }
        }
        st1=st1->nextplace;
    }

    /* Ergaenzung um den Postset von Transition 'tr1' */

    ts1=tr1->postset; ts2=new_tr->postset;
    while (ts1) {
        if (ts1->place->preset->trans!=new_tr) {  /*Weitere Kante hinzufuegen,*/
            (newts=M_allocTSEdge())->prevtrans=NULL;
            newts->trans=new_tr;
            newts->place=ts1->place;
            newts->label=V_MScreate_copy(ts1->label, sub1);
            newts->flags=ts1->flags;
            newts->nexttrans=ts1->place->preset;
            ts1->place->preset=newts->nexttrans->prevtrans=newts;
            if (ts2) {
              while (ts2->nextplace &&
                     ts2->place->numname<ts1->place->numname)   /* aber wo ? */
                ts2=ts2->nextplace;
              if (ts2->place->numname>ts1->place->numname) {    /* Vor ts2 ! */
                if ((newts->prevplace=ts2->prevplace)!=NULL)
                  newts->prevplace->nextplace=newts;
                else
                  new_tr->postset=newts;
                (newts->nextplace=ts2)->prevplace=newts;
              } else {                                       /* Hinter ts2 ! */
                (ts2->nextplace=newts)->prevplace=ts2;
                newts->nextplace=NULL;
              }
            } else {
              newts->prevplace=newts->nextplace=NULL;
              new_tr->postset=newts;
            }
            ts2=newts;
        } else {
            V_MScopy( V_mmLeaf, ts1->label, sub1 );
            V_MSsum( V_mmLeaf, ts1->place->preset->label );
            VSremove( ts1->place->preset->label );
            ts1->place->preset->label = V_MScreate_copy( V_mmLeaf, NULL );
            if ( V_MScardinality( ts1->place->preset->label ) > max ) {
                M_RemoveTransition(net,new_tr,False);
                return NULL;
            }
        }
        ts1=ts1->nextplace;
    }

    /* Sofort testen, ob die neu entstandene Transition schalten kann ! */
    
    /* Speicherprotokoll fuer Schnittmengen initialisieren */
    allocated_range = MYmalloc( (new_tr->vars.ComEntries * MAX_DIM) 
                                * sizeof( Integer* ) );
    anz_allocated_ranges = 0;

    is_not_dead = Check_Transition( new_tr );

    /* Schnittmengen freigeben */
    for( i=0; i<anz_allocated_ranges; i++ )
      MYfree( allocated_range[i] );
    MYfree( allocated_range );

    if ( !is_not_dead ) {
        M_RemoveTransition(net,new_tr,False);
        return NULL;
    }
 
    if ( flag & MNET_SYNCHRONIZE )
    {
      /* Sofort testen, ob die neu entstandene Transition ein Duplikat ist ! */

      M_Transition tr;
      if (new_tr->preset && new_tr->preset->place->postset)
	newst=new_tr->preset->place->postset->nexttrans; 
      else
	newst=NULL;
        /* Die erste ist new_tr !*/
      while (newst) {
        tr=newst->trans;
        if ( Tr_are_equal( new_tr, tr ) ) {
          M_RemoveTransition(net,new_tr,False);   /* Dann loeschen ! */
          return NULL;
        }
        newst=newst->nexttrans;
      }
    }

    /* Referenzen der neuen Transition setzen */

    new_tr->RefEntries = tr1->RefEntries + tr2->RefEntries;
    if ( new_tr->RefEntries ) {
        new_tr->ref = MYmalloc( new_tr->RefEntries * sizeof( Ref ) );
	if (tr2->RefEntries)
          memcpy( new_tr->ref, tr2->ref, tr2->RefEntries * sizeof( Ref ) );
	if (tr1->RefEntries)
	  memcpy( new_tr->ref + tr2->RefEntries, tr1->ref,
                  tr1->RefEntries * sizeof( Ref ) );
    }

    /* Blockverweise zusammensetzen */

    new_tr->BlockEntries = tr1->BlockEntries + tr2->BlockEntries;
    if (new_tr->BlockEntries) {
      new_tr->blocks = MYmalloc(new_tr->BlockEntries * sizeof(BlockNr));
      memcpy(new_tr->blocks, tr2->blocks, tr2->BlockEntries * sizeof(BlockNr));
      memcpy(new_tr->blocks + tr2->BlockEntries, tr1->blocks,
        tr1->BlockEntries * sizeof(BlockNr));
    }  /* if Verweise zum Kopieren vorhanden */
    else
      new_tr->blocks = NULL;  /* NIX */

    /* Markierung der neuen Transition nur, falls vorher eine der */
    /* beiden Transitionen damit markiert war, denn nur bei       */
    /* Channel 0 ist das wichtig                                  */

    if (((tr1->flags & TR_CHAN0_ACCESS) != 0)
       || ((tr2->flags & TR_CHAN0_ACCESS) != 0)) {
      new_tr->flags |= TR_CHAN0_ACCESS;
    }

    /* if die neue Transition hat etwas mit Kanaelen 0 zu tun hat */

    if ((new_tr->flags & TR_CHAN0_ACCESS) != 0) {

      /* Fuer Transitionen im Scope ein Phantom erzeugen */ 
      /* falls sie noch kein Phantom haben               */
      if ( !tr2->PhantomEntries && (tr2->flags & TR_SCOPE) )
        {
          phtr = M_Copy2PTransition(net, tr2);
          M_AddPTransPhantom(tr2, phtr);
          M_AddPTransPhantom(phtr, tr2);
        }
      if ( !tr1->PhantomEntries && (tr1->flags & TR_SCOPE) )
        {
          phtr = M_Copy2PTransition(net, tr1);
          M_AddPTransPhantom(tr1, phtr);
          M_AddPTransPhantom(phtr, tr1);
        }

      /* SYNC-Transition in der Mitte zwischen ihren Phantomen plazieren */
      if ( (tr1->flags & TR_SCOPE) && (tr2->flags & TR_SCOPE) )
        {
          new_tr->x_pos  = (tr1->x_pos + tr2->x_pos) >> 1;
          new_tr->y_pos  = (tr1->y_pos + tr2->y_pos) >> 1;
        }
    }
    else /* ssch */
    {
	if ((tr1->flags & TR_SCOPE) && (tr2->flags & TR_SCOPE))
	{
          new_tr->x_pos  = (tr1->x_pos + tr2->x_pos) >> 1;
          new_tr->y_pos  = (tr1->y_pos + tr2->y_pos) >> 1;
	}
	else if (tr1->flags & TR_SCOPE)
	{
	  new_tr->x_pos = tr1->x_pos;
	  new_tr->y_pos = tr1->y_pos;
	}
    }

    /* einen Teil der ueberlappenden Transitionen im Scope vermeiden */

    tr2->x_pos += 15;
    tr2->y_pos += 15;
    if ( tr1->flags & TR_REF_REC )
      {
        new_tr->y_pos -= 30;
        new_tr->x_pos -= 30;
        new_tr->flags &= ~TR_REF_REC;
      }
    
    /* Phantomverweise fuer die neue Transition zusammensetzen und gleich- */
    /* zeitig Rueckwaertsverweise in den betroffenen Phantomen eintragen   */

    new_tr->PhantomEntries = tr1->PhantomEntries + tr2->PhantomEntries;
    if (new_tr->PhantomEntries) {
      new_tr->phantoms = MYmalloc(new_tr->PhantomEntries*sizeof(M_Transition));
      for( i=0; i<tr1->PhantomEntries; i++ )
        {
          new_tr->phantoms[i]=tr1->phantoms[i];
          M_AddPTransPhantom(tr1->phantoms[i], new_tr);
        }
      for( i=0; i<tr2->PhantomEntries; i++ )
        {
          new_tr->phantoms[i+tr1->PhantomEntries]=tr2->phantoms[i];
          M_AddPTransPhantom(tr2->phantoms[i], new_tr);
        }
    }

    new_tr->meaning   = ConcatStrings( tr1->meaning,   tr2->meaning   );

  return new_tr;
}  /* M_sub_Sync_Scope */


/****************************************************************************/
/* Befindet sich an der Position 'x,y' bzw. in dessen vertikaler Umgebung   */
/* bereits eine Stelle oder Transition ?                                    */
/****************************************************************************/

#ifdef EXACT_POSITIONING
static int M_occupied(M_Net net,int x,int y)
{
    register M_Transition tr;
    register M_Place pl;
    register int erg,yp1,ym1;

    yp1=y+RASTER;
    ym1=y-RASTER;
    erg=0;
    tr=net->Transitions;
    while (tr) {
        if (tr->x_pos==x && tr->y_pos>=ym1 && tr->y_pos<=yp1)
            if (tr->y_pos==y)
                erg|=1;
            else if (tr->y_pos==ym1)
                erg|=2;
            else
                erg|=4;
        tr=tr->nexttrans;
    }
    pl=net->Places;
    while (pl) {
        if (pl->x_pos==x && pl->y_pos>=ym1 && pl->y_pos<=yp1)
            if (pl->y_pos==y)
                erg|=1;
            else if (pl->y_pos==ym1)
                erg|=2;
            else
                erg|=4;
        pl=pl->nextplace;
    }
    return erg;
}
#endif

/******************************************************************/
/* Ruft 'M_sub_Sync_Scope' auf und positioniert die eventuell neu */
/* entstandene Transition.                                        */
/******************************************************************/

static void M_sub_Sync( M_Net net, M_Transition tr1, M_Transition tr2,
                                   Action *a1, Action *a2, Cardinal max )
{
    register M_Transition tr3;

    if ((tr3=M_sub_Sync_Scope(net,tr1,tr2,a1,a2,max))!=NULL) {
#ifdef EXACT_POSITIONING
        int x,y,erg;
      
        x=(tr1->x_pos-tr2->x_pos)/2+tr2->x_pos;
        y=(tr1->y_pos>tr2->y_pos)?tr1->y_pos:tr2->y_pos;
        M_left_and_right(net,x,x-RASTER,2*RASTER);
        if ((erg=M_occupied(net,x+RASTER,y))&1) {
            x -= RASTER;
        } else if (erg&2) {
            x -= RASTER;
            y -= RASTER;
        } else if (erg&4) {
            x -= RASTER;
            y += RASTER;
        }
        tr3->x_pos=x;
        tr3->y_pos=y;
#endif
    }
}


/*****************************************************************************/
/* Anwendung der Synchronisations-Operation bezueglich der Aktion 'action'   */
/* auf das Netz 'net', welches auch den Rueckgabewert darstellt.             */
/* Diese erste Funktion sucht systematisch nach Transitions-Paaren, die sich */
/* synchronisieren koennen, und reicht diese (zusammen mit einigen anderen   */
/* Parametern) an die Funktion 'M_sub_Sync_Scope()' weiter.                  */
/* Der Parameter 'max' spielt hier keine Rolle, er wird unveraendert an die  */
/* Funktion 'M_sub_Sync_Scope()' weitergegeben und auch dort erlaeutert.     */
/*****************************************************************************/

M_Net M_Synchronisation(M_Net net, MultiSet mm, Cardinal max)
{
    M_Transition        tr1,tr2;
    M_Place             pl;
    Cardinal32          i;
    Action              *a1, *a2;

    i=0; pl=net->Places;    /* Durchnummerierung der Stellen notwendig fuer  */
    while (pl) {            /* das Funktionieren von 'sub_Sync_Scope()'. */
        pl->numname=++i;
        pl=pl->nextplace;
    }
    while (mm->anzahl) {
    /* while there are still occurrences of the action pointed by mm
       in the net */
        tr1=net->LastTransition;
        while (tr1) {
        /* for all transition in the net */
             if ( (a1 = MSis_in( mm->nr, mm->arity, tr1->label )) ) {
             /* if the label contains the action pointed by mm,
                let a1 be this action */
                tr2=tr1;
                while (tr2) {
                    if ( (a2 = MSconj_is_in( mm->nr, mm->arity, tr2->label )) )
		    { /* if the conjugated of a1 is on this label */
			if (a2->nr & 1)
                            M_sub_Sync(net,tr2,tr1,a2,a1,max);
			else
                            M_sub_Sync(net,tr1,tr2,a1,a2,max);
		    }
                    tr2=tr2->prevtrans;
                }
            }
            /* else if the label contains the conjugated action of mm* */
	    else if ( (a1 = MSconj_is_in( mm->nr, mm->arity, tr1->label )) ) {
                tr2=tr1;
                while (tr2) {
                    if ( (a2 = MSis_in( mm->nr, mm->arity, tr2->label )) )
		    { /* if the action pointed by mm is in this label */
			if (a2->nr & 1)
                            M_sub_Sync(net,tr2,tr1,a2,a1,max);
			else
                            M_sub_Sync(net,tr1,tr2,a1,a2,max);
		    }
                    tr2=tr2->prevtrans;
                }
            }
            tr1=tr1->prevtrans;
        }
        mm++;
    }
    return net;
}


/*****************************************************************************/
/* Anwendung der Restriktions-Operation bezueglich der Aktion 'action'       */
/* auf das Netz 'net', welches auch den Rueckgabewert darstellt.             */
/*****************************************************************************/

M_Net M_Restriction(M_Net net, MultiSet mm)
{
  M_Transition tr;

  tr = net->LastTransition;
  while (tr) {
    if (MSis_disjunct(mm, tr->label)) {
      tr = tr->prevtrans;
    }
    else {
      tr = M_RemoveTransition(net, tr, False);
    }
  }
  return net;
}  /* M_Restriction */

  /***********************/
  /* PlaceMultiplication */
  /***********************/

void M_PlaceMultiplication(M_Net net1, M_Net net2, PLabel entry, PLabel exit)
{
    register M_Place    pl1,pl2,new_pl;
             M_STEdge   stlast,pstlast;
             M_TSEdge   tslast,ptslast;
             Cardinal   x_dim;
             int        x_pos;

    x_dim=0;
    M_PreparePlaces(net2,entry,entry);
    pl1=net1->LastPlace;
    while (pl1 && pl1->numname) {
        if (pl1->label & exit) {  /* pl1 laeuft durch die net1-exit-Stellen  */
            pl2=net2->LastPlace;  /* pl2 laeuft durch die net2-entry-Stellen */
            while (pl2) {
                x_dim+=2*RASTER;
                new_pl=M_DuplicatePlace(net1,pl1,&stlast,&tslast,
                                        &pstlast,&ptslast);
                if (pl1->label!=pl2->label)
                  new_pl->label=EMPTY;

                if (new_pl->postset!=NULL) {
                    if ((stlast->nexttrans
                         =M_CopyPlacePostset(pl2,new_pl,&stdummy,False))!=NULL)
                        stlast->nexttrans->prevtrans=stlast;
                } else
                    new_pl->postset
                    =M_CopyPlacePostset(pl2,new_pl,&stdummy,False);
                if (new_pl->ppostset!=NULL) {
                    if ((pstlast->nexttrans
                         =M_CopyPlacePostset(pl2,new_pl,&stdummy,True))!=NULL)
                        pstlast->nexttrans->prevtrans=pstlast;
                } else 
                    new_pl->ppostset
                      =M_CopyPlacePostset(pl2,new_pl,&stdummy,True);

                if ( new_pl->preset!=NULL) {
                    if ((tslast->nexttrans
                         =M_CopyPlacePreset(pl2,new_pl,&tsdummy,False))!=NULL)
                        tslast->nexttrans->prevtrans=tslast;
                } else
                    new_pl->preset
                    =M_CopyPlacePreset(pl2,new_pl,&tsdummy,False);
                if (new_pl->ppreset!=NULL) {
                    if ((ptslast->nexttrans
                         =M_CopyPlacePreset(pl2,new_pl,&tsdummy,True))!=NULL)
                        ptslast->nexttrans->prevtrans=ptslast;
                } else
                  new_pl->ppreset
                    =M_CopyPlacePreset(pl2,new_pl,&tsdummy,True);

                new_pl->flags |= pl2->flags;
                M_SumPlaceRefs( new_pl, pl2 );
                pl2=pl2->prevplace;
            }
            pl1=M_RemovePlace(net1,pl1);
        } else
             pl1=pl1->prevplace;
    }

    x_dim -= RASTER;
    x_pos=(int)(net1->x_dim-(x_dim))/2+net1->x_pos;
    if (x_dim>net1->x_dim) {
        net1->x_dim=x_dim;
        net1->x_pos=x_pos;
    }
    while (pl1) {
        x_pos=(pl1->x_pos=x_pos)+2*RASTER;
        pl1->numname=1;
        pl1=pl1->prevplace;
    }
    M_RestoreKillPlaces(net2,entry);
}  /* M_PlaceMultiplication */

  /*************/
  /* Iteration */
  /*************/

M_Net M_Iteration( alt_set_type alt, M_Net enter )
{
    if ( !alt.repeat_net ) {
        /* Sonderfall, damit die Optik von Pseudo-IFs (Schleifen ohne
           repeat_zweig: Der Exit wirkt als IF.) besser aussieht      */
        M_Remove_net( &alt.repeat_net );

	if (!enter)
	  return alt.exit_net;
	else
	  return M_Sequence( enter, alt.exit_net );
    }

    if ( !enter ) { 
      /* falls kein enter-Netz uebergeben wurde, wird ein leeres erzeugt */
      VariableStruct vars;
      enter = M_CreateBasicBox( MScreate_empty_set(), NULL, NULL,
                                Create_empty_Variables( &vars, 1 ), 0 );
    }

    return (flag&BIG_LOOP) ? BigIteration(alt,enter) :
                (flag&MEDIUM_LOOP) ? MediumIteration(alt,enter):
                     SmallIteration(alt,enter);
} /* M_Iteration */

  /***********/
  /* Auswahl */
  /***********/

M_Net M_Choice( M_Net net1, M_Net net2 )
{
    if ( !net1 )
      return net2;
    if ( !net2 )
      return net1;

    M_juxtapose(net1,net2);
    M_PlaceMultiplication(net1,net2,ENTRY,ENTRY);
    M_PlaceMultiplication(net1,net2, EXIT, EXIT);
    return M_DisjointNetUnion(net1,net2);
}  /* M_Choice */

  /********************/
  /* Nebenlaeufigkeit */
  /********************/
M_Net M_ConcurrentComposition( M_Net net1, M_Net net2 )
{
    M_juxtapose(net1,net2);
    net2->x_pos=net1->x_pos;
    net2->y_pos=net1->y_pos;
    net2->x_dim=net1->x_dim;
    net2->y_dim=net1->y_dim;

    return M_DisjointNetUnion(net2,net1);
}  /* M_ConcurrentComposition */

  /***********/
  /* Sequenz */
  /***********/

M_Net M_Sequence(M_Net net1, M_Net net2)
{
    M_move_hv(net2,net1->x_pos-net2->x_pos
                   +(int)(net1->x_dim-net2->x_dim)/2,
                   net1->y_pos-net2->y_pos+net1->y_dim-RASTER);
    if (net2->x_dim>net1->x_dim) {
        net1->x_dim=net2->x_dim;
        net1->x_pos=net2->x_pos;
    }
    net1->y_dim+=net2->y_dim-RASTER;

    M_PlaceMultiplication(net1,net2,ENTRY,EXIT);
    return M_DisjointNetUnion(net1,net2);
}  /* M_Sequence */

  /**************/
  /* Replacement */
  /**************/

M_Net M_Replacement (M_Net net, M_Transition tr, M_Net box)
{
  M_STEdge st;
  M_TSEdge ts;

  M_move_hv(box,tr->x_pos-box->x_pos,
            tr->y_pos-box->y_pos-(box->y_dim-RASTER)/2);
  M_up_and_down(net,tr->y_pos,tr->y_pos,box->y_dim-5*RASTER);
  net->x_dim = tr->x_pos-net->x_pos;

  st=tr->preset;
  while (st) {
    st->place->label=INTERN1;
    st=st->nextplace;
  }
  
  ts=tr->postset;
  while (ts) {
    ts->place->label=INTERN2;
    ts=ts->nextplace;
  }
  
  M_RemoveTransition(net,tr,False);

  M_PlaceMultiplication(box,net,INTERN1,ENTRY);
  M_PlaceMultiplication(box,net,INTERN2, EXIT);
  box->x_dim += net->x_dim;
  box->y_dim  = net->y_dim;
  box->x_pos  = net->x_pos;
  box->y_pos  = net->y_pos;
  return M_DisjointNetUnion(box,net);
}  /* M_Replacement */


/*****************************************************************************/
/* tr->numname & 1 : ' akt' befindet sich in der Beschriftung                */
/* tr->numname & 2 : '^akt' befindet sich in der Beschriftung                */
/*****************************************************************************/

static void set_numname(M_Transition tr,Action *akt)
{
    Action_nr nr;

    tr->numname = 0;
    nr = ( akt->nr & ~(Action_nr)1 );

    if ( MSis_in( nr, akt->arity , tr->label) )
      tr->numname |= 1;
    if ( MSconj_is_in( nr, akt->arity, tr->label) )
      tr->numname |= 2;
    if ( !tr->numname ) tr->numname = 4;
}


/*****************************************************************************/
/* Erzeugung einer Basisbox; die Transition erhaelt als Beschriftung eine    */
/* Kopie des uebergebenen Labels 'label', welches eine Multimenge oder       */
/* eine als Multimenge verkleidete Variable sein kann.                       */
/* Rueckgabewert ist das erzeugte Netz.                                      */
/*****************************************************************************/

M_Net M_CreateBasicBox( MultiSet label, node_ptr guard, Integer *def,
                        Variables v_ptr, RefNr ref )
{
    register M_Place      pl;
    register M_Transition tr;
    register M_STEdge     st;
    register M_TSEdge     ts;
    register M_Net        box;

    (box=M_empty_net())->NumComponents=1;
    box->NumPlaces=2;
    box->NumTransitions=1;
    box->NumPTransitions = 0;  /* Phantomtransitionen */
    box->PTransitions = box->LastPTransition = NULL;
    box->x_pos=box->y_pos=0;
    box->x_dim=1*RASTER;
    box->y_dim=5*RASTER;
    M_InitBlockInNet(box);  /* Bloecke vorbereiten */
    (pl=box->Places=M_allocPlace())->prevplace=NULL;
    pl->numname=1;
    pl->component=1;
    pl->label=ENTRY;
    M_CopySingleRange( pl->range, macro_id_set );
    pl->range[1]=NULL;
    if ( ref ) {
        pl->RefEntries = 1;
        pl->ref = MYmalloc( sizeof( Ref ) );
        (pl->ref)->nr   = ref;
        (pl->ref)->flag = PRE_REF;
    } else
        pl->RefEntries = 0;
    pl->meaning=NULL;
    pl->init_marking=NULL;
    pl->curr_marking=NULL;
    pl->BlockEntries = 0;
    pl->blocks = NULL;
    pl->flags=0;
    pl->ppreset = NULL;  /* Phantomkanten */
    pl->ppostset = NULL;
    pl->preset=NULL;
    pl->x_pos=pl->y_pos=0;
    tr=box->Transitions=box->LastTransition=M_allocTrans();
    tr->prevtrans=tr->nexttrans=NULL;
    tr->numname=1;
    tr->label=label;
    if ( def )
      tr->def = def;
    else
      {
        tr->def=MYmalloc( 2*sizeof( Integer ) );
        tr->def[0] = 2;
        tr->def[1] = -1;
      }
    if ( guard )
      tr->guard = guard;
    else
      {
        tr->guard = MYmalloc( sizeof( node ) );
        tr->guard->typ = N_HEAD;
        tr->guard->op1 = 0;
        tr->guard->op2 = 1;
      }
    if ( ref ) {
        tr->RefEntries = 1;
        tr->ref = MYmalloc( sizeof( Ref ) );
        (tr->ref)->nr   = ref;
        (tr->ref)->flag = 0;
    } else
        tr->RefEntries = 0;
    tr->meaning=NULL;
    tr->BlockEntries = 0;
    tr->blocks = NULL;
      /* Phantomverweise initialisieren */
    tr->PhantomEntries = 0;
    tr->phantoms = NULL;
    tr->links=NULL;
    tr->time_label=NULL;
    tr->flags=0;
    tr->vars.ComEntries=v_ptr->ComEntries;
    tr->vars.ComArray  =v_ptr->ComArray;
    tr->x_pos=0;
    tr->y_pos=2*RASTER;
    st=pl->postset=tr->preset=M_allocSTEdge();
    st->prevplace=st->nextplace=st->prevtrans=st->nexttrans=NULL;
    st->place=pl;
    st->trans=tr;
    st->label=V_MScreate_singleton( 0 );
    st->flags=0;
    ( pl=( (pl->nextplace=box->LastPlace=M_allocPlace())
           ->prevplace=pl ) ->nextplace ) ->nextplace=NULL;
    pl->numname=1;
    pl->component=1;
    pl->postset=NULL;
    pl->label=EXIT;
    M_CopySingleRange( pl->range, macro_id_set );
    pl->range[1]=NULL;
    pl->RefEntries = 0;
    if ( ref ) {
        pl->RefEntries = 1;
        pl->ref = MYmalloc( sizeof( Ref ) );
        (pl->ref)->nr   = ref;
        (pl->ref)->flag = POST_REF;
    } else
        pl->RefEntries = 0;
    pl->meaning=NULL;
    pl->init_marking=NULL;
    pl->curr_marking=NULL;
    pl->BlockEntries=0;
    pl->blocks = NULL;
    pl->flags=0;
    pl->x_pos=0*RASTER;
    pl->y_pos=4*RASTER;
    pl->ppreset = NULL;  /* Phantomkanten */
    pl->ppostset = NULL;
    ts=pl->preset=tr->postset=M_allocTSEdge();
    ts->prevplace=ts->nextplace=ts->prevtrans=ts->nexttrans=NULL;
    ts->place=pl;
    ts->trans=tr;
    ts->label=V_MScreate_singleton( 0 );
    ts->flags=0;
    return box;
}

/*****************************************************************************/
/* Spezialroutine aehnlich 'M_RemoveTransition()' fuer die nachfolgende      */
/* Funktion 'Scoping()'.                                                     */
/* Entfernen der Transition 'tr' aus dem Netz 'net' zusammen mit allen       */
/* Kanten, die in/aus diese/r Transition fuehren.                            */
/* Ausserden werden alle Phantom-Transitionen entfernt, auf die 'tr' zeigt   */
/* und die fuer keine andere Transition mehr benoetigt werden.               */
/* Rueckgabewert ist 'tr->prevtrans'.                                        */
/*****************************************************************************/

static M_Transition RemoveTransition(M_Net net,M_Transition tr)
{
    register M_STEdge     st,stdel;
    register M_TSEdge     ts,tsdel;
             M_Transition tr_help;
             Cardinal     i,j;

    if (net && tr) {
        ts=tr->postset;
        while (ts) {
            if (ts->prevtrans)
                ts->prevtrans->nexttrans=ts->nexttrans;
            else
                ts->place->preset=ts->nexttrans;
            if (ts->nexttrans)
                ts->nexttrans->prevtrans=ts->prevtrans;
            ts=(tsdel=ts)->nextplace;
            M_removeTSEdge(tsdel);
        }
        st=tr->preset;
        while (st) {
            if (st->prevtrans)
                st->prevtrans->nexttrans=st->nexttrans;
            else
                st->place->postset=st->nexttrans;
            if (st->nexttrans)
                st->nexttrans->prevtrans=st->prevtrans;
            st=(stdel=st)->nextplace;
            M_removeSTEdge(stdel);
        }
        if (tr->prevtrans)
            tr->prevtrans->nexttrans=tr->nexttrans;
        else
            net->Transitions=tr->nexttrans;
        if (tr->nexttrans)
            tr->nexttrans->prevtrans=tr->prevtrans;
        else
            net->LastTransition=tr->prevtrans;
        MSremove(tr->label);
        MYfree( tr->def );
        MYfree( tr->guard );
        if ( tr->RefEntries )
          MYfree( tr->ref );
        if ( tr->meaning )
          MYfree( tr->meaning );
        if (tr->BlockEntries) {
          MYfree(tr->blocks);
        }
          /* Phantomverweise killen */
        if (tr->PhantomEntries) {
          for ( i=0; i<tr->PhantomEntries; i++ ) {
            tr_help=tr->phantoms[i];
            if (tr_help->PhantomEntries>1) {
              j=0;
              while( tr_help->phantoms[j] != tr )
                j++;
              tr_help->phantoms[j]=tr_help->phantoms[--tr_help->PhantomEntries];
            } else
              M_RemoveTransition(net, tr_help, True);
          }
          MYfree(tr->phantoms);
        }
        MYfree( tr->vars.ComArray );
        net->NumTransitions--;
        tr_help=tr->prevtrans; 
        M_removeTrans(tr);
        return tr_help;
    } else
      return NULL;
}


/***********************************/
/* Anwendung der Scoping-Operation */
/* (analog zu 'Synchronisation()') */
/***********************************/

M_Net M_Scoping(M_Net net,MultiSet mm,Cardinal max)
{
    register M_Transition tr1,tr2,tr3;
             M_Place      pl;
             Cardinal32   i;
             Action       *a1, *a2;

    i=0; pl=net->Places;    /* Durchnummerierung der Stellen notwendig fuer  */
    while (pl) {            /* das Funktionieren von 'sub_Sync_Scope()'. */
        pl->numname=++i;
        pl=pl->nextplace;
    }
    while (mm->anzahl) {

        /* fuer alle Transitionen numname setzen */
        tr1=tr2=net->LastTransition;
        while (tr2) {
            set_numname(tr2,mm);
            tr2=tr2->prevtrans;
        }

        i=5; /* Bit 0 == 1: zwischen FirstTransition und tr1 koennte noch eine
                            Transition mit *mm            enthalten sein */
             /* Bit 2 == 1: zwischen FirstTransition und tr1 koennte noch eine
                            Transition mit *mm konjugiert enthalten sein */
        while (tr1) {
            if (tr1->numname&2) {
                if (i&1) {
                    i&=~2;
                    tr2=tr1;
                    while (tr2) {
                        if (tr2->numname&1) {
                            i|=2;
                            a1 = MSis_in( mm->nr | 1, mm->arity, tr1->label );
                            a2 = MSconj_is_in( mm->nr | 1, mm->arity, tr2->label );
                            if ((tr3=M_sub_Sync_Scope(net,tr1,tr2,a1,a2,max))!=NULL) { 
                                set_numname(tr3,mm);
                            }
                        }
                        tr2=tr2->prevtrans;
                    }
                    if (!(i&2)) /* Keine mit *mm            gefunden ? */
                        i&=~1;
                }
                tr1=RemoveTransition(net,tr1);
            } else if (tr1->numname&1) {
                if (i&4) {
                    i&=~8;
                    tr2=tr1;
                    while (tr2) {
                        if (tr2->numname&2) {
                            i|=8;
                            a1 = MSconj_is_in( mm->nr | 1, mm->arity, tr1->label );
                            a2 = MSis_in( mm->nr | 1, mm->arity, tr2->label );
                            if ((tr3=M_sub_Sync_Scope(net,tr2,tr1,a2,a1,max))!=NULL) {
                                set_numname(tr3,mm);
                            }
                        }
                        tr2=tr2->prevtrans;
                    }
                    if (!(i&8)) /* Keine mit *mm konjugiert gefunden ? */
                        i&=~4;
                }
                tr1=RemoveTransition(net,tr1);
            } else
                tr1=tr1->prevtrans;
        }
        mm++;
    }
    return net;
}

/*****************************************************************************/
/* Erzeugung eines Netzes mit einer Entry- und einer Exit-Stelle, sowie      */
/* n internen Stellen.                                                       */
/* Rueckgabewert ist das erzeugte Netz.                                      */
/*****************************************************************************/

M_Net M_OnlyPlaces(Cardinal n)
{
    register M_Net      box;
    register M_Place    pl;
    register Cardinal   i;

    i=n+2;
    (box=M_empty_net())->NumComponents=1;
    box->x_pos=box->y_pos=0;
    box->x_dim=box->y_dim=RASTER;
    box->NumTransitions=0;
    box->NumPlaces=1;
    box->NumPTransitions = 0;  /* Phantomtransitionen */
    box->PTransitions = box->LastPTransition = NULL;
    box->Transitions=box->LastTransition=NULL;
    M_InitBlockInNet(box);  /* Bloecke vorbereiten */
    (pl=box->Places=box->LastPlace=M_allocPlace())->preset=NULL;
    pl->ppreset = NULL;  /* Phantomkanten */
    pl->ppostset = NULL;
    pl->postset=NULL;
    pl->nextplace=pl->prevplace=NULL;
    pl->numname=--i;
    pl->component=1;
    pl->label=EMPTY;
    M_CopySingleRange( pl->range, macro_id_set );
    pl->range[1]=NULL;
    pl->RefEntries = 0;
    pl->meaning=NULL;
    pl->init_marking=NULL;
    pl->curr_marking=NULL;
    pl->BlockEntries = 0;
    pl->blocks = NULL;
    pl->flags=0;
    pl->x_pos=0;
    pl->y_pos=i*2*RASTER;
    while (i) {
        (pl=M_DuplicatePlace(box,pl,&stdummy,&tsdummy,&stdummy,&tsdummy))
              ->numname=--i;
        pl->y_pos-=2*RASTER;
        box->y_dim+=2*RASTER;
    }
    pl->label=ENTRY;
    box->LastPlace->label=EXIT;
    return box;
}
