/****************************************************************************/
/* compose.c                                                                */
/*                                                                          */
/* Hilfsroutinen fuer ein M-Netz-Backend von b(pn)2                         */
/*                                                                          */
/* Martin Ackermann, Sommer 95                                              */
/* Burkhard Bieber 1995/96:                                                 */
/*   - BlockTable-Behandlung                                                */
/*   - Sync-/Phantom-Behandlung                                             */
/*   - Schoenheitskorrekturen                                               */
/* Martin Ackermann, Winter 1996/97                                         */
/*   - Prozeduren, andere Referenzen, mehr Effizienz, Fehler beseitigt      */
/****************************************************************************/  

#include <stdio.h>
#include <string.h>
#include "compose.h"
#include "block.h"
#include "move.h"
#include "box_sub.h"
#include "tables.h"
#include "macros.h"
#include "f_parse.h"
#include "evaltree.h"

/****************************************************************************/
/* Globalen Variablen zum Parsen einer Expression                           */
/****************************************************************************/

MultiSet       action_set;                       /* auftretende HL-Aktionen */
VariableStruct expr_vars;                 /* im Guard vorkommende Variablen */
node_ptr       expr_tree;/* Guard als Baum in einem dynamischen Knotenarray */
node_ptr       unprimed_guard;/* Guard-Ergaenzungen fuer unprimed Variablen */
#ifdef UNPRIMED_LOCAL
node_ptr       id_guard;      /* Guard-Ergaenzungen fuer Variablenzugriffe  */
#endif
Integer       *expr_def;               /* Definitionen fuer Guard-Variablen */
bool           expr_contains_chan0;      /* falls chan0-Zugriff stattfindet */

/****************************************************************************/
/* Der (Teil-)baum aus old_tree mit Knoten i als Kopf wird rekursiv an den  */
/* guard angehaengt.                                                        */
/* head gibt den N_HEAD-Knoten in guard an, dessen Knotenanzahl dabei       */
/* erhoeht wird.                                                            */
/* Die postorder Reihenfolge der Knoten bleibt erhalten.                    */
/*                                                                          */
/* Rueckgabe: Der (zu head relative) Index, an dem Knoten i in guard zu     */
/*            liegen kommt.                                                 */
/*                                                                          */
/* Falls FE_EXP auf TRUE gesetzt ist, werden Knoten mit FOREACH-Makros      */
/* expandiert.                                                              */
/* Falls FE_exp auf FALSE gesetzt ist, werden die Kommandos, die sich auf   */
/* VAR- / CONST-Knoten beziehen, angepasst (siehe CounterIndex2NodeIndex).  */
/*                                                                          */
/* Auf jeden Baum, der FOREACH-Makros enthaelt, muss einmal ->CopyTree<-    */
/* mit FE_exp=TRUE angewendet werden. Danach muss wie auf ALLE Baeume       */
/* einmal ->TreeToGuard<- mit FE_exp=FALSE angewendet werden.               */
/****************************************************************************/

Cardinal CopyTree( node_ptr old_tree, Cardinal i, node_ptr guard, Cardinal head,
                   bool FE_exp )
{
  node_ptr u, e;
  Cardinal n1=-1, n2=-1;
  bool     new_node = True;

  e = old_tree + i;

  switch( e->typ ) 
    {
      case N_CONST: 
        {
          n1 = e->op1;
          n2 = e->op2;
          /* Liegt ein Konstanten-Makro vor und ist FE_exp==false ? */
          if ( e->op2 && !FE_exp )
            CounterIndex2NodeIndex( Guard_SetConst, e->op2, guard->op2 );
          break;
        }
      case N_VAR:
        {
          n1 = e->op1;
          n2 = e->op2;
          /* Liegt ein Namens-Makro vor und ist FE_exp==false ? */
          if ( e->op1<0 && !FE_exp )
            CounterIndex2NodeIndex( Guard_SetVar, -(e->op1), guard->op2 );
          break;
        }
      case N_KLAMM:
      case N_NOT:
      case N_SIGN:  
        {
          n1 = CopyTree( old_tree, e->op1, guard, head, FE_exp ); 
          break;
        }
      case N_OR:
      case N_AND: 
      case N_EQ:
      case N_UNEQ:
      case N_LT:
      case N_GT:
      case N_LE:
      case N_GE:
      case N_PLUS:
      case N_MIN:
      case N_MULT:
      case N_DIV:
      case N_MOD:
      case N_REM:
        {
          n1 = CopyTree( old_tree, e->op1, guard, head, FE_exp );
          n2 = CopyTree( old_tree, e->op2, guard, head, FE_exp );
          break;
        }
      case N_MACRO:
        {
          /* Sollen FOREACH-Konstrukte expandiert werden ? */
          if ( FE_exp )
            {
              /* Das globale Array foreach wurde zuvor in InterpretCmds 
                 mit den zugehoerigen Parametern belegt. */
              Integer *foreach_set = foreach[e->op2].set;
              
              /* Enthaelt das foreach_set mehr als einen Wert ? */
              if ( *foreach_set > 1 )
                {
                  /* letztes Element des FOREACH-Wertebereichs abarbeiten */
                  macro_index = *foreach_set - 1;
                  macro_element = foreach_set[ *foreach_set - 1 ];
                  InterpretCmds( foreach[e->op2].cmd, 
                                 foreach[e->op2].cmd_entries, NULL );
                  n2 = CopyTree( old_tree, e->op1, guard, head, FE_exp ); 
                  
                  new_node = False;
                  /* uebrige Werte des FOREACH-Wertebereichs abarbeiten */
                  for( macro_index = *foreach_set - 2; 
                       macro_index >= 1; macro_index-- )
                    {
                      macro_element = foreach_set[macro_index];
                      InterpretCmds( foreach[e->op2].cmd, 
                                     foreach[e->op2].cmd_entries, NULL );
                      n1 = CopyTree( old_tree, e->op1, guard, head, FE_exp );

                      u = guard + guard->op2;
                      u->typ = N_AND;
                      u->op1 = n1;
                      u->op2 = n2;
                      guard[ head ].op1++;
                      n2 = guard->op2++ - head ;
                    }
                }
              else
                {
                  /* Falls der FOREACH-Wertebereich leer ist, muss der
                     FOREACH-Ausdruck durch tt ersetzt werden. */

                  e->typ = N_CONST;
                  n1     = TT;
                  n2     = 0;
                }

              MYfree( foreach_set );
              foreach[e->op2].set = NULL;
            }
          else
            {
              n1 = CopyTree( old_tree, e->op1, guard, head, FE_exp ); 
              n2 = e->op2;
            }
          break;
        }
      default: Msg( FAT, "Internal Error: Unknown node type !" );
    }

  /* Ausser im Falle eines FOREACH-Knotens mit einem nichtleeren Wertebereich
     muss an guard jetzt noch ein neuer Knoten gehaengt werden. */
  if ( new_node )
    {
      u = guard + guard->op2;
      u->typ = e->typ;
      u->op1 = n1;
      u->op2 = n2;
      guard[ head ].op1++;
      guard->op2++;
    }
  return ( guard->op2 - 1 - head );
}

/****************************************************************************/
/* Der Baum in old_tree wird an '&'-Knoten in Teilbaeume aufgetrennt, die   */
/* dann in einer guard-Struktur gespeichert werden.                         */
/* Die Routine funktioniert nur, wenn die '&'-Knoten rechts-assoziativ im   */
/* old_tree stehen. Eine links-assoziative Implementierung waere im         */
/* Gegensatz zu dieser Loesung deutlich aufwendiger.                        */
/*                                                                          */
/* Der Parameter FE_exp wird an CopyTree weitergereicht.                    */
/*                                                                          */
/* Rueckgabe: die neu erzeugte guard-Struktur                               */
/****************************************************************************/

node_ptr TreeToGuard( node_ptr old_tree, bool FE_exp )
{
  Cardinal n, i;
  node_ptr u, e;
  bool     last_and = False;
  node_ptr guard;

  if ( old_tree->op1 > 1 )
    {
      guard = MYmalloc( ( 2*old_tree->op1 ) * sizeof( node ) );
                        /* grosszuegige Abschaetzung der neuen Guard-Groesse */
      guard->op2 = 0;
      n = old_tree->op1 - 1;
      do
        {
          e = old_tree + n;
          if ( e->typ == N_AND )
            {
              i = e->op1; /*  kopiere  linken Teilbaum */
              n = e->op2; /* verfolge rechten Teilbaum weiter nach '&'-Knoten */
            }
          else
            {
              i = n;           /* kopiere den Restbaum... */
              last_and = True; /* ... und breche dann ab  */
            }
          u = guard + guard->op2; /* erzeuge N_HEAD fuer einen Teilbaum */
          u->typ = N_HEAD;
          u->op1 = 1;
          CopyTree( old_tree, i, guard, guard->op2++, FE_exp );
        }
      while( !last_and );
      
      guard = MYrealloc( guard, ( guard->op2 + 1 ) * sizeof( node ) );
    }
  else
    {
      guard = MYmalloc( sizeof( node ) );
      guard->op2 = 0;
    }

  /* haenge Schlusselement an den guard */
  u = guard + guard->op2;
  u->typ = N_HEAD;
  u->op1 = 0;
  guard->op2++;

  MYfree( old_tree );
  return guard;
}   

/****************************************************************************/
/* In den glob_range wird das Intervall [first..last] eingefuegt.           */
/* Nur falls 'do_realloc==True' gilt, wird die Speichergroesse des          */
/* glob_range angepasst.                                                    */
/*                                                                          */
/* Normalerweise setzt man additional = 0 und def = NULL.                   */
/*                                                                          */
/* Um allerdings in ranges einzufuegen, die bereits Bestandteil eines def-  */
/* arrays sind, muss man in additional angeben, wieviele Eintraege hinter   */
/* dem glob_range noch zu dem def-Array gehoeren.                           */
/* Ausserdem liefert man in def den Anfang des ZUGEHOERIGEN def-Arrays,     */
/* dessen Groesse dann automatisch angepasst wird.                          */
/****************************************************************************/

void add_values( Integer* *glob_range, Integer first, Integer last, 
                 bool do_realloc, Cardinal additional, Integer* *def )
{ 
  Integer* range = *glob_range;
  Cardinal doppelte=0;      /* Zaehler fuer Werte, die schon enthalten sind */
  Cardinal neue;            /* Anzahl der neu einzufuegenden Werte          */
  Cardinal alte;
  Integer *alt,*neu;
  Integer  i;

  if ( first <= last )
    {
      alt = range+1;
      while( alt < range + *range )
        {
          if ( *alt <= last && *alt >= first )
            doppelte++; /* zaehle Werte, die schon enthalten sind */
          alt++;
        }
      neue = last - first + 1 - doppelte;
      alte = *range;
      
      if ( neue )
        {
          *range += neue;

          if ( do_realloc )
            *glob_range = MYrealloc( *glob_range, *range * sizeof(Integer) );

          if ( def ) {
            size_t offset= *glob_range - *def;
            *(*def)     += neue;
            *def         = MYrealloc( *def, *(*def) * sizeof(Integer) );
            *glob_range = *def + offset;
              /* durch MYrealloc kann sich die Adresse von glob_range aendern */
          }

          range = *glob_range;
          alt = range + alte - 1 + additional;
          neu = alt + neue;
          while ( (*alt > last && alt > range) || alt > range + alte - 1 )
            *(neu--) = *(alt--); /* verschiebe hintere alte Werte ans Ende */
          for ( i=last; i>=first; i-- )
            *(neu--) = i ; /* kopiere neue Werte in den geschaffenen Platz */
        }
    }
}

/****************************************************************************/
/* Eine neue Variablenstruktur wird angelegt. Sie enthaelt standardmaessig  */
/* die Kommunikationsvariablen @c,@d,@e usw. (insgesamt ComEntries Stueck). */
/****************************************************************************/

Variables Create_empty_Variables( Variables v_ptr, Cardinal ComEntries )
{
  Cardinal  i;
  char      var_name[3];

  if ( GL_ProcDecl )
    strcpy( var_name, "id" ); /* Achtung, ggf. var_name vergroessern ! */
  else
    strcpy( var_name, "@c" ); /* Achtung, ggf. var_name vergroessern ! */

  v_ptr->ComEntries = ComEntries;
  v_ptr->ComArray   = MYmalloc( ComEntries * sizeof( ComArrayEntry ) );
  for( i = 0; i < ComEntries; ++i ) { 
      v_ptr->ComArray[i].var  = StrToVariable( var_name );
      v_ptr->ComArray[i].flag = 0;
      var_name[1]++;
  }
  return v_ptr;
}

/****************************************************************************/
/* In die Baumstruktur 'expr_tree' wird ein neuer Knoten eingefuegt.        */
/****************************************************************************/

Integer add_node( Cardinal typ, Integer op1, Integer op2 )
{ 
  if ( expr_tree->op1 >= expr_tree->op2 )
    {
      expr_tree->op2 = expr_tree->op2 + 100;
      expr_tree = MYrealloc( expr_tree, expr_tree->op2 * sizeof( node ) );
    }

  expr_tree[expr_tree->op1].typ = typ;
  expr_tree[expr_tree->op1].op1 = op1;
  expr_tree[expr_tree->op1].op2 = op2;

  return ((expr_tree->op1)++);
}

/****************************************************************************/
/* In das def-Array von 'tr' wird fuer die Projektion 'pr_nr' der           */
/* Variablen 'com_nr' der Wertebereich 'range' eingefuegt.                  */
/*                                                                          */
/* ( PrintAnno geht davon aus, dass bei mehrdimensionalen Definitionen die  */
/*   einzelnen Mengen mit aufsteigender Projektionsnummer direkt aufeinan-  */
/*   derfolgen im def-Array.                                                */
/*   Durch SET_MACRO kann es allerdings passieren, dass einzelne Mengen     */
/*   einer Definition nachtraeglich nachgetragen werden. Deshalb muss beim  */
/*   Einfuegen aufsteigend nach com_nr und pr_nr sortiert werden. )         */
/*                                                                          */
/* Rueckgabe: True  -> 'range' wurde eingefuegt                             */
/*            False -> fuer 'com_nr' existiert entweder fuer die Projektion */
/*                     'pr_nr' schon eine Definition                        */
/*                       ODER                                               */
/*                     'com_nr' soll jetzt eindimensional (bzw. mehrdimen-  */
/*                     sional) definiert werden, obwohl 'com_nr' bereits    */
/*                     mehrdimensional (bzw. eindimensional) definiert ist  */
/****************************************************************************/

void AddSingleRangeToDef( Integer* *d_ptr, Var com_nr, 
                          Integer pr_nr, Integer *range, bool resize )
{
  Integer *d;

  if ( resize )
    *d_ptr = MYrealloc( *d_ptr, (*(*d_ptr) + *range + 2) * sizeof( Integer ) );

  d = (*d_ptr) + 1;
  while( *d != -1 && (  *d <  (Integer) com_nr || 
                       (*d == (Integer) com_nr && *(d+1) < pr_nr) ) )
    {
      d += 2;
      d += *d;
    }

  if( (*d == (Integer) com_nr) && (*(d+1)==pr_nr) )
    { 
      Cardinal new, v1, v2, end1, end2;

      Integer *new_def=d, *d1=d, *d2=range-2;
  
      new = 3; v1 = 3; v2 = 3;
      end1 = d1[2]-1 + 3; end2 = d2[2]-1 + 3;
      while( v1 < end1 && v2 < end2 )
        {
          if ( d1[v1] < d2[v2] )
            v1++;
          else if ( d1[v1] > d2[v2] )
            v2++;
          else
            {
              new_def[new] = d1[v1];
              new++; v1++; v2++;
            }
        }
      new_def[2] = new-3 + 1;
      memmove( d + new, d + end1, 
               ( *(*d_ptr) - (d-*d_ptr) - end1 ) * sizeof(Integer));
      *(*d_ptr) -= (end1 - new);
    }
  else
    {
      memmove( d + *range + 2, d, ( *(*d_ptr) - (d-*d_ptr) ) * sizeof(Integer));
      *(*d_ptr) += *range + 2;
      d[0] = com_nr;
      d[1] = pr_nr;
      memcpy( d + 2, range, *range * sizeof( Integer ) );
    }
}


/****************************************************************************/
/* Vor net wird decl.init_net gehaengt und an net wird decl.exit_net ge-    */
/* haengt. Neben net wird decl.var_net gesetzt und das gesamte Netz wird    */
/* uber decl.scope_actions gescopet.                                        */
/****************************************************************************/

M_Net Scope( M_Net net, decl_type decl )
{ 
  if ( decl.init_net )
    net = M_Sequence( decl.init_net, net );
  if ( decl.exit_net )
    net = M_Sequence( net, decl.exit_net );
  if ( decl.var_net )
    net = M_ConcurrentComposition( decl.var_net, net );

  if ( !(flag&NO_SCOPING) )
    { 
#ifndef NORMAL_SCOPING
      net = M_Scoping( net, decl.scope_actions, max_arc_vars );
#else
      net = M_Synchronisation( net, decl.scope_actions, max_arc_vars );
      net = M_Restriction( net, decl.scope_actions );
#endif
    }

  return net;
}

/****************************************************************************/
/* Sortiert fuer jede Transition die Referenzen nach der Reihenfolge im     */
/* b(pn)2-Quelltext.                                                        */
/****************************************************************************/

#define vgl( comp, op ) \
  RefToStruct(tr->ref[k].nr)->s_ref.comp \
    op \
  RefToStruct(tr->ref[j].nr)->s_ref.comp

static void SortAllTransRefs(M_Net net )
{
   M_Transition tr;
   Cardinal i, j, k;
   Ref      swap;

   if ( !net )
      return;

   tr = net->Transitions;
   while( tr )
   {
      i=0;
      while( i+1 < tr->RefEntries )
      {
         j=i;
         k=i+1;
         while( k < tr->RefEntries )
         {
            if(   vgl( start_l, <  ) ||
                ( vgl( start_l, == ) && vgl( start_c, < ) ) )
               j=k;
               k++;
         }
         if( i != j )
         {
            swap = tr->ref[i];
            tr->ref[i] = tr->ref[j];
            tr->ref[j] = swap;
         }
         i++;
      }
      tr = tr->nexttrans;
   }
}
 
/****************************************************************************/
/* Wenn das Netz fertig ist, passiert so einiges...                         */
/****************************************************************************/

void HandleNet( M_Net net )
{
    int i;
    Cardinal32 c_transitions,c_places,c_places_phant;

    if (flag&ELIMINATE_DEAD) {
        printf("%u dead transition(s) eliminated.\n", dead_eliminated);
    }

    if (flag&REMOVE_DUPL) {
        printf("Removing duplicates: ");
        M_RemoveDuplicates(net,&c_places,&c_transitions,&c_places_phant);
        printf("%u place(s), %u transition(s),\n", c_places, c_transitions);
        printf("                     ");
        printf("%u duplicate ", c_places_phant);
        printf("place(s) preserved because of different phantoms.\n\n" );
    }

    if (flag&CHECK_NET) {
        printf("Checking consistency of resulting net: ");
        if ((i=M_CheckNet(net))!=0) {
            printf("inconsistent (%i)!\n\n",i);
            exit(5);
        } else
            printf("OK !\n\n");
    }

    if (pep_file) {
        M_move_hv(net, RASTER-net->x_pos, RASTER-net->y_pos);
        SortAllTransRefs(net);
        if (flag & BLOCK_OUTPUT) {
          M_CreateBlockElements(net);
        }  /* if Blockausgabe gewuenscht */
        PrintNet_PEP( net );
        if (net->NumTransitions != trans_anz)
          Msg(INF, "Transitions: counter (%lu) <> real number (%lu) !",
            trans_anz, net->NumTransitions);
    }
}


/****************************************************************************/
/* Zwei Deklarationszeilen werden zusammengefasst:                          */
/* Im init_net stehen alle Initnetze, entweder parallel oder sequentiell.   */
/* Im  var_net stehen alle Datenboxen parallel nebeneinander.               */
/* Im exit_net stehen alle Exitnetze, entweder parallel oder sequentiell.   */
/****************************************************************************/

decl_type Sum_Declarators( decl_type d1, decl_type d2 )
{ 
  
  decl_type d;

  if ( d1.init_net )
    {
      if ( d2.init_net )
        {
          if (flag&PAR_TERM)
            d.init_net = M_ConcurrentComposition( d1.init_net, d2.init_net );
          else
            d.init_net = M_Sequence( d1.init_net, d2.init_net );
        }
      else
        d.init_net = d1.init_net;
    }
  else
    d.init_net = d2.init_net;

  if ( d1.var_net )
    {
      if ( d2.var_net )
        d.var_net = M_ConcurrentComposition( d1.var_net, d2.var_net );
      else
        d.var_net = d1.var_net;
    }
  else
    d.var_net = d2.var_net;

  if ( d1.exit_net )
    {
      if ( d2.exit_net )
        {
          if (flag&PAR_TERM)
            d.exit_net = M_ConcurrentComposition( d1.exit_net, d2.exit_net );
          else
            d.exit_net = M_Sequence( d2.exit_net, d1.exit_net );
        }
      else
        d.exit_net = d1.exit_net;
    }
  else
    d.exit_net = d2.exit_net;
  
  return d;
}

/****************************************************************************/
/* Fuer den Expression-Parser werden die globalen Variablen initialisiert.  */
/****************************************************************************/

void init_ExpressionParser( Variables v_ptr )
{ 
  action_set = MScreate_empty_set();

  if ( !v_ptr )
    Create_empty_Variables( &expr_vars, 1 );/* @c als Kantenbeschriftung */
  else
    {
      expr_vars.ComEntries = v_ptr->ComEntries;
      expr_vars.ComArray   = v_ptr->ComArray;
    }

  expr_tree = MYmalloc( 100 * sizeof( node ) );
  expr_tree->typ = N_HEAD;
  expr_tree->op1 = 1;
  expr_tree->op2 = 100;

  expr_def=MYmalloc( 2*sizeof( Integer ) );
  expr_def[0] = 2;
  expr_def[1] = -1;

  expr_contains_chan0 = False;

#ifdef UNPRIMED_LOCAL
/* Wenn man UNPRIMED_LOCAL im Makefile setzt, so werden die Implikationen
   bei Verwendung von ungestrichenen Variablen LOKAL im Guard eingefuegt.
   Die Pros und Kontras gegenueber GLOBAL sind in [ACK97] aufgezeigt. */
  id_guard = MYmalloc( sizeof( node ) );
  id_guard->typ = N_HEAD;
  id_guard->op1 = 0;
  id_guard->op2 = 1;
#endif

  unprimed_guard = MYmalloc( sizeof( node ) );
  unprimed_guard->typ = N_HEAD;
  unprimed_guard->op1 = 0;
  unprimed_guard->op2 = 1;
}

#ifdef UNPRIMED_LOCAL
/****************************************************************************/
/* An den letzten Teilbaum von expr_tree wird der unprimed_guard angehaengt.*/
/****************************************************************************/

Integer add_UnprimedGuard( void )
{
  node_ptr u = unprimed_guard;

  while ( u->op1 )
    {
      Integer op1, op2;

      while ( expr_tree->op1 + 1 + u->op1 >= expr_tree->op2 )
        {
          expr_tree->op2 = expr_tree->op2 + 100;
          expr_tree = MYrealloc( expr_tree, expr_tree->op2 * sizeof( node ) );
        }

      op1 = expr_tree->op1;
      op2 = expr_tree->op2;
      expr_tree->op2 = op1;
      CopyTree( u, u->op1-1, expr_tree, 0, False );
      expr_tree->op2 = op2;

      expr_tree[expr_tree->op1].typ = N_AND;
      expr_tree[expr_tree->op1].op1 = op1 - 1;
      expr_tree[expr_tree->op1].op2 = (expr_tree->op1)++ - 1;

      u += u->op1;
    }

  if ( unprimed_guard->op1 )
    {
      unprimed_guard = MYrealloc( unprimed_guard, sizeof( node ) );
      unprimed_guard->op1 = 0;
      unprimed_guard->op2 = 1;
    }

  return (expr_tree->op1 - 1);
}
#endif

/****************************************************************************/
/* expr_tree wird in '&'-Baeume aufgeteilt und anschliessend unprimed_guard */
/* und id_guard angehaengt.                                                 */
/****************************************************************************/

void exit_ExpressionParser( void )
{
  Integer anz_expr_nodes;

  if ( expr_vars.ComEntries > MaxAnzComVars )
    Msg( FAT, "A single action produces a transition with too many ComVars !" );

  expr_tree = TreeToGuard( expr_tree, False );

  if ( unprimed_guard->op1 )
    {
      anz_expr_nodes = expr_tree->op2 - 1;
      expr_tree->op2 = anz_expr_nodes + unprimed_guard->op2;
      expr_tree = MYrealloc( expr_tree, expr_tree->op2 * sizeof( node ) );
      memcpy( expr_tree + anz_expr_nodes, unprimed_guard, 
              unprimed_guard->op2 * sizeof( node ) );
    }
  MYfree( unprimed_guard );

#ifdef UNPRIMED_LOCAL
  if ( id_guard->op1 )
    {
      anz_expr_nodes = expr_tree->op2 - 1;
      expr_tree->op2 = anz_expr_nodes + id_guard->op2;
      expr_tree = MYrealloc( expr_tree, expr_tree->op2 * sizeof( node ) );
      memcpy( expr_tree + anz_expr_nodes, id_guard, 
              id_guard->op2 * sizeof( node ) );
    }
  MYfree( id_guard );
#endif
}


/****************************************************************************/
/* Die Transition zu einer b(pn)2-Aktion bzw. einem Prozeduraufruf wird als */
/* Kontrollfluss-Transition und ggf. als Kanal0Zugriff markiert.            */
/*                                                                          */
/* Bei Kanal0Zugriffen innerhalb von Prozeduren wird fuer id die lokale     */
/* Umbenennung eingeschaltet. Falls dann innerhalb eines do-od-Konstrukts   */
/* CHOICE-Zweige mit Kanal0Zugriffen verschmelzen, werden die id-Bezeichner */
/* der verschmelzenden Zweige durchnumeriert.                               */
/****************************************************************************/

void Modify_Transition( M_Transition tr )
{
  tr->flags |= TR_SCOPE;
  if ( expr_contains_chan0 ) 
    {
      tr->flags |= TR_CHAN0_ACCESS;

      /* Findet der Zugriff in einer Prozedur statt ? */
      if ( macro_id_set != dot_set ) 
        {
          /* 'id' muss beim Synchronisieren umbenannt werden */
          tr->vars.ComArray[ StrToCom("id",&(tr->vars),tr) ].flag |= VAR_RENAME;

          /* Innerhalb eines do-od-Konstrukts duerfen Kanten mit einem Gewicht
             bis zu macro_max entstehen, wenn CHOISE-Zweige innerhalb einer
             Prozedur ueber Kanaele der Kapazitaet 0 verschmelzen. */
          if ( (Cardinal) macro_max > max_arc_vars )
            max_arc_vars = macro_max;
        }
    }
}


/****************************************************************************/
/* An Adresse n wird ein Guard aus Kopfelement, Baum mit 3 Knoten und       */
/* Schlusselement aufgebaut:                                                */
/*                                                                          */
/*          op3                                                             */
/*         /   \                                                            */
/*     typ2     typ1                                                        */
/*    /    \   /    \                                                       */
/* op11  op12 op21   -1                                                     */
/*                                                                          */
/* Falls n NULL ist, wird fuer den Baum Speicher alloziert.                 */
/* Der Rueckgabewert ist ein Zeiger auf den erzeugten Baum.                 */
/****************************************************************************/

static node_ptr BasicTree( node_ptr n, Cardinal op3, 
                           Cardinal typ1, Integer op11, Integer op12,
                           Cardinal typ2, Integer op21 )
{
  node_ptr t;

  if ( n )
    t = n;
  else
    t = n = MYmalloc( 5 * sizeof( node ) );

  n->typ=N_HEAD;
  n->op1=4;
  n->op2=5;
  n++;
  n->typ=typ1;
  n->op1=op11;
  n->op2=op12;
  n++;
  n->typ=typ2;
  n->op1=op21;
  n->op2=-1;
  n++;
  n->typ=op3;
  n->op1=1;
  n->op2=2;
  n++;
  n->typ=N_HEAD;
  n->op1=0;

  return t;
}


/****************************************************************************/
/* Behandelt Variablenzugriff fuer die Prozedursyntax.                      */
/*                                                                          */
/* Im Gegensatz zur kompositionellen M-Netz-Semantik in der Theorie wird    */
/* hier anhand der Symboltabelle zwischen dem Zugriff auf lokale bzw.       */
/* globale Variablen unterschieden. Dadurch erhaelt man Transitionen mit    */
/* einfacheren und damit effizienteren Beschriftungen.                      */
/*                                                                          */
/* Rueckgabe: die Kommunikationsnummer von id bzw. id_name                  */
/****************************************************************************/

static ComVar Add_ProcSem( char *action_name, Variables v_ptr,
                           node_ptr *ext_ptr, SymTabEntry* s_entry )
{
  ComVar nr;

  if ( s_entry->s_partype == _REF )
    {
      /* Fuer Referenzparameter in rekursiven Prozeduren muss fuer id die
         lokale Umbenennung eingeschaltet werden. */
      nr = StrToCom( "id", v_ptr, NULL );
      v_ptr->ComArray[nr].flag |= VAR_RENAME;
      
      /* Bei rekursiven Prozeduren haben die Transitionen zum lokalen 
         Durchreichen der RefPars als max. Gewicht (max. Rekursionstiefe)-1. 
         Die maximale Rekursionstiefe wird mit der maximalen Anzahl von 
         Instanzen nach oben abgeschaetzt. */
      if ( macro_max-1 > max_arc_vars )
        max_arc_vars = macro_max-1;
    }
  else /* s_entry->s_partype == _VAR, _CONSTVAR, oder _RES */
    {
      if (ext_ptr && macro_id_set!=dot_set && s_entry->s_blknum>=min_local_blk)
        {
          /* Innerhalb von Prozeduren kann beim Zugriff auf lokale Variablen
             anstelle von id_name direkt id verwendet werden. Eine Wertevorgabe
             der Form id_name in {dot,id} ist dann ueberfluessig. */

          nr = StrToCom( "id", v_ptr, NULL );
        }
      else
        {
          /* Fuege id_name=dot fuer den Zugriff auf globale Variablen zum Guard
             *ext_ptr hinzu. */

          char   id_name[ MAX_IDENT_LENGTH ];
          
          node   sub_tree[] = { 
            {N_HEAD, 4, 5}, 
            {N_VAR, 0, -1}, {N_CONST, DOT, 0}, {N_EQ, 1, 2}, {N_HEAD, 0, 0} };

          strcpy( id_name, "id_" );
          strcat( id_name, action_name  );
          nr = StrToCom( id_name, v_ptr, NULL );

          (*ext_ptr)->op2 += 4;
          *ext_ptr = MYrealloc( *ext_ptr, (*ext_ptr)->op2*sizeof(node) );
          sub_tree[1].op1 = nr;
          memcpy( *ext_ptr + (*ext_ptr)->op2 - 5, sub_tree, 5 * sizeof(node) );
        }
    }

  return nr;
}

/****************************************************************************/
/* Falls noch nicht enthalten, wird in set der zweistellige Aktionsterm     */
/* s_name( 's_name, s_name' ) eingefuegt.                                   */
/*                                                                          */
/* Falls 'v_ptr' == NULL ist, werden keine Kommunikationsnummern ermittelt  */
/* bzw. erzeugt.                                                            */
/* VORSICHT: So geparste Aktionen koennen nicht von VerboseActToStr         */
/*           ausgegeben werden.                                             */
/*                                                                          */
/* Falls die Prozedursyntax verwendet wird, wird der Aktionsterm noch um    */
/* den Parameter id_name erweitert. Der Wertterm fuer die Bindung von       */
/* id_name wird an den Guard in *ext_ptr angehaengt.                        */
/*                                                                          */
/* RUECKGABE: Zeiger auf die Aktion                                         */
/****************************************************************************/

Action *Add_VariableAction( SymTabEntry *s_entry, MultiSet *set, 
                            Variables v_ptr, node_ptr *ext_ptr )
{ 
  Action_nr     nr;
  Action       *ActionPtr;
  char         *action_name = s_entry->s_name; 
  Cardinal      action_arity;

  if ( GL_ProcDecl )
    action_arity = 3;
  else
    action_arity = 2;

  nr = StrToAction( action_name );
  ActionPtr = MSis_in( nr, action_arity, *set );

  if ( !ActionPtr )
  { 
    mmSingle->anzahl    = 1;
    mmSingle->nr        = nr;
    mmSingle->arity     = action_arity;

    if ( v_ptr )
      {
        char real_name[ MAX_IDENT_LENGTH ];

        sprintf( real_name, "'%s", action_name );
        mmSingle->com[0] = StrToCom( real_name, v_ptr, NULL );
        sprintf( real_name, "%s'", action_name );
        mmSingle->com[1] = StrToCom( real_name, v_ptr, NULL );

        /* Fuer Referenzparameter in rekursiven Prozeduren muss fuer 's_name 
           und s_name' die lokale Umbenennung eingeschaltet werden. */
          v_ptr->ComArray[ mmSingle->com[0] ].flag
        = v_ptr->ComArray[ mmSingle->com[1] ].flag
        = ( s_entry->s_partype == _REF ) ? VAR_RENAME : 0;

        if ( GL_ProcDecl )
          mmSingle->com[2] = Add_ProcSem(action_name, v_ptr, ext_ptr, s_entry);
      }

    MSresize( *set, MSlength( *set ) + 1 );
    MSsum( *set, mmSingle );
    ActionPtr = MSis_in( nr, action_arity, *set );

  }
  return ActionPtr;
}


/****************************************************************************/
/* In den Guard-Baum (global in expr_tree) wird eine preprimed Variable     */
/* eingefuegt.                                                              */
/****************************************************************************/

Integer Add_preprimed( SymTabEntry *s_entry )
{
  Action *action;

#ifdef UNPRIMED_LOCAL
  action = Add_VariableAction( s_entry, &action_set, 
                               &expr_vars, &id_guard );
#else
  action = Add_VariableAction( s_entry, &action_set, 
                               &expr_vars, &unprimed_guard );
#endif
  return add_node( N_VAR, action->com[0], -1 );
}


/****************************************************************************/
/* In den Guard-Baum (global in expr_tree) wird anstelle einer unprimed     */
/* Variable eine preprimed Variable eingefuegt.                             */
/*                                                                          */
/* Ausserdem wird in die Guard-Erweiterung (global in unprimed_guard)       */
/* die Bedingung "preprimed = postprimed" eingefuegt; natuerlich nur,       */
/* wenn diese Bedingung noch nicht vorhanden ist.                           */
/****************************************************************************/

Integer Add_unprimed( SymTabEntry *s_entry )
{
  Action *action;
  node_ptr n, sub_tree;

#ifdef UNPRIMED_LOCAL
  action = Add_VariableAction( s_entry, &action_set, 
                               &expr_vars, &id_guard );
#else
  action = Add_VariableAction( s_entry, &action_set, 
                               &expr_vars, &unprimed_guard );
#endif

  /* suche die Guard-Erweiterung nach der benutzten Variable ab */

  n = unprimed_guard;
  while( n->op1 )
    {
      n++;
      while( n->typ != N_HEAD && 
             !(n->typ == N_VAR && (Cardinal)n->op1 == action->com[0]) )
        n++;
      if ( n->typ != N_HEAD )
        break;
    }

  if( n->typ == N_HEAD )
    {
      unprimed_guard->op2 += 4;
      unprimed_guard = MYrealloc( unprimed_guard, 
                                  unprimed_guard->op2*sizeof(node) );
      sub_tree = unprimed_guard + unprimed_guard->op2 - 5;
      BasicTree( sub_tree, N_EQ, N_VAR, action->com[0], -1, 
                                 N_VAR, action->com[1]);
    }

  return add_node( N_VAR, action->com[0], -1 );
}


/****************************************************************************/
/* In den Guard-Baum (global in expr_tree) wird eine postprimed Variable    */
/* eingefuegt.                                                              */
/****************************************************************************/

Integer Add_postprimed( SymTabEntry *s_entry )
{
  Action *action;

#ifdef UNPRIMED_LOCAL
  action = Add_VariableAction( s_entry, &action_set, 
                               &expr_vars, &id_guard );
#else
  action = Add_VariableAction( s_entry, &action_set, 
                               &expr_vars, &unprimed_guard );
#endif

  return add_node( N_VAR, action->com[1], -1 );
}


/****************************************************************************/
/* Falls noch nicht enthalten, wird in set eine Aktion eingefuegt:          */
/* - type =  CHAN_NULL:                            Aktion name? (name  )    */
/* - type =  CHAN_NULL |               CHAN_WRITE: Aktion name! (name  )    */
/* - type = !CHAN_NULL | !CHAN_CHECK:              Aktion name? (name? )    */
/* - type = !CHAN_NULL | !CHAN_CHECK | CHAN_WRITE: Aktion name! (name! )    */
/* - type = !CHAN_NULL |  CHAN_CHECK:              Aktion name??(name??)    */
/* - type = !CHAN_NULL |  CHAN_CHECK | CHAN_WRITE: Aktion name!!(name!!)    */
/*                                                                          */
/* Falls 'v_ptr' == NULL ist, werden keine Kommunikationsnummern ermittelt  */
/* bzw. erzeugt.                                                            */
/* VORSICHT: So geparste Aktionen koennen nicht von VerboseActToStr         */
/*           ausgegeben werden.                                             */
/*                                                                          */
/* Falls die Prozedursyntax verwendet wird, wird der Aktionsterm noch um    */
/* den Parameter id_name erweitert. Der Wertterm fuer die Bindung von       */
/* id_name wird an den Guard in *ext_ptr angehaengt.                        */
/*                                                                          */
/* RUECKGABE: Zeiger auf die Aktion                                         */
/****************************************************************************/

Action *Add_ChannelAction( SymTabEntry *s_entry, MultiSet *set, 
                           Variables v_ptr, node_ptr *ext_ptr, Cardinal type )
{ 
  Action_nr     nr;
  Action       *ActionPtr;
  char         *name = s_entry->s_name; 
  char          action_name[ MAX_IDENT_LENGTH ];
  Cardinal      action_arity;

  if ( GL_ProcDecl )
    action_arity = 2;
  else
    action_arity = 1;

  strcpy( action_name, name );

  if ( type & CHAN_WRITE )
    if ( type & CHAN_CHECK )
      strcat( action_name, "!!" );
    else
      strcat( action_name, "!" );
  else
    if ( type & CHAN_CHECK )
      strcat( action_name, "??" );
    else
      strcat( action_name, "?" );


  nr = StrToAction( action_name );
  ActionPtr = MSis_in( nr, action_arity, *set );

  if ( !ActionPtr )
  { 
    mmSingle->anzahl    = 1;
    mmSingle->nr        = nr;
    mmSingle->arity     = action_arity;

    if ( v_ptr )
      {
        /* Bei Kanal 0 muss bei der Lese- und bei der Schreibaktion der selbe
           Name fuer die Kommunikationsvariable verwendet werden, um einen
           Namenskonflikt zu vermeiden. */
        mmSingle->com[0]    = StrToCom( (type & CHAN_NULL) ? name : action_name,
                                        v_ptr, NULL );

        /* Fuer Referenzparameter in rekursiven Prozeduren muss fuer den ersten
           Aktionsparameter die lokale Umbenennung eingeschaltet werden. */
        v_ptr->ComArray[ mmSingle->com[0] ].flag
          = ( s_entry->s_partype == _REF ) ? VAR_RENAME : 0;

        if ( GL_ProcDecl )
          mmSingle->com[1] = Add_ProcSem( name, v_ptr, ext_ptr, s_entry );
      }

    MSresize( *set, MSlength( *set ) + 1 );
    MSsum( *set, mmSingle );
    ActionPtr = MSis_in( nr, action_arity, *set );

  }
  return ActionPtr;
}


/****************************************************************************/
/* In den Guard-Baum (global in expr_tree) wird eine Channel-Variable       */
/* eingefuegt.                                                              */
/* Das type (siehe Add_ChannelAction) bestimmt den Typ der Variablen        */
/* (Read/Write/IsCompletelyEmpty/IsCompletelyFull).                         */
/****************************************************************************/

Integer Add_Channel( SymTabEntry *s_entry, Cardinal type )
{
  Action *action;

  if ( type & CHAN_NULL ) 
    expr_contains_chan0 = True;
#ifdef UNPRIMED_LOCAL
  action = Add_ChannelAction( s_entry, &action_set, &expr_vars, 
                              &id_guard, type );
#else
  action = Add_ChannelAction( s_entry, &action_set, &expr_vars, 
                              &unprimed_guard, type );
#endif
  return add_node( N_VAR, action->com[0], -1 );
}
