/******************************************************************************/
/* evaltree.c                                                                 */
/*                                                                            */
/* Routinen zum Abarbeiten eines b(pn)2-Syntaxbaumes in ein M-Netz            */
/*                                                                            */
/* Martin Ackermann, Winter 96/97                                             */
/******************************************************************************/

#include "evaltree.h"
#include "frontend.h"
#include "mnet.h"
#include "box_sub.h"
#include "macros.h"
#include "f_parse.h"
#include "block.h"
#include "tables.h"
#include "move.h"

#define ofp stderr

/******************************************************************************/
/* globale Variablen                                                          */
/******************************************************************************/

Cardinal block_num     = 0; /* Schachtelungstiefe des aktuell bearbeiteten 
                               Blocks (wird fuer die Zuordnung von 
                               unnamed_blocks benoetigt) */

Cardinal id_counter    = 1; /* zum Generieren von disjunkten id_sets */

Cardinal min_local_blk = 1; /* zur Unterscheidung von globalen/lokalen Decls */


/******************************************************************************/
/* Funktionen                                                                 */
/******************************************************************************/

static M_Net Com2Net    (TreeNode *node, Cardinal *tags);


/******************************************************************************/
/* Uebersetzen der Menge set_org von der Darstellung des Frontends in die in- */
/* terne Darstellung des Backends. Die Menge set_new wird zurueckgeliefert.   */
/******************************************************************************/

Integer* Set_Parser2Intern( Set set_org )
{
  Integer *set_new;
  
  set_org = set_org->next;

  set_new = MYmalloc( sizeof( Integer ) );
  *set_new = 1;
  
  while ( set_org )
    {
      add_values( &set_new, set_org->low, set_org->high, True, 0, NULL );
      set_org = set_org->next;
    }

  return set_new;
}


/******************************************************************************/
/* Erzeugen eines IdSets mit n netzweit einzigartigen ids. Dazu wird der      */
/* globale Zaehler id_counter weitergezaehlt.                                 */
/******************************************************************************/

static Integer* GenerateIdSet( Cardinal n )
{
  Integer *set = MYmalloc( (n+1) * sizeof( Integer ) );
  Cardinal i;


  /* BG nur zum Testen, ob es ausreicht um nicht-disjunkte id-sets zu bekommen */
  if (flag & NO_DISJUNCT_IDSETS) id_counter=1;


  for ( i=1; i<=n; i++ )
    set[i] = id_counter++;

  *set = n+1;
  return set;
}


/******************************************************************************/
/* An die Namen aller Referenz-Parameter einer Prozedur wird mit '@' der      */
/* Prozedurname angehaengt.                                                   */
/******************************************************************************/

void RenameRefPars( SymTabEntry *ptr, char *proc_name )
{
  while (ptr) {

    /* Jetzt schon alle Referenzen eintragen, damit die Numerierung der Refe-
       renzen mit der Reihenfolge der Parameter im Quelltext uebereinstimmt. */
    StructToRef( (ShortEntryPtr) ptr ); 

    if ( ptr->s_partype == _REF )
      {
        char ref_name[ MAX_IDENT_LENGTH ];

        sprintf( ref_name, "%s@%s", ptr->s_name, proc_name );
        MYfree( ptr->s_name );
        ptr->s_name = MYstrdup( ref_name );
      }

    ptr=ptr->s_parlist;
  }
}


/******************************************************************************/
/* Bearbeiten der Parameter einer Prozedur:                                   */
/* - Fuer VAL-, CONST-, und RES-Parameter wird das Prozedurnetz *net_ptr um   */
/*   eine Datenbox fuer den Parameter erweitert.                              */
/* - Fuer REF-Parameter wird die Menge der Aktionssymbole fuer Synchronisa-   */
/*   tion und Restriktion des uebergeordneten Blocks passend erweitert.       */
/******************************************************************************/

void HandlePars( SymTabEntry *ptr, M_Net *net_ptr )
{
  M_Net     val_net;
  M_TSEdge  ts, ts_last;
  M_STEdge  st, st_last;
  Var      *tpl;

  while (ptr) {
    switch ( ptr->s_partype )
      {
      case _VAL:
      case _CONSTPAR:
      case _RES:
        macro_name     = ptr->s_name;
        macro_ref      = StructToRef( (ShortEntryPtr) ptr );
        macro_decl_set = Set_Parser2Intern( ptr->s_set );
        GL_AddActions = True;
        if ( ptr->s_partype == _VAL )
          val_net = GetBox( VAL_PAR, True );
        else if ( ptr->s_partype == _CONSTPAR )
          val_net = GetBox( CONST_PAR, True );
        else /* ptr->s_partype == _RES */
          val_net = GetBox( RES_PAR, True );
        MYfree( macro_decl_set );

        /* val_net rechts neben *net_ptr plazieren */
        M_move_hv( val_net, 
                   (*net_ptr)->x_pos+(*net_ptr)->x_dim-val_net->x_pos+RASTER,
                   trans_mark[MARK_PAR_PRE]->y_pos-val_net->y_pos+2*RASTER );
        (*net_ptr)->x_dim += val_net->x_dim + RASTER;
        *net_ptr = M_DisjointNetUnion( *net_ptr, val_net );

        /* erzeuge TS-Kante von trans_mark[MARK_PAR_PRE] zur VAL-Stelle */
        (ts=M_allocTSEdge())->prevtrans=NULL;
        if ((ts->nexttrans=(ts->place=(*net_ptr)->LastPlace)->preset)!=NULL)
          ts->nexttrans->prevtrans=ts;
        ((*net_ptr)->LastPlace->preset=ts)->nextplace=NULL;
        ts->trans=trans_mark[MARK_PAR_PRE];
        ts_last=trans_mark[MARK_PAR_PRE]->postset;
        if (ts_last)
          {
            while( ts_last->nextplace )
              ts_last = ts_last->nextplace;
            ts_last->nextplace=ts;
          }
        else
          trans_mark[MARK_PAR_PRE]->postset=ts;
        ts->prevplace=ts_last;

        /* TS-Kante mit (Parametername,proc) beschriften */
        tpl    = MYmalloc( 3 * sizeof( Var ) );
        tpl[0] = 3;
        tpl[1] = StrToCom( ptr->s_name, &(trans_mark[MARK_PAR_PRE]->vars), 
                           trans_mark[MARK_PAR_PRE] );
        trans_mark[MARK_PAR_PRE]->vars.ComArray[ tpl[1] ].flag |= VAR_MINOR;
        tpl[2] = StrToCom(      "proc", &(trans_mark[MARK_PAR_PRE]->vars), 
                                          trans_mark[MARK_PAR_PRE] );
        {
          /* erweitere die Aktion bei MARK_PAR_PRE um die neue ComVar */

          /* Dieser Block muss hier eingeschoben werden, weil nach ArrayToTpl
             nicht mehr auf tpl zugegriffen werden darf. */
          trans_mark[MARK_PAR_PRE]->label->
            com[ trans_mark[MARK_PAR_PRE]->label->arity++ ] = tpl[1];
          if ( trans_mark[MARK_PAR_PRE]->label->arity > MaxArity )
            Msg( FAT, "Too many VAL/RES-parameters for a single procedure !" );
        }
        ts->label=V_MScreate_singleton( ArrayToTpl(&tpl)+V_TUPEL_OFFSET );
        ts->flags=0;

        /* erzeuge ST-Kante von der VAL-Stelle zu trans_mark[MARK_PAR_POST] */
        (st=M_allocSTEdge())->prevtrans=NULL;
        if ((st->nexttrans=(st->place=(*net_ptr)->LastPlace)->postset)!=NULL)
          st->nexttrans->prevtrans=st;
        ((*net_ptr)->LastPlace->postset=st)->nextplace=NULL;
        st->trans=trans_mark[MARK_PAR_POST];
        st_last=trans_mark[MARK_PAR_POST]->preset;
        if (st_last)
          {
            while( st_last->nextplace )
              st_last = st_last->nextplace;
            st_last->nextplace=st;
          }
        else
          trans_mark[MARK_PAR_POST]->preset=st;
        st->prevplace=st_last;

        /* ST-Kante mit (Parametername,proc) beschriften */
        tpl    = MYmalloc( 3 * sizeof( Var ) );
        tpl[0] = 3;
        tpl[1] = StrToCom( ptr->s_name, &(trans_mark[MARK_PAR_POST]->vars), 
                           trans_mark[MARK_PAR_POST] );
        trans_mark[MARK_PAR_POST]->vars.ComArray[ tpl[1] ].flag |= VAR_MINOR;
        tpl[2] = StrToCom(      "proc", &(trans_mark[MARK_PAR_POST]->vars), 
                                          trans_mark[MARK_PAR_POST] );
        if ( ptr->s_partype == _RES )
          {
            /* erweitere die Aktion bei MARK_PAR_POST um die neue ComVar */

            /* Dieser Block muss hier eingeschoben werden, weil nach ArrayToTpl
               nicht mehr auf tpl zugegriffen werden darf. */
            trans_mark[MARK_PAR_POST]->label
              ->com[ trans_mark[MARK_PAR_POST]->label->arity++ ] = tpl[1];
          }
        st->label=V_MScreate_singleton( ArrayToTpl(&tpl)+V_TUPEL_OFFSET );
        st->flags=0;
        
        break;

      case _REF:
        if ( ptr->s_type == _VAR )
          Add_VariableAction( ptr, &spec_actions, NULL, NULL )->nr |= 1;
        else
          {
            Add_ChannelAction( ptr, &spec_actions, NULL, NULL, 
                               0 )->nr |= 1;
            Add_ChannelAction( ptr, &spec_actions, NULL, NULL, 
                               CHAN_WRITE )->nr |= 1;

            /* IsCompletelyEmpty- und IsCompletelyFull-Check nur fuer
               Kanaele / Stacks der Kapazitaet > 0 */
            if ( ptr->s_init_or_cap > 0 )
              {
                Add_ChannelAction( ptr, &spec_actions, NULL, NULL, 
                                   CHAN_CHECK | 0 )->nr |= 1;
                Add_ChannelAction( ptr, &spec_actions, NULL, NULL, 
                                   CHAN_CHECK | CHAN_WRITE )->nr |= 1;
              }
          }
        break;

      default:
        Msg( FAT,"Internal Error: Unknown Parameter type !");  
      }

    ptr=ptr->s_parlist;
  }

  /* Aritaet der Aktionen P@call und P@return in spec_actions anpassen */
  MSis_in( trans_mark[MARK_PAR_PRE ]->label->nr, 1, spec_actions )->arity 
    = trans_mark[MARK_PAR_PRE ]->label->arity;
  MSis_in( trans_mark[MARK_PAR_POST]->label->nr, 1, spec_actions )->arity 
    = trans_mark[MARK_PAR_POST]->label->arity;
}


/******************************************************************************/
/* Makro, um den Quelltext uebersichtlicher zu machen.                        */
/******************************************************************************/

#define GetBoxes( name1, name2, name3 )            \
{ GL_AddActions = True;                            \
  new_decl.var_net       = GetBox( name1, False ); \
  GL_AddActions = False;                           \
  new_decl.init_net      = GetBox( name2, False ); \
  new_decl.exit_net      = GetBox( name3, False ); \
}


/******************************************************************************/
/* Umwandeln der Deklarationen eines Blocks in EIN Initialisierungsnetz, EIN  */
/* Variablennetz und EIN Terminierungsnetz.                                   */
/*                                                                            */
/* Falls bei einer Deklarationsart nicht alle drei Netze verwendet werden,    */
/* gibt es keine Konfigurationsdatei und der entsprechende new_decl-Zeiger    */
/* bleibt NULL. Die Prozeduren Sum_Declarators und Scope muessen das berueck- */
/* sichtigen.                                                                 */
/*                                                                            */
/* Die drei Komplettnetze werden von der uebergeordneten Prozedur mit dem     */
/* Kontrollfluss des Blocks zusammengesetzt. In der Theorie wuerde jede       */
/* Deklaration einzeln mit dem Kontrollfluss des Blocks zusammengesetzt.      */
/*                                                                            */
/* Falls das Frontend zwischen den Deklarationen einen Blockbezeichner ver-   */
/* steckt hat, so wird der Eintrag nach *block_entry kopiert.                 */
/******************************************************************************/

decl_type HandleDecls(SymTabEntry *ptr, SymTabEntry* *block_entry)
{ decl_type new_decl;
  decl_type decls    = { NULL, NULL, NULL, NULL };

  while (ptr) {
    if (ptr->s_partype==_NOPAR)
      {
        new_decl.init_net = NULL;
        new_decl.var_net  = NULL;
        new_decl.exit_net = NULL;

        macro_name = ptr->s_name;
        switch(ptr->s_type) 
        {

        case _UDEC: /* kann nur bei GL_Frags == true auftreten:
                       -> undeklarierte Bezeichner ignorieren */
        case MNET_CONST:
            break;

        case _BLOCKNAME:
                  *block_entry = ptr;
            break;

        case _VAR: 
                  macro_decl_set = Set_Parser2Intern( ptr->s_set );
                  macro_ref      = StructToRef( (ShortEntryPtr) ptr );

                  if (ptr->s_init_or_cap!=NONE) 
                    {
                      macro_init = ptr->s_init_or_cap;
                      GetBoxes( INITVAR_BOX, INITVAR_INIT, INITVAR_TERM );
                    }
                  else
                    {
                      macro_init = ptr->s_set->next->low; 
                        /* min(set) als Standard-Initialisierung */
                      GetBoxes( VAR_BOX, VAR_INIT, VAR_TERM );
                    }

                  MYfree( macro_decl_set );

            break;
        case _CHANNEL:
        case _STACK:
                  macro_decl_set = Set_Parser2Intern( ptr->s_set );
                  macro_ref      = StructToRef( (ShortEntryPtr) ptr );

                  if (ptr->s_init_or_cap==0)
                    {
                      GetBoxes( CHAN0_BOX, CHAN0_INIT, CHAN0_TERM );
                    }
                  else if (ptr->s_init_or_cap==1)
                    {
                      GetBoxes( CHAN1_BOX, CHAN1_INIT, CHAN1_TERM );
                    }
                  else if (ptr->s_type==_STACK)
                    {
                      macro_capacity = ptr->s_init_or_cap;
                      macro_capacity_plus_eins = ptr->s_init_or_cap+1;
                      GetBoxes( STACK_BOX, STACK_INIT, STACK_TERM );
                    }
                  else
                    {
                      macro_capacity = ptr->s_init_or_cap;
                      macro_capacity_plus_eins = ptr->s_init_or_cap+1;
                      GetBoxes( CHAN_BOX, CHAN_INIT, CHAN_TERM );
                    }
            MYfree( macro_decl_set );
            break;

        case _PROC:
        case _PPROC:
                  {
                    Integer  *old_macro_id_set  = macro_id_set;
                    Cardinal  old_min_local_blk = min_local_blk;

                    if ( macro_id_set != dot_set )
                      Msg( FAT, "Nested procedures are not allowed !" );

                    /* Referenz des Prozedurnamen vorab eintragen,
                       damit die Reihenfolge der Ref-Nummern stimmt */
                    StructToRef( (ShortEntryPtr) ptr ); 

                    macro_max = ptr->s_init_or_cap;
                    if ( macro_max==NONE )
                      macro_max = max_proc_inst; /* Default-Wert */

                    macro_new_ids = ptr->s_set?
			(Integer*) ptr->s_set : GenerateIdSet( macro_max );

                    RenameRefPars(ptr->s_parlist,ptr->s_name);

                    (Integer*) ptr->s_set = macro_new_ids;
                    macro_id_set = macro_new_ids;
                    min_local_blk = ptr->s_blknum + 1;
                    macro_body = Block2Net((TreeNode*)ptr->s_body);
                    min_local_blk = old_min_local_blk;
                    macro_id_set  = old_macro_id_set;

                    /* einige Makros nochmal neu setzen, da sie waehrend 
                       Block2Net evtl. ueberschrieben wurden. */
                    macro_name = ptr->s_name;
                    macro_ref  = StructToRef( (ShortEntryPtr) ptr );
		    macro_new_ids = (Integer*) ptr->s_set;

                    trans_mark[MARK_PAR_PRE] = trans_mark[MARK_PAR_POST] = NULL;
                    GetBoxes( (flag&USE_PROC_SET) ? PROC_BOX_SET:PROC_BOX, 
                              PROC_INIT, PROC_TERM );

                    if ( new_decl.var_net )
                      {
                        if ( ptr->s_type==_PPROC )
                          {
                            /* Falls eine parametrisierte Prozedur vorliegt,
                               muessen die Markierungen PAR_PRE und PAR_POST
                               in der Schablone vorkommen, weil sonst keine
                               Datenboxen fuer die Parameter angehaengt werden
                               koennen. */
                            if ( !trans_mark[MARK_PAR_PRE] || 
                                 !trans_mark[MARK_PAR_POST] )
                              Msg( FAT, 
                                   "PAR_PRE and/or PAR_POST not specified !" );
                            HandlePars(ptr->s_parlist,&new_decl.var_net);
                          }
                      }
                    else
                      M_Remove_net( &macro_body );
                  }
          break;

        default:
                  Msg( FAT,"Internal Error: Unknown declaration type !");
        }
    
        decls = Sum_Declarators( decls, new_decl );
      }

    ptr=ptr->s_next;
  }

  return decls;
}


/******************************************************************************/
/* Loeschen der IdSets, die fuer Prozeduren durch HandleDecls erzeugt wurden. */
/******************************************************************************/

static void RemoveIdSets( SymTabEntry *ptr )
{
  while (ptr) {
    switch(ptr->s_type) 
      {
        case _PROC:
        case _PPROC:
          {
            MYfree( ptr->s_set );
            ptr->s_set = NULL;
          }
        break;
      }
    ptr=ptr->s_next;
  }
}


/******************************************************************************/
/* Abarbeiten eines expr-Syntaxknotens:                                       */
/*                                                                            */
/* Der Term wird von der TreeNode-Darstellung des Frontends in die            */
/* ArrayOfInteger-Darstellung des Backends umgewandelt.                       */
/******************************************************************************/

#define _entry_  ((SymTabEntry *)node->op1)
#define _null_ ((((SymTabEntry *)node->op1)->s_init_or_cap==0) ? CHAN_NULL : 0)

Integer HandleExpr(TreeNode *node)
{
  Integer n = -1;

  switch ( node->type )
    {
      case N_CONST  : {
                        n = add_node( N_CONST, 
                                      (int) node->op1, (int) node->op2 );
                        break;
                      }
      case N_VAR    : { 
                        switch ( (int)node->op2 ) {
                          case  _PREPRIMED:
                            n = Add_preprimed(  _entry_ );
                            break;
                          case  _UNPRIMED:
                            n = Add_unprimed(   _entry_ );
                            break;
                          case  _POSTPRIMED: 
                            n = Add_postprimed( _entry_ );
                            break;
                          case _READ:
                            n = Add_Channel( _entry_, _null_ );
                            break;
                          case _WRITE:
                            n = Add_Channel( _entry_, _null_ | CHAN_WRITE  );
                            break;
                          case _EMPTY:
                            n = Add_Channel( _entry_, _null_ | CHAN_CHECK );
                            break;
                          case _FULL:
                            n = Add_Channel( _entry_, 
                                             _null_ | CHAN_CHECK | CHAN_WRITE );
                            break;
                        }
                        break;
                      }
      case N_NOT    :
      case N_SIGN   :
      case N_KLAMM  : {
                        n = add_node( node->type, HandleExpr( node->op1 ), 0 );
                        break;
                      }
      case N_OR     :
      case N_AND    :
      case N_EQ     :
      case N_UNEQ   :
      case N_LT     :
      case N_GT     :
      case N_LE     : 
      case N_GE     : 
      case N_PLUS   : 
      case N_MIN    : 
      case N_MULT   : 
      case N_DIV    : 
      case N_MOD    : {
#ifdef UNPRIMED_LOCAL
/* Wenn man UNPRIMED_LOCAL im Makefile setzt, so werden die Implikationen
   bei Verwendung von ungestrichenen Variablen LOKAL im Guard eingefuegt.
   Die Pros und Kontras gegenueber GLOBAL sind in [ACK97] aufgezeigt. */
                        Integer n1, n2;
                        
                        n1 = HandleExpr( node->op1 );
                        if ( node->type == N_OR )
                          n1 = add_UnprimedGuard();

                        n2 = HandleExpr( node->op2 );
                        if ( node->type == N_OR )
                          n2 = add_UnprimedGuard();

                        n = add_node( node->type, n1, n2 );
#else
                        n = add_node( node->type, HandleExpr( node->op1 ), 
                                      HandleExpr( node->op2 ) );
#endif
                        break;
                      }
      default: Msg( FAT, "Internal Error: Unknown node type in expr !" );
    }

  return n;
}


/******************************************************************************/
/* Abarbeitung eines alt_set-Syntaxknotens:                                   */
/*                                                                            */
/* Abweichend von der Theorie wird mit TATSAECHLICH LEEREN repeat- bzw.       */
/* exit-Zweigen gearbeitet. Das Ergebnis ist das selbe, es wird jedoch        */
/* effizienter erzeugt und die Knoten sind besser positioniert.               */
/******************************************************************************/

static alt_set_type AltSet2Net(TreeNode *node, Cardinal *tags)
{
  alt_set_type set;

  switch(node->type) {

  case _REPEAT:
    set.repeat_net = Com2Net(node->op1, tags);
    set.exit_net   = NULL;
    break;

  case _EXIT:
    set.exit_net   = Com2Net(node->op1, tags);
    set.repeat_net = NULL;
    break;

  case _CHOICE:
    {
      alt_set_type set1, set2;

      set1 = AltSet2Net(node->op1, tags);
      set2 = AltSet2Net(node->op2, tags );
      set.repeat_net = M_Choice( set1.repeat_net, set2.repeat_net );
      set.exit_net   = M_Choice( set1.exit_net  , set2.exit_net   );
      break;
    }

  default:
    Msg( FAT,"Internal Error: Unknown node type in alt_set !");
  }

  return set;
}


/*****************************************************************************/
/* Das Teilnetz ref_net (das nur aus Transitionen bestehen darf) wird zu     */
/* *net_ptr hinzugefuegt.                                                    */
/* Die erste Transition von ref_net wird ueber eine (-proc,~id)-Kante        */
/* wechselseitig mit new_pl verbunden.                                       */
/*****************************************************************************/

char          ren_id[]   = { VAR_RENAME_CHAR, 'i', 'd', 0 };
char          min_proc[] = {  VAR_MINOR_CHAR, 'p', 'r', 'o', 'c', 0 };

static void ConnectRefnet( M_Net ref_net, M_Net *net_ptr, M_Place new_pl )
{
  Var          *tpl1, *tpl2;

  ref_net->x_pos = (*net_ptr)->x_pos;
  ref_net->y_pos = (*net_ptr)->y_pos;
  ref_net->x_dim = (*net_ptr)->x_dim;
  ref_net->y_dim = (*net_ptr)->y_dim;
  
  tpl1    = MYmalloc( 3 * sizeof( Var ) );
  tpl1[0] = 3;
  tpl1[1] = StrToCom( min_proc, &(ref_net->Transitions->vars),
                        ref_net->Transitions );
  tpl1[2] = StrToCom(   ren_id, &(ref_net->Transitions->vars),
                        ref_net->Transitions );
  
  tpl2    = MYmalloc( 3 * sizeof( Var ) );
  tpl2[0] = 3;
  tpl2[1] = tpl1[1];
  tpl2[2] = tpl1[2];
  
  /* Die Transition von ref_net muss ganz vorne stehen, 
     damit CreateEdges funktioniert. */
  *net_ptr = M_DisjointNetUnion( ref_net, *net_ptr );
                      
  new_pl->numname=1; /* damit CreateEdges funktioniert */
  M_CreateEdges( new_pl->numname, 
                 V_MScreate_singleton(ArrayToTpl(&tpl1)+V_TUPEL_OFFSET),
                 new_pl->numname, 
                 V_MScreate_singleton(ArrayToTpl(&tpl2)+V_TUPEL_OFFSET),
                 new_pl, *net_ptr, EDGE_INV );
  new_pl->numname=0; /* damit die Positionierung klappt (s.u.) */
}


/******************************************************************************/
/* Fuegt eine Transition fuer die Referenzdurchreiche von Kanaelen bzw.       */
/* Stacks hinzu.                                                              */
/******************************************************************************/

static M_Net Add_ChannelRefBox( SymTabEntry *arg, SymTabEntry *par,
                                Cardinal type )
{
  M_Net         ref_net;

  if ( type & CHAN_WRITE )
    if ( type & CHAN_CHECK )
      macro_ext = "!!";
    else
      macro_ext = "!";
  else
    if ( type & CHAN_CHECK )
      macro_ext = "??";
    else
      macro_ext = "?";

  if ( arg == par )
    ref_net = GetBox( CHANREF_PAR_REC, True );
  else
    ref_net = GetBox( CHANREF_PAR, True );

  Add_ChannelAction( arg, 
                     &(ref_net->Transitions->label), 
                     &(ref_net->Transitions->vars ),
                     &(ref_net->Transitions->guard), type );

  if ( ref_net->Transitions->RefEntries )
    {
      if ( type & CHAN_WRITE )
        if ( type & CHAN_CHECK )
          ref_net->Transitions->ref->flag = FULL_REF;
        else
          ref_net->Transitions->ref->flag = WRITE_REF;
      else
        if ( type & CHAN_CHECK )
          ref_net->Transitions->ref->flag = EMPTY_REF;
        else
          ref_net->Transitions->ref->flag = READ_REF;
    }

  return ref_net;
}


/******************************************************************************/
/* Fuegt eine Stelle fuer eine Referenzdurchreiche zu *net_ptr hinzu.         */
/******************************************************************************/

static void Add_RefPlace( M_Place *pl_ptr, M_Net *net_ptr, 
                          Cardinal *x_dim, SymTabEntry *ptr )
{
  *pl_ptr = M_DuplicatePlace( *net_ptr, place_mark[MARK_REF_PAR], 
                              &stdummy, &tsdummy, &stdummy, &tsdummy );
  *x_dim += (flag&NO_SCOPING) ? 4*RASTER:2*RASTER;

  /* Name des Referenzparameters als Referenz der Stelle setzen */
  /* fixme: S-Listen muessen bestimmt werden koennen */
  (*pl_ptr)->RefEntries   = 0;
  (*pl_ptr)->ref          = MYmalloc( sizeof( Ref ) );
  ((*pl_ptr)->ref)->nr    = StructToRef( (ShortEntryPtr) ptr );
  ((*pl_ptr)->ref)->flag  = 0;
}


/******************************************************************************/
/* Arbeitet einen Syntaxknoten fuer einen Prozeduraufruf ab:                  */
/*                                                                            */
/* Das eingelesene Aufrufnetz wird ggf. um Referenzdurchreichen und           */
/* Beschriftungen fuer die Parameteruebergabe erweitert.                      */
/******************************************************************************/

static void HandleArgs ( SymTabEntry *proc, ArgListEntry *arglist, 
                         M_Net *net_ptr )
{
  ArgListEntry *arg;
  SymTabEntry  *par;

  MultiSet      return_actions = MScreate_empty_set();

  Cardinal      x_dim = 0;
  int           x_pos;
  M_Place       new_pl = NULL;
  M_Net         ref_net;
  Action       *act;
  ComVar        com;
  Integer       prev, last;
  TreeNode      node = { N_VAR, NULL, (TreeNode *) _UNPRIMED }, *expr;

  M_STEdge      st;
  Cardinal      offset;

  arg   = arglist;
  par   = proc->s_parlist;
  while ( par ) 
    {
      if ( par->s_partype == _REF )
        {
          Add_RefPlace( &new_pl, net_ptr, &x_dim, par );

          macro_name     = par->s_name;
          macro_decl_set = Set_Parser2Intern( par->s_set );
          macro_arg      = arg->ident->s_name;
          macro_ref      = StructToRef( (ShortEntryPtr) par );

          GL_AddActions  = False;

          /* Im Spezialfall von rekursiven Referenz-Parametern kann man
             einige Vereinfachungen vornehmen...
             ( Achtung: Die defs MUESSEN geloescht werden, weil man sonst 
             mehrfache Definitionen fuer die selben Variablen haette. ) */

          if ( arg->ident->s_type == _VAR )
            {
              /* Referenz auf Datenvariable */

              if ( arg->ident == par )
                ref_net = GetBox( VARREF_PAR_REC, True );
              else
                ref_net = GetBox( VARREF_PAR, True );

              Add_VariableAction( arg->ident, 
                                  &(ref_net->Transitions->label), 
                                  &(ref_net->Transitions->vars ),
                                  &(ref_net->Transitions->guard) );
              
              ConnectRefnet( ref_net, net_ptr, new_pl );          
            }
          else if ( par->s_init_or_cap )
            {
              /* Referenz auf Kanal / Stack mit Kapazitaet > 0 */

              ref_net = Add_ChannelRefBox( arg->ident, par, 0 );
              ConnectRefnet( ref_net, net_ptr, new_pl );          
              ref_net = Add_ChannelRefBox( arg->ident, par, CHAN_WRITE );
              ConnectRefnet( ref_net, net_ptr, new_pl );          
              ref_net = Add_ChannelRefBox( arg->ident, par, CHAN_CHECK | 0 );
              ConnectRefnet( ref_net, net_ptr, new_pl );          
              ref_net = Add_ChannelRefBox( arg->ident, par, CHAN_CHECK | 
                                                            CHAN_WRITE );
              ConnectRefnet( ref_net, net_ptr, new_pl );          
            }
          else
            {
              /* Referenz auf Kanal der Kapazitaet 0:

                 Lese- und Schreibdurchreiche muessen jeweils einen eigenen
                 RefPlace bekommen. */

              macro_ext = "?";
              if ( arg->ident == par )
                ref_net = GetBox( CHAN0REF_PAR_REC, True );
              else
                ref_net = GetBox( CHAN0REF_PAR, True );

              ref_net->Transitions->ref->flag = READ_REF;
              
              Add_ChannelAction( arg->ident, 
                                 &(ref_net->Transitions->label), 
                                 &(ref_net->Transitions->vars ),
                                 &(ref_net->Transitions->guard), 
                                 CHAN_NULL );

              ConnectRefnet( ref_net, net_ptr, new_pl );          


              Add_RefPlace( &new_pl, net_ptr, &x_dim, par );
            
              macro_ext = "!";
              if ( arg->ident == par )
                ref_net = GetBox( CHAN0REF_PAR_REC, True );
              else
                ref_net = GetBox( CHAN0REF_PAR, True );

              ref_net->Transitions->ref->flag = WRITE_REF;

              Add_ChannelAction( arg->ident, 
                                 &(ref_net->Transitions->label), 
                                 &(ref_net->Transitions->vars ),
                                 &(ref_net->Transitions->guard),
                                 CHAN_NULL | CHAN_WRITE );
              
              ConnectRefnet( ref_net, net_ptr, new_pl );          
            }

          MYfree( macro_decl_set );
        }
      else if (par->s_partype == _VAL || par->s_partype == _RES)
        {
		/* falls Argument und Parameter ein und dasselbe sind,
		   werden beim Aufruf zwei Schreibzugriffe auf die
		   Werte-Stelle dieses Arguments erfolgen.
		   Beispiel: proc R(val nr) begin R(nr) end */
/*		if (arg->ident == par)
*/		if (max_arc_vars < 2) max_arc_vars = 2;
	}
      arg = arg->next;
      par = par->s_parlist;
    }

  /* Wurden Referenz-Stellen erzeugt ? */
  if ( new_pl )
    {
      /* urspruengliche Stelle im Aufrufnetz entfernen */
      M_RemovePlace( *net_ptr, place_mark[MARK_REF_PAR] );

      x_dim-=RASTER;
      x_pos=(int)((*net_ptr)->x_dim-(x_dim))/2+(*net_ptr)->x_pos;
      if (x_dim>(*net_ptr)->x_dim) {
        (*net_ptr)->x_dim=x_dim;
        (*net_ptr)->x_pos=x_pos;
      }
      
      /* alle Referenzstellen mittig im Aufrufnetz plazieren */
      x_pos+=x_dim-( (flag&NO_SCOPING) ? 3*RASTER:RASTER );
      while (!new_pl->numname) {
        x_pos=(new_pl->x_pos=x_pos)
          -( (flag&NO_SCOPING) ? 4*RASTER:2*RASTER );
        new_pl->numname=1;
        st     = new_pl->postset;
        offset = 0;
        while ( st->nexttrans )
          {
            st->trans->x_pos = offset + new_pl->x_pos + 2*RASTER;
            st->trans->y_pos = offset + new_pl->y_pos;
            offset          += 15;
            st               = st->nexttrans;
          }
        new_pl=new_pl->nextplace;
      }
    }


  init_ExpressionParser( &(trans_mark[MARK_CALL_PAR]->vars) );
  
  /* Aufruf- und Rueckkehrtransition um Aktionsterme und Wertterme fuer die
     Uebergabe und Rueckgabe der Argumente erweitern */
  arg   = arglist;
  par   = proc->s_parlist;
  while ( par ) 
    {
      if ( par->s_partype != _REF ) /* par->s_partype in {_VAL,_CONSTPAR,_RES}*/
        {
          if ( arg->ident )
            {
              node.op1 = (TreeNode *) arg->ident;
              expr = &node;
            }
          else 
            expr = arg->arg;

          com = StrToCom( par->s_name, &expr_vars, NULL );
          trans_mark[MARK_CALL_PAR]->label->
            com[ trans_mark[MARK_CALL_PAR]->label->arity++ ] = com;

          /* 'Parametername = expr' zu expr_tree hinzufuegen
             (im N_EQ-Knoten wird im Knotentyp vorlaeufig die Knotenposition
              des vorherigen N_EQ-Knotens zwischengespeichert -> siehe unten) */
          last = expr_tree->op1 - 1;
          add_node( last, add_node( N_VAR, com, -1 ), HandleExpr( expr ) );
                  
          if ( par->s_partype == _RES )
            {
              act = Add_VariableAction( arg->ident, &return_actions,
                                        &(trans_mark[MARK_RETURN_PAR]->vars ),
                                        &(trans_mark[MARK_RETURN_PAR]->guard) );

              trans_mark[MARK_RETURN_PAR]->label->
                com[ trans_mark[MARK_RETURN_PAR]->label->arity++ ] =act->com[1];
            }
        }
      arg = arg->next;
      par = par->s_parlist;
    }

  /* Falls mehrere Wertterme der Form 'Parametername = expr' erzeugt wurden,
     so werden N_AND-Knoten angehaengt, die diese Wertterme verbinden. 
     Um das effizient zu machen, wird der manipulierte Knotentyp der N_EQ-
     Knoten verwendet. */
  last = expr_tree->op1 - 1;
  if ( last )
    {
      while ( (prev = expr_tree[last].typ) )
        {
          expr_tree[last].typ = N_EQ; /* manipulierten Knotentyp korrigieren */
          add_node( N_AND, prev, expr_tree->op1 - 1 );
          last = prev;
        }
      expr_tree[last].typ = N_EQ;
    }

  exit_ExpressionParser();

  /* urspruenglichen Guard der CALL-Transition ersetzen */  
  MYfree( trans_mark[MARK_CALL_PAR]->guard );
  trans_mark[MARK_CALL_PAR]->guard = expr_tree;

  /* urspruengliche Definitionen der CALL-Transition ersetzen 
     (Ist eigentlich ueberfluessig, da die Defs vorher und nachher jeweils leer
      sind, muss aber aus Gruenden der Speicherverwaltung gemacht werden. */
  MYfree( trans_mark[MARK_CALL_PAR]->def );
  trans_mark[MARK_CALL_PAR]->def = expr_def;

  /* leeres ComArray der Transition ersetzen */  
  trans_mark[MARK_CALL_PAR]->vars.ComEntries = expr_vars.ComEntries;
  trans_mark[MARK_CALL_PAR]->vars.ComArray   = expr_vars.ComArray;

  /* Aktionsterme in action_set zum Label der CALL-Transition hinzufuegen */
  MSresize( trans_mark[MARK_CALL_PAR]->label, 
            MSlength( trans_mark[MARK_CALL_PAR]->label ) + 
            MSlength( action_set ) );
  MSsum( trans_mark[MARK_CALL_PAR]->label, action_set );
  MYfree( action_set );

  Modify_Transition( trans_mark[MARK_CALL_PAR] );

  /* Aktionsterme in return_actions zum Label der RETURN-Trans. hinzufuegen */
  MSresize( trans_mark[MARK_RETURN_PAR]->label, 
            MSlength( trans_mark[MARK_RETURN_PAR]->label ) +
            MSlength( return_actions ) );
  MSsum( trans_mark[MARK_RETURN_PAR]->label, return_actions );
  MYfree( return_actions );
}


/******************************************************************************/
/* Ein com-Syntaxknoten wird abgearbeitet.                                    */
/******************************************************************************/

static M_Net Com2Net(TreeNode *node, Cardinal *tags)
{
  M_Net net = NULL, n1, n2;

  switch(node->type) {

    case _BLOCK:
      net = Block2Net(node);
    break;

    case _EXPR:
      init_ExpressionParser( NULL );
      HandleExpr(node->op1);
      exit_ExpressionParser();
      net = M_CreateBasicBox( action_set, expr_tree, expr_def, &expr_vars, 
                              StructToRef((ShortEntryPtr)node->op2) );
      Modify_Transition( net->Transitions );
    break;

    case _SEQ:
      n1  = Com2Net(node->op1, tags);
      n2  = Com2Net(node->op2, tags);
      net = M_Sequence( n1, n2 );
    break;

    case _PAR:
      *tags |= BLOCK_PARALLEL;
      n1  = Com2Net(node->op1, tags);
      n2  = Com2Net(node->op2, tags);
      net = M_ConcurrentComposition( n1, n2 );
    break;

    case _DO: 
      {
        alt_set_type set;
        M_Net        enter_net = (node->op2) ? Com2Net(node->op2, tags) : NULL;

        set = AltSet2Net(node->op1, tags);

        if ( !set.exit_net )
          set.exit_net   = M_OnlyPlaces(0);
        net = M_Iteration( set, enter_net );
        break;
      }

    case _PCALL:
    case _CALL:
      {
        macro_name    =            ((SymTabEntry*)(node->op1))->s_name;

        macro_new_ids = (Integer*) ((SymTabEntry*)(node->op1))->s_set;
	if (!macro_new_ids)
	{
		macro_new_ids = GenerateIdSet(
			((SymTabEntry*)(node->op1))->s_init_or_cap);
        	((SymTabEntry*)(node->op1))->s_set = (SetRange*) macro_new_ids;
	}
		
        macro_ref     = StructToRef( *((ShortEntryPtr*)(node+1)) );
        GL_AddActions = False;
        net = GetBox( PROC_CALL, True );

        /* Das Aufrufnetz muss eine Aufruf- und eine Rueckkehrtransition 
           besitzen, weil sonst die Parameteruebergbabe nicht eingerichtet 
           werden kann. */
        if ( !trans_mark[MARK_CALL_PAR  ] || 
             !trans_mark[MARK_RETURN_PAR] )
          Msg( FAT, "CALL_PAR and/or RETURN_PAR not specified !" );
        HandleArgs((SymTabEntry*)(node->op1), (ArgListEntry*)(node->op2), &net);
      }
    break;

    default:
      Msg( FAT,"Internal Error: Unknown node type in com !");
  }

  return net;
}

/******************************************************************************/
/* Ein block-Syntaxknoten wird abgearbeitet:                                  */
/*                                                                            */
/* Die Deklarationen werden abgearbeitet, der Kontrollfluss wird erzeugt, das */
/* Scoping wird durchgefuehrt und das Netz wird in ein neues Block-Element    */
/* eingebettet.                                                               */
/******************************************************************************/

M_Net Block2Net(TreeNode *node)
{
  M_Net        net;
  decl_type    decls;
  SymTabEntry *block_entry = NULL;
  BlockNr      bl;
  RefNr        refnr = 0;
  MultiSet     old_spec_actions;
  Cardinal     tags = 0;

  block_num++; /* Schachtelungstiefe der Bloecke mitzaehlen */

  /* ref ist nicht vorhanden, falls GL_Frags und kein BEGIN am Anfang */
  if ( *((ShortEntryPtr*)(node+1)) ) 
    refnr = StructToRef( *((ShortEntryPtr*)(node+1)) );
                              /* Referenz des Blocknamennamen vorab eintragen,
                                 damit die Reihenfolge der Ref-Nummern stimmt */


  /* Da Bloecke geschachtelt sein koennen, kann diese Prozedur indirekt rekursiv
     aufgerufen werden. Deshalb muss die globale Variable spec_actions
     zwischengespeichert und restauriert werden. */
  old_spec_actions = spec_actions;
  spec_actions = MScreate_empty_set();
  node->op1=(TreeNode *)invert_list((SymTabEntry *)(node->op1));
  decls = HandleDecls((SymTabEntry *)node->op1,&block_entry);
  decls.scope_actions = spec_actions;
  spec_actions = old_spec_actions;

  /* com ist nicht vorhanden, falls nur decls & GL_Frags */
  if( node->op2 ) 
    {
      net = Com2Net( node->op2, &tags );
      if ( decls.var_net )
        net = Scope( net, decls );
    }
  else 
    { 
      net = decls.var_net;
      M_Remove_net( &decls.exit_net );
    }

  MSremove( decls.scope_actions );
  if ( GL_ProcDecl )
    RemoveIdSets( (SymTabEntry *)node->op1 );

  if ( refnr ) 
    {
      /* Eingangs-/Ausgangsstellen mit PRE-/POST-Referenz des Blocks versehen */
      M_AddAllPlaceRefs(net, ENTRY, refnr,  PRE_REF);
      M_AddAllPlaceRefs(net, EXIT,  refnr, POST_REF);
    }

  /* Block hinzufuegen */
  if (block_entry)
    add_block(block_entry, &bl, refnr, tags);
  else
    add_unnamed_block(block_num, &bl, refnr, tags);
  M_AddAllPlacesBlock(net, bl);
  M_AddAllTransBlock(net, bl);
  M_AddAllPTransBlock(net, bl);
  M_AppendBlock(net, bl);

  block_num--;

  return net;
}


/******************************************************************************/
/* END OF FILE evaltree.c                                                     */
/******************************************************************************/
