/****************************************************************************/
/* mnet.c                                                                   */
/*                                                                          */
/* elementare Prozeduren zum Umgang mit M-Netzen                            */
/*                                                                          */
/* Burkhard Graves, 26.6.'93                                                */
/* Martin Ackermann, Sommer 95                                              */
/* Burkhard Bieber 1995:                                                    */
/*   - Blocktypen ergaenzt                                                  */
/*   - Sync-/Phantom-Transitionen ergaenzt                                  */
/* Martin Ackermann, Winter 1996/97                                         */
/*   - Prozeduren, andere Referenzen, mehr Effizienz, Fehler beseitigt      */
/****************************************************************************/

  /************/
  /* includes */
  /************/

#include <string.h>
#include "mnet.h"
#include "block.h"
#include "tables.h"
#include "multiset.h"
#include "renaming.h"
#include "annotat.h"
#include "ptrans.h"
#include "evaltree.h"
#include "f_parse.h"
#include "macros.h"

#define WARNING_P_NUM( FUNCT, P_NUMNAME ) \
  fprintf( stderr, \
  "WARNING: NN.%s(): a place numbered %u does not exist\n", FUNCT, P_NUMNAME )

#define WARNING_T_NUM( FUNCT, T_NUMNAME ) \
   fprintf( stderr, \
   "WARNING: NN.%s(): a transition numbered %u does not exist\n", FUNCT, T_NUMNAME )

  /*********************/
  /* globale Variablen */ 
  /*********************/

M_STEdge   stdummy;
M_TSEdge   tsdummy;

Cardinal32 trans_anz = 0;       /* Gesamtanzahl der Transitionen im Netz    */
char     entry_type[20]="{dot}";
Cardinal32 max_trans_anz=INT_MAX;  /* Maximal-Anzahl von Transitionen */


/* Wenn REF_STATISTIC definiert ist, so wird fuer jede Referenz ermittelt
   und ausgegeben, bei welchen Netzelementen sie eingetragen ist. */
typedef struct StatisticEntry_ {
  Cardinal32 element;           /* Zeiger auf das Netzelement */
  bool    element_is_place;  /* zur Unterscheidung Stellen<->Transitionen */
  Cardinal   flag;              /* flag, mit element die Referenz traegt */
} StatisticEntry;

typedef struct RefStatistic_ {
  Cardinal32      entries;      /* Anzahl der Eintraege in der RefStatistic */
  StatisticEntry *use;          /* dyn. Array mit Statistik-Eintraegen */
} RefStatistic;

#ifdef REF_STATISTIC
RefStatistic *statistic;    /* Statistik ueber die Benutzung von Referenzen */
#endif
char *ref_file=NULL;

/********************************************/
/* Alloziert Speicher fuer eine Transition. */
/********************************************/

M_Transition M_allocTrans(void)
{
  M_Transition tr;

  trans_anz++;
  if (trans_anz > max_trans_anz) {
    Msg( FAT, "Maximum number of transitions exceeded !") ;
  }  /* if zuviele Transitionen (Kommandozeilenvorgabe) */
  tr = MYmalloc(sizeof(M_TrElem));
  return tr;
}

/*******************************************/
/* Kopiert die Referenzen von pl1 nach pl2 */
/*******************************************/

static void M_CopyPlaceReferences( M_Place pl2, M_Place pl1  )
{       
  pl2->RefEntries = pl1->RefEntries;
  if ( pl2->RefEntries )
    {
      pl2->ref = MYmalloc( pl2->RefEntries * sizeof( Ref ) );
      memcpy( pl2->ref, pl1->ref, pl2->RefEntries * sizeof( Ref ) );
    }
}

/***************************************/
/* Loescht ein komplettes Kreuzprodukt */
/***************************************/

static void M_RemoveRange( M_Place pl )
{
  Cardinal i;
  for ( i=0; pl->range[i] && i<MAX_DIM; i++ )
    if ( pl->range[i] != EMPTY_SET )
      MYfree(pl->range[i]);
}

/*****************************/
/* Kopiert eine Einzelmenge  */
/*****************************/

void M_CopySingleRange( Integer* *dest_ptr, Integer *src )
{
  size_t anz_bytes;

  if ( src == EMPTY_SET ) 
    *dest_ptr = EMPTY_SET;
  else
    {
      anz_bytes = *src * sizeof( Integer );
      *dest_ptr = MYmalloc( anz_bytes );
      memcpy( *dest_ptr, src, anz_bytes );
    }
}

/***************************************/
/* Kopiert ein komplettes Kreuzprodukt */
/***************************************/

void M_CopyRange( M_Place dest, M_Place src )
{
  Cardinal i;

  for ( i=0; src->range[i] && i<MAX_DIM; i++ )
    M_CopySingleRange( dest->range+i, src->range[i] );
  if ( i<MAX_DIM )
    dest->range[i] = NULL;
}

/********************************/
/* Vergleicht zwei Einzelmengen */
/********************************/

bool M_SingleRangesAreEqual( Integer *r1, Integer *r2 )
{
  Cardinal i;
  for(i=0; i < (Cardinal) r1[0]; i++)
    {
      if ( r1[i] != r2[i] )
        return False;
    }
  return True;
}

/*******************************************/
/* Vergleicht zwei komplette Kreuzprodukte */
/*******************************************/

bool M_RangesAreEqual( M_Place pl1, M_Place pl2 )
{
  Cardinal i; 

  for ( i=0; pl1->range[i] && pl2->range[i] && i<MAX_DIM; i++ )
    if ( !M_SingleRangesAreEqual( pl1->range[i], pl2->range[i] ) )
      return False;

  return ( i==MAX_DIM || (!pl1->range[i] && !pl2->range[i]) );
}

/*************************************************************/
/* Liefert True, falls die Stelle den Wertebereich {dot} hat */
/*************************************************************/

static bool M_RangeIsDot( M_Place pl )
{
  if ( pl->range[1] || (pl->range[0])[0]!=2 || (pl->range[0])[1]!=DOT )
    return False;
  else
    return True;
}

/*****************************************************************************/
/* Entfernt Stellen- und Transitionsduplikate aus dem Netz 'net'.            */
/* In c_places bzw. c_transitions wird die Anzahl der Entfernungen notiert.  */
/*****************************************************************************/

void M_RemoveDuplicates(M_Net net,Cardinal32 *c_places,
                                  Cardinal32 *c_transitions,
                                  Cardinal32 *c_places_phant)
{
#ifdef TRANS_DUPLICATES
    register M_Transition tr1, tr2;
#endif
    register M_Place      pl1;
    register M_TSEdge     ts;
    register M_STEdge     st;
             M_Place      pl2;
             M_TSEdge     ts1,ts2;
             M_STEdge     st1,st2;
             bool      is_dot;

    *c_transitions=*c_places=*c_places_phant=0;
#ifdef TRANS_DUPLICATES
    tr1=net->Transitions;
    while (tr1) {
        if ( tr1->preset ) { /* Transition mit Vorbereich ? */
            for(st=tr1->preset->place->postset; st; st=st->nexttrans) {
                tr2=st->trans;
                if (tr2!=tr1 && Tr_are_equal(tr1,tr2)) {
                  M_RemoveTransition(net,tr2,False);
                    ++*c_transitions;
                }
            }
        }
        else if ( tr1->postset ) { /* Transition mit Nachbereich ? */
            for(ts=tr1->postset->place->preset; ts; ts=ts->nexttrans) {
                tr2=ts->trans;
                if (tr2!=tr1 && Tr_are_equal(tr1,tr2)) {
                  M_RemoveTransition(net,tr2,False);
                    ++*c_transitions;
                }
            }
        }
        else { /* isolierte Transition */
            for(tr2 = tr1->nexttrans; tr2; tr2=tr2->nexttrans) {
                if (tr2!=tr1 && Tr_are_equal(tr1,tr2)) {
                  M_RemoveTransition(net,tr2,False);
                    ++*c_transitions;
                }
            }
        }
        tr1=tr1->nexttrans;
    }
#endif

    /* Es werden nur Stellen mit annotation {dot} entfernt,    */
    /* wobei die Kantenbeschriftung keine Rolle spielt !       */
    /*                                                         */
    /* Wenn an Stellenduplikaten verschiedene Phantom-         */
    /* transitionen haengen, dann wird kein Duplikat entfernt. */
    /*                                                         */
    /* Die Fallunterscheidung [isoliert|nur Vorbereich|...]    */
    /* findet lediglich aus Effizienzgruenden statt, damit     */
    /* nicht komplett alle Stellen verglichen werden muessen.  */

    /* Markierung im numname notwendig um Doppel-  */
    /* zaehlungen bei c_places_phant zu vermeiden  */
    for(pl1=net->Places; pl1; pl1=pl1->nextplace)
      pl1->numname=1;

    pl1=net->LastPlace;
    while (pl1) {

      /* if ( M_RangeIsDot( pl1 ) ) */
      {
        is_dot = M_RangeIsDot( pl1 );

        if ((st=pl1->postset)==NULL) {
            if((ts=pl1->preset)==NULL) {
                /* isolierte Stelle */
                pl2=pl1->prevplace;
                while (pl2) {
                    if (!pl2->preset && !pl2->postset
                                     && M_RangesAreEqual( pl1, pl2 )
                                     && pl1->label==pl2->label) {
                        pl2=M_RemovePlace(net,pl2);
                        ++*c_places;
                    } else {
                        pl2=pl2->prevplace;
                    }
                }
                pl1=pl1->prevplace;
                continue;
            }
            /* Stelle ohne Nachbereich, aber mit Vorbereich */
            ts=ts->trans->postset;
            while (ts) {
                pl2=ts->place;
                ts=ts->nextplace;
                if (pl2!=pl1 && pl1->label==pl2->label
                             && M_RangesAreEqual( pl1, pl2 )
                             && pl2->postset==NULL) {
                    /* Test, ob preset uebereinstimmt */
                    ts1=pl1->preset; ts2=pl2->preset;
                    while (ts1 && ts2 && ts1->trans==ts2->trans &&
 ( (  is_dot && V_MScardinality(ts1->label) == V_MScardinality(ts2->label)) || 
                V_MSis_equal(ts1->label, ts2->label) ) ) {
                        ts1=ts1->nexttrans;
                        ts2=ts2->nexttrans;
                    }
                    if (!ts1 && !ts2) {

		        if (flag & PHANTOM_TRANS_OUTPUT) {
			  /* Test, ob Phantom-preset uebereinstimmt */
			  ts1=pl1->ppreset; ts2=pl2->ppreset;
			  while (ts1 && ts2 && ts1->trans==ts2->trans &&
 ( (  is_dot && V_MScardinality(ts1->label) == V_MScardinality(ts2->label)) || 
                V_MSis_equal(ts1->label, ts2->label) ) ) {
			    ts1=ts1->nexttrans;
			    ts2=ts2->nexttrans;
			  }
                        }

                        if (!ts1 && !ts2) {
                          if (!pl2->numname)
                            --*c_places_phant;
                          M_RemovePlace(net,pl2);
                          ++*c_places;
                        }
                        else {
                          if (pl1->numname || pl2->numname)
                            ++*c_places_phant;
                          pl1->numname=pl2->numname=0;
                        }
                    }
                }
            }
            pl1=pl1->prevplace;
            continue;
        }
        /* Stelle sicher mit Nachbereich, evtl. mit Vorbereich */
        st=st->trans->preset;
        while (st) {
            pl2=st->place;
            st=st->nextplace;
            if (pl2!=pl1 && pl1->label==pl2->label
                         && M_RangesAreEqual( pl1, pl2 ) ) {
                /* Test, ob preset uebereinstimmt */
                ts1=pl1->preset; ts2=pl2->preset;
                while (ts1 && ts2 && ts1->trans==ts2->trans &&
 ( (  is_dot && V_MScardinality(ts1->label) == V_MScardinality(ts2->label)) || 
                V_MSis_equal(ts1->label, ts2->label) ) ) {
                    ts1=ts1->nexttrans;
                    ts2=ts2->nexttrans;
                }
                if (!ts1 && !ts2) {
                    /* Test, ob postset uebereinstimmt */
                    st1=pl1->postset; st2=pl2->postset;
                    while (st1 && st2 && st1->trans==st2->trans &&
 ( (  is_dot && V_MScardinality(st1->label) == V_MScardinality(st2->label)) || 
                V_MSis_equal(st1->label, st2->label) ) ) {
                        st1=st1->nexttrans;
                        st2=st2->nexttrans;
                    }
                    if (!st1 && !st2) {

		        if (flag & PHANTOM_TRANS_OUTPUT) {
			  /* Test, ob Phantom-preset uebereinstimmt */
			  ts1=pl1->ppreset; ts2=pl2->ppreset;
			  while (ts1 && ts2 && ts1->trans==ts2->trans &&
 ( (  is_dot && V_MScardinality(ts1->label) == V_MScardinality(ts2->label)) || 
                V_MSis_equal(ts1->label, ts2->label) ) ) {
			    ts1=ts1->nexttrans;
			    ts2=ts2->nexttrans;
			  }
			}

                        if (!ts1 && !ts2) {

			  if (flag & PHANTOM_TRANS_OUTPUT) {
			    /* Test, ob Phantom-postset uebereinstimmt */
			    st1=pl1->ppostset; st2=pl2->ppostset;
			    while (st1 && st2 && st1->trans==st2->trans &&
 ( (  is_dot && V_MScardinality(st1->label) == V_MScardinality(st2->label)) || 
                V_MSis_equal(st1->label, st2->label) ) ) {
			      st1=st1->nexttrans;
			      st2=st2->nexttrans;
			    }
			  }

                          if (!st1 && !st2) {
                            if (!pl2->numname)
                              --*c_places_phant;
                            M_RemovePlace(net,pl2);
                            ++*c_places;
                          } else {
                            if (pl1->numname || pl2->numname)
                              ++*c_places_phant;
                            pl1->numname=pl2->numname=0;
                          }
                        } else {
                          if (pl1->numname || pl2->numname)
                            ++*c_places_phant;
                          pl1->numname=pl2->numname=0;
                        }
                    }
                }
            }
        }

      } /* if */

      pl1=pl1->prevplace;
    }
}

/*****************************************************************************/
/* Erstellt am Anfang der Transitionenliste des Netzes 'net' eine disjunkte  */
/* Kopie der Transition 'tr' mitsamt allen ein- und ausgehenden Kanten;      */
/* 'numname' wird als Kennzeichen einer Kopie auf 0 gesetzt,                 */
/* 'label' werden auf NULL gesetzt.                                          */
/* Rueckgabewert ist ein Zeiger auf die neue Transition, ausserdem wird in   */
/* '*stlast' bzw. '*tslast' ein Zeiger auf die letzte Kante des Presets bzw. */
/* des Postsets abgelegt.                                                    */
/*****************************************************************************/

M_Transition M_DuplicateTrans(M_Net net,M_Transition tr,
                              M_STEdge *stlast,M_TSEdge *tslast,ComVar *sub)
{
    M_Transition newtr=NULL;

    if (net && tr) {
        ++net->NumTransitions;
	newtr=M_allocTrans();
        (newtr->nexttrans=net->Transitions)->prevtrans=newtr;
        (net->Transitions=newtr)->prevtrans=NULL;
        newtr->numname=0;
        newtr->label=NULL;
        newtr->guard=NULL;
        newtr->def=NULL;
        newtr->RefEntries=0;
        newtr->meaning=NULL;
        newtr->links=NULL;
        newtr->time_label=NULL;
        newtr->BlockEntries = 0;
        newtr->blocks = NULL;
        newtr->PhantomEntries=0;
        newtr->flags=tr->flags;
        newtr->x_pos=tr->x_pos;
        newtr->y_pos=tr->y_pos;
        newtr->preset=M_CopyTransPreset(tr,newtr,stlast,sub,False);
        newtr->postset=M_CopyTransPostset(tr,newtr,tslast,sub,False);
    } else {
        Msg( FAT,"DuplicateTrans: no net or no transition given !");
    }
    return newtr;
}

/*****************************************************************************/
/* Alle in die Transition 'tr' eingehenden Kanten werden in der Kantenmatrix */
/* direkt vor die Stelle dupliziert, von der sie ausgehen. Als Transitions-  */
/* zeiger wird diesen Kanten (ein Zeiger auf die neu in die Transitionenliste*/
/* eingefuegte Transition) 'newtr' zugewiesen.                               */
/* Rueckgabewert ist ein Zeiger auf die erste Kante, '*stlast' wird ein      */
/* Zeiger auf die letzte Kante zugewiesen.                                   */
/*****************************************************************************/

M_STEdge M_CopyTransPreset(M_Transition tr,M_Transition newtr,
                           M_STEdge *stlast,ComVar *sub,bool is_phant)
{
    register M_STEdge st,newst;
    M_STEdge          stfirst;

    stfirst=*stlast=NULL;
    if (tr && newtr && ((st=tr->preset)!=NULL)) {
        (stfirst=newst=M_allocSTEdge())->prevplace=NULL;
        while (newst) {
            newst->prevtrans=NULL;
            newst->label=V_MScreate_copy(st->label, sub);
            newst->flags=st->flags;
            newst->trans=newtr;
            newst->place=st->place;
            if (is_phant)
              {
                newst->nexttrans=st->place->ppostset;
                st->place->ppostset=newst;
              }
            else
              {
                newst->nexttrans=st->place->postset;
                st->place->postset=newst;
              }
            if (newst->nexttrans)
              newst->nexttrans->prevtrans=newst;
            newst=((st=st->nextplace)!=NULL)?
                  ((newst->nextplace=M_allocSTEdge())
                   ->prevplace=newst)->nextplace
                  : ((*stlast=newst)->nextplace=NULL);
        }
    }
    return stfirst;
}

/*****************************************************************************/
/* Alle von der Transition 'tr' ausgehenden Kanten werden in der Kantenmatrix*/
/* direkt vor die Stelle dupliziert, in der sie enden. Als Transitionszeiger */
/* wird diesen Kanten (ein Zeiger auf die neu in die Transitionenliste ein-  */
/* gefuegte Transition) 'newtr' zugewiesen.                                  */
/* Rueckgabewert ist ein Zeiger auf die erste Kante, '*stlast' wird ein      */
/* Zeiger auf die letzte Kante zugewiesen.                                   */
/*****************************************************************************/

M_TSEdge M_CopyTransPostset(M_Transition tr,M_Transition newtr,
                            M_TSEdge *tslast,ComVar *sub, bool is_phant)
{
    register M_TSEdge ts,newts;
    M_TSEdge          tsfirst;

    tsfirst=*tslast=NULL;
    if (tr && newtr && ((ts=tr->postset)!=NULL)) {
        (tsfirst=newts=M_allocTSEdge())->prevplace=NULL;
        while (newts) {
            newts->prevtrans=NULL;
            newts->label=V_MScreate_copy(ts->label, sub);
            newts->flags=ts->flags;
            newts->trans=newtr;
            newts->place=ts->place;
            if (is_phant)
              {
                newts->nexttrans=ts->place->ppreset;
                ts->place->ppreset=newts;
              }
            else
              {
                newts->nexttrans=ts->place->preset;
                ts->place->preset=newts;
              }
            if (newts->nexttrans)
              newts->nexttrans->prevtrans=newts;
            newts=((ts=ts->nextplace)!=NULL)?
                  ((newts->nextplace=M_allocTSEdge())
                   ->prevplace=newts)->nextplace
                  : ((*tslast=newts)->nextplace=NULL);
        }
    }
    return tsfirst;
}

/***************************************************************************/
/* Entfernen der Stelle 'pl' aus dem Netz 'net' zusammen mit allen Kanten, */
/* die in/aus diese/r Stelle fuehren (inkl. Phantomkanten).                */
/* Rueckgabewert ist 'pl->prevplace'.                                      */
/***************************************************************************/

M_Place M_RemovePlace(M_Net net,M_Place pl)
{
  register M_STEdge st, stdel;
  register M_TSEdge ts, tsdel;
  M_Place prevplace=NULL;

  if (net && pl) {
      /* Eingangskanten wech */
    ts = pl->preset;
    while (ts) {
      if (ts->prevplace)
        ts->prevplace->nextplace = ts->nextplace;
      else
        ts->trans->postset = ts->nextplace;
      if (ts->nextplace)
        ts->nextplace->prevplace = ts->prevplace;
      ts = (tsdel = ts)->nexttrans;
      M_removeTSEdge(tsdel);
    }  /* while noch Eingangskanten */
      /* Ausgangskanten wech */
    st = pl->postset;
    while (st) {
      if (st->prevplace)
        st->prevplace->nextplace = st->nextplace;
      else
        st->trans->preset = st->nextplace;
      if (st->nextplace)
        st->nextplace->prevplace = st->prevplace;
      st = (stdel = st)->nexttrans;
      M_removeSTEdge(stdel);
    }  /* while noch Ausgangskanten */
      /* Phantomeingangskanten wech */
    ts = pl->ppreset;
    while (ts) {
      if (ts->prevplace)
        ts->prevplace->nextplace = ts->nextplace;
      else
        ts->trans->postset = ts->nextplace;
      if (ts->nextplace)
        ts->nextplace->prevplace = ts->prevplace;
      ts = (tsdel = ts)->nexttrans;
      M_removeTSEdge(tsdel);
    }  /* while noch Phantomeingangskanten */
      /* Phantomausgangskanten wech */
    st = pl->ppostset;
    while (st) {
      if (st->prevplace)
        st->prevplace->nextplace = st->nextplace;
      else
        st->trans->preset = st->nextplace;
      if (st->nextplace)
        st->nextplace->prevplace = st->prevplace;
      st = (stdel = st)->nexttrans;
      M_removeSTEdge(stdel);
    }  /* while noch Phantomausgangskanten */
      /* Stelle aus der Netzverzeigerung loesen */
    if (pl->prevplace)
      pl->prevplace->nextplace = pl->nextplace;
    else
      net->Places = pl->nextplace;
    if (pl->nextplace)
      pl->nextplace->prevplace = pl->prevplace;
    else
      net->LastPlace = pl->prevplace;
      /* sonstige Aufraeumarbeiten */
    M_RemoveRange(pl);
    net->NumPlaces--;
    prevplace = pl->prevplace;
    if (pl->init_marking)
      MYfree(pl->init_marking);
    if (pl->curr_marking)
      MYfree(pl->curr_marking);
    if (pl->meaning)
      MYfree(pl->meaning);
    if (pl->BlockEntries)
      MYfree(pl->blocks);
    if (pl->RefEntries)
      MYfree(pl->ref);
    M_removePlace(pl);
  }  /* if Netz und Stelle vorhanden */
  return prevplace;
}  /* M_RemovePlace */

/*****************************************************************************/
/* Entfernen der Transition 'tr' aus dem Netz 'net' zusammen mit allen       */
/* Kanten, die in/aus diese/r Transition fuehren.                            */
/* Bei normalen Transitionen werden auch ALLE zugehoerigen Phantome          */
/* geloescht. Das kann bei NORMAL_SCOPING zu Fehlern fuehren, wenn die       */
/* Phantome noch zu anderen Transitionen gehoeren !!!                        */
/* Rueckgabewert ist 'tr->prevtrans'.                                        */
/*****************************************************************************/

M_Transition M_RemoveTransition(M_Net net, M_Transition tr, bool tr_is_phant)
{
    register M_STEdge     st,stdel;
    register M_TSEdge     ts,tsdel;
             M_Transition tr_help;
             Cardinal     i;

    if (net && tr) {
        ts=tr->postset;
        while (ts) {
            if (ts->prevtrans)
                ts->prevtrans->nexttrans=ts->nexttrans;
            else
              if (tr_is_phant)
                ts->place->ppreset=ts->nexttrans;
              else
                ts->place->preset=ts->nexttrans;
            if (ts->nexttrans)
                ts->nexttrans->prevtrans=ts->prevtrans;
            ts=(tsdel=ts)->nextplace;
            M_removeTSEdge(tsdel);
        }
        st=tr->preset;
        while (st) {
            if (st->prevtrans)
                st->prevtrans->nexttrans=st->nexttrans;
            else
              if (tr_is_phant)
                st->place->ppostset=st->nexttrans;
              else
                st->place->postset=st->nexttrans;
            if (st->nexttrans)
                st->nexttrans->prevtrans=st->prevtrans;
            st=(stdel=st)->nextplace;
            M_removeSTEdge(stdel);
        }
        if (tr->prevtrans)
            tr->prevtrans->nexttrans=tr->nexttrans;
        else
          if (tr_is_phant)
            net->PTransitions=tr->nexttrans;
          else
            net->Transitions=tr->nexttrans;
        if (tr->nexttrans)
            tr->nexttrans->prevtrans=tr->prevtrans;
        else
          if (tr_is_phant)
            net->LastPTransition=tr->prevtrans;
          else
            net->LastTransition=tr->prevtrans;
        MSremove(tr->label);
        MYfree( tr->def );
        MYfree( tr->guard );
        if ( tr->RefEntries )
          MYfree( tr->ref );
          /* Blockverweise wegwerfen */
        if (tr->BlockEntries) {
          MYfree(tr->blocks);
        } /* if Blockverweise vorhanden */
          /* Phantomverweise wegwerfen */
        if (tr->PhantomEntries) {
          if (!tr_is_phant) {
            for ( i=0; i<tr->PhantomEntries; i++ ) {
	      if (flag & MNET_SYNCHRONIZE)
		{
		  if ( (tr->phantoms[i])->PhantomEntries == 1 )
		    M_RemoveTransition(net, tr->phantoms[i], True);
		  else
		    {
		      M_Transition ph = tr->phantoms[i];
		      Cardinal      j = 0;

		      while ( ph->phantoms[j] != tr )
			j++;
		      ph->phantoms[j] = ph->phantoms[--(ph->PhantomEntries)];
		    }
		}
	      else
		M_RemoveTransition(net, tr->phantoms[i], True);
            }
          }
          MYfree(tr->phantoms);
        }  /* if Phantomverweise vorhanden */
        MYfree(tr->vars.ComArray);
        if (tr->meaning)   MYfree(tr->meaning);
        tr_help=tr->prevtrans;
        if (tr_is_phant) {
          net->NumPTransitions--;
          MYfree(tr);
        } else {
          net->NumTransitions--;
          M_removeTrans(tr);
        }
        return tr_help;
    } else
      return NULL;
}

/********************************************************/
/* Entfernt das Netz '*net' (schnell) aus dem Speicher. */
/********************************************************/

void M_Remove_net(M_Net *net)
{
  register M_STEdge STPtr, STRem;
  register M_TSEdge TSPtr, TSRem;
  register M_Place PlPtr, PlRem;
  register M_Block BlPtr, BlRem;
  M_Transition TrPtr, TrRem;

  if (*net) {
    PlPtr = (*net)->Places;
    if ((TrPtr = (*net)->Transitions)) {  /* Sind Transitionen da? */
      while (PlPtr) {
          /* normale Kanten muellen */
        TSPtr = PlPtr->preset;
        while (TSPtr) {
          TSPtr = (TSRem = TSPtr)->nexttrans;
          M_removeTSEdge(TSRem);
        }  /* while Eingangskanten */
        STPtr = PlPtr->postset;
        while (STPtr) {
          STPtr = (STRem = STPtr)->nexttrans;
          M_removeSTEdge(STRem);
        }  /* while Ausgangskanten */
          /* Phantomkanten muellen */
        TSPtr = PlPtr->ppreset;
        while (TSPtr) {
          TSPtr = (TSRem = TSPtr)->nexttrans;
          M_removeTSEdge(TSRem);
        }  /* while Phantomeingangskanten */
        STPtr = PlPtr->ppostset;
        while (STPtr) {
          STPtr = (STRem = STPtr)->nexttrans;
          M_removeSTEdge(STRem);
        }  /* while Phantomausgangskanten */
        PlPtr = (PlRem = PlPtr)->nextplace;
        M_RemoveRange(PlRem);
           /* Blockverweise muellen */
        if (PlRem->BlockEntries)
          MYfree(PlRem->blocks);
        if (PlRem->RefEntries)
          MYfree(PlRem->ref);
        if (PlRem->init_marking) MYfree(PlRem->init_marking);
        if (PlRem->curr_marking) MYfree(PlRem->curr_marking);
        if (PlRem->meaning) MYfree(PlRem->meaning);
        M_removePlace(PlRem);
      }  /* while Stellen */
        /* normale Transitionen */
      while (TrPtr) {
        TrPtr = (TrRem = TrPtr)->nexttrans;
        MSremove(TrRem->label);
        MYfree(TrRem->def);
        MYfree(TrRem->guard);
        if (TrRem->RefEntries)
          MYfree(TrRem->ref);
          /* Blockverweise wegwerfen */
        if (TrRem->BlockEntries) {
          /* TrRem->BlockEntries = 0; */
          MYfree(TrRem->blocks);
        }  /* if Blockverweise vorhanden */
          /* Phantomverweise wegwerfen */
        if (TrRem->PhantomEntries) {
          /* TrRem->PhantomEntries = 0; */
          MYfree(TrRem->phantoms);
        }  /* if Phantomverweise vorhanden */
        MYfree(TrRem->vars.ComArray);
        if (TrRem->meaning) MYfree(TrRem->meaning);
        M_removeTrans(TrRem);
      }  /* while Transitionen */
        /* Phantomtransitionen */
      TrPtr = (*net)->PTransitions;
      while (TrPtr) {
        TrPtr = (TrRem = TrPtr)->nexttrans;
        MSremove(TrRem->label);
        MYfree(TrRem->def);
        MYfree(TrRem->guard);
        if (TrRem->RefEntries)
          MYfree(TrRem->ref);
          /* Blockverweise wegwerfen */
        if (TrRem->BlockEntries) {
          /* TrRem->BlockEntries = 0; */
          MYfree(TrRem->blocks);
        }  /* if Blockverweise vorhanden */
          /* Phantomverweise wegwerfen */
        if (TrRem->PhantomEntries) {
          /* TrRem->PhantomEntries = 0; */
          MYfree(TrRem->phantoms);
        }  /* if Phantomverweise vorhanden */
        MYfree(TrRem->vars.ComArray);
        if (TrRem->meaning) MYfree(TrRem->meaning);
        MYfree(TrRem);
      }  /* while Phantomtransitionen */
    } else {  /* Sind wenigstens Stellen da? */
      while (PlPtr) {
        PlPtr = (PlRem = PlPtr)->nextplace;
        M_RemoveRange(PlRem);
          /* Blockverweise muellen */
        if (PlRem->BlockEntries)
          MYfree(PlRem->blocks);
        if (PlRem->RefEntries)
          MYfree(PlRem->ref);
        if (PlRem->init_marking) MYfree(PlRem->init_marking);
        if (PlRem->curr_marking) MYfree(PlRem->curr_marking);
        if (PlRem->meaning) MYfree(PlRem->meaning);
        M_removePlace(PlRem);
      }  /* while Stellen */
    }  /* else, sind wenigstens Stellen da? */
      /* Bloecke verschrotten */
    BlPtr = (*net)->Blocks;
    while (BlPtr) {
      BlPtr = (BlRem = BlPtr)->nextblock;
      M_removeBlock(BlRem);
    }  /* while noch Bloecke vorhanden */
    M_removeHead(*net);
    *net = NULL;
  }  /* if Netz vorhanden */
}  /* M_Remove_net */

/******************************************************************/
/* Numeriert die Stellen und Transitionen des Netzes 'net' durch. */
/* Die Bloecke haben bereits eine Id durch den Verweis auf den    */
/* BlockTable bl->blocknr (Cardinal).                             */
/******************************************************************/

void M_numname_all(M_Net net)
{
  M_Place pl;
  M_Transition tr;
  Cardinal32 i;

  i = 0;
  for(pl = net->Places; pl; pl = pl->nextplace)
    pl->numname = ++i;
  i = 0;
  for(tr = net->Transitions; tr; tr = tr->nexttrans)
    tr->numname = ++i;
  i = 0;
  for(tr = net->PTransitions; tr; tr = tr->nexttrans)
    tr->numname = ++i;
}  /* M_numname_all */

/*****************************************************************************/
/* Prueft die Kanten eines Netzes auf logische Fehler und gibt eine          */
/* Fehlernummer (hoffentlich 0 = kein Fehler) zurueck.                       */
/*****************************************************************************/

static int M_CheckArcs(M_TSEdge ts, M_STEdge st, M_Place pl, int offset)
{
  Cardinal32   inzi;
  M_Transition tr;
  M_TSEdge     ts2;
  M_STEdge     st2;
 
  if (ts && ts->prevtrans)
    return offset-3;
  inzi=0;
  while (ts) {
    if (ts->place!=pl)
      return offset-4;
    if (ts->nexttrans && ts->nexttrans->prevtrans!=ts)
      return offset-5;
    tr=ts->trans;
    if (tr->numname<inzi)
      return offset-6;
    else
      inzi=tr->numname;
    ts2=tr->postset;
    while (ts2 && ts2!=ts)
      ts2=ts2->nextplace;
    if (!ts2)
      return offset-7;
    ts=ts->nexttrans;
  }
  if (st && st->prevtrans)
    return offset-8;
  inzi=0;
  while (st) {
    if (st->place!=pl)
      return offset-9;
    if (st->nexttrans && st->nexttrans->prevtrans!=st)
      return offset-10;
    tr=st->trans;
    if (tr->numname<inzi)
      return offset-11;
    else
      inzi=tr->numname;
    st2=tr->preset;
    while (st2 && st2!=st)
      st2=st2->nextplace;
    if (!st2)
      return offset-12;
    st=st->nexttrans;
  }
  return 0;
}

/*****************************************************************************/
/* Prueft die Transitionen eines Netzes auf logische Fehler und gibt eine    */
/* Fehlernummer (hoffentlich 0 = kein Fehler) zurueck.                       */
/*                                                                           */
/* Kann fuer normale oder fuer Phantomtransitionen aufgerufen werden.        */
/* offset dient dazu, die Fehlercodes fuer Phantome um 100 zu verschieben.   */
/*****************************************************************************/

static int M_CheckTransitions( M_Transition tr, M_Transition LastTransition, 
                               M_Transition LastVergleichTransition, 
                               Cardinal32 NumTransitions, int offset )
{
    Cardinal32   events, inzi;
    M_Place      pl;
    M_TSEdge     ts, ts2;
    M_STEdge     st, st2;

    events=0;
    if (tr!=NULL) {
        if (tr->prevtrans)
            return offset-21;
        while (tr) {
            ++events;
            {
              node_ptr n = tr->guard;
              Integer  anz_nodes = 1;  /* 1 fuer das Schlusselement */
              
              if ( n->typ != N_HEAD )
                return offset-201;  /* Guard muss mit N_HEAD beginnen. */
              while( n->op1 )
                {
                  anz_nodes += n->op1;
                  n += n->op1;
                  if ( n->typ != N_HEAD )
                    return offset-201;
                      /* Hinter jedem Teilbaum muss N_HEAD stehen, entweder 
                         als Kopf des nachfolgenden Teilbaums oder als 
                         Schlusselement. */
                }
              if ( anz_nodes != tr->guard->op2 )
                return offset-202;
                  /* Die Knotenanzahl muss der Guardlaenge entsprechen. */
            }
            {
              Integer *d = tr->def+1;
              Integer  anz_values = 2;  /* 1 Kopf- und 1 Schlusselement */
              
              while( *d != -1 )
                {
                  d += 2;
                  anz_values += *d + 2; 
                  d += *d;
                }
              if ( anz_values != *(tr->def) )
                return offset-203;
                  /* Die Werteanzahl muss der Definitionslaenge entsprechen. */
            }
            {
              Cardinal nr1, nr2;
              M_Transition tr2;

              for( nr1=0; nr1<tr->PhantomEntries; nr1++ )
                {
                  tr2 = LastVergleichTransition;

                  while( (tr2 != tr->phantoms[nr1]) && (tr2->prevtrans) )
                    tr2 = tr2->prevtrans;
                  if (tr2 == tr->phantoms[nr1])
                    {
                      nr2=0;
                      while( (nr2 < tr2->PhantomEntries) && 
                             (tr2->phantoms[nr2] != tr) )
                        nr2++;
                      if (nr2 == tr2->PhantomEntries)
                        return offset-204;
                          /* Phantomverweise muessen symmetrisch sein. */
                    }
                  else
                    return offset-205;
                      /* Zu dem Phantomeintrag muss 1 Transition existieren. */
                }
            }
            if (tr->nexttrans && tr->nexttrans->prevtrans!=tr)
                return offset-22;
            st=tr->preset; ts=tr->postset;
            if (st && st->prevplace)
                return offset-23;
            inzi=0;
            while (st) {
                if (st->trans!=tr)
                    return offset-24;
                if (st->nextplace && st->nextplace->prevplace!=st)
                    return offset-25;
                pl=st->place;
                if (pl->numname<inzi)
                    return offset-26;
                else
                    inzi=pl->numname;
                if (offset)
                  st2=pl->ppostset;
                else
                  st2=pl->postset;
                while (st2 && st2!=st)
                    st2=st2->nexttrans;
                if (!st2)
                    return offset-27;
                st=st->nextplace;
            }
            if (ts && ts->prevplace)
                return offset-28;
            inzi=0;
            while (ts) {
                if (ts->trans!=tr)
                    return offset-29;
                if (ts->nextplace && ts->nextplace->prevplace!=ts)
                    return offset-30;
                pl=ts->place;
                if (pl->numname<inzi)
                    return offset-31;
                else
                    inzi=pl->numname;
                if (offset)
                  ts2=pl->ppreset;
                else
                  ts2=pl->preset;
                while (ts2 && ts2!=ts)
                    ts2=ts2->nexttrans;
                if (!ts2)
                    return offset-32;
                ts=ts->nextplace;
            }
            if (tr->nexttrans==NULL && LastTransition!=tr)
                return offset-33;
            tr=tr->nexttrans;
        }
    } else {
        if (LastTransition) {
            return offset-34;
        }
    }
    if (NumTransitions!=events)
        return offset-35;
    return 0;
}

/*****************************************************************************/
/* Prueft das Netz auf logische Fehler und gibt eine Fehlernummer            */
/* (hoffentlich 0 = kein Fehler) zurueck.                                    */
/*****************************************************************************/

int M_CheckNet(M_Net net)
{
    Cardinal32    conds;
    M_TSEdge      ts;
    M_STEdge      st;
    M_Place       pl;
    int           wert;

    M_numname_all(net);

    conds=0;
    if ((pl=net->Places)!=NULL) {
        if (pl->prevplace)
            return -1;
        while (pl) {
            ++conds;
            if (pl->nextplace && pl->nextplace->prevplace!=pl)
                return -2;

            /* normale Kanten */
            ts=pl->preset; st=pl->postset;
            if ((wert=M_CheckArcs( ts, st, pl, 0 ))) return wert;

            /* Phantom-Kanten */
            ts=pl->ppreset; st=pl->ppostset;
            if ((wert=M_CheckArcs( ts, st, pl, -100 ))) return wert;

            if (pl->nextplace==NULL && net->LastPlace!=pl)
                return -13;
            pl=pl->nextplace;
        }
    } else {
        if (net->LastPlace) {
            return -14;
        }
    }
    if (net->NumPlaces!=conds)
        return -15;

    /* normale Transitionen */
    if ((wert=M_CheckTransitions( net->Transitions,    net->LastTransition,
                                  net->LastPTransition,net->NumTransitions,0 )))
      return wert;

    /* Phantom-Transitionen */
    return M_CheckTransitions( net->PTransitions,  net->LastPTransition,
                               net->LastTransition,net->NumPTransitions, -100 );
}

/********************************************************/
/* ASCII-Ausgabe eines Wertebereichs in das File ofp    */
/*                                                      */
/* Dabei werden Teilfolgen vom Typ (n,n+1,n+2,..,m-1,m) */
/* zusammengefasst zu "n..m".                           */
/********************************************************/

void PrintSingleRange( Integer *range, FILE *ofp )
{
  Integer first, last;
  Integer *z,*ende;
  bool erster_wert;

  fprintf( ofp, "{" );

#ifdef print_uncompressed
  for( first = 1; first <= *range-1; first++ )
    {
      fprintf( ofp, "%d", (int) range[first] );
      if ( first != *range-1 )
        fprintf( ofp, "," );
    }
#else
  erster_wert = True;
  z = range + 1;
  ende = range + *range;
  while( z < ende )
    {
      first = *z;
      if ( !erster_wert )
        fprintf( ofp, "," );
      else
        erster_wert=False;
      if ( first > MAX_POS )
        switch ( first )
          {
          case  FF: fprintf( ofp, "ff"  ); break;
          case  TT: fprintf( ofp, "tt"  ); break;
          case DOT: fprintf( ofp, "dot" ); break;
          }
      else
        {
          while( z+1 < ende && *z == *(z+1)-1 )
            z++;  /* Ende der aktuellen Teilfolge suchen */
          last = *z;
          if ( first == last )
            fprintf( ofp, "%d", (int)first );
          else
            fprintf( ofp, "%d..%d", (int)first, (int)last );
        }
      z++;
    }
#endif

  fprintf( ofp, "}" );
}

/************************************************************************/
/* ASCII-Ausgabe des (Kreuz-)Wertebereichs einer Stelle in das File ofp */
/************************************************************************/

void PrintRange( M_Place pl, FILE *ofp )
{
  Cardinal stelle;

  for(stelle=0; pl->range[stelle] && (stelle < MAX_DIM); stelle++ )
    {
      if ( stelle > 0 )
        fprintf( ofp, " X " );
      PrintSingleRange( pl->range[stelle], ofp );
    }
}

/****************************************************************************/
/* Ein Teil-Baum wird rekursiv in das (bereits offene) File ofp ausgegeben. */
/* ACHTUNG: nr muss auf den Kopf des Baums zeigen, das ist der LETZTE       */
/*          Knoten im Array.                                                */
/*                                                                          */
/* Beispielaufruf: PrintTree( stdout, tree, tree->op1 - 1 )                 */
/****************************************************************************/

#define print( op ) { PrintTree( ofp, tree, v_ptr, n->op1 ); \
                      fprintf( ofp, op );                    \
                      PrintTree( ofp, tree, v_ptr, n->op2 ); \
                      break; }

static void PrintTree( FILE *ofp, node_ptr tree, Variables v_ptr, Integer nr )
{
  node_ptr n = tree + nr;

  switch ( n->typ )
    {
      case N_HEAD   : break;
      case N_CONST  : switch ( n->op1 )
                        {
                          case  FF: fprintf( ofp, "ff"  ); break;
                          case  TT: fprintf( ofp, "tt"  ); break;
                          case DOT: fprintf( ofp, "dot" ); break;
                          default:  fprintf( ofp, "%d", (int)n->op1 );
                        }
                      break;
      case N_VAR    : { char name[ MAX_IDENT_LENGTH ];
                        *name = 0;
                        StrCatVar( n->op1, v_ptr, name );
                        if ( n->op2 != -1 )
                          fprintf( ofp, "pr%d(%s)", (int)n->op2, name );
                        else
                          fprintf( ofp, name );
                        break;
                      }
      case N_KLAMM  : {
                        fprintf( ofp, "(" );
                        PrintTree( ofp, tree, v_ptr, n->op1 );
                        fprintf( ofp, ")" );
                        break;
                      }
      case N_OR     : print( " | " );
      case N_AND    : print( " & " );
      case N_NOT    : {
                        fprintf( ofp, "! " );
                        PrintTree( ofp, tree, v_ptr, n->op1 );
                        break;
                      }
      case N_EQ     : print( "="  );
      case N_UNEQ   : print( "#"  );
      case N_LT     : print( "<"  );
      case N_GT     : print( ">"  );
      case N_LE     : print( "<=" );
      case N_GE     : print( ">=" );
      case N_PLUS   : print( "+"  );
      case N_MIN    : print( "-"  );
      case N_MULT   : print( "*"  );
      case N_DIV    : print( "/"  );
      case N_MOD    : print( "%%" );
      case N_REM    : print( "%%" );
      case N_SIGN   : {
                        fprintf( ofp, "-" );
                        PrintTree( ofp, tree, v_ptr, n->op1 );
                        break;
                      }
      default: Msg( FAT, "Internal Error: Unknown node type !" );
    }   
}

/****************************************************************************/
/* Gibt eine Einzelmenge einer (moeglicherweise mehrdimensionalen)          */
/* Variablendefinition in das offene File ofp aus.                          */
/****************************************************************************/

static void PrintDef( FILE *ofp, Integer *d, M_Transition tr, bool *flag )
{
  char name[ MAX_IDENT_LENGTH ];

  if ( *(d+1) == -1 || *(d+1) == 0 )  /* Erster Wertebereich einer Var.? */
    {
      if ( *flag )
        *flag = False;
      else
        fprintf( ofp, " & " );  /* Mehrere Definitionen mit '&' verbinden */
      *name = 0;
      StrCatVar( *d, &(tr->vars), name );
      fprintf( ofp, name );
    }
  else
    fprintf( ofp, "X" );  /* ein weiterer Teilbereich einer Var. folgt */
  
  PrintSingleRange( d+2, ofp );
}

/****************************************************************************/
/* Die gesamte Annotation einer Transition wird in das File ofp ausgegeben. */
/*                                                                          */
/* Dazu werden erst alle Definitionen von Wertebereichen fuer freie         */
/* Variablen ausgegeben, anschliessend alle Teil-Baeume des Guards.         */
/****************************************************************************/

void PrintAnno( FILE *ofp, M_Transition tr, bool is_phant )
{
  node_ptr t = tr->guard;
  Integer *d = tr->def + 1;
  bool  empty_def = True;
    /* FALSE, falls schon mindestens eine Definition geschrieben worden ist */

  while( *d != -1 )
    {
      PrintDef( ofp, d, tr, &empty_def );
      d += 2;
      d += *d;
    }

  if ( !empty_def )
    fprintf( ofp, ": " );  /* Definitionen durch ':' abgrenzen */

  while( t->op1 )
    {
      if ( t != tr->guard )
        fprintf( ofp, " & " );  /* Teil-Baeume durch '&' verbinden */

      if ( t[t->op1-1].typ == N_OR && tr->guard->op1 != tr->guard->op2-1 )
        /* falls ein ODER in einer Kette von UNDs vorkommt -> Klammern */
        fprintf( ofp, "(" );

      PrintTree( ofp, t, &(tr->vars), t->op1 - 1 );

      if ( t[t->op1-1].typ == N_OR && tr->guard->op1 != tr->guard->op2-1 )
        /* falls ein ODER in einer Kette von UNDs vorkommt -> Klammern */
        fprintf( ofp, ")" );

      t += t->op1;
    }
  
  if ( t == tr->guard )
    fprintf( ofp, "tt" );  /* Ein leerer Baum steht fuer 'tt'. */
}

/*******************************************************************/
/* Fuegt bei allen Stellen mit Label label die Referenz ref hinzu. */
/*******************************************************************/

void M_AddAllPlaceRefs( M_Net net, PLabel label, RefNr ref, int flag )
{
  M_Place  pl;
  
  for(pl = net->Places; pl; pl = pl->nextplace)
    { 
      if ( pl->label & label )
        {
          if (pl->RefEntries++)
            pl->ref = MYrealloc( pl->ref, pl->RefEntries * sizeof (Ref) );
          else
            pl->ref = MYmalloc( sizeof (Ref) );
          pl->ref[pl->RefEntries - 1].nr   = ref;
          pl->ref[pl->RefEntries - 1].flag = flag;
        }
    }
}

/*****************************************************************/
/* Fuegt die Referenzen von src zu den Referenzen von dst hinzu. */
/*****************************************************************/

void M_SumPlaceRefs( M_Place dst, M_Place src )
{
  char *meaning;

  if (src->RefEntries)
    {
      size_t anz_bytes = src->RefEntries * sizeof(Ref);

      if (dst->RefEntries)
        dst->ref = MYrealloc( dst->ref, 
                      (dst->RefEntries + src->RefEntries) * sizeof (Ref) );
      else
        dst->ref = MYmalloc( anz_bytes );

      memcpy( dst->ref + dst->RefEntries, src->ref, anz_bytes );
      dst->RefEntries += src->RefEntries;
    }
  meaning = ConcatStrings(dst->meaning, src->meaning);
  if (dst->meaning)
    MYfree(dst->meaning);
  dst->meaning = meaning;
}

/********************************************************************/
/* Fuegt eine neue Transition am Anfang der Transitionenliste ein.  */
/********************************************************************/

M_Transition M_CreateTr(M_Net net, MultiSet mm, node_ptr guard,
                        Variables v_ptr, int x_pos,int y_pos, RefNr ref)
{
    M_Transition tr;

    (tr=M_allocTrans())->prevtrans=NULL;
    tr->x_pos=x_pos;
    tr->y_pos=y_pos;
    if (net->Transitions) {
        (tr->nexttrans=net->Transitions)->prevtrans=tr;
    } else {
        (net->LastTransition=tr)->nexttrans=NULL;
    }
    (net->Transitions=tr)->preset=(M_STEdge)(tr->postset=NULL);
    tr->label=mm;
    tr->flags=0;
    tr->def=MYmalloc( 2 * sizeof( Integer ) );
    tr->def[0] = 2;
    tr->def[1] = -1;
    if ( guard )
      tr->guard=guard;
    else
      {
        tr->guard = MYmalloc( sizeof ( node ) );
        tr->guard->typ = N_HEAD;
        tr->guard->op1 = 0;
        tr->guard->op2 = 1;
      }
    if ( ref ) {
        tr->RefEntries = 1;
        tr->ref = MYmalloc( sizeof( Ref ) );
        (tr->ref)->nr   = ref;
        (tr->ref)->flag = 0;
    } else
        tr->RefEntries = 0;
    tr->meaning=NULL;
    tr->BlockEntries = 0;
    tr->blocks = NULL;
    tr->PhantomEntries = 0;
    tr->phantoms = NULL;
    tr->links = NULL;
    tr->time_label = NULL;
    tr->vars.ComEntries=v_ptr->ComEntries;
    tr->vars.ComArray  =v_ptr->ComArray;
    ++net->NumTransitions;
    return tr;
}

/************************************/
/* Unterfunktion von M_CreateEdges  */
/************************************/

static M_Place M_CreatePreset(Cardinal32 preset,VarMSet label,
                              M_Place pl,M_Net net,Cardinal flags)
{
    M_STEdge st;

    if (preset) {
      while (preset<pl->numname)
	pl=pl->prevplace;
      while (preset>pl->numname)
	pl=pl->nextplace;
    } else {
      pl=net->Places; /* Entry-Stelle */
    }
    (st=M_allocSTEdge())->prevtrans=NULL;
    if ((st->nexttrans=(st->place=pl)->postset)!=NULL)
      st->nexttrans->prevtrans=st;
    (pl->postset=st)->prevplace=NULL;
    if ((st->nextplace=(st->trans=net->Transitions)->preset)!=NULL)
      st->nextplace->prevplace=st;
    (net->Transitions->preset=st)->label=label;
    st->flags=flags;
    return pl;
}

/************************************/
/* Unterfunktion von M_CreateEdges  */
/************************************/

static M_Place M_CreatePostset(Cardinal32 postset,VarMSet label,
                               M_Place pl,M_Net net,Cardinal flags)
{
    M_TSEdge ts;

    if (postset) {
      while (postset<pl->numname)
	pl=pl->prevplace;
      while (postset>pl->numname)
	pl=pl->nextplace;
    } else {
      pl=net->LastPlace; /* Exit-Stelle */
    }
    (ts=M_allocTSEdge())->prevtrans=NULL;
    if ((ts->nexttrans=(ts->place=pl)->preset)!=NULL)
      ts->nexttrans->prevtrans=ts;
    (pl->preset=ts)->prevplace=NULL;
    if ((ts->nextplace=(ts->trans=net->Transitions)->postset)!=NULL)
      ts->nextplace->prevplace=ts;
    (net->Transitions->postset=ts)->label=label;
    ts->flags=flags;
    return pl;
}

/*********************************************************************/
/* Fuegt fuer eine neuen Transition am Anfang der Transitionenliste  */
/* (entstanden durch M_CreateTr) eine Preset-Kante und eine Postset- */
/* Kante ein zu den Places mit den entsprechenden Nummern.           */
/* Der Zeiger pl auf die Stellenliste kann die Suche vereinfachen.   */
/*********************************************************************/

M_Place M_CreateEdges(Cardinal32  preset, VarMSet  pre_label,
                      Cardinal32 postset, VarMSet post_label,
                      M_Place pl, M_Net net, Cardinal flags)
{
    return M_CreatePostset(postset,post_label,
                           M_CreatePreset(preset,pre_label,pl,net,flags),
                           net,flags);
}  /* M_CreateEdges */

/*****************************************************************************/
/* Erstellt am Anfang der Stellenliste des Netzes 'net' eine disjunkte Kopie */
/* der Stelle 'pl' mitsamt allen ein- und ausgehenden Kanten.                */
/* 'numname' wird als Kennzeichen einer Kopie auf 0 gesetzt.                 */
/* Rueckgabewert ist ein Zeiger auf die neue Stelle, ausserdem wird in       */
/* '*stlast' bzw. '*tslast' ein Zeiger auf die letzte Kante des Postsets     */
/* bzw. des Presets abgelegt.                                                */
/* Die Phantomkanten werden kopiert.                                         */
/*****************************************************************************/

M_Place M_DuplicatePlace(M_Net net,M_Place pl,M_STEdge *stlast,M_TSEdge *tslast,
                         M_STEdge *pstlast,M_TSEdge *ptslast)
{
    M_Place newpl = NULL;

    if (net && pl) {
        ++net->NumPlaces;
	newpl=M_allocPlace();
        (newpl->nextplace=net->Places)->prevplace=newpl;
        (net->Places=newpl)->prevplace=NULL;
        newpl->numname=0;
        newpl->label=pl->label;
        M_CopyRange(newpl,pl);
        M_CopyPlaceReferences(newpl,pl);
        M_CopyPlaceBlockEntries(newpl, pl);
        if (pl->meaning)
          newpl->meaning=MYstrdup(pl->meaning);
        else
          newpl->meaning=NULL;
        if (pl->init_marking)
          newpl->init_marking=MYstrdup(pl->init_marking);
        else
          newpl->init_marking=NULL;
        if (pl->curr_marking)
          newpl->curr_marking=MYstrdup(pl->curr_marking);
        else
          newpl->curr_marking=0;
        newpl->flags=pl->flags;
        newpl->component=pl->component;
        newpl->x_pos=pl->x_pos;
        newpl->y_pos=pl->y_pos;
        newpl->preset  =M_CopyPlacePreset (pl,newpl,tslast,False);
        newpl->postset =M_CopyPlacePostset(pl,newpl,stlast,False);
        newpl->ppreset =M_CopyPlacePreset (pl,newpl,ptslast,True);
        newpl->ppostset=M_CopyPlacePostset(pl,newpl,pstlast,True);
    } else {
        Msg( FAT,"DuplicatePlace: no net or no place given");
    }
    return newpl;
}  /* M_DuplicatePlace */

/*****************************************************************************/
/* Alle in die Stelle 'pl' eingehenden Kanten werden in der Kantenmatrix     */
/* direkt vor die Transition dupliziert, von der sie ausgehen. Als Stellen-  */
/* zeiger wird diesen Kanten (ein Zeiger auf die neu in die Stellenliste     */
/* eingefuegte Stelle) 'newpl' zugewiesen.                                   */
/* Rueckgabewert ist ein Zeiger auf die erste Kante, '*tslast' wird ein      */
/* Zeiger auf die letzte Kante zugewiesen.                                   */
/*****************************************************************************/

M_TSEdge M_CopyPlacePreset(M_Place pl,M_Place newpl,
                           M_TSEdge *tslast,bool do_phant)
{
    register M_TSEdge ts,newts;
    M_TSEdge          tsfirst;

    tsfirst=*tslast=NULL;
    if (pl && newpl && ((ts=((do_phant)?pl->ppreset:pl->preset))!=NULL)) {
        (tsfirst=newts=M_allocTSEdge())->prevtrans=NULL;
        while (newts) {
            newts->prevplace=NULL;
            newts->label=V_MScreate_copy(ts->label, NULL);
            newts->flags=ts->flags;
            newts->place=newpl;
            newts->trans=ts->trans;
            newts->nextplace=ts->trans->postset;
            ts->trans->postset=newts->nextplace->prevplace=newts;
            newts=((ts=ts->nexttrans)!=NULL)?
                  ((newts->nexttrans=M_allocTSEdge())
                   ->prevtrans=newts)->nexttrans
                  : ((*tslast=newts)->nexttrans=NULL);
        }
    }
    return tsfirst;
}

/*****************************************************************************/
/* Alle von der Stelle 'pl' ausgehenden Kanten werden in der Kantenmatrix    */
/* direkt vor die Transition dupliziert, in der sie enden. Als Stellenzeiger */
/* wird diesen Kanten (ein Zeiger auf die neu in die Stellenliste eingefuegte*/
/* Stelle) 'newpl' zugewiesen.                                               */
/* Rueckgabewert ist ein Zeiger auf die erste Kante, '*stlast' wird ein      */
/* Zeiger auf die letzte Kante zugewiesen.                                   */
/*****************************************************************************/

M_STEdge M_CopyPlacePostset(M_Place pl,M_Place newpl,
                            M_STEdge *stlast,bool do_phant)
{
    register M_STEdge st,newst;
    M_STEdge          stfirst;

    stfirst=*stlast=NULL;
    if (pl && newpl && ((st=((do_phant)?pl->ppostset:pl->postset))!=NULL)) {
        (stfirst=newst=M_allocSTEdge())->prevtrans=NULL;
        while (newst) {
            newst->prevplace=NULL;
            newst->label=V_MScreate_copy(st->label, NULL);
            newst->flags=st->flags;
            newst->place=newpl;
            newst->trans=st->trans;
            newst->nextplace=st->trans->preset;
            st->trans->preset=newst->nextplace->prevplace=newst;
            newst=((st=st->nexttrans)!=NULL)?
                  ((newst->nexttrans=M_allocSTEdge())
                   ->prevtrans=newst)->nexttrans
                  : ((*stlast=newst)->nexttrans=NULL);
        }
    }
    return stfirst;
}

/*****************************************************************************/
/* Unterfunktion von M_CreateNetCopy:                                        */
/*                                                                           */
/* Kopiert alle Transitionen bzw. alle Phantom-Trans. von net nach net2.     */
/* Damit die Anpassung der gegenseitigen Verweise zwischen Sync- und         */
/* Phantom-Transitionen funktioniert, muss diese Funktion ZUERST fuer die    */
/* normalen Transitionen aufgerufen werden werden.                           */
/*****************************************************************************/

static void M_CreateTransitionsCopy( M_Transition *Transitions2, 
                                     Cardinal32   *NumTransitions2,
                                     M_Transition *LastTransition2,

                                     M_Transition Transitions,
                                     Cardinal32   NumTransitions,

                                     M_Net        net2,
                                     M_Net        net,
                                     bool      do_phants )
{
    register M_TSEdge ts,ts2;
    register M_STEdge st,st2;
    M_Place           pl,pl2;
    M_Transition      tr,tr2;
    size_t            array_length;

    Cardinal32        i;
    M_Transition      tr_such, *p2;

    if ((*NumTransitions2 = NumTransitions) != 0) {
      tr=Transitions;
      (tr2 = *Transitions2 = ((do_phants)?M_allocPTrans():M_allocTrans()) )
        ->prevtrans = NULL;
      while (tr2) {
        tr2->postset = NULL;
        tr2->x_pos = tr->x_pos;
        tr2->y_pos = tr->y_pos;
        tr2->numname = (Cardinal32)tr;
         tr->numname = (Cardinal32)tr2;
        tr2->label = MScreate_copy(tr->label);
        tr2->guard = MYmalloc(tr->guard->op2 * sizeof(node));
        memcpy(tr2->guard, tr->guard, tr->guard->op2 * sizeof(node));
        tr2->def = MYmalloc(*(tr->def) * sizeof(Integer));
        memcpy(tr2->def, tr->def, *(tr->def) * sizeof(Integer));
        tr2->RefEntries = tr->RefEntries;
        if (tr->RefEntries) {
          array_length = tr->RefEntries * sizeof(Ref);
          tr2->ref = MYmalloc(array_length);
          memcpy(tr2->ref, tr->ref, array_length);
        }
        if (tr->meaning)
          tr2->meaning=MYstrdup(tr->meaning);
        else
          tr2->meaning=NULL;
        if (tr->links)
          tr2->links=MYstrdup(tr->links);
        else
          tr2->links=NULL;
        if (tr->time_label)
	{
          tr2->time_label=(Time_stamp*)MYmalloc(sizeof(Time_stamp));
          tr2->time_label->low=tr->time_label->low;
          tr2->time_label->high=tr->time_label->high;
	}
        else
          tr2->time_label=NULL;
        M_CopyTransBlockEntries(tr2, tr);

        /* Die Zeiger der kopierten Phantom-Transitionen stehen noch nicht fest.
           Deshalb werden bei den normalen Transitionen die phantoms zunaechst 
           kopiert, um dann spaeter angepasst zu werden.

           Wenn im zweiten Schritt die Phantom-Transitionen kopiert werden,
           wird jeder phantoms-Eintrag auf die kopierte Normal-Transition
           gesetzt und der Rueckverweis in der Normal-Transition angepasst. */
        if (do_phants)
          {
            tr2->PhantomEntries = tr->PhantomEntries;
            if (tr->PhantomEntries) {
              array_length = tr->PhantomEntries * sizeof(M_Transition);
              tr2->phantoms = MYmalloc(array_length);
              for( i=0; i<tr->PhantomEntries; i++ )
                {
                  tr_such = net2->Transitions;
                  while( (M_Transition) tr_such->numname != tr->phantoms[i] )
                    tr_such = tr_such->nexttrans;
                  tr2->phantoms[i]=tr_such;
                  p2=tr_such->phantoms; 
                  while( *p2!=tr )
                    p2++;
                  *p2=tr2;
                }
            }  /* if es gibt etwas zu kopieren */
            else {
              tr2->phantoms = NULL;
            }  /* else, keine Verweise */
          }
        else
          M_CopyTransPhantomEntries(tr2, tr);

        tr2->flags = tr->flags;
	if ((tr2->vars.ComEntries = tr->vars.ComEntries))
	{
          array_length = tr->vars.ComEntries * sizeof(ComArrayEntry);
          tr2->vars.ComArray = MYmalloc(array_length);
          memcpy(tr2->vars.ComArray, tr->vars.ComArray, array_length);
	}
	else
	  tr2->vars.ComArray = NULL;
        if ((st = tr->preset) != NULL) {  /* Preset der Trans einrichten */
          (st2 = tr2->preset = (M_STEdge)st->place->prevplace)->prevplace
            = NULL;
          while (st2) {
            st2->trans = tr2;
            st->place->prevplace = (M_Place) st2->nexttrans;
            st2 = ((st = st->nextplace) != NULL) ?
              ((st2->nextplace = (M_STEdge) st->place->prevplace)->prevplace
                = st2)->nextplace
              : (st2->nextplace = NULL);
          }
        }
        else
          tr2->preset=NULL;
        tr2 = ((tr = tr->nexttrans) != NULL) ?
          ((tr2->nexttrans = ((do_phants)?M_allocPTrans():M_allocTrans()) )
           ->prevtrans = tr2) ->nexttrans : 
          ((*LastTransition2 = tr2)->nexttrans = NULL);
      }
    }
    else {
      *Transitions2 = *LastTransition2 = NULL;
    }

    if ((pl=net->Places)!=NULL) {
        if ((tr=Transitions)!=NULL) {
            pl2=net2->Places;
            while (pl) {
/* !!! */       pl->prevplace=(M_Place)((do_phants)?pl2->ppreset:pl2->preset);
                pl=pl->nextplace;
                pl2=pl2->nextplace;
            }
            tr2=*Transitions2;
            while (tr2) {
                if ((ts=tr->postset)!=NULL) { /* Postset der Transition einr.*/
/* !!! */           (ts2=tr2->postset=(M_TSEdge)ts->place->prevplace)
                    ->prevplace=NULL;
                    while (ts2) {
                        ts2->trans=tr2;
/* !!! */               ts->place->prevplace=(M_Place)ts2->nexttrans;
                        ts2=((ts=ts->nextplace)!=NULL)?
/* !!! */                     ((ts2->nextplace=(M_TSEdge)ts->place->prevplace)
                               ->prevplace=ts2)->nextplace
                            : (ts2->nextplace=NULL);
                    }
                }
                tr=tr->nexttrans;
                tr2=tr2->nexttrans;
            }
        }
        if (do_phants)
          {
            (pl=net->Places)->prevplace=NULL; /* Missbrauchte Stellen-Vor-    */
            while (pl->nextplace) {           /* gaenger-Zeiger restaurieren! */
              pl=(pl->nextplace->prevplace=pl)->nextplace;
            }
          }
        else
          {
            for(pl=net->Places, pl2=net2->Places; pl;
                pl=pl->nextplace, pl2=pl2->nextplace)
              pl->prevplace=(M_Place)pl2->ppostset;
          }
    }
}

/*****************************************************************************/
/* Erzeugt eine Kopie des uebergebenen Netzes 'net'.                         */
/* Rueckgabewert ist ein Zeiger auf diese Kopie.                             */
/* In dem Algorithmus werden voruebergehend die Stellen-Vorgaenger-Zeiger im */
/* Original-Netz so verbogen, dass sie auf den Preset bzw. Postset der       */
/* korrespondierenden Stellen in der Kopie zeigen. Dies ist eine Hilfe beim  */
/* Einrichten des Postsets bzw. des Presets der neuen Transitionen.          */
/* Die Zeilen, in denen dieser Kunstgriff eine Rolle spielt, sind mit '!!!'  */
/* markiert.                                                                 */
/*                                                                           */
/* Nach dieser Funktion gilt:                                                */
/* Im der Netz-Kopie zeigt fuer jede Komponente der numname auf die          */
/* zugehoerige Originalkomponente im Original-Netz.                          */
/* Umgekehrt zeigt der numname der Originalkomponente auf die zugehoerige    */
/* Komponente in der Netz-Kopie !!!!                                         */
/*****************************************************************************/

M_Net M_CreateNetCopy(M_Net net)
{
    register M_TSEdge ts,ts2;
    register M_STEdge st,st2;
    M_Place           pl,pl2;
    M_Net             net2;
    int l_count;

    net2 = M_empty_net();
    pl=net->Places;
    net2->NumComponents=net->NumComponents;
    net2->x_pos=net->x_pos;
    net2->y_pos=net->y_pos;
    net2->x_dim=net->x_dim;
    net2->y_dim=net->y_dim;
    
	/* copy Links */
    net2->NumLinks=net->NumLinks;
    net2->MaxNumLinks=net->MaxNumLinks;
    
    if (net2->MaxNumLinks>0)
      {
        net2->Links =
           (linkdef *) MYmalloc ( (net2->MaxNumLinks) * sizeof (linkdef *));
        for (l_count=0;l_count<net2->MaxNumLinks;l_count++)
         {
           if (net->Links[l_count] == NULL)
                net2->Links[l_count] = NULL;
           else
              {
                net2->Links[l_count] = (linkdef) MYmalloc
                      ( (strlen(net->Links[l_count])+1) * sizeof (char));
                strcpy (net2->Links[l_count], net->Links[l_count]);
              };
         }; /* for l_count */
      }; /* if (net2->MaxNumLinks>0) */

    M_CopyBlockInNets(net2, net);  /* Blockgeroedel kopieren */
    if ((net2->NumPlaces=net->NumPlaces)!=0) {
        (pl2=net2->Places=M_allocPlace())->prevplace=NULL;
        while (pl2) {
            pl2->x_pos=pl->x_pos;
            pl2->y_pos=pl->y_pos;
            pl2->component=pl->component;
            pl2->label=pl->label;
            M_CopyRange( pl2, pl );
            M_CopyPlaceReferences( pl2, pl );
            M_CopyPlaceBlockEntries(pl2, pl);
            if (pl->meaning)
              pl2->meaning=MYstrdup(pl->meaning);
            else
              pl2->meaning=NULL;
            if (pl->init_marking)
              pl2->init_marking=MYstrdup(pl->init_marking);
            else
              pl2->init_marking=NULL;
            if (pl->curr_marking)
              pl2->curr_marking=MYstrdup(pl->curr_marking);
            else
              pl2->curr_marking=0;
            pl2->flags=pl->flags;
            pl2->numname=(Cardinal32)pl;
             pl->numname=(Cardinal32)pl2;
            if ((ts=pl->preset)!=NULL) {       /* Preset der Stelle kopieren */
                (ts2=pl2->preset=M_allocTSEdge())->prevtrans=NULL;
                while (ts2) {
                    ts2->place=pl2;
                    ts2->label=V_MScreate_copy(ts->label, NULL);
                    ts2->numname=(Cardinal32)ts;
                     ts->numname=(Cardinal32)ts2;
                    ts2->flags=ts->flags;
                    ts2=((ts=ts->nexttrans)!=NULL)?
                          ((ts2->nexttrans=M_allocTSEdge())
                           ->prevtrans=ts2)->nexttrans
                        : (ts2->nexttrans=NULL);
                }
            } else {
                pl2->preset=NULL;
            }
            if ((st=pl->postset)!=NULL) {     /* PostSet der Stelle kopieren */
                (st2=pl2->postset=M_allocSTEdge())->prevtrans=NULL;
                while (st2) {
                    st2->place=pl2;
                    st2->label=V_MScreate_copy(st->label, NULL);
                    st2->numname=(Cardinal32)st;
                     st->numname=(Cardinal32)st2;
                    st2->flags=st->flags;
                    st2=((st=st->nexttrans)!=NULL)?
                          ((st2->nexttrans=M_allocSTEdge())
                           ->prevtrans=st2)->nexttrans
                        : (st2->nexttrans=NULL);
                }
            } else {
                pl2->postset=NULL;
            }

            if ((ts=pl->ppreset)!=NULL) {         /* Phantom-Preset kopieren */
                (ts2=pl2->ppreset=M_allocTSEdge())->prevtrans=NULL;
                while (ts2) {
                    ts2->place=pl2;
                    ts2->label=V_MScreate_copy(ts->label, NULL);
                    ts2->numname=(Cardinal32)ts;
                     ts->numname=(Cardinal32)ts2;
                    ts2->flags=ts->flags;
                    ts2=((ts=ts->nexttrans)!=NULL)?
                          ((ts2->nexttrans=M_allocTSEdge())
                           ->prevtrans=ts2)->nexttrans
                        : (ts2->nexttrans=NULL);
                }
            } else {
                pl2->ppreset=NULL;
            }
            if ((st=pl->ppostset)!=NULL) {       /* Phantom-PostSet kopieren */
                (st2=pl2->ppostset=M_allocSTEdge())->prevtrans=NULL;
                while (st2) {
                    st2->place=pl2;
                    st2->label=V_MScreate_copy(st->label, NULL);
                    st2->numname=(Cardinal32)st;
                     st->numname=(Cardinal32)st2;
                    st2->flags=st->flags;
                    st2=((st=st->nexttrans)!=NULL)?
                          ((st2->nexttrans=M_allocSTEdge())
                           ->prevtrans=st2)->nexttrans
                        : (st2->nexttrans=NULL);
                }
            } else {
                pl2->ppostset=NULL;
            }
/* !!! */   pl->prevplace=(M_Place)pl2->postset;
            pl2=((pl=pl->nextplace)!=NULL)?
                  ((pl2->nextplace=M_allocPlace())->prevplace=pl2)
                  ->nextplace
                : ((net2->LastPlace=pl2)->nextplace=NULL);
        }
    } else {
        net2->Places=net2->LastPlace=NULL;
    }

      /* Transitionen */
    M_CreateTransitionsCopy( &(net2->Transitions),
                             &(net2->NumTransitions),
                             &(net2->LastTransition),
                             net->Transitions,
                             net->NumTransitions,
                             net2, net, False );

      /* Phantomtransitionen */
    M_CreateTransitionsCopy( &(net2->PTransitions),
                             &(net2->NumPTransitions),
                             &(net2->LastPTransition),
                             net->PTransitions,
                             net->NumPTransitions,
                             net2, net, True );
  return net2;
}

  /**************************************************************************/
  /* Veraenderung von 'net1', so dass 'net1' eine disjunkte Vereinigung der */
  /* Netze 'net1' und 'net2' darstellt. Achtung: Das Kopfelement von 'net2' */
  /* wird geloescht.                                                        */
  /* Ausnahme: Die Bloecke werden vereinigt, NICHT disjunkt vereinigt!!!    */
  /* Rueckgabewert ist 'net1'.                                              */
  /**************************************************************************/
/* TODO: Links missing. */
M_Net M_DisjointNetUnion(M_Net net1, M_Net net2)
{
  if (net1->Places != NULL) {
    if ((net1->LastPlace->nextplace = net2->Places) != NULL) {
      net2->Places->prevplace = net1->LastPlace;
      net1->LastPlace = net2->LastPlace;
    }
  } else {
    net1->Places = net2->Places;
    net1->LastPlace = net2->LastPlace;
  }
  if (net1->Transitions != NULL) {
    if ((net1->LastTransition->nexttrans = net2->Transitions) != NULL) {
      net2->Transitions->prevtrans = net1->LastTransition;
      net1->LastTransition = net2->LastTransition;
    }
  } else {
    net1->Transitions = net2->Transitions;
    net1->LastTransition = net2->LastTransition;
  }
  if (net1->PTransitions != NULL) {
    if ((net1->LastPTransition->nexttrans = net2->PTransitions) != NULL) {
      net2->PTransitions->prevtrans = net1->LastPTransition;
      net1->LastPTransition = net2->LastPTransition;
    }
  } else {
    net1->PTransitions = net2->PTransitions;
    net1->LastPTransition = net2->LastPTransition;
  }
  net1->NumPlaces += net2->NumPlaces;
  net1->NumTransitions += net2->NumTransitions;
  net1->NumPTransitions += net2->NumPTransitions;
  M_JointBlockUnion(net1, net2);  /* Blockmengen vereinigen, Doppelte raus! */
  M_RemoveBlocksFromNet(net2);    /* Blockmenge2 wegschmeissen */
  M_removeHead(net2);
  return net1;
}  /* M_DisjointNetUnion */

/******************************************************/
/* Schreibt die Quelltextreferenz ref in das File ofp */
/******************************************************/

static void PrintReference( RefNr r_nr, int flags, FILE *ofp )
{
  if ( r_nr )
    {
      ShortEntryPtr ref = RefToStruct( r_nr );
      if (flags & PRE_REF)
        fputc( '\'', ofp );
      fprintf( ofp, "(%u,%u;%u,%u)", ref->s_ref.start_l, ref->s_ref.start_c,
                                     ref->s_ref.end_l  , ref->s_ref.end_c    );
      if (flags & POST_REF)
        fputc( '\'', ofp );
    }
  else
    fprintf( ofp, "(0,0;0,0)" );
}

/********************************************************************/
/* Schreibt die Bedeutung der Quelltextreferenz ref in das File ofp */
/********************************************************************/

Cardinal SrcLineEntries = 1; /* Zeilenanzahl des Quelltextes */

void PrintMeaning( RefNr r_nr, int flags, FILE *ofp )
{
    ShortEntryPtr ref;

    if ( !SrcBuffer ) { /* Falls der Quelltext noch nicht eingelesen wurde,
                           wird er EINMALIG eingelesen und aufbewahrt */

        FILE     *src;
        Cardinal  SrcCharEntries = 0;
        Cardinal  i = 0, j = 0;
        int       chr;

        /* Zeilenanzahl und Groesse des Files bestimmen */

        if (!(src = fopen( source, "r" )))
          return;
        while( (chr = fgetc( src )) != EOF ) {
            SrcCharEntries++;
            if ( chr == '\n' ) 
                SrcLineEntries++;
        }
        rewind( src );

        /* Fuer die ermittelte Groesse Speicher anfordern */
        SrcBuffer = MYmalloc( SrcCharEntries + 1 ); /* + 1 fuer EOF */
        SrcLine   = MYmalloc( SrcLineEntries * sizeof( char* ) );

        /* Quelltext einlesen und zeilenweise ablegen */
        SrcLine[0] = SrcBuffer;
        while( (chr= fgetc( src )) != EOF ) {
            if ( chr == '\n' ) {                   /* neue Zeile:         */
                SrcBuffer[j++] = 0;                /* also String beenden */
                SrcLine[++i] = SrcBuffer + j;  /* & naechsten Zeilen- */
            }                                      /* anfang speichern    */
            else SrcBuffer[j++] = chr;
        }
        SrcBuffer[j] = 0;

        fclose( src );
    }

    /* leere Referenz -> nichts ausgeben */
    if ( r_nr )
    {
      ref = RefToStruct( r_nr );
      if ( ref->s_ref.start_l < ref->s_ref.end_l || 
           ( ref->s_ref.start_l == ref->s_ref.end_l && 
             ref->s_ref.start_c < ref->s_ref.end_c ) ) {
        if ( ref->s_ref.end_l <= SrcLineEntries ) {
          if (flags & PRE_REF)
            fputc( '\'', ofp );
          if ( ref->s_ref.start_l == ref->s_ref.end_l ) 
            /* einzeilige Referenz */
            fprintf( ofp, "%.*s", ref->s_ref.end_c - ref->s_ref.start_c,
                    SrcLine[ ref->s_ref.start_l - 1 ] + ref->s_ref.start_c - 1);
          else {
            /* mehrzeilige Referenz */
            Cardinal i = ref->s_ref.start_l - 1;
                    
            fprintf( ofp, "%s\n", SrcLine[ i ] + ref->s_ref.start_c - 1 );
            while ( ++i < ( ref->s_ref.end_l - 1 ) )
              fprintf( ofp, "%s\n", SrcLine[ i ] );
            fprintf( ofp, "%.*s", ref->s_ref.end_c - 1, SrcLine[ i ] );
          }
          switch ( flags ) 
            {
              case POST_REF:
                fputc( '\'', ofp );
                break;
              case INIT_REF:
                fprintf( ofp, "@init" );
                break;
              case TERM_REF:
                fprintf( ofp, "@term" );
                break;
              case ACCESS_REF:
                fprintf( ofp, "@access" );
                break;
              case READ_REF:
                fprintf( ofp, "@read" );
                break;
              case WRITE_REF:
                fprintf( ofp, "@write" );
                break;
              case FULL_REF:
                fprintf( ofp, "@full" );
                break;
              case EMPTY_REF:
                fprintf( ofp, "@empty" );
                break;
              case SHIFT_REF:
                fprintf( ofp, "@shift" );
                break;
              case CALL_REF:
                fprintf( ofp, "@call" );
                break;
              case RETURN_REF:
                fprintf( ofp, "@return" );
                break;
            }
        } 
        else
          Msg( INF, "Reference out of source code: %u,%u;%u,%u", 
               ref->s_ref.start_l, ref->s_ref.start_c,
               ref->s_ref.end_l,   ref->s_ref.end_c );
      } 
      else
        Msg( INF, "Nonsensical reference: %u,%u;%u,%u", 
             ref->s_ref.start_l, ref->s_ref.start_c,
             ref->s_ref.end_l,   ref->s_ref.end_c );
    }
}

/********************************************/
/* Ausgabe von Referenzen in die Datei ofp. */
/********************************************/

static void PrintReferences( FILE *ofp, Cardinal RefEntries, Ref *ref, 
                             bool is_place, Cardinal32 numname )
{
  Cardinal      i;
  bool          nothing_printed;
#ifdef REF_STATISTIC
  RefStatistic *p;
#endif

  if ( RefEntries ) 
    {
#ifdef REF_STATISTIC
      if (ref_file) {
        for (i = 0; i < RefEntries; i++) {
          p = statistic + ref[i].nr;
          p->use = MYrealloc( p->use, (p->entries+1) * sizeof(StatisticEntry) );
          p->use[p->entries].element          = numname;
          p->use[p->entries].flag             = ref[i].flag;
          p->use[p->entries].element_is_place = is_place;
          p->entries++;
        }
      }
#endif
      if ( !(flag & SHORT_FILE) ) {
        nothing_printed = True;
        for (i = 0; i < RefEntries; i++) {
          if (    ( !is_place && ( !ref[i].flag || flag & NO_SCOPING ) )
               || (ref[i].flag & (VAL_REF | INIT_REF | TERM_REF )) )
            {
              if (nothing_printed)
                {
                  fprintf(ofp, "b\"");
                  nothing_printed = False;
                }
              else
                fprintf(ofp, "*");
	      PrintMeaning( ref[i].nr, ref[i].flag, ofp );
            }
        }
        if (!nothing_printed)
          fputc('\"',ofp);
      }
    }    
}

/**********************************************/
/* Ausgabe von Linkdefinitionen in Datei ofp. */
/**********************************************/

static void M_Print_Link_Symbols(FILE *ofp, M_Net net)
{
  int count;

  for (count = 0; count < net->NumLinks; count++)
  {
    if (net->Links[count])
      fprintf (ofp, "%s\n", net -> Links[count]);
  };
} /* M_Print_Link_Symbols */


/**********************************************/
/* Ausgabe von Transitionen in die Datei ofp. */
/**********************************************/

static void M_PrintTransitions( FILE *ofp, M_Transition tr, M_Net net,
                                bool is_phant )
{
  while (tr) {
    if (!(flag & SHORT_FILE))
      fprintf(ofp, "%u ", tr->numname);
    fprintf(ofp, "%d@%d", scale * tr->x_pos, scale * tr->y_pos);
    if (tr->label->anzahl > 0)
      {
        fprintf(ofp, "i\"");
        MSWriteMultiset(ofp, tr->label, &(tr->vars));
        fprintf(ofp,"\"");
      }
    if ((flag & MNET_SYNCHRONIZE)||(!bpn_templates))
      {
	if (tr->meaning)
	  fprintf( ofp, "b\"%s\"", tr->meaning );
	if (tr->links)
	  fprintf( ofp, "l\"%s\"", tr->links );
	if (tr->time_label)
	  fprintf( ofp, "E%.0fL%.0f", tr->time_label->low,tr->time_label->high);
      }
    if ((!(flag & MNET_SYNCHRONIZE))||(!bpn_templates))
      {
	PrintReferences( ofp, tr->RefEntries, tr->ref, False, tr->numname );
	if (tr->flags & TR_REPEAT)
	  fprintf(ofp, "b\"REPEAT\"" );
      }
    if (tr->flags != 0)
      fprintf(ofp, "v%d", tr->flags);
    fprintf(ofp, "g\"");
    PrintAnno(ofp, tr, is_phant);
    fprintf(ofp, "\"");
    PrintBlocks(ofp, net, tr->BlockEntries, tr->blocks);
    if ((flag & PHANTOM_TRANS_OUTPUT) && ((tr->flags & TR_CHAN0_ACCESS) != 0))
    {
      fprintf(ofp, "S");
    }  /* if Blockausgabe und Sync-Trans */
    PrintPhantoms(ofp, net, tr);
    fprintf(ofp, "\n");
    tr = tr->nexttrans;
  }

}

/*******************************************/
/* Ausgabe von TS-Kanten in die Datei ofp. */
/*******************************************/

static void M_PrintTSEdges( FILE *ofp, M_Transition tr, M_Net net )
{
  M_TSEdge ts;
  
  while (tr) {
    for(ts = tr->postset; ts; ts = ts->nextplace) {
      fprintf(ofp, "%u<%up\"", tr->numname, ts->place->numname);
      V_MSWriteMultiset(ofp, ts->label, &(tr->vars));
      fprintf(ofp,"\"");
      if (ts->flags & EDGE_INV)
        fprintf(ofp,"v4");
      fprintf(ofp,"\n");
    }
    tr = tr->nexttrans;
  }
}

/*******************************************/
/* Ausgabe von ST-Kanten in die Datei ofp. */
/*******************************************/

static void M_PrintSTEdges( FILE *ofp, M_Transition tr, M_Net net )
{
  M_STEdge st;
  
  while(tr) {
    for(st = tr->preset; st; st = st->nextplace) {
      fprintf(ofp, "%u>%up\"", st->place->numname, tr->numname);
      V_MSWriteMultiset(ofp, st->label, &(tr->vars));
      fprintf(ofp,"\"");
      if (st->flags & EDGE_INV)
        fprintf(ofp,"v4");
      fprintf(ofp,"\n");
    }
    tr = tr->nexttrans;
  }
}

/****************************************************************************/
/* Unterfunktion zur Ausgabe in ein File mittels der Funktion fkt           */
/****************************************************************************/

static void ausgabe(char *file,M_Net net,void (*fkt)(M_Net,FILE *))
{
    FILE *yyout;

    if ((yyout=fopen(file,"w"))==NULL) {
        fprintf(stderr,"Error: Could not open file %s !\n", file);
        exit(3);
    } else {
        printf("Writing in file %s ...\n",file);
        fkt(net,yyout);
        fclose(yyout);
    }
}

/********************************************************/
/* Ausgabe der Referenz-Informationen in die Datei ofp. */
/********************************************************/

static void M_PrintRefFile(M_Net net, FILE *ofp)
{
  RefNr         i;
  ShortEntryPtr ptr;
  Integer      *range;
  char          type;
  M_Place       pl;
  M_Transition  tr;
  M_Block       bl;
  BlockPtr      bl_ptr;

#ifdef REF_STATISTIC
  Cardinal32 j;
#endif
  
  fprintf(ofp, "PEP\nReferences_BPN2_HL\nFORMAT_N\nDEF\n");
  for(i=1; (ptr=RefToStruct(i)); i++) {
    fprintf(ofp, "%d R\"", i);
    PrintReference(i, 0, ofp);
    fputc('"',ofp);
#ifdef REF_STATISTIC
    if (! statistic[i].entries)
      Msg(INF,"Reference %d unused !",i);
    for(j=0; j<statistic[i].entries; j++)
      {
        if (statistic[i].use[j].element_is_place)
          fputc('P', ofp);
        else
          fputc('T', ofp);
        fprintf(ofp,"%d",statistic[i].use[j].element);
        if (statistic[i].use[j].element_is_place &&
            statistic[i].use[j].flag)
          fprintf(ofp,":%d",statistic[i].use[j].flag);
      }
    MYfree( statistic[i].use );
#endif
    type = ptr->s_type;
    switch ( type )
      {
      case _CHANNEL:
        fputc('C',ofp);
        break;
      case _STACK:
        fputc('S',ofp);
        break;
      case _VAR:
        fputc('V',ofp);
        break;
      case _PROC:
      case _PPROC:
        fputc('U',ofp);
        break;
      case _BEGIN:
        fputc('E',ofp);
        break;
      case _PROCCALL:
      case _ACTION:
        fputc('A',ofp);
        break;
      }
    if ( flag & MNET_SYNCHRONIZE )
      {
        if ( ((SyncRefEntry*)ptr)->s_description )
	  fprintf( ofp, "\"%s\"", ((SyncRefEntry*)ptr)->s_description );
        if ( ((SyncRefEntry*)ptr)->s_meaning     )
	  fprintf( ofp, "b\"%s\"", ((SyncRefEntry*)ptr)->s_meaning );
      }
    else
      {
	switch ( type )
	  {
	    case _BEGIN:
	      bl = net->Blocks;
	      while ((bl_ptr = BlockToStruct(bl->blocknr))->ref!=i)
		bl = bl->nextblock;
	      fprintf(ofp, "\"%d\"", bl_ptr->blocktags);
	      break;
	    case _CHANNEL:
            case _STACK:
	      fprintf(ofp,"\"{%d}",((SymTabEntry*)ptr)->s_init_or_cap);
	    case _VAR:
	      if (type==_VAR)
		fputc('"',ofp);
	      range = Set_Parser2Intern( ((SymTabEntry*)ptr)->s_set );
	      PrintSingleRange(range, ofp);
	      fputc('"',ofp);
	      MYfree(range);
	      break;
	  }
	fprintf(ofp,"b\"");
	PrintMeaning(i, 0, ofp);
	fputc('"',ofp);
      }
    fputc('\n',ofp);
  }  /* while noch Referenzen */

#ifdef REF_STATISTIC
  MYfree( statistic );
#endif

  PrintNet_Blocks_PEP(net, ofp, True);

  fprintf(ofp, "PL\n");
  for(pl = net->Places; pl; pl=pl->nextplace) {
    if (pl->RefEntries)
      {
        fprintf(ofp, "P%u ", pl->numname);
        for (i = 0; i < pl->RefEntries; i++) {
          if ( i>0 )
                fputc('*',ofp);
          else
            fprintf(ofp,"T\"");
          fprintf(ofp, "%u:%u", pl->ref[i].nr, pl->ref[i].flag);
        }
        if ( i>0 )
          fputc('"',ofp);
        fputc('\n',ofp);
      }
  }

  fprintf(ofp, "TR\n");
  for(tr = net->Transitions; tr; tr=tr->nexttrans) {
    if (tr->RefEntries)
      {
        fprintf(ofp, "T%u ", tr->numname);
        for (i = 0; i < tr->RefEntries; i++) {
          if ( i>0 )
            fputc('*',ofp);
          else
                fprintf(ofp,"T\"");
          fprintf(ofp, "%u:%u", tr->ref[i].nr, tr->ref[i].flag);
        }
        if ( i>0 )
          fputc('"',ofp);
        fputc('\n',ofp);
      }
  }

  if ( flag & PHANTOM_TRANS_OUTPUT )
  {
    fprintf(ofp, "PTR\n");
    for(tr = net->PTransitions; tr; tr=tr->nexttrans) {
      if (tr->RefEntries)
      {
        fprintf(ofp, "PT%u ", tr->numname);
        for (i = 0; i < tr->RefEntries; i++) {
          if ( i>0 )
            fputc('*',ofp);
          else
            fprintf(ofp,"T\"");
          fprintf(ofp, "%u:%u", tr->ref[i].nr, tr->ref[i].flag);
        }
        if ( i>0 )
          fputc('"',ofp);
        fputc('\n',ofp);
      }
    }
  }
}

/*********************************************************/
/* Ausgabe des Netzes im HL-Net-Format in die Datei ofp. */
/*********************************************************/

void M_PrintNet(M_Net net, FILE *ofp)
{
  M_Place pl;

  if (!net)
    return;  /* kein Netz, keine Ausgabe */

  /* Stellen und Transitionen mit eindeutigem numerischen Namen versehen */
  M_numname_all(net);

  /* Neues oder altes Format? */
  if (!(flag & BLOCK_OUTPUT) && !(flag & PHANTOM_TRANS_OUTPUT))
    fprintf(ofp, "PEP\nHLNet\nFORMAT_N\n");
  else
    fprintf(ofp, "PEP\nHLNet\nFORMAT_N2\n");

  /* Bloecke ausgeben */
  PrintNet_Blocks_PEP(net, ofp, False);

  /* Link Symbols */
  if (net->NumLinks >0)
  {
    fprintf(ofp, "LS\n");
    M_Print_Link_Symbols(ofp, net);
  };
  
  /* Stellen */
  fprintf(ofp, "PL\n");
  for(pl = net->Places; pl; pl=pl->nextplace) {
    if (!(flag & SHORT_FILE))
      fprintf(ofp, "%u ", pl->numname);
    fprintf(ofp, "%d@%d", scale * pl->x_pos, scale * pl->y_pos);
    switch(pl->label) {
      case ENTRY:
        fprintf(ofp, "e"); 
	if (!(flag & MNET_SYNCHRONIZE))
	  fprintf(ofp, "z\"%s\"y\"%s\"",entry_type,entry_type);
	break;
      case EXIT:
        fprintf(ofp, "x"); break;
      case EMPTY:
      default:
        if (!(flag & SHORT_FILE))
          fprintf(ofp, " ");
        break;
    }  /* switch ueber den Label */
    if ((flag & MNET_SYNCHRONIZE)||(!bpn_templates))
      {
	if (pl->init_marking)
	  fprintf( ofp, "z\"%s\"", pl->init_marking );
	if (pl->curr_marking)
	  fprintf( ofp, "y\"%s\"", pl->curr_marking );
	if (pl->meaning)
	  fprintf( ofp, "b\"%s\"", pl->meaning );
      }
    if (!(flag & MNET_SYNCHRONIZE)||(!bpn_templates))
      PrintReferences( ofp, pl->RefEntries, pl->ref, True, pl->numname );
    fprintf(ofp, "Z\"");
    PrintRange(pl, ofp);
    fprintf(ofp, "\"");
    PrintBlocks(ofp, net, pl->BlockEntries, pl->blocks);
    fprintf(ofp, "\n");
  } /* for Stellen */

  /* Transitionen */
  fprintf(ofp, "TR\n");
  M_PrintTransitions(ofp,net->Transitions,net,False);

  /* Phantomtranstionen */
  if (flag & PHANTOM_TRANS_OUTPUT) {
    fprintf(ofp, "PTR\n");
    M_PrintTransitions(ofp,net->PTransitions,net,True);
  }  /* if Ausgabe der Bloecke und Sync-/Phantomtransitionen gewuenscht */

  /* Kanten von Transitionen zu Stellen */
  fprintf(ofp, "TP\n");
  M_PrintTSEdges(ofp,net->Transitions,net);

  /* Kanten von Stellen nach Transitionen */
  fprintf(ofp, "PT\n");
  M_PrintSTEdges(ofp,net->Transitions,net);

  if (flag & PHANTOM_TRANS_OUTPUT) {
    /* Kanten von Phantomtranstitionen zu Stellen */
    fprintf(ofp, "PTP\n");
    M_PrintTSEdges(ofp,net->PTransitions,net);

    /* Kanten von Stellen nach Phantomtransitionen */
    fprintf(ofp, "PPT\n");
    M_PrintSTEdges(ofp,net->PTransitions,net);
  }
} /* M_PrintNet */

/********************************************************/
/* Ausgabe des Netzes net und ggf. seiner Referenzdatei */
/********************************************************/

void PrintNet_PEP( M_Net net )
{
#ifdef REF_STATISTIC
  RefNr i, first_unused; 

  if ( ref_file )
    {
      /* Tabelle fuer Referenzen-Statistik initialisieren */
      first_unused = 1;
      while( RefToStruct(first_unused) )
        first_unused++;
      statistic = MYmalloc( first_unused * sizeof( RefStatistic ) );
      for( i=1; i<first_unused; i++ )
        { 
          statistic[i].entries = 0;
          statistic[i].use     = MYmalloc( 0 * sizeof( StatisticEntry ) );
        }
    }
#endif

  ausgabe( pep_file, net, M_PrintNet );
  if ( ref_file )
    ausgabe( ref_file, net, M_PrintRefFile );
}


/*****************************************************************************/
/* Veraendert die Stellenliste so, dass alle Stellen, die man von            */
/* 'net->LastPlace' bzw. dem Rueckgabewert (einer Stelle) aus rueckwaerts    */
/* besucht, die Beschriftung 'label1' bzw. 'label2' aufweisen.               */
/*****************************************************************************/

M_Place M_PreparePlaces(M_Net net,PLabel label1,PLabel label2)
{
    register M_Place pl,pl1,pl2;
             M_Place merke;

    pl=net->LastPlace;
    net->LastPlace=merke=pl1=pl2=NULL;
    while (pl) {
        if (pl->label & label1) {
            if (pl1) {
                pl1=pl1->prevplace=pl;
            } else {
                pl1=net->LastPlace=pl;
            }
        } else if (pl->label & label2) {
            if (pl2) {
                pl2=pl2->prevplace=pl;
            } else {
                pl2=merke=pl;
            }
        }
        pl=pl->prevplace;
    }
    if (pl1)
        pl1->prevplace=NULL;
    if (pl2)
        pl2->prevplace=NULL;
    return merke;
}

/*****************************************************************************/
/* Stellt die normale Stellenliste wieder her und loescht dabei Stellen mit  */
/* der Beschriftung 'label'.                                                 */
/* Rueckgabewert ist das uebergebene Netz.                                   */
/*****************************************************************************/

M_Net M_RestoreKillPlaces(M_Net net,PLabel label)
{
    register M_Place pl;

    pl=net->Places;
    while (pl->nextplace) {
        pl=(pl->nextplace->prevplace=pl)->nextplace;
        if (pl->prevplace->label & label)
            M_RemovePlace(net,pl->prevplace);
    }
    if ((net->LastPlace=pl)->label & label)
        M_RemovePlace(net,pl);
    return net;
}

/*****************************************************************************/
/************************** M_empty_net() ************************************/
/*****************************************************************************/

M_Net M_empty_net( void )
  {

    register M_Net net = M_allocHead();

    net->Places = net->LastPlace = NULL;
    net->Transitions = net->LastTransition = NULL;
    net->PTransitions = net->LastPTransition = NULL;
    net->Links = NULL;
    M_InitBlockInNet(net);
    net->NumPlaces = net->NumTransitions = net->NumComponents = 0;
    net->NumPTransitions = 0;
    net->NumLinks = 0;
    net->MaxNumLinks = 0;
    net->x_dim = net->y_dim = 0; 
    net->x_pos = net->y_pos = MAX_INT; 
      /* damit die Suche nach dem Minimum funktioniert */

    return ( net );

  } /* M_empty_net() */

/*****************************************************************************/
/************************ M_fast_insert_block ********************************/
/*****************************************************************************/

M_Block M_fast_insert_block( M_Net net, Cardinal32 B_numname,
                             M_Block *bl_last, char *name )
{
    register M_Block bl,
                     blnew;
    SymTabEntry      entry;  /* Hilfsstruktur */
    BlockNr          blocknr;
 
    if ( ! ( bl = *bl_last ) || ( bl->numname > B_numname ) )
      {
        bl = net->Blocks;
      }
    while ( bl && ( bl->numname < B_numname ) )
      {
        bl = bl->nextblock;
      }
    if ( bl && ( B_numname == bl->numname ) )
      {
        fprintf( stderr,
               "ERROR: HL.M_fast_insert_block(): block '%u' already exists.\n",
                 B_numname );
        exit ( 1 );
      } /* if */
    blnew = ((M_Block) MYmalloc(sizeof(M_BlockElem)));

    entry.s_name = name;
    entry.s_blknum = 0;
    add_block( &entry, &blocknr, 0, 0 );

    if ( bl )
      {
        if ( (blnew->prevblock = bl->prevblock) )
          {
            blnew->prevblock->nextblock = blnew;
          }
        else
          {
            net->Blocks = blnew;
          }
        ( blnew->nextblock = bl ) -> prevblock = blnew;
      }
    else
      {
        if ( (blnew->prevblock = net->LastBlock) )
          {
            net->LastBlock->nextblock = blnew;
          }
        else
          {
            net->Blocks = blnew;
          }
        ( net->LastBlock = blnew ) -> nextblock = NULL;
      }
    blnew->blocknr = blocknr;
    blnew->numname = B_numname;
    blnew->x_pos = blnew->y_pos = 0;
    blnew->x_min = blnew->y_min = MAX_INT;
    blnew->x_max = blnew->y_max = 0;
    blnew->ever_used = False;
    net->NumBlocks++;

    return ( *bl_last = blnew );
} /* M_fast_insert_block() */

/*****************************************************************************/
/************************ M_fast_insert_place ********************************/
/*****************************************************************************/

M_Place M_fast_insert_place( M_Net net, Cardinal32 P_numname,
                             M_Place *pl_last )
{
    register M_Place pl,
                     plnew;

    if ( ! ( pl = *pl_last ) || ( pl->numname > P_numname ) )
      {
        pl = net->Places;
      }
    while ( pl && ( pl->numname < P_numname ) )
      {
        pl = pl->nextplace;
      }
    if ( pl && ( P_numname == pl->numname ) )
      {
        fprintf( stderr,
               "ERROR: HL.M_fast_insert_place(): place '%u' already exists.\n",
                 P_numname );
        exit ( 1 );
      }
    plnew = M_allocPlace();
    if ( pl )
      {
        if ( (plnew->prevplace = pl->prevplace) )
          {
            plnew->prevplace->nextplace = plnew;
          }
        else
          {
            net->Places = plnew;
          }
        ( plnew->nextplace = pl ) -> prevplace = plnew;
      }
    else
      {
        if ( (plnew->prevplace = net->LastPlace) )
          {
            net->LastPlace->nextplace = plnew;
          }
        else
          {
            net->Places = plnew;
          }
        ( net->LastPlace = plnew ) -> nextplace = NULL;
      }
    plnew->component = 0;
    plnew->numname = P_numname;
    plnew->preset = NULL;
    plnew->postset = NULL;
    plnew->ppreset = NULL;
    plnew->ppostset = NULL;
    plnew->label = EMPTY;
    plnew->range[0] = NULL;
    plnew->RefEntries = 0;
    plnew->BlockEntries = 0;
    plnew->blocks = NULL;
    plnew->meaning = NULL;
    plnew->init_marking = NULL;
    plnew->curr_marking = NULL;
    plnew->x_pos = plnew->y_pos = plnew->flags = 0;
    net->NumPlaces++;

    return ( *pl_last = plnew );
} /* M_fast_insert_place() */


/*****************************************************************************/
/*********************** M_fast_insert_transition ****************************/
/*****************************************************************************/

M_Transition M_fast_insert_transition( Cardinal32   * NumTransitionsPtr,
				       M_Transition * TransitionsPtr,
				       M_Transition * LastTransitionPtr,
                                       Cardinal32   T_numname,
                                       M_Transition * tr_last,
				       bool         do_phants )
{
    register M_Transition tr,
                          trnew;

    if ( ! ( tr = *tr_last ) || ( tr->numname > T_numname ) )
      {
        tr = (*TransitionsPtr);
      }
    while ( tr && ( tr->numname < T_numname ) )
      {
        tr = tr->nexttrans;
      }
    if ( tr && ( T_numname == tr->numname ) )
      {
        fprintf( stderr,
     "ERROR: HL.M_fast_insert_transition(): transition '%u' already exists.\n",
                 T_numname );
        exit ( 1 );
      }
    trnew = ((do_phants)?M_allocPTrans():M_allocTrans());
    if ( tr )
      {
        if ( (trnew->prevtrans = tr->prevtrans) )
          {
            trnew->prevtrans->nexttrans = trnew;
          }
        else
          {
            (*TransitionsPtr) = trnew;
          }
        ( trnew->nexttrans = tr ) -> prevtrans = trnew;
      }
    else
      {
        if ( (trnew->prevtrans = (*LastTransitionPtr)) )
          {
            (*LastTransitionPtr)->nexttrans = trnew;
          }
        else
          {
            (*TransitionsPtr) = trnew;
          }
        ( (*LastTransitionPtr) = trnew ) -> nexttrans = NULL;
      }
    trnew->preset = NULL;
    trnew->postset = NULL;
    trnew->numname = T_numname;
    trnew->label = NULL;
    trnew->guard = NULL;
    trnew->def = NULL;
    trnew->RefEntries = 0;
    trnew->ref = NULL;
    trnew->vars.ComEntries = 0;
    trnew->vars.ComArray = NULL;
    trnew->BlockEntries = 0;
    trnew->blocks = NULL;
    trnew->PhantomEntries = 0;
    trnew->phantoms = NULL;
    trnew->meaning = NULL;
    trnew->links=NULL;
    trnew->time_label=NULL;
    trnew->x_pos = trnew->y_pos = trnew->flags = 0;
    (*NumTransitionsPtr)++;

    return ( *tr_last = trnew );
} /* M_fast_insert_transition() */

/*****************************************************************************/
/************************* M_fast_insert_TP_edge *****************************/
/*****************************************************************************/

M_TSEdge M_fast_insert_TP_edge( M_Place      first_place,
				M_Transition first_trans,
                                Cardinal32   P_numname,
                                Cardinal32   T_numname,
                                M_Place      * pl_last,
                                M_Transition * tr_last,
                                M_TSEdge     * ts_last,
				bool         do_phants )
{
    register M_Place      pl;
    register M_Transition tr;
    register M_TSEdge     ts,
                          tsnew;
    register bool         kl1,
                          kl2;

    pl = *pl_last;
    if ( ( ts = *ts_last ) && ( ts->place->numname == P_numname ) )
      {
        pl = ts->place;
      }
    else if ( pl && ( pl->numname == P_numname ) )
      {
        ts = NULL;
      }
    else if ( pl && ts && ( ( kl1 = ( pl->numname < P_numname ) ),
                            ( kl2 = ( ts->place->numname < P_numname ) ),
                            ( kl1 || kl2 ) ) )
      {
        if ( ( ! kl1 ) || ( kl2 && ( pl->numname < ts->place->numname ) ) )
          {
            pl = ts->place;
          }
        else
          {
            ts = NULL;
          }
      }
    else
      {
        pl = first_place;
        ts = NULL;
      }
    while ( pl && ( pl->numname < P_numname ) )
      {
ts = NULL;
        pl = pl->nextplace;
      }

    if ( ! ( pl && ( pl->numname == P_numname ) ) )
      {
        WARNING_P_NUM( "M_fast_insert_TP_edge", P_numname );
        return ( NULL );
      }

    if ( ( ts && ( ts->trans->numname < T_numname ) ) || 
	 ( ts = (do_phants ? pl->ppreset:pl->preset) ) )
      {
        if ( T_numname < ts->trans->numname )
          {
            ts = NULL;
          }
        else
          {
            while ( ts->nexttrans
                    && ( ts->nexttrans->trans->numname < T_numname ) )
              {
                ts = ts->nexttrans;
              }
            if ( ( tsnew = ts->nexttrans )
                 && ( tsnew->trans->numname == T_numname ) )
              {
                *pl_last = ts->place;
                *tr_last = ts->trans;
                return ( *ts_last = ts );
              }
          }
      }
    tr = ( *tr_last && ( (*tr_last)->numname == T_numname ) ) ? *tr_last
                                           : ts ? ts->trans : first_trans;
    while ( tr && ( tr->numname < T_numname ) )
      {
        tr = tr->nexttrans;
      }
    if ( ! ( tr && ( tr->numname == T_numname ) ) )
      {
        WARNING_T_NUM( "M_fast_insert_TP_edge", T_numname );
        return ( NULL );
      }

    tsnew = M_allocTSEdge();
    tsnew->label = NULL;
    tsnew->flags = 0;
    if ( ts )
      {
        if ( (tsnew->nexttrans = ts->nexttrans) )
          {
            tsnew->nexttrans->prevtrans = tsnew;
          }
        ( tsnew->prevtrans = ts ) -> nexttrans = tsnew;
      }
    else
      {
        if ( (tsnew->nexttrans = (do_phants ? pl->ppreset:pl->preset)) )
          {
            tsnew->nexttrans->prevtrans = tsnew;
          }
        ( (do_phants ? pl->ppreset:pl->preset) = tsnew ) -> prevtrans = NULL;
      }
    if ( (ts = tr->postset) )
      {
        if ( P_numname < ts->place->numname )
          {
            ts = NULL;
          }
        else
          {
            while ( ts->nextplace
                    && ( ts->nextplace->place->numname < P_numname ) )
              {
                ts = ts->nextplace;
              }
          }
      }
    if ( ts )
      {
        if ( (tsnew->nextplace = ts->nextplace) )
          {
            tsnew->nextplace->prevplace = tsnew;
          }
        ( tsnew->prevplace = ts ) -> nextplace = tsnew;
      }
    else
      {
        if ( (tsnew->nextplace = tr->postset) )
          {
            tsnew->nextplace->prevplace = tsnew;
          }
        ( tr->postset = tsnew ) -> prevplace = NULL;
      }
    *pl_last = tsnew->place = pl;
    *tr_last = tsnew->trans = tr;

    return ( *ts_last = tsnew );
} /* M_fast_insert_TP_edge() */


/*****************************************************************************/
/************************* M_fast_insert_PT_edge *****************************/
/*****************************************************************************/

M_STEdge M_fast_insert_PT_edge( M_Place      first_place,
				M_Transition first_trans,
                                Cardinal32   P_numname,
                                Cardinal32   T_numname,
                                M_Place      * pl_last,
                                M_Transition * tr_last,
                                M_STEdge     * st_last,
				bool         do_phants )
{
    register M_Place      pl;
    register M_Transition tr;
    register M_STEdge     st,
                          stnew;
    register bool         kl1,
                          kl2;

    pl = *pl_last;
    if ( ( st = *st_last ) && ( st->place->numname == P_numname ) )
      {
        pl = st->place;
      }
    else if ( pl && ( pl->numname == P_numname ) )
      {
        st = NULL;
      }
    else if ( pl && st && ( ( kl1 = ( pl->numname < P_numname ) ),
                            ( kl2 = ( st->place->numname < P_numname ) ),
                            ( kl1 || kl2 ) ) )
      {
        if ( ( ! kl1 ) || ( kl2 && ( pl->numname < st->place->numname ) ) )
          {
            pl = st->place;
          }
        else
          {
            st = NULL;
          }
      }
    else
      {
        pl = first_place;
        st = NULL;
      }
    while ( pl && ( pl->numname < P_numname ) )
      {
        st = NULL;
        pl = pl->nextplace;
      }

    if ( ! ( pl && ( pl->numname == P_numname ) ) )
      {
        WARNING_P_NUM( "fast_insert_PT_edge", P_numname );
        return ( NULL );
      } /* if */

    if ( ( st && ( st->trans->numname < T_numname ) ) || 
	 ( st = (do_phants ? pl->ppostset:pl->postset) ) )
      {
        if ( ( T_numname < st->trans->numname ) )
          {
            st = NULL;
          } /* if */
        else
          {
            while ( st->nexttrans
                    && ( st->nexttrans->trans->numname < T_numname ) )
              {
                st = st->nexttrans;
              } /* while */
            if ( ( stnew = st->nexttrans )
                 && ( stnew->trans->numname == T_numname ) )
              {
                *pl_last = st->place;
                *tr_last = st->trans;
                return ( *st_last = st );
              } /* if */
          } /* else */
      } /* if */
    tr = ( *tr_last && ( (*tr_last)->numname == T_numname ) ) ? *tr_last
                                           : st ? st->trans : first_trans;
    while ( tr && ( tr->numname < T_numname ) )
      {
        tr = tr->nexttrans;
      }
    if ( ! ( tr && ( tr->numname == T_numname ) ) )
      {
        WARNING_T_NUM( "fast_insert_PT_edge", T_numname );
        return ( NULL );
      }
    stnew = M_allocSTEdge();
    stnew->label = NULL;
    stnew->flags = 0;
    if ( st )
      {
        if ( (stnew->nexttrans = st->nexttrans) )
          {
            stnew->nexttrans->prevtrans = stnew;
          }
        ( stnew->prevtrans = st ) -> nexttrans = stnew;
      }
    else
      {
        if ( (stnew->nexttrans = (do_phants ? pl->ppostset:pl->postset)) )
          {
            stnew->nexttrans->prevtrans = stnew;
          }
        ( (do_phants ? pl->ppostset:pl->postset) = stnew ) -> prevtrans = NULL;
      }
    if ( (st = tr->preset) )
      {
        if ( ( P_numname < st->place->numname ) )
          {
            st = NULL;
          }
        else
          {
            while ( st->nextplace
                    && ( st->nextplace->place->numname < P_numname ) )
              {
                st = st->nextplace;
              }
          }
      }
    if ( st )
      {
        if ( (stnew->nextplace = st->nextplace) )
          {
            stnew->nextplace->prevplace = stnew;
          }
        ( stnew->prevplace = st ) -> nextplace = stnew;
      }
    else
      {
        if ( (stnew->nextplace = tr->preset) )
          {
            stnew->nextplace->prevplace = stnew;
          }
        ( tr->preset = stnew ) -> prevplace = NULL;
      }
    *pl_last = stnew->place = pl;
    *tr_last = stnew->trans = tr;

    return ( *st_last = stnew );
} /* M_fast_insert_PT_edge() */


/*****************************************************************************/
/************************* M_insert_link_symbol *****************************/
/*****************************************************************************/
void M_insert_link_symbol(M_Net net, char *ls)
{
  Cardinal32 MaxLinkSymbs = net->MaxNumLinks;
  int count;

  if ( net->NumLinks >= MaxLinkSymbs )
  {
    if (net->Links)
      net->Links = (linkdef *) MYrealloc( net->Links,
                   ( MaxLinkSymbs += 10 ) * sizeof( linkdef *) );
    else
      net->Links = (linkdef *) MYmalloc((MaxLinkSymbs += 10)*sizeof(linkdef*));
    for ( count = MaxLinkSymbs - 10 ; count < MaxLinkSymbs ; count++ )
     {
       net->Links[count] = NULL;
     }
    net->MaxNumLinks = MaxLinkSymbs;
  }
  count = strlen(ls);
  net->Links[net->NumLinks] =
                (linkdef) MYmalloc( (count+1) * sizeof (linkdef) );
  strcpy(net->Links[net->NumLinks], ls);
  net->NumLinks++;
} /* M_insert_link_symbol */

