/*****************************************************************************/
/*  multiset.c                                                               */
/*  Mi.,04.11.1992 - Mi.,04.11.1992                           Stefan Roemer  */
/*  Martin Ackermann, Fruehjahr 95                                           */
/*****************************************************************************/

/*
Mit den Routinen in MULTISET.C werden Multimengen von HL-Aktionen behandelt.

Die Namen der HL-Aktionen werden in einer Symboltabelle namens ActTable 
abgelegt. Die Reihenfolge der Namen in der Symboltabelle ergibt sich aus der 
Reihenfolge des Eintragens, z.B. ActTable = [ BB, A@e, C ], wenn zuerst
'BB', dann 'A@e' und schliesslich 'C' eingetragen wird.
Spaeter werden alle Aktionsnamen nur noch ueber ihren Index in ActTable
angesprochen. Im obigen Beispiel steht also 0 fuer 'BB', 1 fuer 'A@e' und 2 
fuer 'C'.

Eine Multimenge von HL-Aktionen ist ein dynamisches Array von Elementen 
folgenden Typs:

typedef struct Action_
{ Action_nr	nr;		// Aktionsnummer
  Cardinal	arity;		// Aritaet der Aktion
  ComVar	com[MaxArity];	// Kommunikationsvariablen
  Cardinal	anzahl;		// Anzahl des Auftretens in der Multimenge 
} Action, *MultiSet;

Damit alle Routinen effizient arbeiten, werden die Eintraege im Array zunaechst
nach aufsteigender 'nr', dann nach aufsteigender 'arity' und schliesslich nach
aufsteigenden 'com's sortiert. Das Ende des Arrays wird durch einen Eintrag mit
'anzahl' = 0 markiert. 

Beispiel (siehe oben und VAR_MSET.C): Die Multimenge 
{ A@e, A@e, A@e, BB(a,b1), BB(a,a), BB(b2), BB(b2)} 
wird repraesentiert durch das Array 
[ {0,1,[0,x,....,x],2} ,        // fuer 2 mal BB(b2),   d.h. arity = 1
  {0,2,[1,1,x,..,x],1} ,        // fuer 1 mal BB(a,a),  d.h. arity = 2
  {0,2,[1,2,x,..,x],1} ,        // fuer 1 mal BB(a,b1), d.h. arity = 2
  {2,0,[x,......,x],3} ,        // fuer 3 mal A@e,      d.h. arity = 0
  {x,x,[x,......,x],0} ],       // Schluss-Element
wobei 'x' jeweils beliebig ist.

Da man nicht vorhersehen kann, wie oft eine Aktion mit verschiedenen Aritaeten
oder com's in einer Multimenge auftreten wird, koennen die Mengen mmDummy und
mmLeaf nicht mehr dynamisch angepasst werden. Sie werden so dimensioniert, dass
jede Aktion (im Durchschnitt) in MaxComb Varianten auftreten kann. Folglich
muessen die Routinen, die mmDummy, mmLeaf & mmSub zur Aufnahme grosser 
Multimengen verwenden, Ueberlaeufe abfangen. Das ist zur Zeit nur in MSsum 
noetig.

*/

  /************/
  /* includes */
  /************/

#include <stdio.h>
#include <string.h>

#include "frontend.h"
#include "multiset.h"
#include "tables.h"
#include "annotat.h"

/*****************************************************************************/
/* Globale Variablen fuer die Behandlung von Multimengen.                    */
/*****************************************************************************/

Cardinal MaxActions   = ACTION_BUF_SIZE;
Cardinal MaxVars      = ACTION_BUF_SIZE;
Cardinal MaxRefs      = ACTION_BUF_SIZE;
Cardinal MaxTpls      = ACTION_BUF_SIZE;

Cardinal MaxComb      = 1; 
  /* In bpn2-Netzen tritt in jedem multiset (auch wenn es durch MSsum entsteht)
     jede Action stets mit den selben Variablen auf. */

Cardinal ActCounter  = 0;

Cardinal VarCounter  = 0;

Cardinal RefCounter  = 0;

Cardinal TplCounter  = 0;

MultiSet mmDummy    = NULL;       /* Hilfsmenge bei der Berechnung von sum,  */
                                  /* union,...                               */

MultiSet mmLeaf     = NULL;       /* Hilfsmenge fuer externe Berechnungen    */

MultiSet mmSub      = NULL;       /* Hilfsmenge bei der Berechnung einer     */
                                  /* substituierten Kopie                    */

MultiSet mmSingle   = NULL;       /* Hilfsmenge fuer Multimengen mit maximal */
                                  /* einem Eintrag                           */

bool  MSdisplay_style = MS_WIDE;

/*****************************************************************************/
/* Initialisieren der globalen Variablen des Multimengen-Funktionspakets.    */
/*****************************************************************************/

void InitMultiSets()

  {
    if ( flag & MNET_SYNCHRONIZE )
      MaxComb = 10;

    if ( mmDummy )
      {
        MSremove( mmDummy );
      } /* if */
    mmDummy  = MSalloc( MaxActions*MaxComb );
    if ( mmLeaf )
      {
        MSremove( mmLeaf );
      } /* if */
    mmLeaf   = MSalloc( MaxActions*MaxComb );
    if ( mmSub )
      {
        MSremove( mmSub );
      } /* if */
    mmSub   = MSalloc( MaxActions*MaxComb );
    if ( ! mmSingle )
      {
        mmSingle = MSalloc( 1 );
        mmSingle[1].anzahl = 0;
      } /* if */

  } /* InitMultiSets() */

/*****************************************************************************/
/* Kopiert eine HL-Aktion                                                    */
/*                                                                           */
/* Falls sub != NULL ist, werden alle auftretenden Kommunikationsvariablen   */
/* mittels sub substituiert.                                                 */
/*****************************************************************************/

void CopySubAction( Action *a1, Action *a2, ComVar *sub )
{ 
  Cardinal i;

  a1->anzahl	= a2->anzahl;
  a1->nr	= a2->nr;
  a1->arity	= a2->arity;
  
  for( i = 0; i < a2->arity; i++ )
    if ( sub && a2->com[i] < V_CONST_OFFSET )
      a1->com[i] = sub[ a2->com[i] ];
    else
      a1->com[i] = a2->com[i];
}

/*****************************************************************************/
/* Kopiert eine HL-Aktion                                                    */ /*****************************************************************************/

void CopyAction( Action *a1, Action *a2 )
{ 
  CopySubAction( a1, a2, NULL );
}

/*****************************************************************************/
/* Vergleicht zwei HL-Aktionen (s.o.):                                       */
/* 0  ->  gleich                                                             */
/* -  ->  a1 > a2                                                            */
/* +  ->  a1 < a2                                                            */
/*****************************************************************************/

int Comp( Action *a1, Action *a2 )
{
  Cardinal i;

  if ( a1->nr != a2->nr ) 
    return ( (int) a2->nr - a1->nr );
  else 
    if ( a1->arity != a2->arity )
      return ( (int) a2->arity - a1->arity );
    else
    { i = 0;
      while ( i < a1->arity )
      { if ( a1->com[i] != a2->com[i] )
          return ( (int) a2->com[i] - a1->com[i] );
        i++;
      }
    }
  return 0;
}

/*****************************************************************************/
/* Dasselbe wie Comp, vergleicht aber keine Com-Nummern			     */
/*****************************************************************************/

int WeakComp( Action *a1, Action *a2 )
{
  if ( a1->nr != a2->nr ) 
    return ( (int) a2->nr - a1->nr );
  else 
    return ( (int) a2->arity - a1->arity );
}

/*****************************************************************************/
/* Funktion zum Kopieren einer Multimenge in eine bereits bestehende.        */
/*                                                                           */
/* Falls sub != NULL ist, werden alle auftretenden Kommunikationsvariablen   */
/* mittels sub substituiert.                                                 */
/*****************************************************************************/

MultiSet MScopy(MultiSet mm1,MultiSet mm2,ComVar *sub)

  {
    if ( sub )
	{   
	    Action sub_act[2];

	    sub_act[1].anzahl = 0;
	    mm1->anzahl = 0;
	    
	    while ( mm2->anzahl != 0 ) 
		{
		    CopySubAction( sub_act, mm2++, sub );
		    MSsum( mm1, sub_act );
	            /* Durch die Substitution kann sich die Reihenfolge 
                       der Aktionen aendern. Deshalb werden sie einzeln
		       mittels MSsum eingefuegt.                             */
		}
	}
    else
	{   
	    register MultiSet mm1Ptr = mm1;
	    
	    while ( mm2->anzahl != 0 )
		CopyAction( mm1Ptr++, mm2++ );
	    mm1Ptr->anzahl = 0;
	}

    return ( mm1 );

  } /* MScopy() */

/*****************************************************************************/
/* Funktion zur Initialisierung einer leeren Multimenge.                     */
/*****************************************************************************/

MultiSet MScreate_empty_set()

  {                                                       /* Gibt {} zurueck */

    MultiSet mm;

    ( mm = MSalloc( 0 ) )->anzahl = 0;

    return ( mm );

  } /* MScreate_empty_set() */

/*****************************************************************************/
/* Funktion zur Initialisierung einer neuen Multimenge, die als einziges     */
/* Element die uebergebene Aktion mit dem Code `akt` enthaelt.               */
/*****************************************************************************/

MultiSet MScreate_singleton(Action_nr act)

  {

    register MultiSet mm;

    ( mm = MSalloc( 1 ) )->nr = act;
    mm->anzahl = 1;
    mm->arity = 0;
    ( mm + 1 )->anzahl = 0;

    return ( mm );

  } /* MScreate_singleton() */

/*****************************************************************************/
/* Funktion zur Initialisierung einer neuen Multimenge als Kopie einer be-   */
/* reits vorhandenen Multimenge; der Speicherbedarf der neuen Menge ent-     */
/* spricht ihrer tatsaechlichen "Laenge".                                    */
/*****************************************************************************/

MultiSet MScreate_copy( MultiSet mm )
  {

    MultiSet          mm2;        /* Zeiger auf die zu erzeugende Kopie.     */
    register MultiSet mmptr = mm, /* Laufzeiger durch die Originalmenge.     */
                      mm2ptr;     /* Laufzeiger durch die Kopie.             */
    register Cardinal len = 0;    /* Laenge der Originalmenge.               */

                                  /* "Laenge" der Multimenge ermitteln und   */
                                  /* in `len` ablegen.                       */
    while ( mmptr->anzahl )
      {
        len++;
        mmptr++;
      }
                                  /* Speicherplatz fuer die Kopie anfordern. */
    mm2ptr = mm2 = MSalloc( len );
    mmptr = mm;
                                  /* Eintraege einzeln kopieren.             */
    while ((mmptr->anzahl)!=0)
      CopyAction(mm2ptr++, mmptr++);
    mm2ptr->anzahl = 0;
                                  /* Zeiger auf die Kopie zurueckliefern.    */
    return ( mm2 );

  } /* MScreate_copy() */

/*****************************************************************************/
/* Funktion zur Berechnung der Maechtigkeit einer Multimenge                 */
/*****************************************************************************/

Cardinal MScardinality(MultiSet mm)

  {

    Cardinal card = 0;            /* Kardinalitaet der Menge.                */

                                  /* Durch die Menge laufen                  */
                                  /*             und dabei die Anzahlen der  */
                                  /* Vorkommen der einzelnen Aktionen auf-   */
                                  /* summieren.                              */
    while ( mm->anzahl )
      {
        card += (mm++)->anzahl;
      } /* while */

                                  /* Ermittelte Kardinalitaet zurueckgeben.  */
    return ( card );

  } /* MScardinality() */

/*****************************************************************************/
/* Funktion zur Berechnung der Anzahl verschiedener Aktionen in einer Multi- */
/* menge.                                                                    */
/*****************************************************************************/

Cardinal MSlength(MultiSet mm)

  {

    Cardinal len = 0;             /* "Laenge" der Menge.                     */

                                  /* Durch die Menge laufen                  */
                                  /*             und dabei zaehlen, wieviele */
                                  /* Aktionen in der Menge sind.             */
    while ( (mm++)->anzahl )
      {
        len++;
      } /* while */

                                  /* Ermittelte Laenge zurueckliefern.       */
    return ( len );

  } /* MSlength() */

/*****************************************************************************/
/* Funktion zur Feststellung, ob eine Aktion mit der Nummer `akt` und der    */
/* Aritaet 'arity' in der Multimenge `mm` vorkommt.                          */
/* Rueckgabe: Zeiger auf die Aktion ( NULL falls nicht enthalten )           */
/*****************************************************************************/

Action *MSis_in( Action_nr akt, Cardinal arity, MultiSet mm )

  {

    if ( akt && mm )
      {
                                  /* In der Menge diejenige Stelle suchen,   */
                                  /* an der die Aktion `akt` stehen muss.    */
        while ( mm->anzahl && ( mm->nr < akt ) )
          {
            mm++;
          } /* while */

        while ( mm->anzahl && ( mm->nr == akt ) )
          { if ( mm->arity == arity ) return mm;
            mm++;
          }

      } /* if */
                                  /* Falls bei der Parameteruebergabe ein    */
                                  /* Fehler aufgetreten ist, oder die        */
                                  /* Aktionsnummer nicht enthalten ist, wird */
                                  /* NULL als Pointer zurueckgegeben.        */
    return ( NULL );

  } /* MSis_in() */

/*****************************************************************************/
/* Funktion zur Feststellung, ob zwei Multimengen `mm1` und `mm2` gleich     */
/* sind.                                                                     */
/*****************************************************************************/

bool MSis_equal(MultiSet mm1,MultiSet mm2)

  {

                                  /* Die Mengen aktionsweise durchforsten:   */
    while ( mm1->anzahl )
      {
                                  /* Bei der ersten Nicht-Uebereinstimmung   */
        if ( ( mm1->anzahl!=mm2->anzahl ) || ( Comp( mm1++, mm2++ ) != 0 ) )
          {
                                  /* wird die weitere Suche abgebrochen und  */
                                  /* 'False' als Ergebnis geliefert.         */
            return ( False );
          }
      } /* while */

                                  /* An dieser Stelle wurde die erste Menge  */
                                  /* bis zum Ende abgearbeitet; das Ergebnis */
                                  /* des Vergleichs ist 'True', wenn jetzt   */
                                  /* auch die zweite Menge 'leer' ist, sonst */
                                  /* 'False'.                                */
    return ( !mm2->anzahl );

  } /* MSis_equal() */

/*****************************************************************************/
/* Funktion zur Summenbildung zweier Multimengen.                            */
/*****************************************************************************/

MultiSet MSsum(MultiSet mm1,MultiSet mm2)

  {

    register MultiSet mm1Ptr = mm1,
                      mm2Ptr = mm2,
                      mmDum  = mmDummy;
    Cardinal new_length = MSlength( mm1 ) + MSlength( mm2 );

    while ( new_length > MaxActions*MaxComb )
      {
	/* Platz schaffen */
	MaxComb++;
	MSresize( mmDummy, MaxActions*MaxComb );
	MSresize( mmLeaf,  MaxActions*MaxComb );
	MSresize( mmSub,   MaxActions*MaxComb );
        mmDum = mmDummy;
      }

    while ( mm1Ptr->anzahl || mm2Ptr->anzahl )
      {
        if ( mm1Ptr->anzahl && mm2Ptr->anzahl && ( Comp(mm1Ptr,mm2Ptr)==0 ) )
          {
            CopyAction( mmDum, mm1Ptr );
            if ( ( (mmDum++)->anzahl = mm1Ptr->anzahl + mm2Ptr->anzahl ) < mm1Ptr->anzahl )
              {
                Msg( FAT, "Overflow: action name too often used. Remedy: 'action'" );
              } /* if */
            mm1Ptr++, mm2Ptr++;
          } /* if */
        else if ( ( ! mm2Ptr->anzahl ) || ( mm1Ptr->anzahl && ( Comp(mm1Ptr,mm2Ptr)>0 ) ) )
          {
            CopyAction( mmDum++, mm1Ptr++ );
          } /* else if */
        else /* ( ( ! mm1Ptr->anzahl ) || ( Comp(mm1Ptr,mm2Ptr)<0 ) ) */
          {
            CopyAction( mmDum++, mm2Ptr++ );
          } /* else */
      } /* while */

    mmDum->anzahl = 0;
    mmDum = mmDummy;
    mm1Ptr = mm1;
    while ((mmDum->anzahl)!=0)
      CopyAction( mm1Ptr++, mmDum++ );
    (mm1Ptr)->anzahl = 0;

    return ( mm1 );

  } /* MSsum() */

/*****************************************************************************/
/* Funktion zur Differenzbildung zweier Multimengen.                         */
/*****************************************************************************/

MultiSet MSdifference(MultiSet mm1,MultiSet mm2)

  {

    register MultiSet mm1rPtr = mm1, /* Lesezeiger    */
                      mm1wPtr = mm1, /* Schreibzeiger */
                      mm2Ptr  = mm2;

    while ( mm1rPtr->anzahl )
      {
        if ( mm2Ptr->anzahl && ( Comp(mm1rPtr,mm2Ptr)==0 ) )
          {
            CopyAction( mm1wPtr, mm1rPtr );
            mm1wPtr->anzahl = ( mm1rPtr->anzahl > mm2Ptr->anzahl ) 
                              ? mm1rPtr->anzahl - mm2Ptr->anzahl : 0;
            mm1rPtr++;
            mm2Ptr++;
            if ( mm1wPtr->anzahl ) mm1wPtr++;
          } /* if */
        else if ( mm2Ptr->anzahl && ( Comp(mm1rPtr,mm2Ptr)<0 ) )
          {
            mm2Ptr++;
          } /* else if */
        else /* ( ( ! mm2Ptr->anzahl ) || ( Comp(mm1rPtr,mm2Ptr)>0 ) ) */
          {
            CopyAction( mm1wPtr++, mm1rPtr++ );
          } /* else */
      } /* while */

    mm1wPtr->anzahl = 0;

    return ( mm1 );

  } /* MSdifference() */

MultiSet MSweakdifference(MultiSet mm1,MultiSet mm2)
{
	register MultiSet mm1rPtr = mm1, /* Lesezeiger    */
			  mm1wPtr = mm1, /* Schreibzeiger */
			  mm2Ptr  = mm2;

	while (mm2Ptr->anzahl)
	{
	    mm1rPtr = mm1wPtr = mm1;

	    while (mm1rPtr->anzahl)
	    {
		if (WeakComp(mm1rPtr,mm2Ptr)==0)
		{
			CopyAction(mm1wPtr, mm1rPtr);
			mm1wPtr->anzahl = (mm1rPtr->anzahl > mm2Ptr->anzahl)?
				mm1rPtr->anzahl - mm2Ptr->anzahl : 0;
			mm1rPtr++;
			if (mm1wPtr->anzahl) mm1wPtr++;
		}
		else CopyAction(mm1wPtr++, mm1rPtr++);
	    }
	    mm2Ptr++;
	    mm1wPtr->anzahl = 0;

	}

	return mm1;

}

/*****************************************************************************/
/* Funktion zum Entfernen von `num` Paaren `{ akt, ^akt }` aus der Multi-    */
/* menge `mm`. Hilfsfunktion fuer den Synchronisations-Operator.             */
/*                                                                           */
/* Falls sub != NULL ist, werden alle auftretenden Kommunikationsvariablen   */
/* mittels sub substituiert.                                                 */
/*****************************************************************************/

MultiSet MSremove_action_pairs(MultiSet mm,Action *akt,ComVar* sub,Cardinal num)

  {
    Cardinal i;

                                  /* Parameter auf Gueltigkeit testen.       */
    if ( mm && akt && num )
      {
                                  /* Hilfsmenge `{ num akt, num ^akt }`      */
                                  /* in `mmDummy` aufbauen und               */
        mmDummy[0].nr = ( mmDummy[1].nr = akt->nr | 1 ) - 1;
        mmDummy[0].anzahl = mmDummy[1].anzahl = num;
        mmDummy[0].arity  = mmDummy[1].arity  = akt->arity;
        for( i = 0; i < akt->arity; i++ )
	  if ( sub && akt->com[i] < V_CONST_OFFSET )
	    mmDummy[0].com[i] = mmDummy[1].com[i] = sub[ akt->com[i] ];
	  else
	    mmDummy[0].com[i] = mmDummy[1].com[i] = akt->com[i];
        mmDummy[2].anzahl = 0;
                                  /* von der Menge `mm` abziehen.            */
        MSdifference( mm, mmDummy );
      } /* if */

                                  /* Ergebnismenge zurueckliefern.           */
    return ( mm );

  } /* MSremove_action_pairs() */

/*****************************************************************************/
/* Unterfunktion von MSWriteMultiset:                                        */
/* wandelt eine HL-Aktion in einen String um, z.B. A" oder "A(b,c)" etc.     */
/*****************************************************************************/

static char *VerboseActToStr( Action *action, char *string, Variables v_ptr )
{ 
  Cardinal i;

  ActionToStr( action->nr, string );
  if (action->arity)
  { strcat( string, "(" );
    for( i=0; i<action->arity; i++)
    { StrCatVar( action->com[i], v_ptr, string );
      strcat( string, "," );
    }
    string[strlen(string)-1] = ')'; /* ersetze Komma durch Klammer */
  }
  return string;
}

/*****************************************************************************/
/* Funktion zur Ausgabe einer Multimenge ( Datei oder Bildschirm ).          */
/*****************************************************************************/

void MSWriteMultiset(FILE *fileptr,MultiSet mm, Variables v_ptr)

  {

    Cardinal i;                   /* Zaehlvariable.                          */
    char     dum[ MAX_IDENT_LENGTH ];          /* String zur Aufnahme des    */
                                               /* anzuzeigenden Namens.      */

        /* fputc( '{', fileptr ); */
        if ( mm->anzahl )
          {
                                  /* Menge ist nicht leer.                   */
            if ( MSdisplay_style )
              {
                                  /* Ausgabe in Kurzschreibweise.            */
                if ( mm->anzahl > 1 ) fprintf( fileptr, FORMAT_CARD, mm->anzahl );
                fprintf( fileptr, "%s", VerboseActToStr( mm++, dum, v_ptr ) );
                while ( mm->anzahl )
                  {
                    fputc( ',', fileptr );
                    if ( mm->anzahl > 1 ) fprintf( fileptr, FORMAT_CARD, mm->anzahl );
                    fprintf( fileptr, "%s",
                                         VerboseActToStr( mm++, dum, v_ptr ) );
                  } /* while */
              } /* if */
            else /* WIDE */
              {
                                   /* Ausfuehrliche Ausgabe der Menge.       */
                fprintf( fileptr, "%s", VerboseActToStr( mm, dum, v_ptr ) );
                for ( i = 2 ; i <= mm->anzahl ; i++ )
                  {
                    fprintf( fileptr, ",%s", VerboseActToStr( mm, dum, v_ptr ) );
                  } /* for */
                mm++;
                while ( mm->anzahl )
                  {
                    for ( i = 1 ; i <= mm->anzahl ; i++ )
                      {
                        fprintf( fileptr, ",%s", VerboseActToStr( mm, dum, v_ptr ) );
                      } /* for */
                    mm++;
                  } /* while */
              } /* else */
          } /* if */
          /* fprintf( fileptr, "}" ); */

  } /* MSWriteMultiset() */

/*****************************************************************************/
/* Funktion zur Feststellung, ob zwei Multimengen `mm1` und `mm2` disjunkt   */
/* sind. Die Konjugation der Aktionen bleibt unberuecksichtigt.              */
/*****************************************************************************/

bool MSis_disjunct(MultiSet mm1,MultiSet mm2)

  {
    Action_nr act1, act2;

    while ( (mm1->anzahl != 0) && (mm2->anzahl != 0) )
      {
	act1 = mm1->nr; act2 = mm2->nr;

        if ( (act1|=1) < (act2|=1) )
          {
            mm1++;
          }
        else if ( act2 < act1 )
          {
            mm2++;
          }
        else /* act1 == act2 */
          {
            if ( mm1->arity < mm2->arity )
              mm1++;
            else if ( mm2->arity < mm1->arity )
              mm2++;
            else /* mm1->arity == mm2->arity */
              return False;
          }
      } /* while */
    return ( True );

  } /* MSis_disjunct() */

  /*** Dateiende multiset.c ***/

