/*****************************************************************************/
/*  var_mset.c                                                               */
/*  Mi.,04.11.1992 - Mi.,04.11.1992                           Stefan Roemer  */
/*  Martin Ackermann, Fruehjahr 95                                           */
/*****************************************************************************/

/*
Mit den Routinen in VAR_MSET.C werden Multimengen von Kommunikationsvariablen 
behandelt. Sie sind hervorgegangen aus MULTISET.C von Stefan Roemer.

Die Namen der Kommunikationsvariablen werden in einer Symboltabelle namens
VarTable abgelegt. Die Reihenfolge der Namen in der Symboltabelle ergibt sich 
aus der Reihenfolge des Eintragens, z.B. VarTable = [ b2, a, b1 ], wenn zuerst
'b2', dann 'a' und schliesslich 'b1' eingetragen wird.
Spaeter werden alle Variablen nur noch ueber ihren Index in VarTable
angesprochen. Im obigen Beispiel steht also 0 fuer 'b2', 1 fuer 'a' und 2 
fuer 'b1'.

Eine Multimenge von Kommunikationsvariablen ist ein dynamisches Array von
Elementen folgenden Typs:

typedef struct Variable_
{ ComVar	nr;		 // Nummer der Kommunikationsvariablen
  Cardinal	anzahl;		 // Anzahl des Auftretens in der Multimenge 
} Variable, *VarMSet;

Damit alle Routinen effizient arbeiten, werden die Eintraege im Array nach 
aufsteigender 'nr' sortiert. Das Ende des Arrays wird durch einen Eintrag mit 
'anzahl' = 0 markiert. 

Beispiel (s.o.):  Die Multimenge { b1, b1, b1, b2, b2 } wird repraesentiert 
durch das Array [ {0,2} , {2,3} , {x,0} ], wobei 'x' jeweils beliebig ist.
*/


#include <stdio.h>
#include <string.h>
#include "multiset.h"
#include "var_mset.h"
#include "annotat.h"
#include "tables.h"

/*****************************************************************************/
/* Globale Variablen fuer die Behandlung von Multimengen.                    */
/*****************************************************************************/

VarMSet V_mmDummy    = NULL;      /* Hilfsmenge bei der Berechnung von sum,  */
                                  /* union,...                               */

VarMSet V_mmLeaf     = NULL;      /* Hilfsmenge bei der Berechnung der       */
                                  /* Blaetter des Parsetrees.                */

/*****************************************************************************/
/* Initialisieren der globalen Variablen des Multimengen-Funktionspakets.    */
/*****************************************************************************/

void InitVarMSets()

  {

    if ( V_mmDummy )
      {
        VSremove( V_mmDummy );
      } /* if */
    V_mmDummy  = VSalloc( MaxVars );
    if ( V_mmLeaf )
      {
        VSremove( V_mmLeaf );
      } /* if */
    V_mmLeaf  = VSalloc( MaxVars );

  } /* InitVarMSets() */


/*****************************************************************************/
/* Funktion zum Kopieren einer Multimenge in eine bereits bestehende.        */
/*****************************************************************************/

static Var SubstTpl( Var nr, ComVar *sub )

{
  Var *v_org = TplToArray( nr - V_TUPEL_OFFSET );
  Var *v_neu = MYmalloc( *v_org * sizeof( Var ) );
  Cardinal i;
  
  *v_neu = *v_org;
  for ( i=1; i<*v_org; i++ )
    {
      if ( v_org[i] < V_CONST_OFFSET )
	v_neu[i] = sub[ v_org[i] ];
      else
	v_neu[i] = v_org[i];
    }
  return ArrayToTpl( &v_neu ) + V_TUPEL_OFFSET;
}		      


VarMSet V_MScopy(VarMSet mm1, VarMSet mm2, ComVar *sub )

  {
    if (sub )
	{
	    Variable sub_var[2];

	    sub_var[1].anzahl = 0;
	    mm1->anzahl = 0;
	    
	    while ( (sub_var[0].anzahl = mm2->anzahl) != 0 ) 
		{
		    if ( mm2->nr < V_CONST_OFFSET )
		      sub_var[0].nr = sub[ mm2->nr ];
		    else 
		      {
			if ( mm2->nr >= V_TUPEL_OFFSET )
			  sub_var[0].nr = SubstTpl( mm2->nr, sub );
			else
			  sub_var[0].nr = mm2->nr;
		      }

		    mm2++;
		    V_MSsum( mm1, sub_var );
	            /* Durch die Substitution kann sich die Reihenfolge 
                       der Aktionen aendern. Deshalb werden sie einzeln
		       mittels MSsum eingefuegt.                             */
		}
	}
    else
	{
	    register VarMSet mm1Ptr = mm1;

	    while ( (mm1Ptr->anzahl = mm2->anzahl) != 0 )
	      (mm1Ptr++)->nr = (mm2++)->nr;
	}

    return ( mm1 );

  } /* V_MScopy() */

/*****************************************************************************/
/* Funktion zur Initialisierung einer neuen Multimenge, die als einziges     */
/* Element die uebergebene Aktion mit dem Code `akt` enthaelt.               */
/*****************************************************************************/

VarMSet V_MScreate_singleton(ComVar var)

  {

    register VarMSet mm;

    ( mm = VSalloc( 1 ) )->nr = var;
    mm->anzahl = 1;
    ( mm + 1 )->anzahl = 0;

    return ( mm );

  } /* V_MScreate_singleton() */


/*****************************************************************************/
/* Funktion zur Initialisierung einer neuen Multimenge als Kopie einer be-   */
/* reits vorhandenen Multimenge; der Speicherbedarf der neuen Menge ent-     */
/* spricht ihrer tatsaechlichen "Laenge".                                    */
/*****************************************************************************/

VarMSet V_MScreate_copy( VarMSet mm, ComVar *sub )

  {

    VarMSet           mm2;        /* Zeiger auf die zu erzeugende Kopie.     */
    register VarMSet  mmptr = mm; /* Laufzeiger durch die Originalmenge.     */
    register Cardinal len = 0;    /* Laenge der Originalmenge.               */
    Variable          sub_var[2];

                                  /* "Laenge" der Multimenge ermitteln und   */
                                  /* in `len` ablegen.                       */
    while ( mmptr->anzahl )
      {
        len++;
        mmptr++;
      }
                                  /* Speicherplatz fuer die Kopie anfordern. */
    mm2 = VSalloc( len );
    mmptr = mm;
    
    sub_var[1].anzahl = 0;
    mm2->anzahl = 0;
                                  /* Eintraege einzeln kopieren.             */
    while ( (sub_var[0].anzahl = mmptr->anzahl) !=0 ) {
      if ( sub && (mmptr->nr < V_CONST_OFFSET) )
        sub_var[0].nr = sub[ mmptr->nr ];
      else
	{
	  if ( sub && (mmptr->nr >= V_TUPEL_OFFSET) )
	    sub_var[0].nr = SubstTpl( mmptr->nr, sub );
	  else
	    sub_var[0].nr = mmptr->nr;
	}
      mmptr++;
      V_MSsum( mm2, sub_var );
	            /* Durch die Substitution kann sich die Reihenfolge 
                       der Aktionen aendern. Deshalb werden sie einzeln
		       mittels MSsum eingefuegt.                             */
    }
                                  /* Zeiger auf die Kopie zurueckliefern.    */
    return ( mm2 );

  } /* V_MScreate_copy() */


/*****************************************************************************/
/* Funktion zur Berechnung der Maechtigkeit einer Multimenge                 */
/*****************************************************************************/

Cardinal V_MScardinality( VarMSet mm)

  {

    Cardinal card = 0;            /* Kardinalitaet der Menge.                */

                                  /* Durch die Menge laufen ( in Zweier-     */
                                  /* schritten ) und dabei die Anzahlen der  */
                                  /* Vorkommen der einzelnen Aktionen auf-   */
                                  /* summieren.                              */
    while ( mm->anzahl )
      {
        card += (mm++)->anzahl;
      } /* while */

                                  /* Ermittelte Kardinalitaet zurueckgeben.  */
    return ( card );

  } /* V_MScardinality() */

/*****************************************************************************/
/* Funktion zur Berechnung, wie oft eine Aktion mit dem Aktionscode `akt` in */
/* der Multimenge `mm` enthalten ist.                                        */
/*****************************************************************************/

Cardinal V_MScount(ComVar var, VarMSet mm)

  {

    if ( var && mm )
      {
                                  /* In der Menge diejenige Stelle suchen,   */
                                  /* an der die Aktion `akt` stehen muss.    */
        while ( mm->anzahl && ( mm->nr < var ) )
          {
            mm++;
          } /* while */
                                  /* Rueckgabe: falls die Aktion in der      */
                                  /* Menge enthalten ist, wird die Anzahl    */
                                  /* ihrer Vorkommen geliefert, sonst null.  */
        return ( ( var == mm->nr ) ? mm->anzahl : 0 );
      } /* if */
                                  /* Falls bei der Parameteruebergabe ein    */
                                  /* Fehler aufgetreten ist, wird null als   */
                                  /* ermittelte Anzahl zurueckgegeben.       */
    return ( 0 );

  } /* MScount() */


/*****************************************************************************/
/* Funktion zur Feststellung, ob zwei Multimengen `mm1` und `mm2` gleich     */
/* sind.                                                                     */
/*****************************************************************************/

bool    V_MSis_equal( VarMSet mm1, VarMSet mm2)

  {

                                  /* Die Mengen aktionsweise durchforsten:   */
    while ( mm1->anzahl )
      {
                                  /* Bei der ersten Nicht-Uebereinstimmung   */
        if ( (mm1->anzahl != mm2->anzahl) || ( (mm1++)->nr != (mm2++)->nr ) )
          {
                                  /* wird die weitere Suche abgebrochen und  */
                                  /* 'False' als Ergebnis geliefert.         */
            return ( False );
          }
      } /* while */

                                  /* An dieser Stelle wurde die erste Menge  */
                                  /* bis zum Ende abgearbeitet; das Ergebnis */
                                  /* des Vergleichs ist 'True', wenn jetzt   */
                                  /* auch die zweite Menge 'leer' ist, sonst */
                                  /* 'False'.                                */
    return ( !mm2->anzahl );

  } /* MSis_equal() */


/*****************************************************************************/
/* Funktion zur Summenbildung zweier Multimengen.                            */
/*****************************************************************************/

VarMSet V_MSsum( VarMSet mm1, VarMSet mm2)

  {

    register VarMSet mm1Ptr = mm1,
                     mm2Ptr = mm2,
                     mmDum  = V_mmDummy;

    while ( mm1Ptr->anzahl || mm2Ptr->anzahl )
      {
        if ( mm1Ptr->anzahl && mm2Ptr->anzahl && (mm1Ptr->nr == mm2Ptr->nr) )
          {
            mmDum->anzahl	= mm1Ptr->anzahl;
            mmDum->nr		= mm1Ptr->nr;
            if ( ( (mmDum++)->anzahl = mm1Ptr->anzahl + mm2Ptr->anzahl ) < mm1Ptr->anzahl )
              {
                Msg( FAT, "Overflow: variable name too often used. Remedy: 'Variable'" );
              } /* if */
            mm1Ptr++, mm2Ptr++;
          } /* if */
        else if ( ( ! mm2Ptr->anzahl ) || ( mm1Ptr->anzahl && ( mm1Ptr->nr < mm2Ptr->nr ) ) )
          {
            mmDum->anzahl	= mm1Ptr->anzahl;
            (mmDum++)->nr	= (mm1Ptr++)->nr;
          } /* else if */
        else /* ( ( ! mm1Ptr->anzahl ) || ( mm1Ptr->nr > mm2Ptr->nr ) ) */
          {
            mmDum->anzahl	= mm2Ptr->anzahl;
            (mmDum++)->nr	= (mm2Ptr++)->nr;
          } /* else */
      } /* while */

    mmDum->anzahl = 0;
    mmDum = V_mmDummy;
    mm1Ptr = mm1;
    while ( (mm1Ptr->anzahl = mmDum->anzahl)!=0 )
      (mm1Ptr++)->nr = (mmDum++)->nr;

    return ( mm1 );

  } /* V_MSsum() */


/*****************************************************************************/
/* Funktion zur Ausgabe einer Multimenge ( Datei oder Bildschirm ).          */
/*****************************************************************************/

char *ComVarToStr( ComVar nr, char *string, Variables v_ptr )
{
  string[0] = 0;
  StrCatVar( nr, v_ptr, string );
  return string;
}

void V_MSWriteMultiset( FILE *fileptr, VarMSet mm, void *v_ptr )

  {

    Cardinal i;                   /* Zaehlvariable.                          */
    char     dum[ MAX_IDENT_LENGTH ];   /* String zur Aufnahme des           */
                                        /* anzuzeigenden Namens.             */

        /* fputc( '{', fileptr ); */
        if ( mm->anzahl )
          {
                                  /* Menge ist nicht leer.                   */
            if ( MSdisplay_style )
              {
                                  /* Ausgabe in Kurzschreibweise.            */
                if ( mm->anzahl > 1 ) fprintf( fileptr, FORMAT_CARD, mm->anzahl );
                fprintf( fileptr, "%s", ComVarToStr( (mm++)->nr, dum, v_ptr ) );
                while ( mm->anzahl )
                  {
                    fputc( ',', fileptr );
                    if ( mm->anzahl > 1 ) 
			fprintf( fileptr, FORMAT_CARD, mm->anzahl );
                    fprintf( fileptr, "%s", ComVarToStr((mm++)->nr, dum, v_ptr) );
                  } /* while */
              } /* if */
            else /* WIDE */
              {
                                   /* Ausfuehrliche Ausgabe der Menge.       */
                fprintf( fileptr, "%s", ComVarToStr( mm->nr, dum, v_ptr ) );
                for ( i = 2 ; i <= mm->anzahl ; i++ )
                  {
                    fprintf( fileptr, ",%s", ComVarToStr(mm->nr, dum, v_ptr) );
                  } /* for */
                mm++;
                while ( mm->anzahl )
                  {
                    for ( i = 1 ; i <= mm->anzahl ; i++ )
                      {
                        fprintf( fileptr, ",%s", ComVarToStr(mm->nr, dum, v_ptr) );
                      } /* for */
                    mm++;
                  } /* while */
              } /* else */
          } /* if */
          /* fprintf( fileptr, "}" ); */

  } /* MSWriteMultiset() */


/*****************************************************************************/
/* END OF FILE multiset.c                                                    */
/*****************************************************************************/
