/*****************************************************************************/
/*  parser.h                                                                 */
/*  C-Headerdatei: Hilfsroutinen fuer den 'box expression'-Parser            */
/*  Di.,27.10.1992 - Mi.,04.11.1992                           Stefan Roemer  */
/*****************************************************************************/


#ifndef __PARSER_H__
#define __PARSER_H__

#include "boxtypes.h"

/*****************************************************************************/
/* Datentyp Symboltabelle fuer die Aktions-/Variablennamen ( Parser ).       */
/*****************************************************************************/

typedef char        * SymTabEntry;
typedef SymTabEntry * SymTab;
typedef struct
  {
    char     * FuncName;
    Function FuncMappings;
  } FuncTabEntry,
    * FuncTab;

extern SymTab  SymbolTable;
extern SymTab  VarTable;
extern FuncTab FunctionTable;

/*****************************************************************************/
/* Initialisierungsfunktion, die die vom Parser benutzten globalen Variablen */
/* ( insbesondere die Symboltabellen ) einrichtet. Diese Funktion ist unbe-  */
/* dingt zu Beginn des Hauptprogramms ( vor "Benutzung" des Parsers ) aufzu- */
/* rufen!                                                                    */
/*****************************************************************************/

extern void InitAllSymbolTables(void);
  /* GLOBAL SymTab   SymbolTabelle, (schreibend)
                     VarTable;      (schreibend)
            Cardinal MaxActions,    (schreibend)
                     MaxVars,       (schreibend)
                     MaxFunctions,  (schreibend)
                     ActCounter,    (schreibend)
                     VarCounter,    (schreibend)
                     FuncCounter;   (schreibend)  */

/* Aenderung 6.6.1999: yyerror jetzt in boxxpars.y         -Christian Stehno */

/*****************************************************************************/
/* Funktion zum Erzeugen eines neuen Knotens fuer den Syntaxbaum des 'box    */
/* expressions's. In diesen neuen Knoten wird der Operationscode `Operation` */
/* kopiert, sowie die Zeiger auf die drei Sohnknoten, `Ausdruck1`,           */
/* `Ausdruck2` und `Ausdruck3`, von denen einzelne auch NULL sein koennen.   */
/* Zurueckgeliefert wird ein Zeiger auf diesen neu eingerichteten Knoten.    */
/*****************************************************************************/

extern TTree GenNewNode(Cardinal Operation,
                           TTree Ausdruck1,
                           TTree Ausdruck2,
                           TTree Ausdruck3,
                        int fl, int fc, int ll, int lc);

/*****************************************************************************/
/* Funktion ( als Makro realisiert ) zum Entfernen des Knotens `node` des    */
/* Syntaxbaums aus dem Hauptspeicher.                                        */
/*****************************************************************************/

#define DeleteNode( node ) free( (char *) node )


/*****************************************************************************/
/* Funktion zum Entfernen eines ganzen Syntaxbaumes aus dem Speicher,        */
/* beginnend bei der Wurzel `root` ( call-by-reference-Parameter ).          */
/* Nach Beendigung dieser Funktion gilt `*root == NULL`.                     */
/*****************************************************************************/

extern void RemoveTTree(TTree *root);


extern Time_stamp *CreateTimeStamp(Time low, Time high);
/*****************************************************************************/
/* Funktion zur Umwandlung eines Aktionsnamens `Name` (String) in den ent-   */
/* sprechenden Aktionscode (Cardinal bzw. Action), der als Ergebnis zurueck- */
/* geliefert wird. Falls der Name noch nicht in der Aktions-Symboltabelle    */
/* vorhanden ist, wird dieser hinzugefuegt und ein entsprechender Code fuer  */
/* ihn berechnet.                                                            */
/*****************************************************************************/

extern Action StrToAction(char *Name);
  /* GLOBAL: SymTab   SymbolTable; (schreibend)
             Cardinal ActCounter;  (schreibend)
             MultiSet mmDummy,     (schreibend)
                      mmLeaf       (schreibend) */

/*****************************************************************************/
/* Funktion zur Umwandlung eines Aktionscodes `akt` in den entsprechenden    */
/* Namen, der in der bereitgestellten Zeichenkette `str` abgelegt wird.      */
/* `str` sollte also ein Zeiger auf einen zur Aufnahme des Aktionsnamens ge- */
/* nuegend gross initialisierten Speicherbereich sein; `str` wird schliess-  */
/* lich als Funktionsergebnis geliefert.                                     */
/*****************************************************************************/

extern char *ActionToStr(Action akt,char *str);
  /* GLOBAL: SymTab SymbolTable; (lesend)  */


/*****************************************************************************/
/* Funktion zum Umwandeln eines Variablennamens `Name` (String) in den ent-  */
/* sprechenden Variablencode (Var bzw. Cardinal), der als Funktionsergebnis  */
/* geliefert wird. Falls der Variablenname bisher noch nicht existiert, so   */
/* wird ein neuer Code fuer diesen ermittelt und der Name in die Variablen-  */
/* Symboltabelle aufgenommen.                                                */
/*****************************************************************************/

extern Var StrToVariable(char *Name);
  /* GLOBAL: SymTab   VarTable;   (schreibend)
             Cardinal VarCounter; (schreibend)  */


/*****************************************************************************/
/* Funktion zum Umwandeln des Variablencodes `var` in den entsprechenden     */
/* Variablennamen `str`, der in den dafuer zu initialisierenden Speicher-    */
/* bereich kopiert wird. Rueckgabe ist der Zeiger `str`.                     */
/*****************************************************************************/

extern char *VariableToStr(Var var,char *str);
  /* GLOBAL: SymTab VarTable; (lesend)  */


/*****************************************************************************/
/* Funktion zum Unwandeln eines Funktionsnamens `Name` (String) in den ent-  */
/* sprechenden Funktionscode (FuncCode bzw. Cardinal), der als Ergebnis      */
/* geliefert wird. Falls der Funktionsname bisher noch nicht existiert, so   */
/* wird ein neuer Code fuer diesen ermittelt und der Name in die Funktions-  */
/* Symboltabelle mitsamt der Funktionsdeklaration `Mappings` aufgenommen.    */
/*****************************************************************************/

extern FuncCode StrToFunction(char *Name,Function Mappings);
  /* GLOBAL: FuncTab  FunctionTable;   (schreibend)
             Cardinal FunctionCounter; (schreibend)  */


/*****************************************************************************/
/* Funktion zum Umwandeln des Funktionscodes `func` in den entsprechenden    */
/* Variablennamen `str`, der in den dafuer zu initialisierenden Speicher-    */
/* bereich kopiert wird. Rueckgabe ist der Zeiger `str`.                     */
/*****************************************************************************/

extern char *FunctionToStr(FuncCode funccode,char *str);
  /* GLOBAL: FuncTab FunctionTable; (lesend)  */


/*****************************************************************************/
/* Funktion zur Freigabe des von der Symboltabelle `table` belegten Haupt-   */
/* speicherplatzes; `table` ist hiernach undefiniert!                        */
/*****************************************************************************/

extern void RemoveSymbolTable(SymTab *table);


/*****************************************************************************/
/* Funktion zur Freigabe des von der Funktionen-Symboltabelle belegten Haupt-*/
/* speicherplatzes.                                                          */
/*****************************************************************************/

extern void RemoveFunctionTable(void);
  /* GLOBAL: FuncTab FunctionTable; (schreibend)  */


/*****************************************************************************/
/* Prozedur zur Darstellung der Symboltabelle `table` auf dem Bildschirm.    */
/* Nur zu Analysezwecken eingerichtet!                                       */
/*****************************************************************************/

extern void PrintSymbolTable(SymTab table);

/*****************************************************************************/


#endif


/*****************************************************************************/
/* END OF FILE parser.h                                                      */
/*****************************************************************************/

