/****************************************************************/
/*                                                              */
/*            C O D E - G E N E R A T O R  for B(PN)^2          */
/*                                                              */
/*            Universitaet Hildesheim                           */
/*            Institut fuer Informatik                          */
/*            Stephan Melzer                                    */
/****************************************************************/
/*            Diplomarbeit                                      */
/*            File        : local_shm.c
              Version     : 1.0
	      Last Update : 18.10.1994
 ****************************************************************/
/*            Module information :                              */
/* The routines of the memory management, called local_memory_management
   for shared memory access. An internal garbage block collector 
   controls the free memory blocks by a linear block chaining.
   Memory management functions : initialize_shm, switch_page, 
                                 actual_page, release_shm
   Control functions : shm_alloc, shm_free
   Garbage routines : remove_block
*/
/****************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "local_shm.h"
#include "configure.h"
#include "includes.h"
#include "misc.h"

dataspace *page;

/* Default page is page #0 */

int *active_page = (int*)-1;

garbage_block* garbage;

int anker_id = -1;

void error(const char* s)
{
  fprintf(stderr, "Shared Memory Error : ");
  fprintf(stderr, s);
  fprintf(stderr, "\n");
  exit(EXIT_FAILURE);
}

void initialize_shm_control(void)
{
  char* anker_ptr;
  int i;

  if ((anker_id = shmget(IPC_PRIVATE,
			 sizeof(int) + 2 * sizeof (dataspace) + 
	                 MAX_GARBAGE * sizeof(garbage_block),
  		         shm_create_bits
                         ) ) == -1 )
  {
    perror("local_shm.initialize_shm_control().shmget()");
    error ("cannot get anker memory segment");
  }

  if ((anker_ptr = shmat(anker_id, (char*) 0, 0)) == (char*) -1)
  {
    perror("local_shm.initialize_shm_control().shmat()");
    error ("cannot attach anker memory segment");
  }

#ifdef DEBUG
  (void) memset(anker_ptr, 2, sizeof(int) + 2 * sizeof(dataspace) + 
	MAX_GARBAGE * sizeof(garbage_block));/* reiner Testzweck */
#endif /* DEBUG */

  active_page = (int*) anker_ptr;
  (*active_page) = 0;
  anker_ptr += sizeof(int);
  page = (dataspace*) anker_ptr;
  page[0].shm_size = 4096;
  page[0].shm_id = -1;
  page[0].shm_begin = (char*)-1;
  page[1].shm_size = 4096;
  page[1].shm_id = -1;
  page[1].shm_begin = (char*)-1;
  anker_ptr += 2 * sizeof(dataspace);
  garbage = (garbage_block*) anker_ptr;
  for (i=0; i < MAX_GARBAGE; i++)
    garbage[i].is_used = FALSE;
}

void release_shm_control(void)
{
  if (anker_id > -1) {
    if (active_page != (int*)-1) {
      if (shmdt((char*) active_page) == -1)
	perror("local_shm.release_shm_control().shmdt()");
    }
    if (shmctl(anker_id, shm_remove_bits, (struct shmid_ds *) 0) == -1)
      perror("local_shm.release_shm_control().shmctl()");
  }
}

garbage_block* mem_block(void)
{
  int i;

  for (i=0; i < MAX_GARBAGE; i++)
    if (! garbage[i].is_used)
    {
      garbage[i].is_used = TRUE;
      return (garbage_block*) &(garbage[i]);
    }

  error("too much free blocks");

  /* wg. gcc-warning: diese Anweisung wird nie erreicht - */
  /* in error() wird exit() aufgerufen                    */
  return (garbage_block*) NULL; 
    
}

void free_block(garbage_block* g)
{
  g->is_used = FALSE;
}


void initialize_shm(int side)
/* Shared Memory Allocation (static) in two phases.
   side : Number of the specified data page
*/
{
  /* First phase : prepare and allocate the data segment */
  if ((page[side].shm_id = shmget(IPC_PRIVATE, page[side].shm_size,
                                  shm_create_bits
                                 ) ) == -1)
  {
    perror("local_shm.initialize_shm().shmget()");
    error ("cannot get shared memory segment");
  }

  /* Second phase : attach the memory segment and get the 
     initial address */
  if ((page[side].shm_begin = shmat(page[side].shm_id, (char*) 0, 0
				   ) ) == (char*) -1)
  {
    perror("local_shm.initialize_shm().shmat()");
    error ("cannot attach shared memory segment");
  }

  /* Initialize local memory management of this page */
  page[side].free_blocks = mem_block();
  page[side].free_blocks->block_begin = page[side].shm_begin;
  page[side].free_blocks->size = page[side].shm_size;
  page[side].free_blocks->next = (garbage_block*) NULL;
}

void switch_page(int side)
/* Changing the momentary data segment 
   side : Number of the specified data page
*/
{
  (*active_page) = side;
}

int actual_page(void)
/* Resulting the momentary data segment 
*/
{
  return (*active_page);
}

garbage_block* get_free_block(garbage_block* L, int need_size)
/* Searching for an empty, suitable free block. Returning the pointer
   to the found block.
   L : Pointer to a linear list of garbage blocks
   need_size : minimal size of needed block
*/
{
  while (L != nil_block)
    if (L->size >= need_size)
      return L;
    else
      L = L->next;
  /* Otherwise: there is not such a block */
  return nil_block;
}

void remove_block(garbage_block** L, garbage_block* g)
/* Removing of a free block out of the free garbage block list
   L : Pointer to the garbage block list
   g : removal block
*/
{
  garbage_block* dummy = *L;

  if (dummy == nil_block)
  {
    error ("garbage_block_list is empty");
  }
  if (dummy == g)
  /* It is the first block, ... */
  {
    *L = (*L)->next;
    free_block(g);
    return;
  }
  /* ... or an other */
  while (dummy->next != nil_block)
    if (dummy->next == g)
    {
      dummy->next = dummy->next->next;
      free_block (g);
      return;
    }
    else
      dummy = dummy->next;
}

char* shm_alloc(int need_size)
/* Allocating shared memory of the actual data page. Returning the
   address of the segment.
   need_size : Size of the ordered memory segment
*/
{
  garbage_block* dummy = page[(*active_page)].free_blocks;

  while (dummy != nil_block)
    if (dummy->size > need_size)
    {
      char* result = dummy->block_begin;

      /* Minimize the block by the allocated segment */
      dummy->block_begin += need_size;
      dummy->size -= need_size;
      return result;
    }
    else if (dummy->size == need_size)
    {
      char* result = dummy->block_begin;
    
      /* Eat the whole free block */
      remove_block(&(page[(*active_page)].free_blocks), dummy);
      return result;
    }
    else
      dummy = dummy->next;

  error("insufficient memory");

  /* wg. gcc-warning: diese Anweisung wird nie erreicht - */
  /* in error() wird exit() aufgerufen                    */
  return (char*) -1;
}
	    
void x_shm_free(void* ptr, int size, int side)
/* Deallocating the previous data segment of the specified page.
   
   [xxxxxxxxxxyyyyyyyyyyyyyyzzzzzzzzzzzz]
    |         |             |
    in_front  ptr	    at_tail

   Unify touching other free block.
   ptr : Pointer of the specified data segment
   size : Size of previous ordered segment
   side : Page of this pointer
*/
{
  garbage_block* dummy = page[side].free_blocks,
    * in_front = nil_block,
    * at_tail  = nil_block;
  int blocks_inside = FALSE;

  if (size == 0)
    return;

  while (dummy != nil_block)
  {
    if (((char*) ptr) + size == dummy->block_begin)
      at_tail = dummy;
    if (dummy->block_begin + dummy->size == ((char*) ptr))
      in_front = dummy;
    if (((int) ptr <= (int) dummy->block_begin) &&
	((int) (((char*) ptr) + size) > (int) dummy->block_begin))
      blocks_inside = TRUE;
    dummy = dummy->next;
  }

  if (blocks_inside)
    /* The garbage block list is still coherent */
    error ("intersecting blocks addressed");

  if (at_tail != nil_block)
    /* Eat the next block */
  {
    size += at_tail->size;
    remove_block(&(page[side].free_blocks), at_tail);
  }

  if (in_front != nil_block)
    /* Eat the previous page */
  {
    in_front->size += size;
    return;
  }

  dummy = mem_block ();
  dummy->size = size;
  dummy->block_begin = (char*) ptr;
  dummy->next = page[side].free_blocks;
  page[side].free_blocks = dummy;

}

void shm_free(void* ptr, int size)
/* Deallocating the data segment. Specify the associated page
   ptr : Pointer of the specified data segment
   size : Size of previous ordered segment
*/
{
  if (ptr == NULL)
    return;

  if ( ((int) page[0].shm_begin <= (int) ptr) &&
       (((int) page[0].shm_begin) + page[0].shm_size > (int) ptr) )
    x_shm_free(ptr, size, 0);
  else if ( ((int) page[1].shm_begin <= (int) ptr) &&
	    (((int) page[1].shm_begin) + page[1].shm_size > (int) ptr) )
    x_shm_free(ptr, size, 1);
  else
    error("invalid shared memory address");
}
	
void release_shm(int side)
/* Releasing of an allocated data segement (also in two phases)
  side : Data page
*/
{
  if (page[side].shm_id > -1) {
    if (page[side].shm_begin != (char*)-1) {
      if (shmdt(page[side].shm_begin) == -1)
	perror("local_shm.release_shm().shmdt()");
    }
    if (shmctl(page[side].shm_id, shm_remove_bits,(struct shmid_ds *) 0) == -1)
      perror("local_shm.release_shm().shmctl()");
  }
}

#ifdef DEBUG

void dump(int side)
/* Garbage block overview of a data page 
   side : Data page
*/
{
  garbage_block *dummy = page[side].free_blocks;

  while (dummy != nil_block)
  {
    fprintf (stderr, "Block at %d, size %d, offset %d \n",
	    (int)dummy->block_begin,
	    dummy->size, 
	    - ((int) page[side].shm_begin) + ((int) dummy->block_begin));
    dummy = dummy->next;
  }
}
 
void hex_dump(int side, int offset, int size)
/* hexadecimal memory overview.
  side : Data page
  offset : beginning address
  size : Upper bound of the specified segment
*/
{
  char* ptr = page[side].shm_begin + offset;
  int counter;

  for (counter = 0; counter < size; counter ++, ptr++)
    fprintf (stderr, " Address : %d Value : %d \n", (int) ptr, (int) *ptr);
}   

#endif /* DEBUG */

/*-------------------------------------------------------------------------*/
/* End of file local_shm.c */
/*-------------------------------------------------------------------------*/
