/*-------------------------------------------------------------------------*/
/*  misc.c                                                                 */
/*-------------------------------------------------------------------------*/
/*  B(PN)^2 -> C Codegenerator                                              */
/*                                                                         */
/*  Johannes Jaeger --- Sommer 1997                                        */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include "configure.h"
#include "misc.h"
#include "bool.h"
#include "includes.h"
#include "error.h"

/* globale Variable fuer Informationen, */
/* die von einem generierten Programm   */
/* zur Laufzeit in die Protokolldatei   */
/* geschrieben werden                   */
ProtMsg PMsg = {
    "",
    (RefList*)NULL
};

/* globale Variable, mit der die Eintragung     */
/* von Referenzen in PMsg gesteuert wird.       */
/* Beim Testen von Guarded Commands (s.         */
/* gc_control.c) bzw. zu synchronisierenden     */
/* Aktionen (s. scheduler.c) wird diese auf     */
/* TRUE gesetzt, damit keine Referenzen         */
/* in PMsg eingetragen werden, sonst auf FALSE. */
boolean Testing = FALSE;

long Path_Max;

/* default options settings */
boolean generate_executable = TRUE;
boolean info_on             = FALSE;

     /* Eingabedatei */
FILE *BPN2_File = (FILE*) NULL,
     /* das generierte C-Programm */
     *C_File    = (FILE*) NULL,
     /* der generierte Headerfile */
     *H_File    = (FILE*) NULL,
     /* Ausgabedatei fuer B(PN)^2-Symbole */
     *Info_File = (FILE*) NULL;


char *BPN2_fname = (char*) NULL,
     *C_fname    = (char*) NULL,
     *H_fname    = (char*) NULL;

char Info_fname[1024];

     /* Pfad fuer die Runtime-Bibliothek */
char *lib_path   = (char*) NULL;



/*-------------------------------------------------------------------------*/
/* malloc() mit Fehlermeldung */
void* Malloc( size_t size )
{
  void *ptr;
  if ( !(ptr = malloc(size)) )
    runtime_error(ERR_NOMEM);
  return ptr;
}

/*-------------------------------------------------------------------------*/
void* Realloc(void *ptr, size_t size)
{
  void *resptr;
  if ( !(resptr = realloc(ptr, size)) )
    runtime_error(ERR_NOMEM);
  return resptr;
}

/*-------------------------------------------------------------------------*/
static void BuildFileName(char *fname, const char *suffix)
{
  strcpy(fname, BPN2_fname);
  strcpy((char*) &(fname[strlen(fname) - strlen(BPN2_SUFFIX)]), suffix);
}

/*-------------------------------------------------------------------------*/
static void SetFileNames()
{
  /* build C_fname from BPN2_fname */
  BuildFileName(C_fname, C_SUFFIX);
  /* build H_fname from BPN2_fname */
  BuildFileName(H_fname, H_SUFFIX);
}

/*-------------------------------------------------------------------------*/
/* Speicher fuer Dateinamen anfordern, diese Initialisieren */
static void InitFileNames()
{
  C_fname    = (char*) Malloc( (size_t)(strlen(BPN2_fname) + 1) );
  H_fname    = (char*) Malloc( (size_t)(strlen(BPN2_fname) + 1) );
  SetFileNames();
}


/*-------------------------------------------------------------------------*/
void InitPathNames(const char *path)
{
  /* maximale Pfadlaenge des aktuellen Dateisystems ermitteln */
  if ( (Path_Max = pathconf(path, _PC_PATH_MAX)) == -1) {
#ifdef DEBUG
    fprintf(stderr, "Path = %s -- Path_Max = %ld\n", path, Path_Max);
#endif /*DEBUG */
    /* Defaultwert */
    Path_Max = 2048;
  }
  else
    Path_Max = 2 * Path_Max;

  lib_path   = (char*) Malloc( (size_t) Path_Max * sizeof(char) );

  /* get environment variable if set */
  if (getenv("BPN2_CODEGENERATOR_LIB") == (char*) NULL)
    strcpy(lib_path, STD_LIB_PATH);
  else
    strcpy(lib_path, getenv("BPN2_CODEGENERATOR_LIB"));
}

/*-------------------------------------------------------------------------*/
void Init(const char *Prog_path)
{
  /* Exithandler installieren */
  if (atexit(&CleanUp) != 0)
    perror("misc.Init().atexit()");

  /* Dateinamen initialisieren */
  InitFileNames();
  InitPathNames(Prog_path);
}

/*-------------------------------------------------------------------------*/
/* calloc() mit Fehlermeldung */
void* Calloc(size_t nmemb, size_t size)
{
  void *ptr;

  if ( !(ptr = calloc(nmemb, size)) )
    runtime_error(ERR_NOMEM);

  return ptr;
}

/*-------------------------------------------------------------------------*/
char* upper_name(char* name)
{
  int i;

  for (i=0; i < (int) strlen(name); i++)
    name[i] = toupper(name[i]);
  return name;
}


/*-------------------------------------------------------------------------*/
void OpenFile(FILE **fp, const char *fname, const char *mode)
{
  if (*fp != (FILE*) NULL)
    return;

  if ( (*fp = fopen(fname, mode)) == (FILE*) NULL )
  {
    perror("misc.OpenFile().fopen()");
    exec_error(ERR_CANT_OPEN_FILE, fname);
  }
}

/*-------------------------------------------------------------------------*/
void CloseFile(FILE **fp)
{
  if (*fp == (FILE*) NULL)
    return;
  if (fclose(*fp) != 0)
    perror("misc.CloseFile().fclose()");
  *fp = (FILE*) NULL;
}

/*-------------------------------------------------------------------------*/
static void CleanName(char **Name)
{
  if (*Name != (char*) NULL)
  {
    free(*Name);
    *Name = (char*) NULL;
  }
}

/*-------------------------------------------------------------------------*/
/* Exithandler */
void CleanUp(void)
{
#ifdef DEBUG
  fprintf(stderr, "\nCalling CleanUp()\n");
#endif /* DEBUG */
  CleanName(&C_fname);
  CleanName(&H_fname);
  CleanName(&lib_path);

  CloseFile(&BPN2_File);
  CloseFile(&C_File);
  CloseFile(&H_File);
  CloseFile(&Info_File);
}

/*-------------------------------------------------------------------------*/
/* prueft, ob value innerhalb von scope liegt */
boolean legalvalue(int value, Tscope scope)
{
  unsigned int j = 0;

  while (j < scope.num_bounds) {
    if ( (scope.bounds[j].low <= value) && (scope.bounds[j].high >= value) )
      return TRUE;
    j++;
  }
  return FALSE;
}

/*-------------------------------------------------------------------------*/
/* erzeugt einen Zufallswert innerhalb von scope */
int createValue(Tscope scope)
{
  int tmpval;

  int upper_bound = scope.bounds[scope.num_bounds-1].high;
  int lower_bound = scope.bounds[0].low;

  do {
    tmpval = ( (upper_bound != lower_bound) ?
	       (rand() % (upper_bound - lower_bound)) + lower_bound :
	       lower_bound);
  } while (!legalvalue(tmpval, scope));
  return(tmpval);
}

/*-------------------------------------------------------------------------*/
/* Werte einer Referenzstruktur setzen */
void SetRef(Reference *ref, int startl, int startc,
	    int endl, int endc, int RNum)
{
  ref->start_l = startl;
  ref->start_c = startc;
  ref->end_l   = endl;
  ref->end_c   = endc;
  ref->RefNum  = RNum;
}

/*-------------------------------------------------------------------------*/
/* Haengt an die uebergebene Liste RefListHead ein neues Element    */
/* vom Typ RefList (s. includes.h) an und trgt die uebergebenen    */
/* Werte dort ein; der Listenanfang RefListHead wird zurueckgegeben */
RefList* AddRefListElem(RefList* RefListHead,
			int startl, int startc, int endl, int endc, int RNum)
{
  RefList* result, *tmpListHead;

  if (!RefListHead) /* Liste ist noch leer */
    result = tmpListHead = (RefList*)Malloc(sizeof(RefList));
  else {
    result = tmpListHead = RefListHead;
    while (tmpListHead->nextRef)  /* tmpListHead auf letztes Element setzen */
      tmpListHead = tmpListHead->nextRef;
    tmpListHead->nextRef = (RefList*)Malloc(sizeof(RefList));
    tmpListHead = tmpListHead->nextRef;
  }

  /* Werte in neues Element eintragen */
  tmpListHead->nextRef = (RefList*) NULL;
  tmpListHead->ref.start_l = startl;
  tmpListHead->ref.start_c = startc;
  tmpListHead->ref.end_l   = endl;
  tmpListHead->ref.end_c   = endc;
  tmpListHead->ref.RefNum  = RNum;

  return result;
}

/*-------------------------------------------------------------------------*/
/* Speicherfreigabe fuer eine Refrenzliste */
void FreeRefList(RefList** refs)
{
  if (*refs == (RefList*) NULL)
    return;
  FreeRefList(&((*refs)->nextRef));
  free(*refs);
  *refs = (RefList*)NULL;
}

/*-------------------------------------------------------------------------*/
/* Initialisierung der globalen Variablen PMsg */
void InitPMsg(void)
{
  PMsg.Msg[0] = '\0';
  FreeRefList(&(PMsg.refs));
}

/*-------------------------------------------------------------------------*/
/* erzeugt ein neues Listenelement, traegt die Werte ein */
/* und haengt es an list dran */
BucketSwitch* addbuckets(BucketSwitch* list, int old, int new)
{
  BucketSwitch* tmpentry;
  BucketSwitch* newentry = (BucketSwitch*) Malloc(sizeof(BucketSwitch));

  newentry->oldbucket = old;
  newentry->newbucket = new;
  newentry->next = (BucketSwitch*) NULL;

  if (list == (BucketSwitch*) NULL)
    return newentry;
  else {
    tmpentry = list;
    while (tmpentry->next)
      tmpentry = tmpentry->next;
    tmpentry->next = newentry;
    return list;
  }
}

/*-------------------------------------------------------------------------*/
/* End of file misc.c */
/*-------------------------------------------------------------------------*/
