/****************************************************************/
/*                                                              */
/*            C O D E - G E N E R A T O R  for B(PN)^2          */
/*                                                              */
/*            Universitaet Hildesheim                           */
/*            Institut fuer Informatik                          */
/*            Stephan Melzer                                    */
/****************************************************************/
/*            Diplomarbeit                                      */
/*            File        : symboltable.c                             
              Version     : 1.0
	      Last Update : 18.10.1994
 ****************************************************************/
/*            Module information :                              */
/* This module contains the symboltable-handling for the compiler.
   The table which is hashtable based, allows the handling of a 
   nested block structures. This structure provides unique ident-
   ifier for the generated code. THis module is seperated into two 
   parts :
   Part #1 parse-time module : 
   General : init_hashtab, hash_fkt, rehash_fkt
   Operations on the table : insert, insert_function, overriding_
                             insert, set_properties, get_properties,
                             get_procedure_type, read_write_access, 
			     entry_used, lookup, allowed_access, 
			     recovery_entry
   Part #2 code generation 
   Symboltable File : generate_symbol_file, print_hashtab
   Copying of the switch table : needed_memory, add_hash_table  */
   
/****************************************************************/

#include <stdlib.h>
#include <string.h>
#include "symboltable.h"
#include "misc.h"


#define WRITE2INT(X, Y, Z) fprintf(c_code, X, Y, Z)
#define WRITE4INT(U, V, W, X, Y) fprintf(c_code, U, V, W, X, Y)


int buckets_in_use;
entry hashtab[MAX_IDENT + 1];      /* see not : termination */


/********************** PART #1 parse-time module ***************/

void init_hashtab(void)
/* Initializing of the hashtable to make all buckets available  
*/
{ 
  int i;
  for (i = 0; i <= MAX_IDENT; i++)
    hashtab[i].bucket_free = TRUE;
  buckets_in_use = 0;
}

int hash_fkt(char* symbol)
/* The used hash function which mapps a string to an integer in the
   range {0, MAX_IDENT} :
   {a, ..., z, A, ..., Z, _, 0, ..., 9}* -> {0, ..., MAX_IDENT}
   symbol : Identifier
*/
{
  int result = 0,
      i;
  for (i = 0; i < (int)strlen(symbol); i++)
    result += symbol[i] - 'A';
  return result % MAX_IDENT;
}

int rehash(int bucket)
/* The used rehash function with an offset hash_offset
   bucket : number of the last attempt to find a free bucket
*/
{
  return (bucket + hash_offset) % MAX_IDENT;
}

int insert(char* symbol)
/* This function maps an entry (no function) to a bucket using the 
   hash- and rehash-function. The bucket number is the result, in 
   case of an error -1 is returned.
   symbol : Identifier
*/
{
  if (buckets_in_use == MAX_IDENT)  /*  (*)  */
  {
    fprintf(stderr, "Internal Error : Symboltable overflow");
    return (-1);
  }
  else
  {
    int	bucket_ptr = hash_fkt(symbol);

    buckets_in_use++;
    /* rehash until free bucket is found; termination guarentee
       see (*) above and the specified hash_offset */
    while (! hashtab[bucket_ptr].bucket_free)
      bucket_ptr = rehash(bucket_ptr);
    strcpy(hashtab [bucket_ptr].sym, symbol);
    hashtab [bucket_ptr].bucket_free = FALSE;
    hashtab [bucket_ptr].access_allowed = TRUE;
    hashtab [bucket_ptr].covered_entry = -1;
    hashtab [bucket_ptr].access_noticed = FALSE;
    return bucket_ptr;
  }
}
	
void insert_function(char* symbol, char* params)
/* This function maps an entry of a function identifier to a bucket 
   using the hash- and rehash-function. 
   symbol : Identifier
   params : String which indicates the formal parameter list 
            of the function with its procedure name symbol
*/
{
  if (buckets_in_use == MAX_IDENT)  /*   (*)   */
    fprintf(stderr, "Internal Error : Symboltable overflow");
  else
  {
    int	bucket_ptr = hash_fkt(symbol);

    buckets_in_use++;	  
    /* rehash until free bucket is found; termination guarentee
       see (*) above and the specified hash_offset */
    while (! hashtab [bucket_ptr].bucket_free)
      bucket_ptr = rehash(bucket_ptr);
    strcpy(hashtab [bucket_ptr].sym, symbol);
    hashtab [bucket_ptr].bucket_free = FALSE;
    hashtab [bucket_ptr].access_allowed = TRUE;
    hashtab [bucket_ptr].covered_entry = -1;
    hashtab [bucket_ptr].access_noticed = FALSE;
    hashtab [bucket_ptr].func_params = (char*) Malloc((strlen(params) + 1) *
                                                       sizeof(char));
    (void)strcpy(hashtab [bucket_ptr].func_params, params);
    hashtab [bucket_ptr].data_type  = FUNCTION;
    hashtab [bucket_ptr].channel_size  = -1;
  }
}

int lookup(char* symbol)
/* This function returns the bucket number of an already inserted
   symbol. If this symbol is not found -1 returned.
   symbol : Identifier
*/
{
  int bucket_ptr = hash_fkt(symbol);

  if (hashtab[bucket_ptr].bucket_free)
    return (-1);       
  while ((strcmp(symbol, hashtab [bucket_ptr].sym) != 0) ||
       (! hashtab [bucket_ptr].access_allowed )) 
  {
    bucket_ptr = rehash(bucket_ptr);
    if (hashtab[bucket_ptr].bucket_free)
      return (-1);    
  }
  return bucket_ptr;
}

void set_properties(int bucket, int type, Tscope scope, 
		    int chan_size, int init, int s_or_c, int RNum)
/* This function specifies the datatype of the identifier which
   is mapped to this bucket.
   bucket : number of the bucket which contains the identifer of 
            the following type
   type   : number which indicates the datatype 
   l_bound, r_bound : in case of an ordered dataset, the upper and
                      lower bounds of the underlying set
   chan_size : specifies the capacity of channel, in case of a plain
               datatype the size is -1
   init : in case of a plain datatype, the initializing value 
*/
{
  if (hashtab[bucket].bucket_free)
    fprintf(stderr, "Internal Error : Hash-Bucket empty");
  if ((type == CHANNEL_TYPE) && (chan_size < 0))
    fprintf(stderr, "Internal Error : Invalid channel capacity");
  if (! hashtab[bucket].access_allowed)
    fprintf(stderr, "Internal Error : No access on this bucket");
  hashtab[bucket].scope = scope;
  hashtab[bucket].data_type   = type;
  hashtab[bucket].channel_size= chan_size;
  hashtab[bucket].init_value  = init;
  hashtab[bucket].stack_or_chan = s_or_c;
  hashtab[bucket].RefNum = RNum;
}

void get_properties(int bucket, int* type, Tscope* scope,
	            int* chan_size, int* init, int* s_or_c)
/* This function gives known about the datatype of the identifier which
   is mapped to this bucket.
   bucket : number of the bucket which contains the identifer of 
            the following type
   *type   : number which indicates the datatype 
   *chan_size : specifies the capacity of channel, in case of a plain
                datatype the size is -1
   *init : in case of a plain datatype, the initializing value 
*/

{
  if (hashtab[bucket].bucket_free)
    fprintf(stderr, "Internal Error : Hash-Bucket empty");
  if (! hashtab[bucket].access_allowed)
    fprintf(stderr, "Internal Error : No access on this bucket");
  *scope = hashtab[bucket].scope;
  *type = hashtab[bucket].data_type;
  *init	= hashtab[bucket].init_value;
  *chan_size = hashtab[bucket].channel_size;
  *s_or_c = hashtab[bucket].stack_or_chan;
}

char* get_procedure_type(int bucket)
/* This function gives known about the prototype of the function which
   is mapped to this bucket. The result is the string which indicates 
   the type of the function.
   bucket : number of the bucket which contains the identifer of 
            the function
*/ 
{
  if (hashtab[bucket].bucket_free)
    fprintf(stderr, "Internal Error : Hash-Bucket empty");
  if (! hashtab[bucket].access_allowed)
    fprintf(stderr, "Internal Error : No access on this bucket");
  if (hashtab[bucket].data_type != FUNCTION)
    fprintf(stderr, "Internal Error : Hash-Bucket contains no procedure");
  return hashtab[bucket].func_params;
}

void read_write_access(int bucket)
/* This function is used to indicate the use of the identifer. The 
   statistic of the access to symbols is used to implement warnings
   of unused symbols.
   bucket : number of the bucket which is accessed
*/
{
  if (hashtab[bucket].bucket_free)
    fprintf(stderr, "Internal Error : Hash-Bucket empty");
  if (! hashtab[bucket].access_allowed)
    fprintf(stderr, "Internal Error : No access on this bucket");
  hashtab[bucket].access_noticed = TRUE;
}

int entry_used(int bucket)
/* This function asks for the access to the identifier mapped to this
   bucket. 
   bucket : number of the bucket which is accessed
*/
{
  if (hashtab[bucket].bucket_free)
    fprintf(stderr, "Internal Error : Hash-Bucket empty");
  if (! hashtab[bucket].access_allowed)
    fprintf(stderr, "Internal Error : No access on this bucket");
  return hashtab[bucket].access_noticed;
}

int allowed_access(int bucket)
/* This function asks for the visibility of the underlying bucket. In case 
   of a nested block structure, inner blocks can cover identifiers of outer
   blocks.
   bucket : number of the bucket which is accessed
*/
{
  if (hashtab[bucket].bucket_free)
    fprintf(stderr, "Internal Error : Hash-Bucket empty");
  return hashtab[bucket].access_allowed;
}

int overriding_insert(char* symbol, int type, Tscope scope, 
		      int chan_size, int init, int s_or_c, int RNum)
/* This functions features the nested block based mapping symbols with 
   the same identifer to different unique bucket, by hidding the outer 
   bucket.
   symbol, type, u_bound, l_bound, chan_size, init : see insert
*/
{
  int used_bucket = lookup(symbol);
 
  if (used_bucket >= 0)   /* already existing symbol */
  {
    int	new_bucket = insert(symbol);

    if (new_bucket == -1)
      return -1;
    hashtab[used_bucket].access_allowed = FALSE;
    hashtab[new_bucket].covered_entry  = used_bucket;
    set_properties(new_bucket, type, scope, chan_size, init,
		   s_or_c, RNum);
    return new_bucket;
  }
  else
  {
    int new_bucket = insert(symbol);

    if (new_bucket == -1)
      return -1;
    set_properties(new_bucket, type, scope, chan_size, init,
		   s_or_c, RNum);
    return new_bucket;
  }
}

void recover_entry(int bucket)
/* This functions recovers a hidden identifier 
  bucket : number of the bucket
*/
{
  if (hashtab[bucket].bucket_free)
    fprintf(stderr, "Internal Error : Hash-Bucket empty");
  if (! hashtab[bucket].access_allowed)
    fprintf(stderr, "Internal Error : No access on this bucket");
  hashtab[hashtab[bucket].covered_entry].access_allowed = TRUE;
  hashtab[bucket].access_allowed = FALSE;
}
	
/***************** PART #2 code generating part **********************/

void print_hashtab(FILE* ptr)
/* This function copies the symbol table into an extra file, which can be
   optionally disabled.
   ptr : Pointer to file descriptor of the info file
*/
{
  int i;

  fprintf(ptr, 
  "  *********************************************************************\n");
  fprintf(ptr, 
  "  *                  I D E N T I F I E R    T a b l  e                *\n");
  fprintf(ptr, 
  "  *********************************************************************\n");
  fprintf(ptr, "\n");

  for (i = 0; i < MAX_IDENT; i++)
    if (! hashtab[i].bucket_free)
    {
      fprintf(ptr, "Identifier: %20s  Typ :", hashtab[i].sym);
      if (hashtab[i].data_type == FUNCTION)
      {
	unsigned int j;
  
        fprintf(ptr, " PROCEDURE (");
        for (j = 0; j < strlen(hashtab[i].func_params); j++)
        {
	  if (hashtab[i].func_params[j] == '0')
	    fprintf(ptr, "const bool");
	  else if (hashtab[i].func_params[j] == '1')
	    fprintf(ptr, "const int");
	  else if (hashtab[i].func_params[j] == '2')
	    fprintf(ptr, "var bool");
	  else if (hashtab[i].func_params[j] == '3')
	    fprintf(ptr, "var int");
	  else
	    fprintf(ptr, "undef");
	  if ((j+1) != strlen(hashtab[i].func_params))
	    fprintf(ptr, ", ");
        }
	fprintf(ptr, ") \n");
      }
      else
      {
	if (hashtab[i].channel_size >= 0)
	  fprintf(ptr, " %s (%d) OF ",
		  hashtab[i].stack_or_chan == CHAN_VAR?"CHANNEL":"STACK",
		  hashtab[i].channel_size);
        fprintf(ptr, " %s \n", 
		hashtab[i].data_type == BOOLEAN_TYPE ? "BOOLEAN" : "NUMERIC");
      }
    }
}

void needed_memory(int* mem, int* size)
/* This function calculates the dataspace size for the whole data segment.
   *mem : Number of bytes needed to store all variables
   *size : Number of different identifiers which are used
*/
{
  int i,
      result = 0;

  *size = 0;
  for (i = 0; i < MAX_IDENT; i++)
    if ( (! hashtab[i].bucket_free) && 
	 (hashtab[i].data_type != FUNCTION))
    {
      (*size)++;
      if (hashtab[i].channel_size <= 0) /*channel 0 and plain vars */
	result++;
      else result += hashtab[i].channel_size + 1;
    }
  *mem = result * sizeof(int);
}

void add_hash_table(FILE* c_code)
/* This function generates the code which specifies at run time the size
   of the shared memory segment and initialize the local_memory_managment.
   c_code : pointer to the file descriptor o fthe c_code file
*/
{
  int mem_size, 
      non_function_size,
      i;
  unsigned int j;

  needed_memory(&mem_size, &non_function_size);
  WRITE2INT("generate_memory_table (%d, max (%d, 1)); \n",
	    non_function_size, mem_size);
  for (i=0; i < MAX_IDENT; i++)
    if (! hashtab[i].bucket_free) 
      if (hashtab[i].data_type != FUNCTION)
      {
	fprintf(c_code, "{\n  Tscope scope;\n");
	fprintf(c_code, "  scope.num_bounds = %u;\n",
		hashtab[i].scope.num_bounds );
	for (j = 0; j < hashtab[i].scope.num_bounds; j++)
	{
	  fprintf(c_code, "  scope.bounds[%u].low = %d;\n", j,
		  hashtab[i].scope.bounds[j].low );
	  fprintf(c_code, "  scope.bounds[%u].high = %d;\n", j,
		  hashtab[i].scope.bounds[j].high );
	}
	fprintf(c_code,
		"initialize_variable(%d, \"%s\", %d, %i, scope, %d, %d);\n}\n",
		i,
		hashtab[i].sym,
		(hashtab[i].channel_size >= 0 ? hashtab[i].channel_size : 0),
		hashtab[i].init_value,
		hashtab[i].stack_or_chan,
		hashtab[i].RefNum);
      }
}

/*-------------------------------------------------------------------------*/
/* End of file symboltable.c */
/*-------------------------------------------------------------------------*/
