/*-------------------------------------------------------------------------*/
/*  translate.c                                                            */
/*-------------------------------------------------------------------------*/
/*  B(PN)^2 -> C Codegenerator                                              */
/*                                                                         */
/*  Johannes Jaeger --- Sommer 1997                                        */
/*-------------------------------------------------------------------------*/
/* In diesem Modul wird der Ackermann-Parsebaum nach dem Modell des Parsers*/
/* von Stephan Melzer in einen Melzer-Parsebaum uebertragen                */

#include "bool.h"
#include "includes.h"
#include "error.h"
#include "list.h"
#include "tree_handling.h"
#include "frontend.h"
#include "symboltable.h"
#include "translate_misc.h"

/* globaler Zaehler fuer die Referenznummern von */
/* - begin Statements
   - Variablendeklarationen
   - Prozedurkoepfen
   - Prozedurparametern
   - atomaren Aktionen
*/
int RefCount = 1;

/*-----------------------------------------------------------------------*/
static char* empty_parameter_list()
{
  char* result = (char*) Malloc(sizeof(char));
  result[0] = '\0';
  return result;
}

/*-----------------------------------------------------------------------*/
/* erzeugt einen String mit Parametertypen */
char* parameter_list(SymTabEntry* ptr)
{
  char* result = empty_parameter_list();

  while (ptr) {
    if ( (ptr->s_partype == _REF) || (ptr->s_partype == _RES) ) {
      result = (char*)Realloc((void*)result, sizeof(char));
      (void)strcat( result, FORMAL_VAR_INT);
    }
    else {
      result = (char*)Realloc((void*)result, sizeof(char));
      (void)strcat( result, FORMAL_CONST_INT);
    }
    ptr = ptr->s_parlist;
  }
  return result;
}
/*-----------------------------------------------------------------------*/
/* erzeugt einen Knoten fuer einen Ausdruck */
static SemNode* expr_node(TreeNode* node)
{
  switch (node->type)
  {
    case N_CONST:
    {
      if ( (int)(node->op1) == TT )
        return create_const_leaf(BOOLEAN, "C", TRUE);
      else
        if ( (int)(node->op1) == FF )
	  return create_const_leaf(BOOLEAN, "C", FALSE);
        else
	  return create_const_leaf(NUMERIC, "C", (int)(node->op1) );
    }
    case N_VAR:
    {
      switch ((int)(node->op2))
      {
	case _PREPRIMED:
	{
	  return(create_ident_leaf( ((SymTabEntry*)node->op1)->s_name,
				    CONST, "'X") );
	  break;
	}
        case _POSTPRIMED:
	{
	  return(create_ident_leaf( ((SymTabEntry*)node->op1)->s_name,
				    VAR, "X'") );
	  break;
	}
        case _READ:
	{
	  return(create_channel_leaf( ((SymTabEntry*)node->op1)->s_name,
				      CONST, "X?"));
	  break;
	}
        case _WRITE:
	{
	  return(create_channel_leaf( ((SymTabEntry*)node->op1)->s_name,
				      VAR, "X!"));
	  break;
	}
        case _EMPTY:
	{
	  return(create_channel_leaf( ((SymTabEntry*)node->op1)->s_name,
				      CONST, "X??"));
	  break;
	}
        case _FULL:
	{
	  return(create_channel_leaf( ((SymTabEntry*)node->op1)->s_name,
				      CONST, "X!!"));
	  break;
	}
        default:
	{
     	  return(create_ident_leaf( ((SymTabEntry*)node->op1)->s_name,
				    CONST, "'X") );
	  break;
	}
      }
      break;
    }
    case N_KLAMM:
    {
      return(expr_node(node->op1));
      break;
    }
    case N_OR:
    {
      SemNode *l,*r;
      l = expr_node(node->op1);
      r = expr_node(node->op2);

      return( checked_type(l, r, CONST, BOOLEAN,
			   l->semantic.expr_constraint.is_const &&
			   r->semantic.expr_constraint.is_const ?
			   CONST : VAR, "OR", OR_) );
      break;
    }
    case N_AND:
    {
      SemNode *l, *r;
      l = expr_node(node->op1);
      r = expr_node(node->op2);
      return( checked_type(l, r, UNDEF, BOOLEAN,
			   l->semantic.expr_constraint.is_const &&
			   r->semantic.expr_constraint.is_const ?
			   CONST : VAR, "AND", AND_) );
      break;
    }
    case N_NOT:
    {
      return( checked_type_single(expr_node(node->op1), BOOLEAN,
				  CONST, "NOT", NOT_) );
      break;
    }
    case N_EQ:
    {
      return( checked_equation( expr_node(node->op1), expr_node(node->op2)) );
      break;
    }
    case N_UNEQ:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, BOOLEAN, CONST, "#", UNEQU_) );
      break;
    }
    case N_LT:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, BOOLEAN, CONST, "<", LESS_) );
      break;
    }
    case N_GT:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, BOOLEAN, CONST, ">", GREA_) );
      break;
    }
    case N_LE:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, BOOLEAN, CONST, "<=", ELESS_) );
      break;
    }
    case N_GE:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, BOOLEAN, CONST, ">=", EGREA_) );
      break;
    }
    case N_PLUS:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, NUMERIC, CONST, "+", PLUS_) );
      break;
    }
    case N_MIN:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, NUMERIC, CONST, "-", MINUS_) );
      break;
    }
    case N_MULT:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, NUMERIC, CONST, "*", MULT_) );
      break;
    }
    case N_DIV:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, NUMERIC, CONST, "/", DIV_) );
      break;
    }
    case N_MOD:
    {
      return( checked_type( expr_node(node->op1), expr_node(node->op2),
			    CONST, NUMERIC, CONST, "%", MOD_) );
      break;
    }
    case N_SIGN:
    {
      return( checked_type_single(expr_node(node->op1), NUMERIC,
				  CONST, "-", UNARYMINUS_) );
      break;
    }
    default: break;
  }

  /* sollte nie erreicht werden */
  return((SemNode*)NULL);
}


/*-----------------------------------------------------------------------*/
/* forward decl. */

static SemNode* block_node(TreeNode* node, list** decls, int RNum);
static SemNode* com_node(TreeNode* node);

/*-----------------------------------------------------------------------*/
/* Knoten fuer einen do-od-Zweig erzeugen */
static SemNode* altset_node(TreeNode* node)
{
  switch (node->type)
  {
    case _REPEAT:
    {
      return( control_repeat(com_node(node->op1), SEQ_) );
      break;
    }
    case _EXIT:
    {
      return( control_exit(com_node(node->op1), SEQ_) );
      break;
    }
    case _CHOICE:
    {
      SemNode* sn1, *sn2;
      sn1 = altset_node(node->op1);
      sn2 = altset_node(node->op2);
      return( join_control(sn1, sn2, CHOICE_) );
      /*
      return( join_control( altset_node(node->op1),
			    altset_node(node->op2), CHOICE_ ) );
			    */
      break;
    }
  }
  /* sollte nie erreicht werden */
  return((SemNode*)NULL);
}

/*-----------------------------------------------------------------------*/
static SemNode* sem_node_listEntry(ArgListEntry* arg)
{
  SemNode* result = (SemNode*) NULL;

  if (arg->ident) {
      if ( (arg->ident->s_type == _CHANNEL) || (arg->ident->s_type == _STACK) )
	result = init_sem_node_list_channel(arg->ident->s_name);
      else
	result = init_sem_node_list(
             create_ident_leaf(arg->ident->s_name, CONST, "'X" )
                                   );
  }
  else
    result =init_sem_node_list( expr_node(arg->arg) );

  return result;
}

/*-----------------------------------------------------------------------*/
/* erzeugt Knoten fuer die Parameter eines Prozeduraufrufs */
static SemNode* sem_node_list(ArgListEntry* arglist)
{
  SemNode* result = (SemNode*) NULL;
  ArgListEntry* arg = arglist;

  while (arg) {
    if (result)
      result = append_sem_node_list(result, sem_node_listEntry(arg));
    else
      result = sem_node_listEntry(arg);
    arg = arg->next;
  }
  return result;
}

/*-----------------------------------------------------------------------*/
/* erzeugt Knoten fuer die verschiedenen com-Bestandteile */
/* eines B(PN)^2-Programms */
static SemNode* com_node(TreeNode* node)
{
  if ( node == (TreeNode*)NULL )
    return( (SemNode*)NULL );

  switch (node->type)
  {
    case _BLOCK: /* Programm-/Prozedurblock */
    {
      list* dummy;
      return( block_node(node, &dummy, 0) );
      break;
    }
    case _EXPR: /* Ausdruck */
    {
      SemNode* result;
      int TmpRefCount = RefCount++;
      result = create_checked_atomic_command(expr_node(node->op1),
					     EXPR_BEGIN_);
      /* Referenzwerte der atomaren Aktion uebernehmen */
      SetRef(&(result->ref), ((SymTabEntry*)(node->op2))->s_ref.start_l,
	     ((SymTabEntry*)(node->op2))->s_ref.start_c,
	     ((SymTabEntry*)(node->op2))->s_ref.end_l,
	     ((SymTabEntry*)(node->op2))->s_ref.end_c,
	     TmpRefCount
	     );
      return result;
      break;
    }
    case _SEQ: /* Sequenz */
    {
      SemNode* sn1, *sn2;
      sn1 = com_node(node->op1);
      sn2 = com_node(node->op2);
      return( join_control(sn1, sn2, SEQ_) );
      /*
      return( join_control(com_node(node->op1), com_node(node->op2), SEQ_) );
      */
      break;
    }
    case _PAR: /* Parallelkonstrukt */
    {
      SemNode* sn1, *sn2;
      sn1 = com_node(node->op1);
      sn2 = com_node(node->op2);
      return( join_control(sn1, sn2, PARA_) );
      /*
      return( join_control(com_node(node->op1), com_node(node->op2), PARA_) );
      */
      break;
    }
    case _DO: /* do-od-Konstrukt */
    {
      SemNode* sn1, *sn2;
      sn2 = com_node(node->op2);
      sn1 = altset_node(node->op1);
      return( finite_loop_control(sn2, sn1, DO_) );
      /*
      return( finite_loop_control(com_node(node->op2),
				  altset_node(node->op1), DO_) );
				  */
      break;
    }
    case _CALL: /* Prozeduraufruf ohne Parameter */
    {
      SemNode* result;
      int TmpRefCount = RefCount++;

      result = create_procedure_call( ((SymTabEntry*)(node->op1))->s_name,
				     empty_sem_node_list(), IDENT_);
      /* Referenzwerte des Prozeduraufrufs uebernehmen */
      SetRef(&(result->ref), (*((ShortEntryPtr*)(node+1)))->s_ref.start_l,
	     (*((ShortEntryPtr*)(node+1)))->s_ref.start_c,
	     (*((ShortEntryPtr*)(node+1)))->s_ref.end_l,
	     (*((ShortEntryPtr*)(node+1)))->s_ref.end_c,
	     TmpRefCount
	    );
      return(result);
      break;
    }
    case _PCALL: /* Prozeduraufruf mit Parametern */
    {
      SemNode* result;
      int TmpRefCount = RefCount++;

      result = create_procedure_call( ((SymTabEntry*)(node->op1))->s_name,
				  sem_node_list( (ArgListEntry*)(node->op2) ),
				  IDENT_);
      /* Referenzwerte des Prozeduraufrufs uebernehmen */
      SetRef(&(result->ref), (*((ShortEntryPtr*)(node+1)))->s_ref.start_l,
	     (*((ShortEntryPtr*)(node+1)))->s_ref.start_c,
	     (*((ShortEntryPtr*)(node+1)))->s_ref.end_l,
	     (*((ShortEntryPtr*)(node+1)))->s_ref.end_c,
	     TmpRefCount
	    );
      return(result);
      break;
    }
    default: break;
  }
  /* sollte nie erreicht werden */
  return( (SemNode*)NULL );
}

/*-----------------------------------------------------------------------*/
/* den Typ SetRange (Ackermann) in Tscope umsetzen */
static data_description create_set(SetRange* set)
{
  data_description result;
  int highestvalue;

  highestvalue = set->high;
  result.scope.num_bounds = 0;

  do {
    set = set->next;
    result.scope.bounds[result.scope.num_bounds].low = set->low;
    result.scope.bounds[result.scope.num_bounds].high = set->high;
    result.scope.num_bounds++;
  } while ( (highestvalue > set->high) && 
	    (result.scope.num_bounds < MAX_INTERVALS) );  
  /* Warnung wegen zuvieler Intervalle ausgeben */
  if (result.scope.num_bounds == MAX_INTERVALS)
    compile_error(ERR_TOO_MANY_RANGES);
  return result;
}

/*-----------------------------------------------------------------------*/
/* die Variablen- und Prozedurdeklarationen eines Blocks */
/* uebertragen */
list* translate_decls(SymTabEntry* ptr, int BeginRNum)
/* BeginRNum: falls translate_decls() im Rahmen einer Prozedur-
              definition mit Parametern aufgerufen wird, enthaelt
	      BeginRNum die Referenznummer des Begin-Statements
	      dieser Prozedur;
	      sonst enthaelt BeginRNum den Wert 0.
*/
/* Die Referenznummerierung einer Prozedurdefinition muss in der
   Reihenfolge
     Prozedurkopf, Parameter, Begin-Statement des Prozedurblocks
     und lokale Variablen
   vorgenommen werden.
   Die Uebernahme einer Prozedurdefinition in den Melzer-Parsebaum
   geschieht allerdings in der Reihenfolge
     Prozedurkopf, Begin-Statement des Prozedurblocks, Parameter
     und lokale Variablen.
   Aus diesem Grund muss die Referenznummerierung des Begin-Statements
   besonders behandelt werden - siehe Kommentare im Quelltext.
*/
{
  list* newlist;
  list* mainlist = (list*)-1;
  data_description init_data, data_set;
  SetRange *set;

  while (ptr)
  {
      switch (ptr->s_type)
      {
        case _VAR:  /* Variable */
	{
          int VarRefNum;
	  newlist = empty_list();
	  newlist = append_list(newlist, ptr->s_name, ptr->s_partype);
	  set = ptr->s_set->next;

          /* ohne Initialisierungswert */
          if (ptr->s_init_or_cap == NONE)
	  {
            init_data.data_type = -1;
	    init_data.init_value = 0;
	  }
          /* initialisiert mit TRUE */
	  else if (ptr->s_init_or_cap == TT)
	  {
            init_data.data_type = BOOLEAN_TYPE;
	    init_data.init_value = 1;
	  }
          /* initialisiert mit FALSE */
	  else if (ptr->s_init_or_cap == FF)
	  {
            init_data.data_type = BOOLEAN_TYPE;
	    init_data.init_value = 0;
	  }
	  else
	  /* mit Initialisierungswert */
	  {
            init_data.data_type = USER_DEFINED;
	    init_data.init_value = ptr->s_init_or_cap;
	  }


          if ( (set->low == FF) || (set->low == TT) )
	    data_set = create_bool_set();
	  else
            data_set = create_user_defined_set( create_set(ptr->s_set) );

	  VarRefNum = RefCount++;
          /* wenn dies die erste lokale Variable innnerhalb einer Prozedur- */
	  /* deklaration ist, sind VarRefNum und das Argument BeginRNum     */
          /* identisch - VarRefNum u. RefCount muessen  also inkrementiert  */
	  /* werden (s.o.) */
	  if (VarRefNum == BeginRNum) {
	    VarRefNum++;
	    RefCount++;
	  }
	  /* in Symboltabelle eintragen */
	  mainlist = hide_identifier( mainlist, newlist,
                     create_initialized_typed_set(data_set, init_data),
				      VarRefNum);
	  break;
          
	}
        case _CHANNEL:  /* Kanal- */
        case _STACK:    /* bzw. Stack-Variable */
	{
	  int VarRefNum;

	  newlist = empty_list();
	  newlist = append_list(newlist, ptr->s_name, ptr->s_partype);
	  set = ptr->s_set->next;

          if ( (set->low == FF) || (set->low == TT) )
	    data_set = create_bool_set();
	  else
	    data_set = create_user_defined_set( create_set(ptr->s_set) );

	  VarRefNum = RefCount++;
	  VarRefNum = RefCount++;
          /* wenn dies die erste lokale Variable innnerhalb einer Prozedur- */
	  /* deklaration ist, sind VarRefNum und das Argument BeginRNum     */
          /* identisch - VarRefNum u. RefCount muessen  also inkrementiert  */
	  /* werden (s.o.) */
	  if (VarRefNum == BeginRNum) {
	    VarRefNum++;
	    RefCount++;
	  }

	  /* in Symboltabelle eintragen */	  
	  mainlist = hide_identifier( mainlist, newlist,
                     create_typed_channel(data_set, ptr->s_init_or_cap,
					  (ptr->s_type == _CHANNEL) ?
					  CHAN_VAR : STACK_VAR), VarRefNum);
	  break;
 	}
        case _PROC: /* Prozedur ohne Parameter */
	{
	  list* dummy;
	  ProcEntry *TmpProcs = (ProcEntry*) NULL;
	  int ProcHeadRefNum = RefCount++;
	  
	  if ( (ptr->s_init_or_cap == MAX_INT) ||
               (ptr->s_init_or_cap == 1)       ||
	       (ptr->s_init_or_cap == 0)         ) 
	  { /* _max_ = 1 -- nur eine Prozedur-Instanz */
            /* Prozedur in Symboltabelle eintragen */
	    insert_function(ptr->s_name, empty_parameter_list());
            /* Prozedurblock uebersetzen; Prozedurdeklaration */
            /* in Liste aufnehmen */
	    TmpProcs = AddProcEntry(TmpProcs, ptr->s_name, (list*) NULL,
				 block_node((TreeNode*)ptr->s_body, &dummy,0));
            /* in Prozedurkopfliste aufnehmen */
	    Procedures = AddProcHead(Procedures, ptr->s_name,
				     1, 0, TmpProcs, ProcHeadRefNum);
	  }
	  else { /* _max_ > 1 -- mehrere Prozedur-Instanzen */
	    int i, SavedCount;
	    char name[MAX_IDENT_LENGTH+3],
	         number[3];

            /* Prozedur mit Originalnamen in Symboltabelle eintragen */
	    insert_function(ptr->s_name, empty_parameter_list());
            SavedCount = RefCount;
	    for (i = 1; i <= ptr->s_init_or_cap; i++) {
	      RefCount = SavedCount;
	      (void)sprintf(number, "%d", i);
	      (void)strcpy(name, ptr->s_name);
	      (void)strcat(name, number);
            /* Prozedur mit Name+Nummer in Symboltabelle eintragen */
	      insert_function(name, empty_parameter_list());
            /* Prozedurblock uebersetzen; Prozedurdeklaration */
            /* in Liste aufnehmen */
	      TmpProcs = AddProcEntry(TmpProcs, name, (list*) NULL,
			     block_node((TreeNode*)ptr->s_body, &dummy,0));
	    }
            /* in Prozedurkopfliste aufnehmen */
            Procedures = AddProcHead(Procedures, ptr->s_name,
				     ptr->s_init_or_cap,
				     0, TmpProcs, ProcHeadRefNum);
	  }

	  break;
        }
        case _PPROC:
        {
	  ProcEntry *TmpProcs = (ProcEntry*) NULL;
	  SemNode* TmpNode;
	  list* TmpList;
          SymTabEntry *tmpptr;
          int j = 0,
	      BegRNum;
	  int ProcHeadRefNum = RefCount++;

	  /* Anzahl der Parameter ermitteln */
	  tmpptr = ptr->s_parlist;
	  while (tmpptr) {
            j++;
	    tmpptr = tmpptr->s_parlist;
	  }
	  /* Referenznummer fuer das Begin-Staement   */
	  /* des Prozedurblocks ermitteln - wird      */
	  /* an nachfolgenden Aufruf von block_node() */
	  /* und dort an translate_decls() uebergeben */
	  BegRNum = ProcHeadRefNum + j + 1;

	  if ( (ptr->s_init_or_cap == MAX_INT) ||
               (ptr->s_init_or_cap == 1)       ||
	       (ptr->s_init_or_cap == 0)         ) 
	  { /* _max_ = 1 -- nur eine Prozedur-Instanz */
            char *param_list = parameter_list(ptr->s_parlist);
            /* Prozedur in Symboltabelle eintragen */
	    insert_function(ptr->s_name, param_list);
            /* Prozedurblock uebersetzen */
	    TmpNode = block_node((TreeNode*)ptr->s_body, &TmpList, BegRNum);
	    /*	    TmpNode->ref.RefNum = BegRNum; */
            /* in Liste aufnehmen */
	    TmpProcs = AddProcEntry(TmpProcs, ptr->s_name, TmpList, TmpNode);
            /* in Prozedurkopfliste aufnehmen */
            Procedures = AddProcHead(Procedures, ptr->s_name,
				     1, j, TmpProcs, ProcHeadRefNum);
            free(param_list);
	  }
	  else { /* _max_ > 1 -- mehrere Prozedur-Instanzen */
	    int i, SavedCount;
	    char name[MAX_IDENT_LENGTH+4],
	         number[4];
            char *param_list = parameter_list(ptr->s_parlist);

            /* Prozedur mit Originalnamen in Symboltabelle eintragen */
	    insert_function(ptr->s_name, param_list);
            free(param_list);
	    SavedCount = RefCount;
	    for (i = 1; i <= ptr->s_init_or_cap; i++) {
	      RefCount = SavedCount;
	      (void)sprintf(number, "%d", i);
	      (void)strcpy(name, ptr->s_name);
	      (void)strcat(name, number);
            /* Prozedur mit Name+Nummer in Symboltabelle eintragen */
	      param_list = parameter_list(ptr->s_parlist);
	      insert_function(name, parameter_list(ptr->s_parlist));
	      free(param_list);
            /* Prozedurblock uebersetzen; Prozedurdeklaration */
            /* in Liste aufnehmen */
	      TmpNode = block_node((TreeNode*)ptr->s_body, &TmpList, BegRNum);
	      /*	      TmpNode->ref.RefNum = BegRNum; */
	      TmpProcs = AddProcEntry(TmpProcs, name, TmpList, TmpNode);
	    }
            /* in Prozedurkopfliste aufnehmen */
             Procedures = AddProcHead(Procedures, ptr->s_name,
				     ptr->s_init_or_cap,
				      j, TmpProcs, ProcHeadRefNum);
	  }
	  break;
        }
        default: break;
      }
      ptr = ptr->s_next;
  } /* while() */

  return(mainlist);
}

/*-----------------------------------------------------------------------*/
/* Uebersetzung eines Blocks */
static SemNode* block_node(TreeNode* node, list** decls, int RNum)
{
  list* TmpList;
  SemNode* TmpNode;
  int TmpRefCount;
#ifndef PRINT_TREE
  static TreeNode* treenode = (TreeNode*) NULL;
#endif

  /* Parameter RNum enthaelt bei Aufruf aus einer Prozedur-     */
  /* deklaration heraus die Referenznummer des Begin-Statements */
  if (RNum == 0)
    TmpRefCount = RefCount++;
  else
    TmpRefCount = RNum;

  if (node->op1)
  {
#ifndef PRINT_TREE
    if (treenode != node) {
      treenode = node;
      /* die Symbolliste von Ackermann mu */
      /* einmal(!) invertiert werden       */
      node->op1 = (TreeNode*) invert_list( (SymTabEntry*)(node->op1) );
    }
#endif
    /* Deklaratioonen eines Blocks uebersetzen */
    TmpList = translate_decls((SymTabEntry*)(node->op1), RNum);
  }
  else
    TmpList = (list*) -1;

  /* falls eine Prozedurdeklaration keine lokalen Variablen enthaelt */
  /* enthaelt der globale Zaehler RefCount jetzt die Referenznummer  */
  /* des Begin-Statements                                            */
  if (TmpRefCount == RefCount)
    RefCount++;

  /* den Block uebersetzen */
  TmpNode = control_block( com_node(node->op2), TmpList);

  /* Referenz des Begin-Statements uebernehmen */
  SetRef(&(TmpNode->ref), (*((ShortEntryPtr*)(node+1)))->s_ref.start_l,
	 (*((ShortEntryPtr*)(node+1)))->s_ref.start_c,
	 (*((ShortEntryPtr*)(node+1)))->s_ref.end_l,
	 (*((ShortEntryPtr*)(node+1)))->s_ref.end_c,
	 TmpRefCount
	);

  recover_hidden_identifier(TmpList);

  /* Liste zurueckgeben; wird beim Aufruf */
  /* fuer Prozedurbloecke benoetigt (s.o.)*/
  *decls = TmpList;

  return(TmpNode);
}

/*-----------------------------------------------------------------------*/
/* uebertraegt den Ackermann-Parsebaum in */
/* einen Melzer-Parsebaum */
SemNode* translateParseTree(TreeNode* tree)
{
  list* dummy;
#ifdef DEBUG
  fprintf(stderr, "\ntranslateParseTree()\n");
#endif /* DEBUG */

  /* Symboltabelle initialisieren */
  init_hashtab();

  return (SemNode*) block_node(tree, &dummy, 0);
}

/*-----------------------------------------------------------------------*/
/* End of file translate.c */
/*-----------------------------------------------------------------------*/
