/****************************************************************/
/*                                                              */
/*            C O D E - G E N E R A T O R  for B(PN)^2          */
/*                                                              */
/*            Universitaet Hildesheim                           */
/*            Institut fuer Informatik                          */
/*            Stephan Melzer                                    */
/****************************************************************/
/*            Diplomarbeit                                      */
/*            File        : var_mem_manage.c                    
              Version     : 1.0
	      Last Update : 18.10.1994
 ****************************************************************/
/*            Module information :                              */
/* This module organizes the shared access to the data by preparing,
   modifing and rewriting copies of the actual data memory. A switch-
   table stores a mapping of internal and user defined identifiers,
   out of this the runtime protocol is generated.
   The following routines are applied in every atomic command (in
   detail in its bollean expression) :
   get_pre_primed_value, get_post_primed_value
   The routines 
   get_pre_primed_variables and get_post_primed_variable
   are used for a non-destructive read (ex. channel access)
*/
/****************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/shm.h> 

#include "bool.h"
#include "error.h"
#include "var_mem_manage.h"
#include "local_sem.h"
#include "local_shm.h"
#include "includes.h"
#include "misc.h"



#define NO_CHECK    0
#define EMPTY_CHECK 1
#define FULL_CHECK  2


int sem_prot = SEMPROT_NUM;

char* free_segment; 
int next_table_index;
int table_id = -1; 
unsigned int table_size;
int unknown;

#ifdef NO_SHARED_INDEX
int index_switch[HASH_TABLE_SIZE]; 
#else
int* index_switch = (int*) -1;
int index_switch_id = -1;
#endif /* NO_SHARED_INDEX */

/* Tabelle fuer die B(PN)^2-Variablen im shared memory */
mem_entry* table = (mem_entry*)-1;

static int queue_is_empty(int* queue);
static int queue_is_full(int* queue);


/* File descriptor for the protocol file */
FILE* prot_file = (FILE*) NULL;
char Prot_fname[1024] = "";
boolean proto2file   = FALSE,
     proto2stdout = FALSE;

void sleepy(int j)
{
  sleep(j);
}

/*-------------------------------------------------------------------------*/
void generate_memory_table(int size, int x_memory_size)
/* This routine initialize the data segment which stores
   the data of the B(PN)^2 - program instead of the module shm.c.
   Latter organizes the additional data. 
   This data segment can be allocated in fixed size, because the size 
   of B(PN)^2 data segment can be calculated a priori.
   size : number of entries (number of different identifiers)
   x_memory_size : size of data needed by the sum over all sizes
                   of each used variable
*/ 
{
  /*--------------------------------------------------*/ 
  if ( (table_id = shmget( IPC_PRIVATE,
                           x_memory_size + size * sizeof(mem_entry),
                           shm_create_bits
                         )
       ) == -1)
  {
    perror("var_mem_manage.generate_memory_table().shmget()");
    error("cannot get shared static memory segment"); 
  }

  if ( (table = (mem_entry*) shmat(table_id, (char*)0, 0)) == (mem_entry*)-1) 
  {
    perror("var_mem_manage.generate_memory_table().shmat()");
    error ("cannot attach shared static memory segment"); 
  } 
  /*--------------------------------------------------*/ 
#ifndef NO_SHARED_INDEX
  if ( (index_switch_id = shmget( IPC_PRIVATE,
                           (int)(HASH_TABLE_SIZE) * sizeof(int),
                           shm_create_bits
                         )
       ) == -1)
  {
    perror("var_mem_manage.generate_memory_table().shmget()-no2");
    error("cannot get shared static memory segment"); 
  }

  if ( (index_switch = (int*) shmat(index_switch_id, (char*)0, 0)) ==(int*)-1) 
  {
    perror("var_mem_manage.generate_memory_table().shmat()");
    error ("cannot attach shared static memory segment"); 
  } 
#endif 
  /*--------------------------------------------------*/ 

  free_segment = ((char*) table) + size * sizeof(mem_entry); 
  next_table_index = 0;
  table_size = size;
#ifdef NO_SHARED_INDEX
  (void) memset(index_switch, UNUSED, HASH_TABLE_SIZE * sizeof(int));
#else
  {
    int i;
    for (i = 0; i < HASH_TABLE_SIZE; i++)
      index_switch[i] = UNUSED;
  }
#endif
} 

/*-------------------------------------------------------------------------*/
void erase_memory_table(void) 
/* This routine deallocates the required memory for B(PN)^2
*/
{ 
  if (table_id > -1) {
    if (table != (mem_entry*)-1) {
      if (shmdt((char *) table) == -1) 
	perror("var_mem_manage.erase_memory_table().shmdt()");
    }
    if (shmctl(table_id, shm_remove_bits, (struct shmid_ds*) 0) == -1)
      perror("var_mem_manage.erase_memory_table().shmctl()");
  }
  /*--------------------------------------------------*/ 
#ifndef NO_SHARED_INDEX
  if (index_switch_id > -1) {
    if (index_switch != (int*)-1) {
      if (shmdt((char *) index_switch) == -1) 
	perror("var_mem_manage.erase_memory_table().shmdt()");
    }
    if (shmctl(index_switch_id, shm_remove_bits, (struct shmid_ds*) 0) == -1)
      perror("var_mem_manage.erase_memory_table().shmctl()");
  }
#endif
} 
 
/*-------------------------------------------------------------------------*/
void initialize_variable(int hash_number, char var_name[], 
                         int channel_size, int init_value,
			 Tscope scope, int s_or_c, int RNum)
/* This function creates a new entry in the switch table.
   Concerning the parameters see above.
*/
{ 
  mem_entry* new_entry;

  if (index_switch[hash_number] != UNUSED)
    error("variable already initialized");
  else {
    index_switch[hash_number]= next_table_index;
    new_entry = &(table[next_table_index]);
    new_entry->value     = (int*) free_segment; 
    free_segment += (channel_size == 0 ? sizeof(int) : 
		       (channel_size + 1) * sizeof(int)); 
    new_entry->stack_or_chan = s_or_c;
    new_entry->RefNum = RNum;
    new_entry->checked_flag = NO_CHECK;
    new_entry->fifo_head = 0; 
    new_entry->fifo_tail = channel_size;
    new_entry->fifo_size = channel_size; 
    new_entry->is_locked = FALSE;
    new_entry->scope     = scope;
    strcpy (new_entry->name, var_name);
    if (channel_size == 0) {
	*(new_entry->value) = init_value;
	new_entry->init = init_value;
    }
    next_table_index++;
  }
}

/*----------------------------------------------------------------------*/
void lock(int semaphore, SET vars)
/* This function locks the access to the variables in the set vars.
   This critical section is controled by a semaphore.
   semaphore : pointer to the key of the table semaphore
   vars : Set of varibales to be locked
*/
{
  int i, j, index, goon;

  do
  {
    goon = TRUE;
    sem_wait(semaphore);
    for (i=0; (i < INT_WIDTH * BIT_LENGTH) && goon; i++)
      if (is_element(vars, i))
      {
	index = index_switch[i];
        if (table[index].is_locked)
	{
#ifdef DEBUG
	  fprintf(stderr, "Var.: %s ist bereits gelockt\n", table[index].name);
#endif
	  for (j=0; j < i; j++)
	    if (is_element(vars, j))
              table[index_switch[j]].is_locked = FALSE;
          sem_signal(semaphore);
	  sleepy(1);
	  goon = FALSE;
        }
        else
          table[index].is_locked = TRUE;
      }
  } while (!goon);

  sem_signal(semaphore);
}

/*-------------------------------------------------------------------------*/
void unlock(int semaphore, SET vars)
/* This function locks the access to the variables in the set vars.
   This critical section is controled by a semaphore.
   semaphore : pointer to the key of the table semaphore
   vars : Set of varibales to be locked
*/
{
  int i;

  sem_wait(semaphore);
  for (i=0; i < INT_WIDTH * BIT_LENGTH; i++)
    if (is_element(vars, i))
      table[index_switch [i]].is_locked = FALSE;
  sem_signal(semaphore);
}
 
/*-------------------------------------------------------------------------*/
int get_value_from_table(int index) 
/* This function gets the actual value of the variable indexed by hash bucket
  number "index".
   index : number of the hash bucket
*/
{  
  if (index == UNUSED) 
    error ("variable is not initialized"); 
  return *(table[index].value); 
} 
 
/*-------------------------------------------------------------------------*/
void set_value_from_table (int index, int new_value) 
/* This function sets the value of the variable indexed by a
   hash bucket number.
   index : number of the hash bucket
*/
{
  if (index == UNUSED) 
    error ("variable is not initialized"); 

  if ( legalvalue(new_value, table[index].scope) )
    *(table[index].value) = new_value; 
  else
    warning_illegal_value(table[index].name, new_value, table[index].scope);
}

/*-------------------------------------------------------------------------*/
int addone(int index, int pos)
/* Cyclic addition modulo channelsize.
  index : number of the hash_number
  pos : pos in the queue
*/
{
  if (pos < table[index].fifo_size)
    return pos + 1;
  else
    return 0;
}

/*-------------------------------------------------------------------------*/
int act_value_from_table(int index) 
{  
  if (index == UNUSED) 
    error("variable is not initialized"); 
 
  if (table[index].fifo_size == 0) 
    error("variable is not channel"); 

  if ( queue_is_empty( (int*)&table[index] ) ) 
  {
    unknown = TRUE;
    (void)fprintf(stderr, "\nVariable %s\n", table[index].name);
    runtime_warning(ERR_IS_EMPTY); 
    return UNKNOWN; 
  }
  else {
    if (table[index].stack_or_chan == CHAN_VAR)
      return (table[index].value[table[index].fifo_head]);
    else
      return (table[index].value[table[index].lifo_ptr]);
  }
} 
 
/*-------------------------------------------------------------------------*/
void enqueue_value_from_table(int index, int new_value) 
{  
  if (index == UNUSED) 
    error("variable is not initialized"); 
 
  if (table[index].fifo_size == 0) 
    error("variable is not channel"); 

  if ( queue_is_full( (int*)&table[index] ) ) {
    unknown = TRUE;
    (void)fprintf(stderr, "\nVariable %s\n", table[index].name);
    runtime_warning(ERR_IS_FULL); 
  }
  else
  {
    if ( legalvalue(new_value, table[index].scope) )
    {
      if (table[index].stack_or_chan == CHAN_VAR) {
	table[index].fifo_tail = addone(index, table[index].fifo_tail);
	table[index].value[table[index].fifo_tail] = new_value;
      } else {
	table[index].lifo_ptr++;
	table[index].value[table[index].lifo_ptr] = new_value;
      }
    }
    else
      warning_illegal_value(table[index].name, new_value, table[index].scope);
  }
} 

/*-------------------------------------------------------------------------*/
int dequeue_value_from_table(int index) 
{
  int result;

  if (index == UNUSED) 
    error("variable is not initialized"); 

  if (table[index].fifo_size == 0) 
    error("variable is not channel");

  if ( queue_is_empty( (int*)&table[index] ) ) 
  {
    unknown = TRUE; 
    (void)fprintf(stderr, "\nVariable %s\n", table[index].name);
    runtime_warning(ERR_IS_EMPTY); 
    result = UNKNOWN; 
  } 
  else 
  {
    if ( table[index].stack_or_chan == CHAN_VAR ) {
      result = table[index].value[table[index].fifo_head];
      table[index].fifo_head = addone(index, table[index].fifo_head);
    } else {
      result = table[index].value[table[index].lifo_ptr];
      table[index].lifo_ptr--;
    }
  }
  return result; 
}


/*-------------------------------------------------------------------------*/
int* get_addr_in_table(int index)
/* This function returns the address of the variable within the table.
   index : see above
*/
{
  if (table[index].fifo_size <= 0)
    return table[index].value;
  else
    return (int*) &(table[index]);
}

/*-------------------------------------------------------------------------*/
int get_memory_table_size(void)
/* The number of the available variables is returned. 
*/
{
  return table_size;
}

/*-------------------------------------------------------------------------*/
container* allocate_local_memory_for_copy(int size)
/* This function allocates memory outside the shared memory, in order to
   get memory to copy the actual variable setting.
   size : number of available variables
*/
{
  return (container*) Calloc((size_t) size, sizeof(container));
}

/*-------------------------------------------------------------------------*/
void get_pre_primed_variables(container* copy)
/* This function prepares a container to store the pre primed variables
   of all available variables. 
   copy : Pointer to a container which contains values of the pre primed
          variables
*/
{
  int c;

  for (c = 0; c < get_memory_table_size(); c++)
  {
    copy[c].is_used = FALSE;
    if (table[c].fifo_size == 0)
      copy[c].value = *(table[c].value); 
    else
      if ( queue_is_empty( (int*)&table[c] ) )
	copy[c].value = UNKNOWN;
      else
	copy[c].value = act_value_from_table(c);
  }
}

/*-------------------------------------------------------------------------*/
void get_post_primed_variables(container* copy)
/* This function prepares a container to store the post primed variables
   of all available variables. 
   copy : Pointer to a container which contains values of the post primed
          variables
*/
{
  int c;

  for (c = 0; c < get_memory_table_size(); c++)
  {
    copy[c].is_used = FALSE;
    if (table[c].fifo_size == 0)
      copy[c].value = UNKNOWN; 
    else
      if ( queue_is_full( (int*)&table[c] ) )
	copy[c].value = FULL;
      else
	copy[c].value = UNKNOWN;
  }
}

/*-------------------------------------------------------------------------*/
void initialize_protokoll(void)
/* This function initializes the protocol file for the runtime system.
   prot_name : string of the protocol file
*/
{
  if (!proto2file)
    return;
  OpenFile(&prot_file, Prot_fname, "w+");
  CloseFile(&prot_file);
}

/*-------------------------------------------------------------------------*/
void print_container(container* pre, container* post)
{
  int c;

  for (c=0; c < get_memory_table_size(); c++)
  {
    if (pre[c].is_used)
      printf ("PRE %s : %i ", table[c].name, pre[c].value);
    if (post[c].is_used)
      printf ("POST %s : %i \n", table[c].name, post[c].value);
  }
} 

/*-------------------------------------------------------------------------*/
void print_var_table()
{
  unsigned int c;

  for (c=0; c < table_size; c++)
    printf ("Var : %s VALUE : %i access %i \n",
	    table[c].name, *(table[c].value), table[c].is_locked);
}

/*-------------------------------------------------------------------------*/
void rewrite_copies(container* pre, container* post)
/* This function actualize the values of the varaiables in the way of 
   modified container. The changing of the value setting invokes a 
   protocol entry. 
   prot_name : string of the protocol file
   pre : pointer to a container which indicates the pre values of the varaibles
   post : pointer to a container which indicates the post values of the
          variables
*/
{
  int c;
  /* fuer Variablennamen und -werte */
  char transition_label[MULTI_LABEL_SIZE] = "";
  /* fuer Variablen-Referenznummern und Flags */
  char VarRefs[2*MULTI_LABEL_SIZE] = "T\"";
  /* fuer Referenznummer(n) der atomaren Aktion(en) */
  char ComRefs[MULTI_LABEL_SIZE] = "";
  /* fuer Zeilen-/Spaltenreferenzen */
  char LineColRefs[MULTI_LABEL_SIZE] = "";

  sem_wait(sem_prot);

  for (c = 0; c < get_memory_table_size(); c++)
    if ((pre[c].is_used) && (table[c].fifo_size != 0))
    { /* Lesezugriff auf Stack/Channel */
      char dummy_string[SINGLE_LABEL_SIZE];
      /* Variablenname, Referenznummer und gelesener Wert */
      sprintf(dummy_string, "%s(%d)?%d ", table[c].name,table[c].RefNum,
              dequeue_value_from_table(c));
      strcat(transition_label, dummy_string);
      /* Referenznummer und Flag */
      sprintf(dummy_string, "%d:%d*", table[c].RefNum, SC_READ);
      strcat(VarRefs, dummy_string);
    }
    else if ((pre[c].is_used) && (!post [c].is_used) && 
	     (pre[c].value != SYNCHRONIZED))
    { /* Lesezugriff auf Variable */
      char dummy_string[SINGLE_LABEL_SIZE];
	
      /* Variablenname, Referenznummer und zweimal "Vorwert" */
      /* (nur Lesezugriff, keine Aenderung) */
      sprintf (dummy_string, "%s(%d)_%d_%d ", table[c].name,table[c].RefNum, 
               pre[c].value, pre[c].value);
      strcat (transition_label, dummy_string);
      /* Referenznummer und Flag */
      sprintf(dummy_string, "%d:%d*", table[c].RefNum, VAR_ACCESS);
      strcat(VarRefs, dummy_string);
    }
    else if ((pre[c].is_used) && (post [c].is_used) && 
	     (pre[c].value == SYNCHRONIZED))
    { /* Synchronisation */
      char dummy_string[SINGLE_LABEL_SIZE];
      /* */
      /* Variablenname, Referenznummer und Wert zweimal ausgeben */
      sprintf (dummy_string, "%s(%d)!%d %s(%d)?%d ",
	       table[c].name, table[c].RefNum, post[c].value,
	       table[c].name, table[c].RefNum, post[c].value);
      strcat (transition_label, dummy_string);
      /* Referenznummern und Flags */
      sprintf(dummy_string, "%d:%d*%d:%d*",
	      table[c].RefNum, SC_READ,
	      table[c].RefNum, SC_WRITE);
      strcat(VarRefs, dummy_string);
      post[c].is_used = FALSE;
      pre[c].is_used = FALSE;
      /* fuer detailierte Ausgabe */
      (void)strcpy(PMsg.Msg, "synchronisation");
    }

  for (c = 0; c < get_memory_table_size(); c++)
    if (post[c].is_used)
    { /* Wert der Variablen geaendert */
      if (table[c].fifo_size == 0)
      {
	char dummy_string[SINGLE_LABEL_SIZE];
	
      /* Variablenname, Referenznummer, Vor- und Nachwert */
        sprintf (dummy_string, "%s(%d)_%d_%d ", 
                 table[c].name, table[c].RefNum, pre[c].value, post[c].value);
        strcat(transition_label, dummy_string);
	/* Referenznummer und Flag */
        sprintf(dummy_string, "%d:%d*",table[c].RefNum, VAR_ACCESS);
        strcat(VarRefs, dummy_string);
	set_value_from_table(c, post[c].value);
      }
      else
      { /* Schreibzugriff auf Stack/Channel */
	char dummy_string[SINGLE_LABEL_SIZE];
	
	/* Variablenname, Referenznummer und geschriebener Wert */
        sprintf(dummy_string, "%s(%d)!%d ", table[c].name,
		table[c].RefNum, post[c].value);
        strcat(transition_label, dummy_string);
	/* Referenznummer und Flag */
        sprintf(dummy_string, "%d:%d*",table[c].RefNum, SC_WRITE);
        strcat(VarRefs, dummy_string);
	enqueue_value_from_table(c, post[c].value);
      }
    }
   /* Ausgabe fuer X?? bzw. X!! */
    else if ( !(pre[c].is_used) && (table[c].fifo_size > 0) &&
	      (table[c].checked_flag != NO_CHECK))
    {
      char dummy_string[SINGLE_LABEL_SIZE];
      /* Abfrage auf "vollkommen leer" */
      if (table[c].checked_flag & EMPTY_CHECK) {
	/* Variablenname, Referenznummer Ergebnis */
	sprintf(dummy_string, "%s(%d)??%s ",
		table[c].name, table[c].RefNum,
		queue_is_empty( (int*)&table[c] )?"tt":"ff");
        strcat(transition_label, dummy_string);
	table[c].checked_flag &= ~EMPTY_CHECK;
	/* Referenznummer und Flag */
        sprintf(dummy_string, "%d:%d*",table[c].RefNum, SC_EMPTY);
        strcat(VarRefs, dummy_string);
      }
      /* Abfrage auf "vollkommen voll" */
      if (table[c].checked_flag & FULL_CHECK) {
	/* Variablenname, Referenznummer Ergebnis */
	sprintf(dummy_string, "%s(%d)!!%s ",
		table[c].name, table[c].RefNum,
		queue_is_full( (int*)&table[c] ) ? "tt" : "ff");
	table[c].checked_flag &= ~FULL_CHECK;
	/* Referenznummer und Flag */
        sprintf(dummy_string, "%d:%d*",table[c].RefNum, SC_FULL);
        strcat(VarRefs, dummy_string);
      }
    }


  if (proto2file || proto2stdout) {/*Protokoll in Datei und/oder nach stdout*/
    char dummy_string[SINGLE_LABEL_SIZE];

    /* in PMsg sind Zeilen-/Spalten-Referenzen und */
    /* Referenznummern der atomaren Aktion(en) gesammelt */
    RefList* Refs = PMsg.refs;
    while (Refs != (RefList*)NULL) {
      /* Zeilen-/Spalten-Referenz */
      sprintf(dummy_string, "R\"(%d,%d;%d,%d)\"",
	      Refs->ref.start_l,
	      Refs->ref.start_c,
	      Refs->ref.end_l,
	      Refs->ref.end_c
             );
      strcat(LineColRefs, dummy_string);
      /* Referenznummer der atomaren Aktion */
      sprintf(dummy_string, "%d:0", Refs->ref.RefNum);
      strcat(ComRefs, dummy_string);
      strcat(ComRefs, Refs->nextRef?"*":"");
      Refs = Refs->nextRef;
    }
    /* Zusammenfassen zu einem String wie z.B. */
    /* T"2:256*2:512*7:0*8:0" , d.h.,          */
    /* Synchronisation ueber die Variable mit Referenznummer 2  */
    /* in den atomaren Aktionen mit den Referenznummern 7 und 8 */
    strcat(ComRefs, "\"");
    strcat(VarRefs, ComRefs);

    if (proto2file) { /* Protokoll in Datei */
      OpenFile(&prot_file, Prot_fname, "a+");
      /* Prozess-ID ausgeben */
      fprintf(prot_file, "PID: %d\t", (int) getpid());
      /* Referenz der Form T"....." ausgeben */
      fprintf(prot_file, VarRefs);
#ifdef DETAIL_PROTO
      /* ggf. zustzlich Zeilen-/Spalten-Referenz */
      /* in der Form R"(5,3;5,11)" ausgeben ...   */
      fprintf(prot_file, "\t");
      fprintf(prot_file, LineColRefs);
      /* sowie eine Erklaerung ... */
      if (strlen(PMsg.Msg) > 0)
	fprintf(prot_file, " %s", PMsg.Msg);
      /* und Variablenname(n) mit Vor- und Nachwert(en) */
      if (strlen(transition_label) > 0)
	fprintf(prot_file, " %s", transition_label);
#endif /* DETAIL_PROTO */
      fprintf(prot_file, "\n");

      CloseFile(&prot_file);
    }

    if (proto2stdout) { /* Protokoll nach stdout ausgeben */
      /* Prozess-ID ausgeben */
      fprintf(stdout, "PID: %d\t", (int) getpid());
      /* Referenz der Form T"....." ausgeben */
      fprintf(stdout, VarRefs);
#ifdef DETAIL_PROTO
      /* ggf. zustzlich Zeilen-/Spalten-Referenz */
      /* in der Form R"(5,3;5,11)" ausgeben ...   */
      fprintf(stdout, "\t");
      fprintf(stdout, LineColRefs);
      /* sowie eine Erklaerung ... */
      if (strlen(PMsg.Msg) > 0)
	fprintf(stdout, " %s", PMsg.Msg);
      /* und Variablenname(n) mit Vor- und Nachwert(en) */
      if (strlen(transition_label) > 0)
	fprintf(stdout, " %s", transition_label);
#endif /* DETAIL_PROTO */
      fprintf(stdout, "\n");
    }

  } /* if (proto2file || proto2stdout) */ 
  
  InitPMsg();

  sem_signal(sem_prot);
}

/*-------------------------------------------------------------------------*/
void rewrite_copies_noprot(container* post)
/* This function actualize the values of the varaiables in the way of 
   modified container.
   pre : pointer to a container which indicates the pre values of the varaibles
   post : pointer to a container which indicates the post values of the
          variables
*/
{
  int c;

  sem_wait(sem_prot);

  for (c = 0; c < get_memory_table_size(); c++)
    if (post[c].is_used)
    {
      if (table[c].fifo_size == 0)
	set_value_from_table(c, post[c].value);
      else
	enqueue_value_from_table(c, post[c].value);
    }

  sem_signal(sem_prot);
}

/*-------------------------------------------------------------------------*/
void deallocate_local_memory_copy(container* page)
/* This function disposes the local allocated memory for the containers.
   page : pointer to a container which has to be given free
*/
{
  free(page);
}

/*-------------------------------------------------------------------------*/
int get_pre_primed_value(container* pre, int hash_number, int* flag)
/* This function allows the access to a variable inside a container.
   The actual value is returned and the flag indicates wether the access
   is allowed or not.
   pre : pointer to a container which contains the pre values
   hash_number : number of the hash bucket of the specified variable
   flag : is set to TRUE if the access is correct
*/
{
  int index = index_switch[hash_number];

  pre[index].is_used = TRUE;
  *flag = *flag && (pre[index].value != UNKNOWN);  /* ???? */
  return (pre[index].value == UNKNOWN ? 0 : pre[index].value);
}

/*-------------------------------------------------------------------------*/
int getpreval_noprot(container* pre, int hash_number, int* flag)
{
  int index = index_switch[hash_number];

  *flag = *flag && (pre[index].value != UNKNOWN);  /* ???? */
  return (pre[index].value == UNKNOWN ? 0 : pre[index].value);
}

/*-------------------------------------------------------------------------*/
int get_post_primed_value(container* post, int hash_number)
/* This function allows the access to a variable inside a container.
   The actual value is returned and the flag indicates wether the access
   is allowed or not.
   pre : pointer to a container which contains the pre values
   hash_number : number of the hash bucket of the specified variable
   flag : is set to TRUE if the access is correct
*/
{
  int index = index_switch[hash_number];

  if (post[index].value == UNKNOWN)
    post[index].value = createValue(table[index].scope);
  return (post[index].value);
}

/*-------------------------------------------------------------------------*/
int get_synchro(container* pre, container* post, int hash_number, int* flag)
{
  int index = index_switch[hash_number];

  *flag = *flag && (pre[index].value == SYNCHRONIZED);
  post[index].is_used = TRUE;
  return post[index].value; 
}

/*-------------------------------------------------------------------------*/
void set_synchro(container* pre, container* post, int hash_number, 
	         int new_value, int* flag)
{
  int index = index_switch[hash_number];

  *flag = *flag && (pre[index].value != SYNCHRONIZED);
  pre[index].is_used = TRUE;
  post[index].value = new_value;
  pre[index].value = SYNCHRONIZED;
}

/*-------------------------------------------------------------------------*/
#ifdef OLD_VERSION
void set_post_primed_value(container* post, int hash_number, 
                           int new_value, int* flag)
/* This function allows the setting of a variable inside a container.
   If there was already a writing access to this variable, then check
   the coherence of both value and indicates this by the flag.
   post : pointer to a container which contains the post values
   hash_number : number of the hash bucket of the specified variable
   new_value : post value of the variable
   flag : is set to FALSE if a possible previous post value does not
          fit to this new_value
*/
{
  int index = index_switch[hash_number];

  if (post[index].is_used)
  {
    if (post[index].value == new_value)
      return;
    else
      *flag = FALSE;
  }
  else
  {
    post[index].is_used = TRUE;
    post[index].value   = new_value;
  }
}

#else /* OLD_VERSION */

/* ACHTUNG -- muss unbedingt getestet werden !! -- jj */

/*-------------------------------------------------------------------------*/
void set_post_primed_value(container* post, int hash_number, 
                           int new_value, int* flag)
/* This function allows the setting of a variable inside a container.
   If there was already a writing access to this variable, then check
   the coherence of both value and indicates this by the flag.
   post : pointer to a container which contains the post values
   hash_number : number of the hash bucket of the specified variable
   new_value : post value of the variable
   flag : is set to FALSE if a possible previous post value does not
          fit to this new_value
*/
{
  int index = index_switch[hash_number];

  if (post[index].is_used)
  {
    if (post[index].value == new_value)
      return;
    else
      *flag = FALSE;
  }
  else
    if (legalvalue(new_value, table[index].scope))
    {
      post[index].is_used = TRUE;
      post[index].value   = new_value;
    }
    else
      *flag = FALSE;
}

#endif /* OLD_VERSION */

/*-------------------------------------------------------------------------*/
int top_fifo(container* pre, int hash_number, int *flag)
{
  int index = index_switch[hash_number];

  if (pre[index].value == UNKNOWN)
  {
    if (table[index].checked_flag == NO_CHECK) {
      (void)fprintf(stderr, "\nVariable %s\n", table[index].name);
      runtime_warning(ERR_IS_EMPTY); 
    }
    *flag = FALSE;
    return 0;
  }
  else
  {
    pre[index].is_used = TRUE;
    return (pre[index].value);
  }
}

/*-------------------------------------------------------------------------*/
void push_fifo(container* post, int hash_number, int new_value, int* flag)
{
  int index = index_switch[hash_number];

  if (post[index].value == FULL) {
    *flag = FALSE;
    if (table[index].checked_flag == NO_CHECK) {
      (void)fprintf(stderr, "\nVariable %s\n", table[index].name);
      runtime_warning(ERR_IS_FULL );
    }
  }
  else
  {
    if (post[index].is_used)
    {
      if (post[index].value == new_value)
	return;
      else
	*flag = FALSE;
    }
    else
    {
      post[index].is_used = TRUE;
      post[index].value   = new_value;
    }
  }
}	    
	
/*-------------------------------------------------------------------------*/
int* get_addr(int hash_number)
/* This function returns the address belonging to a variable specified by
   hash bucket number.
   hash_number : number of a hash bucket
*/
{
  int index = index_switch[hash_number];
  return get_addr_in_table(index);
}

/*-------------------------------------------------------------------------*/
void re_initialize(SET ident_set, boolean only2proto)
/* Re-initializes variables at block beginning.
   ident_set : Set of variables belonging to this block
   only2prot : TRUE -> wird am Ende eines B(PN)^2-Blocks aufgerufen,
                 um fuer jede lokale Variable eine Ausgabe in
	         die Protokolldatei zu generieren,
               FALSE -> wird zu Beginn eines B(PN)^2-Blocks aufgerufen,
                 um alle lokalen Variablen zu initialisieren und zu 
                 protokollieren
*/
{
  int i;
  char tmpstr[50], protostr[1024]= "";

  if (is_empty(ident_set) ||
      ((only2proto) && (!proto2file) && (!proto2stdout))
     )
    return;

  for (i=0; i < INT_WIDTH * BIT_LENGTH; i++)
    if (is_element(ident_set, i))
    {
      int index = index_switch[i];

      if (!only2proto) { /* Variablen initialisieren */
	*(table[index].value) = table[index].init;
	table[index].fifo_head = 0;
	table[index].fifo_tail = table[index].fifo_size;
	table[index].lifo_ptr = -1;
	table[index].checked_flag = NO_CHECK;
        /* Protokollausgabe: */
        /* Prozess-ID, Referenznummer und Flag fuer Initialisierung */
	sprintf(tmpstr, "PID: %d\tT\"%d:%d\"\n",
		(int) getpid(), table[index].RefNum, VAR_INIT);
      }
      else
        /* Protokollausgabe: */
        /* Prozess-ID, Referenznummer und Flag fuer Blockterminierung */
	sprintf(tmpstr, "PID: %d\tT\"%d:%d\"\n",
		(int) getpid(), table[index].RefNum, VAR_TERM);
      strcat(protostr, tmpstr);
    }

  sem_wait(sem_prot);

  if (proto2file) {
    OpenFile(&prot_file, Prot_fname, "a+");
    (void)fprintf(prot_file, protostr);
    CloseFile(&prot_file);
  }
  if (proto2stdout)
    (void)fprintf(stdout, protostr);

  sem_signal(sem_prot);
}

/*------------------------------------------------------------------------*/
void PCall2Prot(char* PName, Reference ref, int RNum, int Flag)
/*
   wird vor Aufruf einer Prozedur und nach Beendigung einer Prozedur
   aufgerufen.
   PName : Prozedurbezeichner
   ref   : Struktur mit Referenznummer des Prozeduraufrufs sowie dessen
           Zeilen-/Spalten-Referenzen
   RNum  : Referenznummer des aufgerufenen Prozedurkopfes
   Flag  : gibt an, ob Aufruf oder Beendigung einer Prozedur
*/
{
  if ((!proto2file) && (!proto2stdout))
    return;

  sem_wait(sem_prot);

  if (proto2file) {
    OpenFile(&prot_file, Prot_fname, "a+");
    /* Prozess-ID, Referenznummer des Prozedurkopfes, Flag und */
    /* Referenznummer des Prozeduraufrufs */
#ifndef DETAIL_PROTO
    (void)fprintf(prot_file, "PID: %d\tT\"%d:%d*%d:0\"\n",
		  (int)getpid(), RNum, Flag, ref.RefNum);
#else /* DETAIL_PROTO */
    (void)fprintf(prot_file, "PID: %d\tT\"%d:%d*%d:0\"",
		  (int)getpid(), RNum, Flag, ref.RefNum);
    /* Zeilen-/Spaltenreferenz des Prozeduraufrufs und */
    /* Prozedurbezeichner */
    if (Flag == PROC_CALL)
      (void)fprintf(prot_file, "\tR\"(%d,%d;%d,%d)\" call of %s\n",
		    ref.start_l, ref.start_c,ref.end_l, ref.end_c, PName);
    else
      (void)fprintf(prot_file, "\tR\"(%d,%d;%d,%d)\" return from %s\n",
		    ref.start_l, ref.start_c,ref.end_l, ref.end_c, PName);
#endif /* DETAIL_PROTO */
    CloseFile(&prot_file);
  }

  if (proto2stdout) {
    /* Prozess-ID, Referenznummer des Prozedurkopfes, Flag und */
    /* Referenznummer des Prozeduraufrufs */
#ifndef DETAIL_PROTO
    (void)fprintf(stdout, "PID: %d\tT\"%d:%d*%d:0\"\n",
		  (int)getpid(), RNum, Flag, ref.RefNum);
#else /* DETAIL_PROTO */
    (void)fprintf(stdout, "PID: %d\tT\"%d:%d*%d:0\"",
		  (int)getpid(), RNum, Flag, ref.RefNum);
    /* Zeilen-/Spaltenreferenz des Prozeduraufrufs und */
    /* Prozedurbezeichner */
    if (Flag == PROC_CALL)
      (void)fprintf(stdout, "\tR\"(%d,%d;%d,%d)\" call of %s\n",
		    ref.start_l, ref.start_c,ref.end_l, ref.end_c, PName);
    else
      (void)fprintf(stdout, "\tR\"(%d,%d;%d,%d)\" return from %s\n",
		    ref.start_l, ref.start_c,ref.end_l, ref.end_c, PName);
#endif /* DETAIL_PROTO */
  }

  sem_signal(sem_prot);
}

/************************ USER - ROUTINES for QUEUEs ********************/

/*-------------------------------------------------------------------------*/
int add_one(int* queue, int pos)
{
  mem_entry* q = (mem_entry*) queue;

  if (pos < q->fifo_size)
    return pos + 1;
  else
    return 0;
}

/*-------------------------------------------------------------------------*/
static int queue_is_empty(int* queue)
{
  mem_entry* q = (mem_entry*) queue;

  if (q->stack_or_chan == CHAN_VAR)
    return (add_one(queue, q->fifo_tail) == q->fifo_head);
  else
    return (q->lifo_ptr == -1);
}

/*-------------------------------------------------------------------------*/
static int queue_is_full(int* queue)
{
  mem_entry* q = (mem_entry*) queue;

  if (q->stack_or_chan == CHAN_VAR) {
    return (add_one(queue, add_one(queue, q->fifo_tail)) == q->fifo_head);
  } else
    return (q->lifo_ptr == (q->fifo_size - 1));
}

/*-------------------------------------------------------------------------*/
void enqueue(int* queue, int value)
{
  mem_entry* q = (mem_entry*) queue;

  if (!queue_is_full(queue))
  {
    if ( legalvalue(value, q->scope) )
    {
      if (q->stack_or_chan == CHAN_VAR) {
	q->fifo_tail = add_one(queue, q->fifo_tail);
	q->value[q->fifo_tail] = value;
      } else {
	q->lifo_ptr++;
	q->value[q->lifo_ptr] = value;
      }
    }
    else
      warning_illegal_value(q->name, value, q->scope);
  }
  else {
    (void)fprintf(stderr, "\nVariable %s\n", q->name);
    runtime_warning(ERR_IS_FULL); 
  }
}

/*-------------------------------------------------------------------------*/
int dequeue(int* queue)
{
  mem_entry* q = (mem_entry*) queue;
  int result = 0;

  if (!queue_is_empty(queue))
  {
    if (q->stack_or_chan == CHAN_VAR) {
      result = q->value[q->fifo_head];
      q->fifo_head = add_one(queue, q->fifo_head);
    } else {
      result = q->value[q->lifo_ptr];
      q->lifo_ptr--;
    }
  }
  else {
    (void)fprintf(stderr, "\nVariable %s\n", q->name);
    runtime_warning(ERR_IS_EMPTY); 
  }

  return result;
}

/*-------------------------------------------------------------------------*/
int empty_check(int* queue)
{
  mem_entry* q = (mem_entry*) queue;

  q->checked_flag |= EMPTY_CHECK;

  if (q->stack_or_chan == CHAN_VAR)
    return (add_one(queue, q->fifo_tail) == q->fifo_head);
  else
    return (q->lifo_ptr == -1);
}

/*-------------------------------------------------------------------------*/
int full_check(int* queue)
{
  mem_entry* q = (mem_entry*) queue;

  q->checked_flag |= FULL_CHECK;

  if (q->stack_or_chan == CHAN_VAR) {
    return (add_one(queue, add_one(queue, q->fifo_tail)) == q->fifo_head);
  } else
    return (q->lifo_ptr == (q->fifo_size - 1));
}

/*-------------------------------------------------------------------------*/
int NoOfElements(int* queue)
{
  mem_entry* q = (mem_entry*) queue;

  if (q->stack_or_chan == CHAN_VAR) {
    int h = q->fifo_head;
    int t = q->fifo_tail;
    int i = 1;

    while(h != t) {
      if (h < q->fifo_size)
	h++;
      else 
	h = 0;
      if (i < q->fifo_size)
	i++;
      else 
	i = 0;
    }
    return i;
  } else
    return(q->lifo_ptr + 1);
}

/*-------------------------------------------------------------------------*/
/* End of file var_mem_manage.c */
/*-------------------------------------------------------------------------*/
