/*****************************************************************************/
/* compose.c                                                                 */
/*****************************************************************************/
/* Compose two or more nets into one larger net.			     */
/*                                                            Stefan Schwoon */
/*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "memlib.h"
#include "mnet.h"
#include "readHL.h"
#include "mtupel.h"

/*****************************************************************************/

static void usage(char *myname)
{
  fprintf(stderr,
    "\n%s -- compose one or more nets into one larger net\n\n"

    "Usage: %s -o outfile <infile1> [modifier1] [infile2 [modifier2] ...]\n\n"

    "     Syntax of modifiers:\n"
    "       modifier -> expr | expr:expr     op    -> + | - | *\n"
    "       expr     -> op value expr        value -> float | max\n"
    "       expr     -> empty                max   -> # netnum\n"
    "       netnum   -> 1 | .. | number of infiles\n"
    "     Examples:\n"
    "       *0.5                divide all coords by two\n"
    "       +#1:                shift x-coords by maximum x-extent of net1\n"
    "                           (place net to the right of net1)\n\n"

    "Infiles may be either high level or low level nets, but all files\n"
    "must be of the same type. (You cannot mix high and low level nets.)\n"
    "The nets will be composed into one large net (transforming their\n"
    "coordinate-space according to their respective modifiers) and the\n"
    "result is written to outfile.\n\n"
 
    "Version 0.9 (22.09.1997) Stefan Schwoon\n\n", myname, myname);

  exit(1);
} /* usage() */


/*****************************************************************************/

M_Block	*BlArray;

MPTupel make_blocktupel(char *blocklist)
{
	MPTupel blocktupel = MPT_Create();
	MTupel	mt;
	int	i;

	if (!blocklist) return blocktupel;

	mt = read_MTupel(&blocklist);
	for (i=0; i<mt->count; i++)
		MPTT_AddValue(blocktupel, BlArray[MT_Place(mt,i)]);
	free(mt->place); free(mt);

	return blocktupel;

}

void decode_block_refs(M_Net net)
{
	register M_Block	bl;
	register M_Place	pl;
	register M_Transition	tr;
	Cardinal32		i = 0;

	for (bl = net->Blocks; bl; bl = bl->nextplace)
	  if (bl->numname > i)
            i = bl->numname;

	BlArray = MYmalloc((i+1) * sizeof(M_Block));

	for (bl = net->Blocks; bl; bl = bl->nextplace)
		BlArray[bl->numname] = bl;

	for (bl = net->Blocks; bl; bl = bl->nextplace)
		bl->blocktupel = make_blocktupel(bl->blocklist);
	for (pl = net->Places; pl; pl = pl->nextplace)
		pl->blocktupel = make_blocktupel(pl->blocklist);
	for (tr = net->Transitions; tr; tr = tr->nexttrans)
		tr->blocktupel = make_blocktupel(tr->blocklist);
	for (tr = net->PhTransitions; tr; tr = tr->nexttrans)
		tr->blocktupel = make_blocktupel(tr->blocklist);

	free(BlArray);

}

void decode_phantom_refs(M_Net net)
{
	register M_Transition	tr;
	register M_Transition	ptr;
	M_Transition		*TrArray;
	MTupel			mt;
	Cardinal32		i = 0;

	for (tr = net->Transitions; tr; tr = tr->nexttrans)
		if (tr->numname > i)  i = tr->numname;

	TrArray = MYmalloc((i+1) * sizeof(M_Transition));

	for (tr = net->Transitions; tr; tr = tr->nexttrans)
		TrArray[tr->numname] = tr;

	for (ptr = net->PhTransitions; ptr; ptr = ptr->nexttrans)
	{
		ptr->partners = MPT_Create();
		if (!ptr->corrtran) continue;
		mt = read_MTupel(&ptr->corrtran);
		for (i=0; i<mt->count; i++)
			MPTT_AddValue(ptr->partners, TrArray[MT_Place(mt,i)]);
		free(mt->place); free(mt);
	}

	free(TrArray);

}

void rename_node(char **name, int i)
{
	char *tmp = MYmalloc(strlen(*name)+4);
	sprintf(tmp,"%s.%u",*name,i);
	free(*name);
	*name = tmp;
}

void rename_nodes(M_Net net, int i)
{
	register M_Block	bl;
	register M_Place	pl;
	register M_Transition	tr;

	for (bl = net->Blocks; bl; bl = bl->nextplace)
		rename_node(&bl->charname,i);
	for (pl = net->Places; pl; pl = pl->nextplace)
		rename_node(&pl->charname,i);
	for (tr = net->Transitions; tr; tr = tr->nexttrans)
		rename_node(&tr->charname,i);
	for (tr = net->PhTransitions; tr; tr = tr->nexttrans)
		rename_node(&tr->charname,i);

}
/*****************************************************************************/

#define CO_X 1
#define CO_Y 0

void parse_mod(char *mod, M_Net *net, int numnets,
	       int xflag, float *scale, float *add)
{
	char op, *c = mod;

	while ((op = *c))
	{
		if (!strchr("*-+",op))
		{
			fprintf(stderr,"syntax error in modifier %s\n", mod);
			exit(1);
		}
		else
		{
			float val;
			if (*++c == '#')
			{
				int num = 0;
				while (isdigit(*++c)) num = num*10 + *c - '0';
				if (!num || num > numnets)
				{
					fprintf(stderr, "invalid net number in "
						" modifier %s\n", mod);
					exit(1);
				}
				val = xflag? net[num-1]->x : net[num-1]->y;
			}
			else
			{
				val = 0;
				while (isdigit(*c)) val = val*10 + *c++ - '0';
				if (*c == '.')
				{
					float div = 1;
					while (isdigit(*++c))
						val += (*c-'0') * (div /= 10);
				}
			}

			switch(op)
			{
				case '*': *scale *= val;
					  break;
				case '+': *add += val;
					  break;
				case '-': *add -= val;
					  break;
			}
		}
	}
}

float xscale, xadd, yscale, yadd;

void shift_coords(t_coords *co)
{
	if (!co) return;
	co->x *= xscale; co->x += xadd;
	co->y *= yscale; co->y += yadd;
}

void shift_place(M_Place pl)
{
	shift_coords(pl->co);
	shift_coords(pl->absname);
	shift_coords(pl->absbed);
}

void shift_trans(M_Transition tr)
{
	shift_coords(tr->co);
	shift_coords(tr->absname);
	shift_coords(tr->absbed);
	shift_coords(tr->absval);
}				
	
void shift_edge(M_Edge st)
{
	shift_coords(st->midpoint);
	shift_coords(st->absweight);
	shift_coords(st->absins);
}

void translate_coords(M_Net *net, char **modifier, int numnets)
{
	int i;
	M_Block bl;
	M_Place pl;
	M_Transition tr;
	M_STEdge st;
	M_TSEdge ts;
	M_Text tx;
	
	for (i = 0; i < numnets; i++)
	{
		char *c;
		if (!modifier[i]) continue;
		xscale = yscale = 1;
		xadd = yadd = 0;

		if ((c = strchr(modifier[i],':')))
		{
			*c = '\0';
			parse_mod(modifier[i],net,numnets,CO_X,&xscale,&xadd);
			parse_mod(++c,net,numnets,CO_Y,&yscale,&yadd);
		}
		else
		{
			parse_mod(modifier[i],net,numnets,CO_X,&xscale,&xadd);
			parse_mod(modifier[i],net,numnets,CO_Y,&yscale,&yadd);
		}
		
		for (bl = net[i]->Blocks; bl; bl = bl->nextplace)
			shift_place(bl);
		for (pl = net[i]->Places; pl; pl = pl->nextplace)
			shift_place(pl);
		for (tr = net[i]->Transitions; tr; tr = tr->nexttrans)
			shift_trans(tr);
		for (tr = net[i]->PhTransitions; tr; tr = tr->nexttrans)
			shift_trans(tr);
		for (pl = net[i]->Places; pl; pl = pl->nextplace)
		{
			for (ts = pl->preset; ts; ts = ts->nexttrans)
				shift_edge(ts);
			for (st = pl->postset; st; st = st->nexttrans)
				shift_edge(st);
			for (ts = pl->phpreset; ts; ts = ts->nexttrans)
				shift_edge(ts);
			for (st = pl->phpostset; st; st = st->nexttrans)
				shift_edge(st);
		}
		for (tx = net[i]->Texts; tx; tx = tx->nexttext)
			shift_coords(tx->co);

	}
}

/*****************************************************************************/
/* Main program.                                                             */
/*****************************************************************************/

int main(int argc, char **argv)
{
	#define MAX_NETS 64

	int	i, numbl, numpl, numtr, numph, numtx;
	int	numnets = 0;
	char    *outname = NULL;
	char	*modifier[MAX_NETS];
	clock_t time;
	bool outputflag = False;

	M_Net		net[MAX_NETS];
	M_Net		enet = NULL;
	M_Block 	*BlArray;
	M_Block		bl;
	M_Place 	*PlArray;
	M_Place 	pl;
	M_Transition 	*TrArray, *PhArray;
	M_Transition 	tr, ph;
	M_Text		*TxArray;
	M_Text		tx;

	time = clock();

	if (argc < 2) usage(argv[0]);

	for (i = 0; i < MAX_NETS; modifier[i++] = NULL);
	
	for (i = 1; i < argc; i++)

		if (!strcmp(argv[i],"-o"))
			outputflag = True;
		else if (outputflag)
		{
			outputflag = False;
			outname = argv[i];
		}
		else if (strchr(":*-+",*argv[i]))
		{
			if (!numnets || modifier[numnets-1])
			{
				fprintf(stderr,"modifier %s in wrong position;"
						" aborting.\n", argv[i]);
				exit(1);
			}
			modifier[numnets-1] = argv[i];
		}
		else
		{
			net[numnets] = read_HLnet(argv[i]);
			if (numnets && strcmp(net[0]->type,net[numnets]->type))
			{
				fprintf(stderr,"can't compose nets of different"
					" types; aborting.\n");
				exit(1);
			}
			numnets++;
		}
			
	if (!numnets)
	{
		fprintf(stderr,"no nets specified, nothing to do!\n");
		exit(1);
	}
	
	if (!outname)
	{
		fprintf(stderr, "no output file specified; aborting.\n");
		exit(1);
	}

	/* beware, evil hack */
	for (i = 0; i < numnets*2; i++)
	{
		enet = net[i/2];
		decode_phantom_refs(enet);
		tr = enet->Transitions;
		enet->Transitions = enet->PhTransitions;
		enet->PhTransitions = tr;
	}

	for (i = 0; i < numnets; i++) decode_block_refs(net[i]);
	if (numnets > 1) for (i = 0; i < numnets; i++) rename_nodes(net[i],i+1);

	translate_coords(net,modifier,numnets);

	numbl = numpl = numtr = numph = numtx = 0;
	for (i = 0; i < numnets; i++)
	{
		enet = net[i];
		numbl += enet->NumBlocks;
		numpl += enet->NumPlaces;
		numtr += enet->NumTransitions;
		numph += enet->NumPhTransitions;
		numtx += enet->NumTexts;
	}

	BlArray = MYmalloc(numbl * sizeof(M_Block));
	PlArray = MYmalloc(numpl * sizeof(M_Place));
	TrArray = MYmalloc(numtr * sizeof(M_Transition));
	PhArray = MYmalloc(numph * sizeof(M_Transition));
	TxArray = MYmalloc(numtx * sizeof(M_Text));

	while(numbl) BlArray[numbl--] = NULL;
	while(numpl) PlArray[numpl--] = NULL;
	while(numtr) TrArray[numtr--] = NULL;
	while(numph) PhArray[numph--] = NULL;
	while(numtx) TxArray[numtx--] = NULL;

	for (i = 0; i < numnets; i++)
	{
		enet = net[i];
		for (bl = enet->Blocks; bl; bl = bl->nextplace)
			(BlArray[numbl] = bl)->numname = ++numbl;
		for (pl = enet->Places; pl; pl = pl->nextplace)
			(PlArray[numpl] = pl)->numname = ++numpl;
		for (tr = enet->Transitions; tr; tr = tr->nexttrans)
			(TrArray[numtr] = tr)->numname = ++numtr;
		for (ph = enet->PhTransitions; ph; ph = ph->nexttrans)
			(PhArray[numph] = ph)->numname = ++numph;
		for (tx = enet->Texts; tx; tx = tx->nexttext)
			TxArray[numtx++] = tx;

	}
		
	write_HLnet(outname,
		    enet->type,
		    numbl, BlArray,
		    numpl, PlArray,
		    numtr, TrArray,
		    numph, PhArray,
		    numtx, TxArray);

/*	printf("Time needed: %f sec.\n\n",(float)(clock()-time)/CLOCKS_PER_SEC);
*/
	return 0;
}

/*****************************************************************************/
/* END OF FILE compose.c                                                     */
/*****************************************************************************/

