/*****************************************************************************/
/* main.c 								     */
/*****************************************************************************/
/* Produce fc2 file from hl-net and slist.   				     */
/*                                                            Stefan Schwoon */
/*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "mnet.h"
#include "readHL.h"
#include "read_slist.h"
#include "rw_fc2.h"
#include "mtupel.h"

/*****************************************************************************/

static void usage(char *myname)
{
  fprintf(stderr,
    "\n%s -- create .fc2 file from a list of s-invariants\n\n"

    "Usage: %s <infile> [FileOptions]\n\n"

    "     FileOptions:\n"
    "      -o  <filename>     specify an alternate .fc2 file\n"
    "      -sl <filename>     specify a file to read s-lists from\n\n"

    "All file names not explicitly stated on the command line will\n"
    "default to the input files' basenames plus appropriate extensions.\n\n"

    "Version 0.9.0 (12.12.1997) Stefan Schwoon\n\n", myname, myname);

  exit(1);
} /* usage() */


/*****************************************************************************/

void init_net(M_Net net)
{
	M_Place		pl;
	M_Transition	tr;
	int		i;
	MPTupel		pllist;

	/* isolierte Plaetze */
	for (pl = net->Places; pl; pl = pl->nextplace) if (pl->invnum < 0)
	{
		MPTT_AddValue(net->invplist,pllist = MPT_Create());
		create_unmarked(net,pllist);

		pl->invnum = net->NumInv++;
		pl->index = 1;
		MPTT_AddValue(pllist,pl);
	}

	/* Transitionen initialisieren */
	for (tr = net->Transitions; tr; tr = tr->nexttrans)
	{
		tr->index = MYmalloc(net->NumInv * sizeof(int));
		for (i = 0; i < net->NumInv; tr->index[i++] = -1);
	}

} /* init_net */

void insert_transitions(M_Net net)
{
	MPTupel		trlist;
	M_Place		pl;
	M_Transition	tr;
	M_Edge		st,ts;
	int		i,*p;

	net->invtlist = MPT_Create();
	for (i = 0; i < net->NumInv; i++)
		MPTT_AddValue(net->invtlist,MPT_Create());

	for (pl = net->Places; pl; pl = pl->nextplace)
	{
		if (!pl->index) continue;
		for (ts = pl->preset, st = pl->postset; ts || st;
			ts? (ts = ts->nexttrans) : (st = st->nexttrans))
		{
			tr = ts? ts->trans : st->trans;
			if (*(p = tr->index + pl->invnum) >= 0) continue;
			net->numbridges += (tr->invcounter++ == 0);
			trlist = MT_PlaceMPT(net->invtlist,pl->invnum);
			*p = trlist->count;
			MPTT_AddValue(trlist,tr);
		}
	}

} /* insert_transitions */

void make_inner_nets(M_Net net)
{
	MPTupel		pllist;
	MPTupel		trlist;
	M_Place		plfrom,plto;
	M_Transition	tr;
	M_Edge		st,ts;
	int		i,j;

	for (i = 0; i < net->NumInv; i++)
	{
		pllist = MT_PlaceMPT(net->invplist,i);
		trlist = MT_PlaceMPT(net->invtlist,i);
		for (j = 0; j < trlist->count; j++)
		{
			tr = MT_PlaceT(trlist,j);
			plfrom = plto = MT_PlaceP(pllist,0);

			for (st = tr->preset; st; st = st->nextplace)
				if (st->place->invnum == i) plfrom = st->place;
			
			for (ts = tr->postset; ts; ts = ts->nextplace)
				if (ts->place->invnum == i) plto = ts->place;
			
			MT_AddValue(plfrom->edges,tr->index[i]);
			MT_AddValue(plfrom->to,plto->index);
		}
	}

} /* make_inner_nets */

/*****************************************************************************/
/* Main program.                                                             */
/*****************************************************************************/

int main(int argc, char **argv)
{

	FILE	*fileptr;
	int	i;
	char    *llnet = NULL, *outfile, *sinv;
	char    **dptr;
	char	structname[80];
	char	tmpname[80], *idx;
	M_Net   net = NULL;
	clock_t time;

	dptr = &llnet;
	for (i=1; i<argc; i++)
		if (!strcmp(argv[i],"-o"))
			dptr = &outfile;
		else if (!strcmp(argv[i],"-sl"))
			dptr = &sinv;
		else
		{
			if (!dptr) usage(argv[0]);
			*dptr = argv[i];

			strcpy(tmpname,argv[i]);
			idx = strrchr(tmpname,'.');
			if (!idx) strcat(tmpname,".");
			idx = strrchr(tmpname,'.');

			if (dptr == &llnet)
			{
				*idx = '\0';
				strcpy(structname,tmpname);
				strcpy(idx,".fc2");
				outfile = MYstrdup(tmpname);
				strcpy(idx,".slist");
				sinv = MYstrdup(tmpname);
			}
			dptr = NULL;
		}

	if (!llnet) usage(argv[0]);

	time = clock();

	net = read_HLnet(llnet);
	if (strcmp(net->type,"PetriBox") && strcmp(net->type,"PTNet"))
	{
		fprintf(stderr,"Error: net in file %s is "
				"not of type PetriBox\n",llnet);
		exit(1);
	}

	if (!(fileptr = fopen(sinv,"r")) || fclose(fileptr))
	{
		fprintf(stderr,"Warning: %s not found\n",sinv);
		net->NumInv = 0;
		net->invplist = MPT_Create();
	}
	else
		read_slist(net,sinv);

	init_net(net);
	insert_transitions(net);
	make_inner_nets(net);

	write_fc2(net,outfile,structname);

/*	printf("Time needed: %f sec.\n\n",(float)(clock()-time)/CLOCKS_PER_SEC);
*/
	return 0;
}

/*****************************************************************************/
/* END OF FILE simplify.c                                                    */
/*****************************************************************************/
