/*****************************/
/* basic_lf.c                */
/* Burkhard Graves, 26.10.94 */
/*****************************/

#include "basic_lf.h"
#include "mca.h"
#include "mc.tab.h"

/********************/
/* Globale Variable */
/********************/

Set *mark;  /* fuer RemoveLE(), BFirst(), BFirstSingle() */

/**************************************************************************/
/* Wenn 'tree' die Wurzel eines Syntaxbaums einer Formel ist, die eine    */
/* Konjunktion von positiven und negativen Stellenaussagen ist, so werden */
/* die jeweiligen Stellenaussagen in 'gamma_plus' und 'gamma_minus'       */
/* abgelegt.                                                              */
/**************************************************************************/

static void calc_gammas(TreeNode *tree,Set *gamma_plus,Set *gamma_minus)
{
    assert(tree->info==AND
           || tree->info==PLACE
           || (tree->info==NOT && tree->right->info==PLACE));
    if (tree->info==AND) {
        calc_gammas(tree->left,gamma_plus,gamma_minus);
        calc_gammas(tree->right,gamma_plus,gamma_minus);
    } else if (tree->info==PLACE)
        SetInsertElement(gamma_plus,(unsigned)(tree->left));
    else if (tree->info==NOT)
        SetInsertElement(gamma_minus,(unsigned)(tree->right->left));
}

/**************************************************************************/
/* Entfernt aus der Konfiguration 'c' das Ereignis 'event'                */
/* sowie rekursiv alle Ereignisse, die diesem Ereignis kausal folgen.     */
/**************************************************************************/

void RemoveGE(Set *c,Event *event)
{
    Edge *c_edge,*e_edge;

    SetRemoveElement(c,event->numname);
    e_edge=event->postset;
    while (e_edge) {
        c_edge=e_edge->condition->postset;
        while (c_edge) {
            if (SetIsElement(c,c_edge->event->numname)) {
                RemoveGE(c,c_edge->event);
                break;
            }
            c_edge=c_edge->nextevent;
        }
        e_edge=e_edge->nextcondition;
    }
}

/**************************************************************************/
/* Sucht in der Konfiguration 'c' nach einem maximalen Ereignis           */
/* (bzgl. der Kausalrelation), das eine mit 'place' beschriftete          */
/* Bedingung im Vorbereich hat.                                           */
/* Existiert ein derartiges Ereignis, so wird es mitsamt allen            */
/* Ereignissen, die ihm kausal folgen aus der Konfiguration 'c' entfernt, */
/* und es wird '1' zurueckgeliefert.                                      */
/* Existiert ein derartiges Ereignis nicht, so wird '0' zurueckgeliefert. */
/**************************************************************************/

static int ge_in_pre_c(Set *c,unsigned place)
{
    unsigned event;
    Edge *edge;

    event=bprocess->num_events+1;
    while (--event) {
        event=SetGetElementPrev(c,event);
        if (!event)
            break;
        edge=(bprocess->events+event)->preset;
        while (edge) {
            if (edge->condition->pl_numname==place) {
                RemoveGE(c,bprocess->events+event);
                return 1;
            }
            edge=edge->nextcondition;
        }
    }
    return 0;
}

/**************************************************************************/
/* Sucht in der Konfiguration 'c' nach einem maximalen Ereignis           */
/* (bzgl. der Kausalrelation), das eine mit 'place' beschriftete          */
/* Bedingung im Nachbereich hat.                                          */
/* Existiert ein derartiges Ereignis, so wird es mitsamt allen            */
/* Ereignissen, die ihm kausal folgen aus der Konfiguration 'c' entfernt, */
/* und es wird '1' zurueckgeliefert.                                      */
/* Existiert ein derartiges Ereignis nicht, so wird '0' zurueckgeliefert. */
/**************************************************************************/

static int ge_in_post_c(Set *c,unsigned place)
{
    unsigned event;
    Edge *edge;

    event=bprocess->num_events+1;
    while (--event) {
        event=SetGetElementPrev(c,event);
        if (!event)
            break;
        edge=(bprocess->events+event)->postset;
        while (edge) {
            if (edge->condition->pl_numname==place) {
                RemoveGE(c,bprocess->events+event);
                return 1;
            }
            edge=edge->nextcondition;
        }
    }
    return 0;
}

/**************************************************************************/
/* Liefert die Konfiguration, die in Last({},pl,{c2}) enthalten ist,      */
/* oder 'NULL', falls Last({},pl,{c2})={} ist.                            */
/* Bemerkung: Optimierte Version von BLast(), fuer den Fall, dass 'gamma' */
/*            nur eine positive Stellenaussage 'pl' enthaelt.             */
/**************************************************************************/

Set *BLastSingle(Set *c2,unsigned pl)
{
    Set *c,*mark;

    mark=Mark(c=SetDuplicate(c2));
    if (!SetIsElement(mark,pl))
        if (!ge_in_pre_c(c,pl)) {
            SetDelete(c);
            c=NULL;
        }
    SetDelete(mark);
    return c;
}

/**************************************************************************/
/* Liefert die Konfiguration, die in Last({},gamma,{c2}) enthalten ist,   */
/* oder 'NULL', falls Last({},gamma,{c2})={} ist.                         */
/* Bemerkung: 'gamma_plus' bzw. 'gamma_minus' enthaelt die positiven bzw. */
/*            negativen Stellenaussagen von 'gamma'.                      */
/**************************************************************************/

Set *BLast(Set *c2,Set *gamma_plus,Set *gamma_minus)
{
    Set *c,*cut,*mark,*help;

    c=SetDuplicate(c2);
    cut=SetCreate();
    mark=SetCreate();
    help=SetCreate();
    for (;;) {
        CutInSet(cut,c);
        LCondInSet(mark,cut);

#ifdef OUTPUT
    fprintf(yyout,"\n       c: ");
    SetDisplay(yyout,c);
    fprintf(yyout,"\n     cut: ");
    SetDisplay(yyout,cut);
    fprintf(yyout,"\n    mark: ");
    SetDisplay(yyout,mark);
#endif

        SetCopy(help,gamma_plus);
        SetDifference(help,mark);

#ifdef OUTPUT
    fprintf(yyout,"\n   help1: ");
    SetDisplay(yyout,help);
#endif

        if (!SetIsEmpty(help)) {
            if (!ge_in_pre_c(c,SetGetElementNext(help,1))) {
                SetDelete(c);
                c=NULL;
                break;
            }
        } else {
            SetCopy(help,gamma_minus);
            SetIntersection(help,mark);

#ifdef OUTPUT
    fprintf(yyout,"\n   help2: ");
    SetDisplay(yyout,help);
#endif

            if (!SetIsEmpty(help)) {
                if (!ge_in_post_c(c,SetGetElementNext(help,1))) {
                    SetDelete(c);
                    c=NULL;
                    break;
                }
            } else
                break;
        }
    }
    SetDelete(help);
    SetDelete(mark);
    SetDelete(cut);
    return c;
}

/**************************************************************************/
/* Liefert Last(C1,gamma,C2).                                             */
/* Bemerkung: 'gamma' ist eine Konjunktion von positiven und negativen    */
/*            Stellenaussagen, enthaelt also keine temporalen Operatoren. */
/**************************************************************************/

SList *BasicLast(SList *C1,TreeNode *gamma,SList *C2)
{
    Set *gamma_plus,*gamma_minus,*c;
    SList *slist;

    slist=NULL;
    if (C2) {
        gamma_plus=SetCreate();
        gamma_minus=SetCreate();
        calc_gammas(gamma,gamma_plus,gamma_minus);

#ifdef OUTPUT
    fprintf(yyout,"\n  gamma+: ");
    SetDisplay(yyout,gamma_plus);
    fprintf(yyout,"\n  gamma-: ");
    SetDisplay(yyout,gamma_minus);
#endif

        while ((C2=C2->next)!=NULL) {
            if ((c=BLast(C2->set,gamma_plus,gamma_minus))!=NULL) {
                if (!LowerEqual(c,C1)) {

#ifdef OUTPUT
    fprintf(yyout,"\n-b-ZwErg: ");
    SetDisplay(yyout,c);
#endif

                    if (slist)
                        SListMaxInsert(slist,c);
                    else
                        slist=SListCreate(c);
                } else {
                    SetDelete(c);
                }
            }
        }
        SetDelete(gamma_minus);
        SetDelete(gamma_plus);
    }

#ifdef OUTPUT
    if (slist) {
        fprintf(yyout,"\n--b--Erg:\n");
        SListDisplay(yyout,slist);
    }
#endif

    return slist;
}

/**************************************************************************/
/* Entfernt aus der Konfiguration 'c' die Geschichte des Ereignisses      */
/* 'event'. Die durch die Konfiguration 'c' determinierte Markierung      */
/* 'mark' wird entsprechend aktualisiert.                                 */
/**************************************************************************/

void RemoveLE(Set *c,Event *event)
{
    Edge *c_edge,*e_edge;

    SetRemoveElement(c,event->numname);
    e_edge=event->preset;
    while (e_edge) {
        c_edge=e_edge->condition->preset;
        if (c_edge && SetIsElement(c,c_edge->event->numname))
            RemoveLE(c,c_edge->event);
        SetRemoveElement(mark,e_edge->condition->pl_numname);
        e_edge=e_edge->nextcondition;
    }
    e_edge=event->postset;
    while (e_edge) {
        SetInsertElement(mark,e_edge->condition->pl_numname);
        e_edge=e_edge->nextcondition;
    }
}

/**************************************************************************/
/* Sucht in der Konfiguration 'c' nach einem maximalen Ereignis           */
/* (bzgl. der Kausalrelation), das eine mit 'place' beschriftete          */
/* Bedingung im Vorbereich hat.                                           */
/* Existiert ein derartiges Ereignis, so wird seine Geschichte aus der    */
/* Konfiguration 'c' entfernt, und es wird '1' zurueckgeliefert.          */
/* Existiert ein derartiges Ereignis nicht, so wird '0' zurueckgeliefert. */
/**************************************************************************/

static int le_in_pre_c(Set *c,unsigned place)
{
    unsigned event;
    Edge *edge;

    event=0;
    while ((event=SetGetElementNext(c,++event))!=0) {
        edge=(bprocess->events+event)->preset;
        while (edge) {
            if (edge->condition->pl_numname==place) {
                RemoveLE(c,bprocess->events+event);
                return 1;
            }
            edge=edge->nextcondition;
        }
    }
    return 0;
}

/**************************************************************************/
/* Sucht in der Konfiguration 'c' nach einem maximalen Ereignis           */
/* (bzgl. der Kausalrelation), das eine mit 'place' beschriftete          */
/* Bedingung im Nachbereich hat.                                          */
/* Existiert ein derartiges Ereignis, so wird seine Geschichte aus der    */
/* Konfiguration 'c' entfernt, und es wird '1' zurueckgeliefert.          */
/* Existiert ein derartiges Ereignis nicht, so wird '0' zurueckgeliefert. */
/**************************************************************************/

static int le_in_post_c(Set *c,unsigned place)
{
    unsigned event;
    Edge *edge;

    event=0;
    while ((event=SetGetElementNext(c,++event))!=0) {
        edge=(bprocess->events+event)->postset;
        while (edge) {
            if (edge->condition->pl_numname==place) {
                RemoveLE(c,bprocess->events+event);
                return 1;
            }
            edge=edge->nextcondition;
        }
    }
    return 0;
}

/**************************************************************************/
/* Liefert die Konfiguration, die in First({},pl,{c2}) enthalten ist,     */
/* oder 'NULL', falls First({},pl,{c2})={} ist.                           */
/* Bemerkung: Optimierte Version von BFirst(), fuer den Fall, dass 'gamma'*/
/*            nur eine positive Stellenaussage 'pl' enthaelt.             */
/**************************************************************************/

Set *BFirstSingle(Set *c2,unsigned pl)
{
    Set *c;

    if (SetIsElement(initial_marking,pl))
        return SetCreate();
    c=SetDuplicate(c2);
    mark=SetDuplicate(initial_marking);
    if (!le_in_post_c(c,pl)) {
        SetDelete(c);
        c=NULL;
    } else
        SetEOR(c,c2);
    SetDelete(mark);
    return c;
}

/**************************************************************************/
/* Liefert die Konfiguration, die in First({},gamma,{c2}) enthalten ist,  */
/* oder 'NULL', falls First({},gamma,{c2})={} ist.                        */
/* Bemerkung: 'gamma_plus' bzw. 'gamma_minus' enthaelt die positiven bzw. */
/*            negativen Stellenaussagen von 'gamma'.                      */
/**************************************************************************/

Set *BFirst(Set *c2,Set *gamma_plus,Set *gamma_minus)
{
    Set *c,*help;

    c=SetDuplicate(c2);
    mark=SetDuplicate(initial_marking);
    help=SetCreate();
    for (;;) {

#ifdef OUTPUT
    fprintf(yyout,"\n       c: ");
    SetDisplay(yyout,c);
    fprintf(yyout,"\n    mark: ");
    SetDisplay(yyout,mark);
#endif

        SetCopy(help,gamma_plus);
        SetDifference(help,mark);

#ifdef OUTPUT
    fprintf(yyout,"\n   help1: ");
    SetDisplay(yyout,help);
#endif

        if (!SetIsEmpty(help)) {
            if (!le_in_post_c(c,SetGetElementNext(help,1))) {
                SetDelete(c);
                c=NULL;
                break;
            }
        } else {
            SetCopy(help,gamma_minus);
            SetIntersection(help,mark);

#ifdef OUTPUT
    fprintf(yyout,"\n   help2: ");
    SetDisplay(yyout,help);
#endif

            if (!SetIsEmpty(help)) {
                if (!le_in_pre_c(c,SetGetElementNext(help,1))) {
                    SetDelete(c);
                    c=NULL;
                    break;
                }
            } else
                break;
        }
    }
    if (c) {
        SetCopy(help,c2);
        SetDifference(help,c);
        SetDelete(c);
        SetDelete(mark);
        return help;
    } else {
        SetDelete(help);
        SetDelete(mark);
        return c;
    }
}

/**************************************************************************/
/* Liefert First(C1,gamma,C2).                                            */
/* Bemerkung: 'gamma' ist eine Konjunktion von positiven und negativen    */
/*            Stellenaussagen, enthaelt also keine temporalen Operatoren. */
/**************************************************************************/

SList *BasicFirst(SList *C1,TreeNode *gamma,SList *C2)
{
    Set *gamma_plus,*gamma_minus,*c;
    SList *slist;

    slist=NULL;
    if (C2) {
        gamma_plus=SetCreate();
        gamma_minus=SetCreate();
        calc_gammas(gamma,gamma_plus,gamma_minus);

#ifdef OUTPUT
    fprintf(yyout,"\n  gamma+: ");
    SetDisplay(yyout,gamma_plus);
    fprintf(yyout,"\n  gamma-: ");
    SetDisplay(yyout,gamma_minus);
#endif

        while ((C2=C2->next)!=NULL) {
            if ((c=BFirst(C2->set,gamma_plus,gamma_minus))!=NULL) {
                if (!LowerEqual(c,C1)) {

#ifdef OUTPUT
    fprintf(yyout,"\n-b-ZwErg: ");
    SetDisplay(yyout,c);
#endif

                    if (slist)
                        SListMinInsert(slist,c);
                    else
                        slist=SListCreate(c);
                } else {
                    SetDelete(c);
                }
            }
        }
        SetDelete(gamma_minus);
        SetDelete(gamma_plus);
    }

#ifdef OUTPUT
    if (slist) {
        fprintf(yyout,"\n--b--Erg:\n");
        SListDisplay(yyout,slist);
    }
#endif

    return slist;
}

/**************************************************************************/
/* (aehnlich 'le_in_post_c()')                                            */
/* Sucht in der Konfiguration 'c' nach einem minimalen NON-CUTOFF         */
/* Ereignis (bzgl. der Kausalrelation), das eine mit 'place' beschriftete */
/* Bedingung im Nachbereich hat.                                          */
/* Existiert ein derartiges Ereignis, so wird seine Geschichte aus der    */
/* Konfiguration 'c' entfernt, und es wird '1' zurueckgeliefert.          */
/* Existiert ein derartiges Ereignis nicht, so wird '0' zurueckgeliefert. */
/**************************************************************************/

static int sub_search_for_corr_conf(Set *c,unsigned place)
{
    unsigned event;
    Edge *edge;

    event=0;
    while ((event=SetGetElementNext(c,++event))!=0
           && !SetIsElement(cutoffs,event)) {
        edge=(bprocess->events+event)->postset;
        while (edge) {
            if (edge->condition->pl_numname==place) {
                RemoveLE(c,bprocess->events+event);
                return 1;
            }
            edge=edge->nextcondition;
        }
    }
    return 0;
}

/**************************************************************************/
/* (aehnlich 'BFirst()')                                                  */
/* Liefert eine keine cutoff-Events enthaltende Konfiguration, die eine   */
/* 'gamma_plus' entsprechende Markierung beschreibt.                      */
/**************************************************************************/

Set *SearchForCorrConf(Set *gamma_plus)
{
    Set *c,*help;
    SList *slist;

    c=NULL;
    mark=SetCreate();
    help=SetCreate();

    slist=max_configs;
    while ((slist=slist->next)!=NULL) {

        c=SetDuplicate(slist->set);
        SetCopy(mark,initial_marking);

        for (;;) {
            SetCopy(help,gamma_plus);
            SetDifference(help,mark);

            if (SetIsEmpty(help))
                break;
            if (!sub_search_for_corr_conf(c,SetGetElementNext(help,1))) {
                SetDelete(c);
                c=NULL;
                break;
            }
        }
        if (c)
            break;

    }
    if (c) {
        SetCopy(help,slist->set);
        SetDifference(help,c);
        SetSwap(c,help);
    }
    SetDelete(help);
    SetDelete(mark);
    SetDelete(gamma_plus);
    return c;
}

/**************************/
/* END OF FILE basic_lf.c */
/**************************/
