/*
This version of PEP contains new functions that read high level
or low levels nets from files. The source code to these functions
can be found in the "net_routines" directory in the PEP source tree.
In general, if you have been using routines from the "neunetze"
directory, we suggest that you use the new functions instead.

In particular, the net_routines directory contains the following files:

	neunetze.c	A lot of useful functions. Nothing has changed here.

	convertPEP.c	Many routines for creating and manipulating nets,
			writing nets into various formats and others.
			It also contains a call to the aforementioned
			new functions.

	readPEP.c	Reads high level or low level nets. The use of 
			its main function, read_PEP_net, is described
			in detail below.

	readlib.c	Several useful functions
	memlib.c	used by readPEP.c

The net_routines/g++ directory contains versions of these source files
compiled for g++.


How to use read_PEP_net:

Suppose you need to read a high/low level net stored in a file using
the PEP net format. There are now two ways to do this.

- You can use read_PEP out of convertPEP.c. read_PEP stores the net
  in data structures described in neunetze.h. This routine should
  be sufficient for most needs and, in fact, many of the tools in
  the PEP package use it. If you have been using this function
  before, you can continue to use it.

  There might be two disadvantages to this function, however.
  First, the data structures from neunetze.h might not suit your needs.
  Second, read_PEP returns only a select group of fields in each record
  and discards the rest. In particular, information on blocks and phantom
  transitions gets lost.

  To use read_PEP, you need to include convertPEP.o, neunetze.o,
  readPEP.o, readlib.o and memlib.o into the list of object files linked
  into your executable.

- Alternatively, you can use read_PEP_net from readPEP.c. This routine
  is more flexible, allows you to use your own data structures to store
  net components in, but requires more work on your side.
  read_PEP_net expects you to specify which part of the information
  contained in a file you are interested in and where it should be
  stored. For each type of component in a net (places, blocks, transitions,
  arcs etc.) you may specify a "hook" function that is called by read_PEP_net
  each time a component of that type has been read.

  Consider the following (very basic) example: A net (high level or low
  level) is read from a file, and the names, numbers and coordinates of all
  the places are extracted and printed to standard output. Additionally,
  all the rest of the information is printed _as it appears in the file_.
  This feature is useful if you need to copy sections of a net without
  creating an extra data structure.

  To compile this example, type
	"gcc example.c readPEP.o readlib.o memlib.o -o example"
  (or something equivalent) in your shell.

  Other, more complex (and useful) examples can be found in the sources
  to hl2ll, simplify, slist and convertPEP.c (which is in fact an interface
  to this function).

  To use read_PEP, you need to include readPEP.o, readlib.o and memlib.o
  into the list of object files linked into your executable.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "readPEP.h"
#include "readlib.h"

int		autonumbering;
int		placecount;

char		*name, *rest;
int		ident;
t_coords	*xy;

int insert_place();

/* In place_dest, we tell the reading function which fields
   we want to read and where information is to be stored.
   E.g., the first lines means "if you encounter a textual
   identifier for a place, read it in and place a pointer to
   the string in `name'". Information from fields not covered
   in this array is collected in a separate string (see below).
*/

t_dest place_dest[] =
	{ { '\'',&name },	/* Textual identifier	*/
	  { '"', &name },	/*     "		*/
	  { '@', &xy },		/* Coordinates		*/
	  { '0', &ident },	/* numeric identifier 	*/
	  {  0 ,  0 } };	/* terminate the list   */
  

/* Here, we tell read_PEP_net which sections to read. All the sections
   not mentioned will be ignored. For each section, you must give:
   - the section identifier ("PL") as it appears in the file
   - the function to call each time a component in this section
     has been read (this function must be of type int)
   - a pointer of type char** to specify where to place a pointer
     to additional information (note that if you specify NULL here,
     any additional information will be discarded)
   - a pointer to the accompanying t_dest struct (place_dest in
     this example)
*/

t_blockdest netdest[] =
	{ { "PL", insert_place, &rest, place_dest },
	  { NULL, NULL, NULL, NULL } };



int main(int argc, char** argv)
{
	placecount = 0;		/* inititialize some variables ... */
	autonumbering = 1;

	if (argc < 2)
	{
		fprintf(stderr, "usage: %s <pepfilename>\n", argv[0]);
		exit(1);
	}

	read_PEP_net(argv[1],netdest); 		/* ... and go! */

	/* filetype contains the type of net (the second line in a PEP file) */
	printf("Net was of type %s.\n", filetype);
	exit(0);
}


/* PEPerror is a pointer to an error handling function. By default, it points
   to PEPerror_ in readPEP.c. You might want to change this, though. */

/* This function gets called each time a place has been read from the file.
   name, rest, ident and xy point to (or contain) appropriate information
   as declared above. Note that you do not need to "strdup" strings because
   read_PEP_net already does it for you.

   You have access to char* blocktype here in case this function is intended
   to handle objects from different sections and you need to distinguish
   between these (might be useful for handling both transitions and phantom
   transitions).
*/

int insert_place()
{
	static char buffer[10];

	placecount++;
	if (ident && ident != placecount) autonumbering = 0;
	if (!ident && autonumbering) ident = placecount;
	if (!ident) PEPerror("missing place identifier");

	if (!name)
	{
		sprintf(buffer, "PL%d", ident);
		name = strdup(buffer);
	}

	printf("%d %s",ident,name);
	if (xy) printf(" (%d@%d)", xy->x, xy->y);
	if (rest) printf(" %s", rest);
	printf("\n");

	/* Return any non-zero value here to indicate an error condition
	   and to abort the read.					 */
	return 0;

}

