/*****************************************************************************/
/* readlib.c								     */
/*****************************************************************************/
/*				    Stefan Roemer,  06.02.1995 - 06.02.1995  */
/*				    Stefan Schwoon, April 1997		     */
/*****************************************************************************/
/* Some auxiliary functions for reading text files.                          */
/*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include "readlib.h"
#include "readPEP.h"

/*****************************************************************************/
/* Global variables.                                                         */
/*****************************************************************************/

char *NNinput_file; /* Name of file being processed.        */
int   NNinput_line; /* Current line of input.               */

char *sbuf = NULL;  /* Buffer for storing strings in.       */
int  sballoc = 0;   /* Amount of memory allocated for sbuf. */

/*****************************************************************************/
/* Abort program due to faulty input file. You can set the PEPerror variable */
/* to a different function if you want errors to be treated differently,     */
/* PEPerror_ ist just the default behaviour.                                 */
/*****************************************************************************/
 
void PEPerror_(char *msg)
{
        fprintf(stderr, "ERROR: syntax error in file '%s', line %d: %s.\n",
                        NNinput_file, NNinput_line, msg);
        exit (10);
} /* PEPerror_() */
 
void (*PEPerror)(char*) = PEPerror_;
 
char errbuf[1024];

/*****************************************************************************/
/* ReadCharComment                                                           */
/* Get next useful character in file (whitespace and comments are ignored.   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

char ReadCharComment( FILE * file )

#else

char ReadCharComment( file )

  FILE * file;

#endif

{
	register char ch = '\0';

	do {
		while (!feof(file) && strchr(" \t\b", ch = getc(file)))
		  skip;
		if (ch == '%')
		{
			while (!feof(file) && ((ch = getc(file)) != '\n'))
			  skip;
			if (ch == '\n')
			{
				NNinput_line++;
				ch = '\0';
			} /* if */
		} /* if */
	} while (!ch);

	return ch;

} /* ReadCharComment() */

/*****************************************************************************/
/* ReadCmdToken                                                              */
/* Read the next string (an alphanumeric word) from a text file; leading     */
/* whitespace and comments are ignored. The word is stored in the global     */
/* variable sbuf.                                                            */
/*****************************************************************************/

void free_sbuf()
{
	free(sbuf);
}

char* ReadCmdToken(FILE *file)
{
	register int  len = 0;
	char *string = sbuf;

	if (!string)
	{
	  atexit(free_sbuf);
	  sbuf = string = (char*)MYmalloc(sballoc = 512);
	}

	if (!isalnum(*string++ = ReadCharComment(file)))
		Error((errbuf,"ReadCmdToken: alphanumerical string expected"));

	len++;
	while (isalnum(*string = getc(file)) || *string == '_')
	{
		string++;
		len++;
		if (len >= sballoc)
		{
			sbuf = (char*)MYrealloc(sbuf,sballoc += 512);
			string = sbuf + len;
		}
	} /* while */

	ungetc(*string, file);
	*string = '\0';
  return sbuf;

} /* ReadCmdToken() */

/*****************************************************************************/
/* ReadNewline                                                               */
/* Find the next newline in a file.					     */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void ReadNewline( FILE * file )

#else

void ReadNewline( file )

  FILE          * file;

#endif

{
	while (!feof(file) && getc(file) != '\n')
	  skip;
	NNinput_line++;

} /* ReadNewline() */

/*****************************************************************************/
/* ReadWhiteSpace					 		     */
/* Return the next char in a file, ignoring whitespace.                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

char ReadWhiteSpace(FILE *file)

#else

char ReadWhiteSpace( file )

  FILE * file;

#endif

{
	register int ch = '\0';

	while (!feof(file) && isspace(ch = getc(file)))
		if (ch == '\n')
			NNinput_line++;

	return ( ch );

} /* ReadWhiteSpace() */

/*****************************************************************************/
/* ReadNumber                                                                */
/* Treat the following chars in a file as a number and returns its value.    */
/* Leading whitespace is ignored. An error is issued if no digit is found.   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

int ReadNumber( FILE * file )

#else

int ReadNumber( file )

  FILE * file;

#endif

{
	register int number = 0,
	             digit,
	             vorz = 1;

	if ((digit = ReadWhiteSpace(file)) == '-')
	{
		vorz = -1;
		digit = ReadWhiteSpace(file);
	}

	if (isdigit(digit))
		number = digit - '0';
	else
		Error((errbuf,"ReadNumber: digit expected"));

	while (isdigit(digit = getc(file)))
		number = number * 10 + digit - '0';

	ungetc(digit, file);
	return(vorz*number);

} /* ReadNumber() */

/*****************************************************************************/
/* ReadDouble                                                                */
/* Treat the following chars in a file as a double and return its value.     */
/* Leading whitespace is ignored. An error is issued if no digit is found.   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

double ReadDouble( FILE * file )

#else

double ReadDouble( file )

  FILE * file;

#endif

{
	double value=0;

	if (!fscanf(file, "%lf", &value))
	{
	  // Empty time label!
	  return 0;
	}
        else
	  return value;

} /* ReadDouble() */

/*****************************************************************************/
/* RL_ReadEnclString                                                         */
/* Read a string enclosed by (single or double) quotes. Leading whitespace   */
/* is ignored. The string is stored in the global variable sbuf. The string  */
/* in sbuf does not contain the quotes, only the chars enclosed by them.     */
/*****************************************************************************/

char* RL_ReadEnclString(FILE *file)
{
	register char *str = sbuf;
	register int  len = 0;
	char          delimiter;

	if (!str) sbuf = str = (char*)MYmalloc(sballoc = 512);

	if ((delimiter = ReadCharComment(file)) != '\'' && delimiter != '"')
		Error((errbuf,"RL_ReadEnclString: string leading ' or \" expected"));

	while (!feof(file) && (*str = getc(file)) != delimiter)
	{
		str++;
		len++;
		if (len >= sballoc)
		{
			sbuf = (char*)MYrealloc(sbuf,sballoc += 512);
			str = sbuf + len;
		}
	} /* while */

	if (*str != delimiter)
		Error((errbuf,"RL_ReadEnclString: closing %c is missing",delimiter));
	*str = '\0';
  return sbuf;

} /* RL_ReadEnclString() */

/*****************************************************************************/
/* Einlesen einer Zeichenkette aus einer Textdatei; fuehrende Leerzeichen    */
/* werden ueberlesen und nicht dem String zugeordnet.                        */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

unsigned char ReadAlphaName( FILE * file, char * string, int maxlen )

#else

unsigned char ReadAlphaName( file, string, maxlen )

  FILE * file;
  char * string;
  int  maxlen;

#endif

  {

    register int ch;
    register int  len = 0;

    if ( ( *string = ReadWhiteSpace( file ) ) == '^' )
      {
        len++;
        *++string = ReadWhiteSpace( file );
      } /* if */
    if ( ! isalnum( *( string++ ) ) )
      {
        PEPerror( "ReadAlphaName: alphanumerical string expected" );
      } /* if */
    len++;
    while( ( isalnum( *string = getc( file ) ) || ( *string == '_' )
            || ( *string == '?' ) || ( *string == '!' ) ) && ( len < maxlen ) )
      {
        string++;
        len++;
      } /* while */
    ungetc( *string, file );
    if ( len == maxlen )
      {
        while ( isalnum( ch = getc( file ) ) )
          skip;
        ungetc( ch, file );
      } /* if */
    *string = '\0';
    
    return ( len );

  } /* ReadAlphaName() */

/*****************************************************************************/
/* ReadCoordinates                                                           */
/* Read a pair of coordinates (two numbers separated by @). Whitespace       */
/* before or between the numbers and the @ is ignored.                       */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void ReadCoordinates( FILE * file, int * x, int * y )

#else

void ReadCoordinates( file, x, y )

  FILE * file;
  int  * x;
  int  * y;

#endif

{
	*x=ReadNumber(file);
	if (ReadWhiteSpace(file) != '@')
		Error((errbuf, "ReadCoordinates: `@' expected"));
	*y=ReadNumber(file);
}

/*****************************************************************************/
/* TestSymbol								     */
/* Checks, if the next char in file is symbol. Leading whitespace is ignored.*/
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool TestSymbol( FILE * file, char symbol )

#else

bool TestSymbol( file, symbol )

  FILE * file;
  char symbol;

#endif

  {

    register char ch;

    if ( ( ch = ReadWhiteSpace( file ) ) == symbol )
      {
        return ( True );
      } /* if */
    else
      {
        ungetc( ch, file );
        return ( False );
      } /* else */

  } /* TestSymbol() */

bool TestDigit( FILE * file )
  {
 
    register int ch;
 
    ch = ReadWhiteSpace( file );
    ungetc( ch, file );
 
    if ( isdigit ( ch ) )
      {
        return ( True );
      } /* if */
    else
      {
        return ( False );
      } /* else */
 
  } /* TestDigit() */
 

/*****************************************************************************/
/* Checks, if next char in file is symbol. If not an error is raised.        */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void ReadSymbol( FILE * file, char symbol )

#else

void ReadSymbol( file, symbol )

  FILE * file;
  char symbol;

#endif

  {

    register char ch;

    if ( ( ch = ReadWhiteSpace( file ) ) != symbol )
      {
        Error((errbuf, "ReadSymbol(): '%c' expected here, but '%c' found",
	       symbol, ch ));
      } /* if */

  } /* ReadSymbol() */


/*****************************************************************************/
/* Read a name from file.                                                    */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

unsigned char ReadName( FILE * file, char * string, int maxlen )

#else

unsigned char ReadName( file, string, maxlen )

  FILE * file;
  char * string;
  int  maxlen;

#endif

  {

    register unsigned char len = 1;

    *string = ReadWhiteSpace( file );
    if (feof(file))
    {
      *string='\0';
      return 0;
    }
    while(    ! isspace( *++string = getc( file ) )
           && ( *string != ',' ) && ( (int) ++len < maxlen ) )
      skip;
    if ( len == maxlen )
      {
        while( ! isspace( getc( file ) ) )
          skip;
      } /* if */
    if ( *string == ',' )
      {
        ungetc( *string, file );
      } /* if */
    *string = '\0';
    
    return ( len );

  } /* ReadName() */
