/*-------------------------------------------------------------------------*/
/*--                                                                     --*/
/*-- READ_UNF.C                                                          --*/
/*--                                                                     --*/
/*-- Routinen zum Einlesen einer Entfaltung aus einer ASCII-Datei.       --*/
/*-- Basiert auf READ_UNF.C von S. Roemer.                               --*/
/*-- Angepasst und erweitert von Johannes Jaeger - Sommer 1994           --*/
/*--                                                                     --*/
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "rw_prefix.h"
#include "readlib.h"

/* einige Fehlermeldungen */
#define E_READLABEL      "ReadPlaceLabel(): Unknown place label"
#define E_RDUNF_NOCOND   "readUNF(): Net has no conditions"

/* Zeichendefinitionen etc. */
#define SPACE_CHAR ' '
#define TAB_CHAR   '\t'
#define MINUS_CHAR '-'
#define ASCII_NULL '\0'
#define NULL_DIGIT '0'
#define NUMBER_BASE 10
#define COMMA_CHAR ','
#define DOT_CHAR   '.'
#define NEW_LINE '\n'

/* Laenge der lokalen Variablen char name[] in readUNF() */
#define NAME_LEN 300

/* eine neu ins Netz eingefuegte Stelle enthaelt keine Markierung */
#define INITIAL_MARK 0

/* eingefuegte Kanten erhalten Gewicht 1 */
#define TP_WEIGHT 1
#define PT_WEIGHT 1


/*-------------------------------------------------------------------------*/
/*-- wegen einer fehlenden allgemeinen Spezifikation fuer das Format der --*/
/*-- ASCII-Datei, die eine Entfaltung enthaelt, muessen die folgenden    --*/
/*-- '#defines' bei einer Layout-Aenderung ggf. angepasst werden und das --*/
/*-- Modul neu compiliert werden                                         --*/

/* die ersten 7 Zeilen der zu lesenden Datei interessieren nicht */
#define HEAD_LINES 7

/* Zeilen vor den Transitions-Eintraegen */
#define LINES_BEFORE_TRANS 5

/* Zeilen vor den Stellen-Eintraegen */
#define LINES_BEFORE_PLACE 3

/* Zeichen direkt hinter einem numname-Eintrag einer Transition/Stelle */
#define NUMNAME_DELIMITER '.'

/* der Pfeil '-->', mit dem postsets und presets gekennzeichnet sind */
#define FIRST_ARROW_CHAR '-'
#define SECOND_ARROW_CHAR '-'
#define THIRD_ARROW_CHAR '>'

/* Klammern um die component-Eintraege einer Stelle */
#define LEFT_COMP_BRACKET '<'
#define RIGHT_COMP_BRACKET '>'

/* Klammern um die label-Eintraege einer Transition/Stelle */
#define LEFT_LABEL_BRACKET '('
#define RIGHT_LABEL_BRACKET ')'

/* mehrere postset-Nummern fuer eine Stelle werden durch Komma getrennt */
#define POSTSET_LIST_DELIMITER ','

/* moegliche Label-Werte */
#define NULL_LABEL '0'
#define E_LABEL    'e'
#define X_LABEL    'x'
#define I_LABEL    'i'
#define D_LABEL    'd'

/* das erste Wort, das der Anzahl von Cut-Off-Punkten folgen muss */
#define CUTOFF_TEXT "cutoff"

/* Text im Eintrag fuer max.Konfigurationen */
#define MAXCONFIG_TEXT "maximal"

/*-------------------------------------------------------------------------*/
/*-- Testen, ob das naechste relevante Zeichen numerisch ist             --*/
/*-- Parameter :                                                         --*/
/*--   FILE *file : die Textdatei, aus der gelesen wird                  --*/
/*--                                                                     --*/
/*-- Rueckgabe : die gelesene Zahl vom Typ Cardinal32                    --*/

#if defined(__STDC__) || defined(__cplusplus)

static bool TestNum( FILE * file )

#else

static bool TestNum( file )
FILE * file;

#endif
{
  int digit;

  /* Leerzeichen und Minus-Zeichen ueberlesen */
  if ( ( digit = ReadWhiteSpace( file ) ) == MINUS_CHAR )
    digit = ReadWhiteSpace( file );

  /* gelesenes Zeichen eine Zahl? */
  if ( isdigit( digit ) )
  {
    /* ja: zurueck damit und True zurueckgeben */
    ungetc( digit, file );
    return ( True );
  }
  else
    /* nein: False zurueckgeben */
    return ( False );

} /* TestNum() */

/*-------------------------------------------------------------------------*/
/*-- in der Textdatei bis zum naechsten Komma bzw. Punkt einschliesslich --*/
/*-- weiterlesen                                                         --*/
/*-- Parameter :                                                         --*/
/*--   FILE *file : die Textdatei, aus der gelesen wird                  --*/

#if defined(__STDC__) || defined(__cplusplus)

static char ReadToCommaDot( FILE * file )

#else

static char ReadToCommaDot( file )
FILE * file;

#endif
{
  char ch, ret_char;

  /* alles, was kein Komma, Punkt oder EOF ist, ueberlesen */
  while ( ( ( ch = getc( file ) ) != EOF ) &&
          ( ch != DOT_CHAR ) && ( ch != COMMA_CHAR ) 
        )
    skip;

  ret_char = ch;

  /* falls das zuletzt gelesene Zeichen kein NEW_LINE ist, */
  /* zurueck damit                                         */
  if ( ( ch = getc( file ) ) != NEW_LINE )
    ungetc( ch, file );

  return (ret_char);

} /* ReadToCommaDot() */

/*-------------------------------------------------------------------------*/
/*-- Lesen einer Stellen-Beschriftung aus einer Textdatei.               --*/
/*-- wird lediglich zum Weiterlesen in der Textdatei verwendet, Inhalt   --*/
/*-- interessiert hier nicht                                             --*/

#if defined(__STDC__) || defined(__cplusplus)

static PLabel ReadPlaceLabel( FILE * file )

#else

static PLabel ReadPlaceLabel( file )
FILE     * file;

#endif
{
  PLabel label=EMPTY;

  ReadSymbol( file, LEFT_LABEL_BRACKET );
  switch ( getc( file ) )
  {
    case NULL_LABEL : label = EMPTY;
                      break;
    case E_LABEL    : label = ENTRY;
                      break;
    case X_LABEL    : label = EXIT;
                      break;
    case I_LABEL    : label = INTERRUPT;
                      break;
    case D_LABEL    : label = DELETE;
                      break;
    default         : NNError( E_READLABEL );
                      break;
  } /* switch */
  ReadSymbol( file, RIGHT_LABEL_BRACKET );

  return ( label );

} /* ReadPlaceLabel() */

/*-------------------------------------------------------------------------*/
/*-- Interface-Routine zum Lesen einer Entfaltung aus einer ASCII-Datei. --*/
/*-- Parameter:                                                          --*/
/*--   FILE* unffile : Filedescriptor der Textdatei mit einer Entfaltung --*/
/*--   LIST_POINTER *CutOffPoints:                                       --*/
/*--                    Adresse eines LIST_POINTERS, ueber den eine Liste--*/
/*--                    der Namen der Cut-Off-Punkte zurueckgegeben wird --*/
/*--   LIST_POINTER *MaxConfig:                                          --*/
/*--                    Adresse eines LIST_POINTERS, ueber den eine Liste--*/
/*--                    der Namen einer max. Konfiguration zurueckgegeben--*/
/*--                    wird, falls keine Cutoffpunkte existieren        --*/
/*--                                                                     --*/
/*-- Rueckgabe:                                                          --*/
/*--     Zeiger vom Typ LMPNet auf das eingelesene Netz                  --*/

LMPNet readUNF( FILE* unffile,
                LIST_POINTER *CutOffPoints,
                LIST_POINTER *MaxConfig   )
{
  LMPNet        net = NULL;
  Cardinal32    numplaces,
                numtransitions,
                numcutoffs,
                numname,
                presetnum,
                postsetnum,
                count;
  Place         pl;
  Transition    tr;
  char          name[NAME_LEN+1];
  bool          has_preset;

  /* Die ersten Zeilen der Datei ueberlesen */
  for (count=HEAD_LINES;count;count--)
    ReadNewline( unffile );

  /* Stellenanzahl einlesen */
  if (( numplaces = ReadNumber( unffile ) ))
  {
    /* Zeiger auf Cut-Off-Point-Liste zunaechst auf NULL  setzen */
    *CutOffPoints = NULL;

    /* Netzheader anlegen */
    net = empty_net();

    /* eine Zeile weiter lesen */
    ReadNewline( unffile );

    /* Transitionenanzahl einlesen */
    numtransitions = ReadNumber( unffile );

    /* die Zeilen vor den Eintraegen fuer die Transitionen ueberlesen */
    for (count=LINES_BEFORE_TRANS;count;count--)
      ReadNewline( unffile );

    /* ab hier zeilenweise Transitionen einlesen */
    for ( count = 1 ; count <= numtransitions ; count++ )
    {
      /* als erstes numname lesen */
      numname = ReadNumber( unffile );

      /* bis zum Zeichen hinter numname weiterlesen */
      ReadSymbol( unffile, NUMNAME_DELIMITER );

      /* eventname lesen (Transitionsname in der Entfaltung) */
      ReadAlphaName( unffile, name, NAME_LEN );

      /* jetzt neue Transition ins net einfuegen */
      /* und numname eintragen                  */
      (tr = insert_transition( net, name ))->numname  = numname;

      /* corresponding transition name in original net */
      ReadAlphaName( unffile, name, NAME_LEN );
      tr->inscript = MYstrdup( name );

      /* eine Zeile weiter */
      ReadNewline( unffile );
    } /* for */

    /* die Zeilen vor den Eintraegen fuer die Stellen ueberlesen */
    for (count=LINES_BEFORE_PLACE;count;count--)
      ReadNewline( unffile );
    pl = NULL;

    /* ab hier zeilenweise Stellen, ggf. mit pre- und postset, einlesen */
    for ( count = 1 ; count <= numplaces ; count++ )
    {
      /* Annahme, es gibt einen preset-Eintrag zur aktuellen Stelle */
      presetnum = ReadNumber( unffile );

      /* ist es ein preset-Eintrag? */
      if ( TestSymbol( unffile, FIRST_ARROW_CHAR ) )
      {
        has_preset = True;
        /* restl. Zeichen des Pfeils lesen */
        ReadSymbol( unffile, SECOND_ARROW_CHAR );
        ReadSymbol( unffile, THIRD_ARROW_CHAR );
        /* numname der Stelle einlesen */
        numname = ReadNumber( unffile );
      }
      else
      {
        has_preset = False;
        /* die erste Zahl war numname der Stelle */
        numname = presetnum;
      }

      /* das Zeichen hinter numname lesen */
      ReadSymbol( unffile, NUMNAME_DELIMITER );

      /* label und component lesen, werden nicht verwendet */
      ReadPlaceLabel( unffile );
      ReadSymbol( unffile, LEFT_COMP_BRACKET );
      ReadNumber( unffile );
      ReadSymbol( unffile, RIGHT_COMP_BRACKET );

      /* name der Stelle (Bedingung) lesen */
      ReadAlphaName( unffile, name, NAME_LEN );
      /* neue Stelle mit Markierung einfuegen */
      pl = insert_place( net, name, INITIAL_MARK );

      /* name der Stelle im Ursprungsnetz lesen, wird nicht verwendet */
      ReadAlphaName( unffile, name, NAME_LEN );

      /* falls die Stelle ein preset hat, entspr. T/S-Kante einfuegen */
      if ( has_preset )
      {
        insert_TP_con(
          net,
          pl->name,
          /* mit presetnum Transition suchen */
          ( find_T_num( net, presetnum ) )->name,
          /* Kantengewicht */
          TP_WEIGHT
        );
      }

      /* gibt es postset-Eintraege ? */
      if ( TestSymbol( unffile, FIRST_ARROW_CHAR ) )
      {
        ReadSymbol( unffile, SECOND_ARROW_CHAR );
        ReadSymbol( unffile, THIRD_ARROW_CHAR );

        /* alle postset-Eintraege lesen und entspr. Kante(n) einfuegen */
        do
        {
          /* numname einer Transition im Nachbereich lesen */
          postsetnum = ReadNumber( unffile );

          /* S/T-Kante einfuegen */
          insert_PT_con(
            net,
            pl->name,
            /* mit presetnum Transition suchen */
            ( find_T_num( net, postsetnum ) )->name,
            /* Kantengewicht */
            PT_WEIGHT
          );
        }
        while ( TestSymbol( unffile, POSTSET_LIST_DELIMITER ) );

      } /* if ( TestSymbol() )*/

    } /* for ( count ... ) */

    /* ggf. Cut-Off-Punkte einlesen */
    /* gibt es einen numerischen Eintrag ? */
    if ( TestNum(unffile) )
    {
      /* Annahme, der Eintrag betrifft Cut-Off-Punkte */
      if (( numcutoffs = ReadNumber(unffile) ))
      {
        /* jetzt das naechste Wort lesen */
        ReadAlphaName( unffile, name, NAME_LEN );
        /* ist es tatsaechlich ein Eintrag fuer Cut-Off-Punkte */
        if ( !strcmp(name, CUTOFF_TEXT) )
        {
          GL_numCutoffs = numcutoffs;
          /* eine Zeile weiter */
          ReadNewline( unffile );

          /* alle Namen der Cut-Off-Punkte lesen und in */
          /* der Liste ablegen                          */
          while (numcutoffs--)
          {
            /* Namen des Cut-Off-Punkts lesen */
            ReadAlphaName( unffile, name, NAME_LEN );

            /* in die Liste eintragen */
            InsertListElem( CutOffPoints, name );

            /* bis zum evtl. naechsten Eintrag weiterlesen */
            ReadToCommaDot(unffile);
          }

        } /* if (!strcmp()) */

      }  /* if (numcutoffs...) */
      else
        /* eine Zeile weiter */
        ReadNewline( unffile );

    } /* if ( TestNum() ) */

    /* falls Entfaltung keine Cutoffpunkte enthaelt */
    /* ggf. eine max. Konfiguration einlesen        */
    if ( !(*CutOffPoints) )
    {
      ReadNewline( unffile );

      /* gibt es Liste der max. Konfigurationen? */
      if ( TestNum(unffile) )
      {
        /* Anzahl ueberlesen */
        ReadNumber(unffile);

        /* tatsaechlich Eintrag fuer max. Konfigurationen? */
        ReadAlphaName( unffile, name, NAME_LEN );
        if ( !strcmp(name, MAXCONFIG_TEXT) )
        {
          /* eine Zeile weiter */
          ReadNewline( unffile );

          do
          {
            /* Name eines Ereignisses lesen */
            ReadAlphaName( unffile, name, NAME_LEN );

            /* in die Liste eintragen */
            InsertListElem( MaxConfig, name );

          } while ( ReadToCommaDot(unffile) == COMMA_CHAR );

        } /* if ( !strcmp() ) */

      } /* if ( TestNum() ) */

    } /* if ( !(*CutOffPoints) ) */

  } /* if ( numplaces = ... ) */
  else
    NNError( E_RDUNF_NOCOND );

  /* Datei schliessen */
  fclose( unffile );

  /* Zeiger auf eingelesenes Netz zurueckgeben */
  return ( net );

} /* readUNF() */

/*-------------------------------------------------------------------------*/
/*--                                                                     --*/
/*-- Ende READ_UNF.C                                                     --*/
/*--                                                                     --*/
/*-------------------------------------------------------------------------*/
