/*****************************************************************************/
/* FILE pep2prod.c                                                           */
/*****************************************************************************/
/* Konvertierungsprogramm zur Umwandlung von Petri-Netzen aus PEP in das     */
/* Dateiformat des PROD Programms                                            */
/*                                             Carola Pohl, 24.02.97         */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/*                             header files                                  */
/*                                                                           */
/*****************************************************************************/

#include <string.h>
#include "neunetze.h"
#include "convertPEP.h"
#include "readlib.h"
#include "memlib.h"


/*****************************************************************************/
/*                                                                           */
/*                               defines                                     */
/*                                                                           */
/*****************************************************************************/

#define MEMOUT {fprintf(stderr, "out of memory\n"); exit(1);}

/*****************************************************************************/
/**************************** usage_err() ************************************/
/*****************************************************************************/
/* prints the message *msg and information about the usage of pep2prod to the*/
/* standard output and halts the program returning an error exit status      */
/*****************************************************************************/

#if defined (__STDC__) || defined (__cplusplus)

static void usage_err( char * msg )

#else

static void usage_err( msg )

char * msg;

#endif

{

	printf( "\n%s\n\n", msg );
	printf( "Usage: \n" );
	printf( "\t\t pep2prod [path]PEP_file [path]PROD_file \n\n");
	printf( "where PEP_file is a file containing a net in the PEP format ");
	printf( "and PROD_file\n" );
	printf( "is the target file. Path may be a valid pathname.\n" );
	printf( "\n\n" );

	exit(1);

} /* usage_err() */

/*****************************************************************************/
/************************** write_PROD() *************************************/
/*****************************************************************************/
/* writes the net *net represented in the PEP format into a file called      */
/* *filename in the PROD representation                                      */
/*****************************************************************************/
/* Caution: If no valid DOS filename is found behind an optional path in     */
/* *filename, the net name printed into the PROD file may be void!           */
/*****************************************************************************/

#if defined (__STDC__) || defined (__cplusplus)

static void write_PROD( char * filename, LMPNet net )

#else

static void write_PROD( filename, net )

char * filename;
LMPNet net;

#endif

{
	FILE       *file;
	Place      curr_pl;
	Transition curr_tr;
	TSEdge     curr_ts;
	STEdge     curr_st;

	num_net_nodes( net );

	if ( ! ( file = fopen( filename, "w" ) ) ) {
		PEPerror( "Cannot open PROD_file." );
	} /* if */

	/******************** write place list *******************************/

        curr_pl=net->Places;

	while (curr_pl) {
		fprintf(file,"#place %s",curr_pl->name);
		fprintf(file," lo(<.1.>) hi(<.1.>)");
		if (curr_pl->marks) {
			fprintf(file," mk(");
			if (curr_pl->marks >1) {
				fprintf(file,"%d",curr_pl->marks);
			} /* if */
			fprintf(file,"<.1.>)");
		} /* if */
		fprintf(file,"\n");
		curr_pl=curr_pl->nextplace;
	} /* while */

	/********************* write transition list *************************/

	curr_tr=net->Transitions;

	while (curr_tr) {
		fprintf(file,"#trans %s\n",curr_tr->name);
		curr_st=curr_tr->preset;
		if (curr_st) {
			fprintf(file,"  in  { ");
		} /* if */

		while (curr_st) {
			if (curr_st->weight) {
				fprintf(file,"%s: ",curr_st->place->name);
				if (curr_st->weight > 1) {
					fprintf(file,"%d",curr_st->weight);
				} /* if */
				fprintf(file,"<.x.>; ");
			} /* if */
			curr_st=curr_st->nextplace;
			if (!curr_st){
				fprintf(file,"}\n");
			} /*if*/

		} /* while */
		
		curr_ts=curr_tr->postset;
		if (curr_ts) {
			fprintf(file,"  out { ");
		} /* if */

		while (curr_ts) {
			if (curr_ts->weight) {
				fprintf(file,"%s: ",curr_ts->place->name);
				if (curr_ts->weight > 1) {
					fprintf(file,"%d",curr_ts->weight);
				} /* if */
				fprintf(file,"<.x.>; ");
			} /* if */
			curr_ts=curr_ts->nextplace;
			if (!curr_ts){
				fprintf(file,"}\n");
			} /*if*/

		} /* while */

		fprintf(file,"#endtr\n");
		curr_tr=curr_tr->nexttrans;

	} /* while */
    fclose(file);

} /* write_PROD() */


#if defined (__STDC__) || defined (__cplusplus)

static void parse_args(int argc, char **argv, char **ll_np, char **prod_np)

#else

static void parse_args(argc, argv, ll_np, prod_np )

int argc;
char **argv;
char **ll_np;
char **prod_np;

#endif

{
	char *name,*d, *e;

	if (argc == 1) {
		usage_err( "name required!" );
	}
	if (argv[1][0] == '-') {
		usage_err( "unknown option!" );
	}
	d = strchr(argv[1], '.');
	if (d) {
		*ll_np = MYstrdup(argv[1]);
	} else {
		name = MYmalloc(strlen(argv[1]) + 8);
		sprintf(name, "%s%s", argv[1], ".ll_net");
		*ll_np = name;
	}
	if (2 < argc) {
		if ((e = strchr(argv[2], '.'))) {
			*prod_np = MYstrdup(argv[2]);
		} else {
			name = MYmalloc(strlen(argv[2]) + 6);
			sprintf(name, "%s%s", argv[2], ".prod");
			*prod_np = name;
		}
	} else {
		if (d) {
			argv[1][d - argv[1]] = '\0';
		}
		name = MYmalloc(strlen(argv[1]) + 6);
		sprintf(name, "%s%s", argv[1], ".prod");
		*prod_np = name;
	}
}


/*****************************************************************************/
/*************************** main ********************************************/
/*****************************************************************************/
/* creates a file called *argv[2] containing the ll-net *argv[1] as a        */
/* PEP file in the PROD format if exactly 2 valid parameters are given, else */
/* a message about correct usage is printed to the standard output           */
/*****************************************************************************/

#if defined (__STDC__) || defined (__cplusplus)

int main( int argc, char * argv[] )

#else

int main( argc, argv )

int argc;
char * argv[];

#endif

{

	char * prod_name = NULL;
	char * ll_name = NULL;
	FILE * file;
	LMPNet net;

	parse_args( argc, argv, &ll_name, &prod_name );

	if (( file = fopen( prod_name, "w" ) )) {
		fclose( file );
	} /* if */
	else {
		free(ll_name);
		free(prod_name);
		usage_err( "Cannot create PROD_file. Possibly wrong path." );
	} /* else */
	CM_storeReferences = False;

	net = read_PEP( ll_name );

	if ( ! net ) {
		free(ll_name);
		free(prod_name);
		usage_err( "Could not read PEP net." );
	} /* if */

	write_PROD( prod_name, net );

	remove_net( &net );
	free(ll_name);
	free(prod_name);

	exit(0);

} /* main() */


/******************************************************************************/
/* END OF FILE pep2prod.c                                                      */
/******************************************************************************/

