/**************************************************************************/
/* CSC399 Individual Project, University of Newcastle upon Tyne           */
/* PEP2SMV - Converter from PEP .ll_net files to SMV specifications       */
/* (C) Oct. 1996 - Apr. 1997 Guido Wimmel                                 */
/*                                                                        */
/* ChangeLog.h: C++ class for generating/optimising expressions           */
/*              for unchanged states                                      */
/*                                                                        */
/* see project dissertation, section 5.3 for compression algorithm        */
/**************************************************************************/

class ChangeLog
{
public:
  /* constructor: entr = number of entries, compr: generate compressed */
  /* expressions or uncompressed ones */ 
  ChangeLog(Cardinal32 entr, bool compr=false); 
  ~ChangeLog();
  
  /* setName: set name for a variable (index idx) */
  inline void setName(Cardinal32 idx, char *name); 
  
  /* logChange: variable with index idx has been changed */
  /*            (e.g. next(Pi)=1), do not include in 'unchanged' */
  /*            expression */
  void logChange(Cardinal32 idx); 

  /* genUnchanged: generate 'unchanged' expressions and write them to f */
  /*               then reset information obtained by logChange */
  void genUnchanged(FILE *f);      

  /* genDefines: generate defines for 'unchanged' expressions (used by */
  /*             compression algorithm) */
  void genDefines(FILE *f);        

private:
  /* recursive utility routines */
  void genUnchanged1(FILE *f, Cardinal32 i, Cardinal32 l, Cardinal32 r);
  void genDefines1(FILE *f, Cardinal32 i, Cardinal32 l, Cardinal32 r);

  /* private variables */
  Cardinal32 entries; /* number of variables */
  Cardinal32 tree_size; /* size of array for binary tree */
  bool compress; /* use compression algorithm ? */
  char **names; /* names of the variables */
  bool *unchanged; /* without compression: contains information if */
                   /* variable was changed */
  bool *unchanged_tree; /* with compression: 'unchanged' tree (see */
                        /*  project dissertation */
};

/**************************************************************************/
/*                        ChangeLog                                       */
/**************************************************************************/
/* constructor: allocate space for variables, initialise                  */
/*  entr: number of entries in table (=number of variables)               */
/*  compr: use compression algorithm if true                              */
/**************************************************************************/

ChangeLog::ChangeLog(Cardinal32 entr, bool compr)
{
  Cardinal32 i;

  /* --- store entr and compr in private variables --- */
  entries=entr;
  compress=compr;

  /* --- allocate space for variable names --- */
  names=new char *[entries];
  for (i=0;i<entries;i++) names[i]="<not set>";

  if (!compress) /* no compression used */
    {
      /* allocate space for 'unchanged' array */
      /* fill with true */
      unchanged = new bool[entries];
      for (i=0;i<entries;i++) unchanged[i]=true;
    }
  else /* compression used */
    {
      /* determine size of unchanged_tree array (first power of 2 */
      /* larger than number of entries */
      tree_size=1; while (tree_size<entries) tree_size<<=1;

      /* allocate unchanged_tree (2*tree_size, because of leaves) */
      unchanged_tree = new bool[2*tree_size];
      /* fill with true (all variables initially 'unchanged') */
      for (i=0;i<2*tree_size;i++) unchanged_tree[i]=true;
    }
}

/**************************************************************************/
/*                        ~ChangeLog                                      */
/**************************************************************************/
/* destructor: free space for variables allocated                         */
/**************************************************************************/

ChangeLog::~ChangeLog()
{
  delete[] names;
  if (!compress)
    delete[] unchanged;
  else
    delete[] unchanged_tree;
}

/**************************************************************************/
/*                        setName                                         */
/**************************************************************************/
/* setName: set name for variable with index idx                          */
/**************************************************************************/

inline void ChangeLog::setName(Cardinal32 idx, char *name)
{
  names[idx]=name;
}

/**************************************************************************/
/*                        logChange                                       */
/**************************************************************************/
/* register change for variable with index idx, so that it is not         */
/* included in 'unchanged' expressions generated by genUnchanged()        */
/**************************************************************************/

void ChangeLog::logChange(Cardinal32 idx)
{
  /* --- if no compression, simply set unchanged[idx] to false --- */
  if (!compress) unchanged[idx]=false;
  else /* compression */
    {
      Cardinal32 l=0,r=entries-1,i=0,mid;

      /* --- descend unchanged_tree from root to leaf corresponding --- */
      /*     variable number idx, set nodes to false */
      /*     i is current node, l..r are indizes of range of variables */
      /*     this node corresponds to */
      unchanged_tree[0]=false;
      do
	{
	  mid=(l+r)/2;
	  if (idx<=mid) /* descend to left */
	    {i=i*2+1;r=mid;}
	  else /* descend to right */
	    {i=i*2+2;l=mid+1;}
	  if (l<=r) unchanged_tree[i]=false;
	} while (l<r);
    }
}

/**************************************************************************/
/*                     genUnchanged                                       */
/**************************************************************************/
/* generate 'unchanged' expressions for variables for which logChange()   */
/* was not called; reset information obtained from logChange()            */
/* write to file f                                                        */
/**************************************************************************/

void ChangeLog::genUnchanged(FILE *f)
{
  if (!compress)
    {
      /* --- if no compression: output all variable names for which --- */
      /*     unchanged[] is still true */
      for (Cardinal32 i=0;i<entries;i++)
	if (unchanged[i])
	    fprintf(f,"next(%s)=%s&",names[i],names[i]);
	else
	  /* set unchanged to true to reset information */
	  unchanged[i]=true;
    }
  else /* --- compression: call recursive function genUnchanged1() --- */
    genUnchanged1(f,0,0,entries-1);
}

/**************************************************************************/
/*                     genUnchanged1                                      */
/**************************************************************************/
/* compression algorithm: recursively generate 'unchanged' expressions    */
/* write to file f                                                        */
/* i: index of current node in array unchanged_tree                       */
/* l: (left) first variable index of range of variables node i belongs to */
/* r: (right) last variable index of range of variables node i belongs to */
/**************************************************************************/

void ChangeLog::genUnchanged1(FILE *f, Cardinal32 i,
			      Cardinal32 l, Cardinal32 r)
{
  Cardinal32 mid;
  
  if (l>r) return; /* shouldn't happen! */

  if (!unchanged_tree[i])
    {
      /* if range of variables were 'changed', split range of places in */
      /* halves and call genUnchanged1 recursively with left and right  */
      /* son in unchanged_tree */
      unchanged_tree[i]=true;
      if (l==r) return;
      mid=(l+r)/2;
      genUnchanged1(f,i*2+1,l,mid); /* left son */
      genUnchanged1(f,i*2+2,mid+1,r); /* right son */
    } else {
      /* range was 'unchanged': output 'next(Pi)=Pi&' */
      /*  if it was only one place */
      /*  and 'u_Pi_Pj&' if it was more than one */
      if (l==r) fprintf(f,"next(%s)=%s&",names[l],names[l]);
      else 
	  fprintf(f,"u_%s_%s&",names[l],names[r]);
    } 
}

/**************************************************************************/
/*                     genDefines                                         */
/**************************************************************************/
/* generate defines needed by compression algorithm, write to f           */
/**************************************************************************/
void ChangeLog::genDefines(FILE *f)
{
  /* call recursive procedure if compression is used, nothing to be done */
  /*  otherwise */
  if (compress) genDefines1(f,0,0,entries-1);
}

/**************************************************************************/
/*                     genDefines1                                        */
/**************************************************************************/
/* compression algorithm: recursively generate defines for 'unchanged'    */
/* expressions, write to file f                                           */
/* i: index of current node in array unchanged_tree                       */
/* l: (left) first variable index of range of variables node i belongs to */
/* r: (right) last variable index of range of variables node i belongs to */
/**************************************************************************/

void ChangeLog::genDefines1(FILE *f, Cardinal32 i, 
			    Cardinal32 l, Cardinal32 r)
{
  Cardinal32 mid;
  
  if (l>=r) return; /* don't generate define for only one place */

  /* --- split range into halves, generate defines for both halves --- */
  mid=(l+r)/2;
  genDefines1(f,i*2+1,l,mid); /* left half */
  genDefines1(f,i*2+2,mid+1,r); /* right half */

  /* --- generate define for current node --- */
  /* use defines for left and right halves for that */
  /* ('DEFINE u_Pi_Pj := u_Pi_Pmid & u_P(mid+1)_Pj;') */
  /* if left or right half is only one place, use */
  /* next(Pi)=Pi instead of u_Pi_Pj */

  fprintf(f,"u_%s_%s:=",names[l],names[r]);
  if (l>=mid) 
    fprintf(f,"next(%s)=%s&",names[l],names[l]);
  else
    fprintf(f,"u_%s_%s&",names[l],names[mid]);
  if (mid+1>=r)
    fprintf(f,"next(%s)=%s;\n",names[r],names[r]);
  else
    fprintf(f,"u_%s_%s;\n",names[mid+1],names[r]);
}

/**************************************************************************/
/* END OF FILE ChangeLog.h                                                */
/**************************************************************************/
