/**************************************************************************/
/* CSC399 Individual Project, University of Newcastle upon Tyne           */
/* PEP2SMV - Converter from PEP .ll_net files to SMV specifications       */
/* (C) Oct. 1996 - Apr. 1997 Guido Wimmel                                 */
/*                                                                        */
/* gencomp.c - Calculate one-token-invariants (components)                */
/*             from a .ll-net file                                        */
/*                                                                        */
/* 28.09.97: Changed backsubst (optimisation, recursion is terminated     */
/*             earlier when non_overlap is specified) -- Guido Wimmel     */
/**************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "neunetze.h"
#include "convertPEP.h"

#include "Components.h"

/* size for input line buffer (spec-file) */
#define BUFSIZE 1000
/* no pivot (used for pivcol) */
#define NO_PIVOT 100000

/* macros for accessing matrices */
#define C(x,y) (c[(x)*num_p+(y)])
#define A(x,y) (a[(x)*n+(y)])

/* variable for command line option */
bool non_overlap=false;

char *c; /* Connection Matrix */
char *a; /* Matrix for solving equations */
Place *pl_col; /* Places corresponding to columns of A */
char *sol; /* solution */
int m,n,num_p,num_t; /* Dimensions of Matrices A and C*/
int *pivcol; /* Pivot colums for each row */

/* component lists and place lists */
Component *first_comp,*last_comp,*comp;
Cardinal32 compid;
PlaceList *first_placel,*last_placel;
ComponentList *complist;

/**************************************************************************/
/*                           print_C                                      */
/**************************************************************************/
/* print Matrix C (connection matrix), labelled with places and           */
/* transitions                                                            */
/**************************************************************************/

void print_C(LMPNet net)
{
  int i,j;
  Place pl;
  Transition tr;

  printf("Connection matrix:\n");
  for (j=0,pl=net->Places;j<num_p;j++,pl=pl->nextplace)
    printf("\t%s",pl->name);
  for (i=0,tr=net->Transitions;i<num_t;i++,tr=tr->nexttrans)
    {
      printf("\n%s:",tr->name);
      for (j=0;j<num_p;j++)
	printf("\t%d",C(i,j));
    }
  printf("\n");
}  

/**************************************************************************/
/*                           print_C                                      */
/**************************************************************************/
/* print Matrix A (coefficient matrix for linear equation system)         */
/* labelled with places and transitions                                   */
/* (Matrix A does not contain columns for the places that are not         */
/*  variables in the equations, i.e. the places with initial marking      */
/*  1 (because m(p)=1 for one place p that is marked and                  */
/*     m(p)=0 for all other marked places)                                */
/**************************************************************************/

void print_A()
{
  int i,j;
   
  printf("Matrix A:\n");
  for (j=0;j<n-1;j++) printf("\t%s",pl_col[j]->name);
  for (i=0;i<m;i++)
    {
      printf("\n");
      for (j=0;j<n;j++)
	printf("\t%d",a[i*n+j]);
    }
  printf("\n");
}

/**************************************************************************/
/*                         backsubst                                      */
/**************************************************************************/
/* recursive backwards substitution in the equation system (to calculate  */
/* _all_ solutions)                                                       */
/* (i,j) is current element in matrix A (j: variable, i: row of matrix)   */
/* pl is the marked place that was used to find possible components       */
/*  it is in                                                              */
/**************************************************************************/

void backsubst(int i,int j,Place pl)
{
  int jj;

  /* If a component has already be found, don't look for more solutions */
  /* if non_overlap is specified */
  if ( (non_overlap) && (pl->components != NULL)) return; 

  /* j=-1: sol[] is a solution ! */
  if (j==-1) {
    /* make new component */
    comp=new Component;
    comp->id = compid++;
    comp->next=NULL;
    sprintf(comp->name,"comp_%d",comp->id);
	    
    /* add it to component list */
    if (first_comp==NULL) first_comp=comp;
    else last_comp->next=comp;
    last_comp=comp;
	    
    /* create a place list with only the place pl */
    /* in it */
    first_placel=last_placel=new PlaceList;
    first_placel->pl=pl;
    first_placel->next=NULL;
	    
    /* add component to the list of components the place pl */
    /* is in */
    complist=new ComponentList;
    complist->comp=comp;
    complist->next=(pl->components);
    pl->components=complist;
	    
    /* add places in solution to place list where sol[...]=1 */
    for (jj=0;jj<n-1;jj++) 
      if (sol[jj]==1) 
	{
	  PlaceList* placel=new PlaceList;
	  placel->pl=pl_col[jj];
	  placel->next=NULL;
	  last_placel->next=placel;
	  last_placel=placel;
	  
	  /* add Component to Component list for this place */
	  complist=new ComponentList;
	  complist->comp=comp;
	  complist->next=(pl_col[jj]->components);
	  pl_col[jj]->components=complist;
	}

    /* store pointer to place list in component */
    comp->places = first_placel;
  }
  else
    /* no solution, recurse further */
    {
      if (j>pivcol[i]) /* j is a free variable */
	{
	  /* assign both possible solutions (0 and 1) to the free variable */
	  /* and call backsubst recursively (with variable j-1, same row)  */
	  /* if non-overlap: solution can only be 1 if place is not yet in a */
	  /* component */
	  sol[j]=0;backsubst(i,j-1,pl); 
	  
	  if ( (!non_overlap) || (pl_col[j]->components==NULL))
	    {
	      sol[j]=1;backsubst(i,j-1,pl);
	    }
	}
      else /* (i,j) is a pivot */
	{ 
	  /* -> calculate value of variable j */
	  sol[j]=A(i,n-1);for (jj=j+1;jj<n-1;jj++) sol[j]-=A(i,jj)*sol[jj];

	  /* solution must only be 1 or 0, and */
	  /* if non-overlap: place must not be in a component already */
	  /* if solution is 1 (place is to be added to a component) */
	  if ( ((sol[j]==1) && ( (!non_overlap) || (pl_col[j]->components==NULL)))
	       ||
	       (sol[j]==0) )

	    /* proceed recursively with variable j-1, previous row of matrix  */
	   {
	     backsubst(i-1,j-1,pl);
	   }
	}
    }   
}

/**************************************************************************/
/*                   calculate_Components                                 */
/**************************************************************************/
/* calculate components (one-token-P-invariants) in net,                  */
/* using Gaussian elimination                                             */
/**************************************************************************/

Component *calculate_Components(LMPNet net)
{
  char h;
  Place pl;
  Transition tr;
  TSEdge ts;
  STEdge st;
  int unmarked_p;
  int i,j,piv_i,piv_j,piv_ii,lastpiv_i;
  
  /* --- Prepare empty Component list for solution --- */
  first_comp=last_comp=NULL;compid=0;
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace) pl->components=NULL;

  /* --- Assign numbers to places and transitions --- */
  num_net_nodes(net);

  /* --- Build connection matrix ---- */
  num_t=net->NumTransitions;
  num_p=net->NumPlaces;
  c = new char[num_t*num_p];
  for (i=0;i<num_t;i++) for (j=0;j<num_p;j++) C(i,j)=0;
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
    {
      for (ts=tr->postset;ts;ts=ts->nextplace)
	C(tr->numname,ts->place->numname)+=1;
      for (st=tr->preset; st;st=st->nextplace)
	C(tr->numname,st->place->numname)-=1;
    }
  
  /* --- count unmarked places, allocate space for matrix A --- */
  unmarked_p=0;
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
      if (pl->marks==0) unmarked_p++;
  m=num_t; n=unmarked_p+1;
  a = new char[m*n];

  /* --- allocate space for solution, place references for --- */
  /*     columns of A, pivot columns of A                      */
  sol = new char[n-1];
  pl_col = new Place[n-1];
  pivcol = new int[m];
  
  /* --- Determine columns to be used for Matrix A --- */
  j=0;
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    if (pl->marks==0) pl_col[j++]=pl;

  /* --- for each marked place, determine components it is in --- */
 
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    if (pl->marks>0)
      {
	/* --- Copy colums for unmarked places from C to A --- */
	for (j=0;j<n-1;j++) for (i=0;i<m;i++)
	  A(i,j)=C(i,pl_col[j]->numname);
		
	/* --- determine column n of a --- */
	for (i=0;i<m;i++)
	  A(i,n-1)=-C(i,pl->numname);

	/* --- Do the Gausselim --- */
	/* (piv_i,piv_j) = pivot row,column */
	lastpiv_i=piv_i=piv_j=0;
	for (i=0;i<m;i++) pivcol[i]=NO_PIVOT;
	while ((piv_i<m)&&(piv_j<(n-1)))
	  {
	    /* Search for pivot in column piv_j */
	    piv_ii=piv_i;
	    while ((A(piv_ii,piv_j)==0)&&(piv_ii<m)) piv_ii++;
	    
	    /* pivot found */
	    if (piv_ii<m)
	      {
		pivcol[piv_i]=piv_j; /* store pivot column */
		/* swap rows */
	       if (piv_ii != piv_i) for (j=piv_j;j<n;j++)
		 {h=A(piv_i,j);A(piv_i,j)=A(piv_ii,j);A(piv_ii,j)=h;}

	       /* scale pivot to 1 */
	       if (A(piv_i,piv_j)==-1) for (j=piv_j;j<n;j++)
		 A(piv_i,j)=-A(piv_i,j);

	       /* Eliminate */
	       for (i=piv_i+1;i<m;i++)
		 if (A(i,piv_j)!=0) 
		   for (j=piv_j+1;j<n;j++)
		       A(i,j)-=A(piv_i,j)*A(i,piv_j);
	       piv_i++;piv_j++;
	     }
	   else
	       piv_j++; /* no pivot found, try next column */
	 }
	i=m-1;

	/* Be careful, i might get less than 0, if empty matrix */
	/* but Cardinal32 i (now int!) */
	/* search for first pivot row or last col. of A != 0 */
	while ((pivcol[i]==NO_PIVOT)&&(A(i,n-1)==0)) i--;

	/* if last col. of A != 0 before a pivot was found */
	/* there is no solution -- otherwise call backsubst */
	/* to find one */
	
	if (pivcol[i]!=NO_PIVOT) 
	  backsubst(i,n-2,pl);
      }
  printf("\n");
	   
  /* free allocated memory */
  delete[] a;
  delete[] c;
  delete[] pl_col;
  delete[] sol;
  delete[] pivcol;
  return NULL;
}

/**************************************************************************/
/*                 print_componentInfo                                    */
/**************************************************************************/
/* print information about components in the net:                         */
/* - components and places that belong to them                            */
/* - for all places: components they belong to                            */
/* verify if components are really one-token-P-invariants, and report     */
/*  error if they are not                                                 */
/**************************************************************************/

void print_componentInfo(LMPNet net,Component *firstcomp)
{
  Place pl;
  Transition tr;
  TSEdge ts;
  STEdge st;
  Component *comp;
  ComponentList *complist;
  PlaceList *placel;
  int i;
  
  /* --- print all components and their place list --- */
  printf("Components information:\n");
  for (comp=firstcomp;comp!=NULL;comp=comp->next)
    {
      printf("Component %s (%d): ",comp->name,comp->id);
      /* count how many places are initially marked in this component */
      i=0;
      for (placel=comp->places;placel!=NULL;placel=placel->next)
	{
	  printf(" %s",placel->pl->name);
	  if (placel->pl->marks>0) i++;
	}
      printf("\n");
      /* report error if not exactly one place is marked */
      if (i!=1) printf("---Error: %d Places initially marked in "
		       " component %s, should be 1!\n",i,comp->name);
    }
  
  /* --- print all places and the components they are in --- */
  printf("For places:\n");
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    {
      printf("Place %s: ",pl->name);
      for (complist=pl->components;complist!=NULL;complist=complist->next)
	printf(" %s (%d)",complist->comp->name,complist->comp->id);
      printf("\n");
    }

  /* --- verify if components are one-token-invariants: --- */
  /* for all transitions: */
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
    {
      /* set visited to 0 for all components */
      for (comp=firstcomp;comp!=NULL;comp=comp->next) comp->visited=0;

      for (ts=tr->postset;ts;ts=ts->nextplace)
	/* for every place in the preset, increase visited */
	/*  for every component it is in */
	for (complist=ts->place->components;complist!=NULL;complist=complist->next) 
	  complist->comp->visited++;
      for (st=tr->preset;st;st=st->nextplace)
	/* for every place in the postset, decrease visited */
	/* for every component it is in */
	for (complist=st->place->components;complist!=NULL;complist=complist->next) 
	  complist->comp->visited--;
    
      /* if visited is not 0 for a component, there were not as many */
      /* places in this component in the preset as there were in the */
      /* postset --> report error */
      for (comp=firstcomp;comp!=NULL;comp=comp->next) 
	if (comp->visited!=0)
	  printf("---Error: %s is not a Component (transition %s)\n",
		 comp->name,tr->name);
    }
}

/**************************************************************************/
/*                           main                                         */
/**************************************************************************/
           
int main( int argc, char *argv[] )
{
  LMPNet net;
  char *myname, *cp, *filename=NULL;
  char outputfilename[256]="";
  FILE *outputfile;

  /* --- parse command line --- */
  myname=*(argv++);
  argc--;
  
  /* print usage if no command line parameters given */
  if (argc==0)
    {
      printf( "PEP2SMV - getComponents\n" );
      printf( "Calculates all 1-invariants of a Petri net\n"); 
      printf( "(C) 1996/97 G. Wimmel, University of Newcastle - TU Muenchen\n\n");
      printf( "Usage: %s inputfile [-o=<output file name] "\
              "[-r]\n"\
              "       -r restricts computation to "\
              "non-overlapping 1-invariants\n",
	      myname);
      exit ( 1 );
    }
  
  /* read command line options */
  for (;argc>0;argc--,argv++)
    {
      if ((*argv)[0]=='-')
	{
	  if (strncmp(*argv,"-o=",3)==0)
	    strcpy(outputfilename,*argv+3);
	  else if (strncmp(*argv,"-r",2)==0)
	    non_overlap=true;
	  else
	  {
	    fprintf(stderr,"Illegal option: %s.\n",*argv);exit(10);
	  }
	}
      else
	{
	  if (filename==NULL) filename=*argv;
	  else
	    {fprintf(stderr,"Command line error.\n");exit(10);}
	}
    }
  
  /* --- read Petri net file */
  CM_use_imarks = True; /* store initial marking in struct Place */
  net = read_PEP( filename );
  if (place_number(net)+trans_number(net)==0)
    {fprintf(stderr, "gencomp error: net must have at least" \
	     "one place and one transition.\n");exit(10);}
  
  /* --- generate output file name (extension .comp) if necessary */
  if (strlen(outputfilename)==0)
    {
      strcpy(outputfilename,filename);
      cp=strrchr(outputfilename,'.');
      if (cp==NULL) strcat(outputfilename,".comp");
      else strcpy(cp,".comp");
    }
  
  /* --- open output file --- */
  if ( ! ( outputfile = fopen(outputfilename, "w" ) ) )
    {
      fprintf(stderr,"gencomp error: "\
	      "Could not open file %s for writing\n",outputfilename);
      exit(10);
    }
  
  /* --- calculate components --- */
  calculate_Components(net);
  
  /* --- print information about components --- */
  print_componentInfo(net,first_comp);

  /* --- write component list to output file --- */
  fprintf(outputfile,"PEP\nS_List\nFORMAT_N2\nLIST\n");
  for (comp=first_comp;comp!=NULL;comp=comp->next)
    {
      fprintf(outputfile,"\"");
      for (PlaceList* placel=comp->places;placel!=NULL;placel=placel->next)
	{
	  fprintf(outputfile,"%s",placel->pl->name);
	  if (placel->next) fprintf(outputfile,",");
	}
      fprintf(outputfile,"\"\n");
    }
   
  if (outputfile!=stdout)
    fclose(outputfile);

   remove_net( &net );
  
  exit ( 0 );
} /* main() */

/**************************************************************************/
/* END OF FILE gencomp.c                                                  */
/**************************************************************************/
