/**************************************************************************/
/* CSC399 Individual Project , University of Newcastle upon Tyne          */
/* PEP2SMV - Converter from PEP .ll_net files to SMV specifications       */
/* (C) Oct. 1996 - Apr. 1997 Guido Wimmel                                 */
/* Oct. 1997: Pseudo S-Invariant support added by Stefan Schwoon	  */
/*                                                                        */
/* pep2smv.c - generate SMV representation from PEP .ll_net files         */
/**************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "convertPEP.h"

#include "ChangeLog.h"

/* size of Table for processes to be generated (model 1) */
#define MAXTABSIZE 1000
/* size for input line buffer (spec-file) */
#define BUFSIZE 1000

/* global variables for command line options */
char *myname;
char *filename=NULL,*specfilename=NULL, *componentfilename=NULL;
int model=5,option_run_smv=0,option_use_pseudo=0;
char *specification;
FILE *outputfile;
char *smv_options[20];

struct processTableElement
/* table of processes to be generated for transitions */
/* (models 1,2) */
{
  Cardinal32 preset_num;   /* number of places in preset */
  Cardinal32 postset_num;  /* number of places in postset */
  Cardinal32 selfloop_num; /* number of places in both pre- and postset */
};

inline bool operator==(struct processTableElement s, 
		       struct processTableElement t)
{
  return ((s.preset_num==t.preset_num)&&(s.postset_num==t.postset_num)&&
	  (s.selfloop_num==t.selfloop_num));
}

/**************************************************************************/
/*                         copy_spec                                      */ 
/**************************************************************************/
/* Copy specification in command line parameter -s=... to outputfile      */
/* (if specified)                                                         */
/* Copy specifications in specfile, line by line (if specified)           */
/* Add SPEC for every specification                                       */
/**************************************************************************/

void copy_spec(FILE *outputfile) 
{
  FILE *specfile;
  char buf[BUFSIZE];
  
  if (specification!=NULL)
    {
      fprintf(outputfile,"SPEC\n%s\n",specification);
    } 
  if (specfilename!=NULL)
    {
      if ( ! ( specfile = fopen(specfilename, "r" ) ) )
	{
	  fprintf(stderr,"PEP2SMV error: Could not open "\
		  "specification file %s\n",specfilename);
	  fclose(outputfile);exit(10);
	} /* if */
      do {
	  fgets(buf,BUFSIZE,specfile);
	  if ((strlen(buf)>1)&&(buf[0]!='%'))
	    fprintf(outputfile,"SPEC\n%s\n",buf);
	} while (!feof(specfile));
      fclose(specfile);
    }
}

/**************************************************************************/
/*                           write_SMV1                                   */
/**************************************************************************/
/* write SMV representation for net models 1 or 2 to file                 */
/* alternative=false: model 1 / alternative=true: model 2                 */
/**************************************************************************/

void write_SMV1( LMPNet net, FILE *file, bool alternative )
{
  Cardinal32 i,j;
  Place pl;
  Transition tr;
  TSEdge ts;
  STEdge st;
  processTableElement cur;
  processTableElement procTab[MAXTABSIZE]; /* table for process types */
  Cardinal32 p=0; /* Position in Table */
  bool found;

  /* --- header --- */
  fprintf(file,"-- generated by PEP2SMV from %s ",NNinput_file);
  if (!alternative) 
    fprintf(file,"(model 1)\n");
  else fprintf(file,"(model 2)\n");
  fprintf(file,"-- (C) 1996-1997 Guido Wimmel\n\n");

  /* --- MODULE main --- */
  /* --- VAR declaration --- */
  fprintf(file,"MODULE main\nVAR\n");

  /* model 2: declaration of 'fired: boolean;' */
  if (alternative) fprintf(file,"fired: boolean;\n");
  
  /* List place variables 'Pi:boolean;' */
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    {
      fprintf(file,"%s: boolean;\n",pl->name);
      pl->visited_pre=pl->visited_post=NULL;
    }

  /* Transition processes 'Tr: process t_i_j_k(...);' */
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
    {
      cur.preset_num=cur.postset_num=cur.selfloop_num=0;
      
      /* determine how many places there are in the */
      /* preset, postset and both in pre- and postset */
      for (st=tr->preset;st;st=st->nextplace)
 	{
      	  cur.preset_num++;
	  st->place->visited_pre=tr;
	}
      for (ts=tr->postset;ts;ts=ts->nextplace)
 	{
	  cur.postset_num++;
	  ts->place->visited_post=tr;
	  /* place in postset was 'visited from' the preset */
          /* -> it is in a selfloop */
	  if (ts->place->visited_pre==tr) cur.selfloop_num++;
	}
      if ((cur.preset_num+cur.postset_num)==0) 
	  fprintf(stderr,"PEP2SMV warning: Transition %s isolated, \
             will be ignored.\n",tr->name);
      else
	{
	  /* write transition process declaration 'Tr: process tr_i_j_k(' */
	  cur.preset_num -= cur.selfloop_num;
	  cur.postset_num -= cur.selfloop_num;
	  fprintf(file,"%s: process tr_%d_%d_%d(",
		  tr->name,cur.preset_num,cur.postset_num,
		  cur.selfloop_num);
	  
	  /* write parameters: places in preset (excluding postset) */
	  for (st=tr->preset;st;st=st->nextplace)
	    if (st->place->visited_post!=tr)
		fprintf(file,"%s,",st->place->name);
	  
	  /* write parameters: places in postset (excluding preset) */
	  for (ts=tr->postset;ts;ts=ts->nextplace)
	    if (ts->place->visited_pre!=tr)
	      fprintf(file,"%s,",ts->place->name);
	  
	  /* write parameters: places in selfloop (preset and postset) */
	  for (ts=tr->postset;ts;ts=ts->nextplace)
      	    {
	      if (ts->place->visited_pre==tr)
		fprintf(file,"%s,",ts->place->name);
	    }
	  
	  /* model 2: 'fired' as a parameter */
	  if (alternative) fprintf(file,"fired,");
	  fseek(file,-1,SEEK_CUR); /* overwrite last "," */
	  fprintf(file,");\n");
	  
	  /* Enter transition module type into table procTab if it is new */
	  found=False;
	  for (i=0;(!found)&&(i<p);i++) found|=(procTab[i]==cur);
	  if (!found) 
	    if (p<MAXTABSIZE) procTab[p++]=cur;
	    else 
	      {fprintf(stderr,"PEP2SMV error: Table full.\n");exit(10);}
	}
      
    } /* for (transitions) */

  /* --- ASSIGN declaration --- */
  /* Initial states, 'init(Pi):=...;' */
  fprintf(file,"\nASSIGN\n");
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    fprintf(file,"init(%s):=%d;\n",pl->name,pl->marks);

  /* --- DEFINE declaration --- */
  /* 'DEFINE deadlock:=!(Ta.enabled|Tb.enabled|...)' */
  fprintf(file,"DEFINE\ndeadlock:=!(");
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
	fprintf(file,"%s.enabled|",tr->name);
  fseek(file,-1,SEEK_CUR);
  fprintf(file,");\n");

  /* --- FAIRNESS declaration --- */
  /* (only model 2): */
  /* 'FAIRNESS (Ta.running|Tb.running|...)&(fired|deadlock)' */
  if (alternative)
    {
      fprintf(file,"\nFAIRNESS\n(");
      for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
	fprintf(file,"%s.running|",tr->name);
      fseek(file,-1,SEEK_CUR);
      fprintf(file,")&(fired|deadlock)\n\n");
    }
      
  /* --- SPEC declaration --- */
  copy_spec(file);
  
  /* --- MODULE definitions for transition processes */
  fprintf(file,"\n");
  for (i=0;i<p;i++) /* for all entries in procTab */
    {
      /* 'MODULE tr_i_j_k(in0,...,ini,out0,...,outj,inout0,...,inoutk[,fired])' */  
      /* (fired only for model 2) */
      fprintf(file,"MODULE tr_%d_%d_%d(",
	      procTab[i].preset_num,procTab[i].postset_num,
	      procTab[i].selfloop_num);
      for (j=0;j<procTab[i].preset_num;j++) fprintf(file,"in%d,",j);
      for (j=0;j<procTab[i].postset_num;j++) fprintf(file,"out%d,",j);
      for (j=0;j<procTab[i].selfloop_num;j++) fprintf(file,"inout%d,",j);
      if (alternative) fprintf(file,"fired,");
      fseek(file,-1,SEEK_CUR);

      /*'ASSIGN */
      /*  next(in...):=case */
      /*   enabled: 0; */
      /*   1: in...; */
      /*  esac;' */
      fprintf(file,")\nASSIGN\n");
      for (j=0;j<procTab[i].preset_num;j++)
	fprintf(file,"next(in%d):=case\n  enabled: 0;\n  1:in%d;\nesac;\n",j,j);

      /* next(out...):=case */
      /*  enabled: 1; */
      /*  1: out...; */
      /* esac; */
      for (j=0;j<procTab[i].postset_num;j++)
	fprintf(file,"next(out%d):=case\n  enabled: 1;\n  1:out%d;\nesac;\n",j,j);
      if (alternative) fprintf(file,"fired:=enabled;\n");

      /* 'DEFINE enabled:=in0&...&ini&inout0&...&inoutk;' */
      fprintf(file,"DEFINE\nenabled:=");
      for (j=0;j<procTab[i].preset_num;j++) fprintf(file,"in%d&",j);
      for (j=0;j<procTab[i].selfloop_num;j++) fprintf(file,"inout%d&",j);
      fseek(file,-1,SEEK_CUR);
      /* if no transitions in preset -> 'enabled:=1;' */
      if (procTab[i].preset_num+procTab[i].selfloop_num == 0)
	fprintf(file,"1");
      fprintf(file,";\n");

      /* model 1: add 'FAIRNESS running' */
      if (!alternative) fprintf(file,"FAIRNESS\nrunning\n\n");
    }
}

/**************************************************************************/
/*                           write_SMV3                                   */
/**************************************************************************/
/* write SMV representation for net model 3 to file                       */
/**************************************************************************/

void write_SMV3( LMPNet net, FILE *file)
{
  Cardinal32 i,j,k;
  Place pl;
  Transition tr;
  TSEdge ts;
  STEdge st;
  /* table for type of transition modules to be generated */
  /* (dependent on number of places in preset) */
  Cardinal32 preset_num[MAXTABSIZE]; 
  Cardinal32 cur_preset_num;
  Cardinal32 p=0; /* Position in Table */
  bool found;

  /* --- header --- */ 
  fprintf(file,"-- generated by PEP2SMV from %s (model 3)\n",NNinput_file);
  fprintf(file,"-- (C) 1996-1997 Guido Wimmel\n\n");

  /* --- MODULE main */
  /* --- VAR declaration */
  fprintf(file,"MODULE main\nVAR\nsel_trans:{");
  
  /* sel_trans: set of all transitions */
  /* 'sel_trans: {Ta,Tb,...,0,XXX};' */
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
      fprintf(file,"%s,",tr->name);
  fprintf(file,"0,XXX};\n");
  
  /* Module instances for transitions */
  /* 'Tr: tr_i(...);' */
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
    {
      /* write parameters: places in preset */
      cur_preset_num=0;
      for (st=tr->preset;st;st=st->nextplace) cur_preset_num++;
      fprintf(file,"%s: tr_%d(",tr->name,cur_preset_num);
      for (st=tr->preset;st;st=st->nextplace)
	fprintf(file,"%s.marked,",st->place->name);
      
      /* write parameter: transition name */
      fprintf(file,"%s);\n",tr->name);

      /* Enter transition module type into table, if it is new */
      found=False;
      for (i=0;(!found)&&(i<p);i++) found|=(preset_num[i]==cur_preset_num);
      if (!found) 
	if (p<MAXTABSIZE) preset_num[p++]=cur_preset_num;
	else {fprintf(stderr,"PEP2SMV error: Table full.\n");exit(10);}
    }
  
  /* Module instances for places */
  /* 'Pl: place({trans. in preset},{trans. in postset}, */
  /*            init. marking, sel_trans);' */
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    {
      fprintf(file,"%s: place({",pl->name);
      for (ts=pl->preset,k=0;ts;ts=ts->nexttrans,k++)
	fprintf(file,"%s,",ts->trans->name);

      /* if no trans. in preset, write {XXX}, no transition name will be */
      /* in this set */
      if (k>0) fseek(file,-1,SEEK_CUR); else fprintf(file,"XXX");
      fprintf(file,"},{");
      for (st=pl->postset,k=0;st;st=st->nexttrans,k++)
	fprintf(file,"%s,",st->trans->name);
      if (k>0) fseek(file,-1,SEEK_CUR); else fprintf(file,"XXX");
      fprintf(file,"},%d,sel_trans);\n",pl->marks);
    }

  /* --- ASSIGN declaration --- */
  /* 'sel_trans:=Ta.enabled_name union Tb.enabled_name union ...;' */
  fprintf(file,"ASSIGN\nsel_trans:=");
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
    fprintf(file,"%s.enabled_name union ",tr->name);
  fseek(file,-7,SEEK_CUR);
  fprintf(file,";\n\n"); 

  /* --- SPEC declaration --- */
  copy_spec(file);

  /* --- DEFINE declaration --- */
  /* 'DEFINE deadlock:=!(Ta.enabled|Tb.enabled|...);' */
  fprintf(file,"DEFINE\ndeadlock:=!(");
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
	fprintf(file,"%s.enabled|",tr->name);
  fseek(file,-1,SEEK_CUR);
  fprintf(file,");\n");

  /* --- FAIRNESS declaration */
  /* 'FAIRNESS !(sel_trans=0)|deadlock' */
  fprintf(file,"FAIRNESS\n!(sel_trans=0)|deadlock\n\n");

  /* --- MODULE definition for place modules */
  fprintf(file,
	  "MODULE place(from_trans,to_trans,initmark,sel_trans)\n" \
	  "VAR\nmarked: boolean;\n" \
	  "ASSIGN\n" \
          "init(marked):=initmark;\n"\
          "next(marked):=case\n" \
	  "sel_trans in from_trans: 1;\n"\
          "sel_trans in to_trans: 0;\n" \
	  "1: marked;\n"\
          "esac;\n");
  
  /* --- MODULE definitions for transition modules */
  /* 'MODULE tr_i(mark0,...,mark(i-1),name) */
  /*  DEFINE */
  /*   enabled:=mark0&...&mark(i-1); */
  /*   enabled_name:=case */
  /*    enabled: name; */
  /*    1: 0; */
  /*   esac;' */      
  fprintf(file,"\n");
  for (i=0;i<p;i++)
    {
      fprintf(file,"MODULE tr_%d(",preset_num[i]);
      for (j=0;j<preset_num[i];j++) fprintf(file,"mark%d,",j);
      fprintf(file,"name)\nDEFINE\nenabled:=");
      for (j=0;j<preset_num[i];j++) fprintf(file,"mark%d&",j);
      fseek(file,-1,SEEK_CUR);
      if (preset_num[i]==0) fprintf(file,"1");
      fprintf(file,
	      ";\nenabled_name:=case\nenabled: name;\n1:0;\nesac;\n\n");
    }
}

/**************************************************************************/
/*                       read_components                                  */
/**************************************************************************/
/* read a componentfile (file name: componentfilename)                    */
/* add pointers to component list to the places in the net                */
/* return pointer to first component, NULL if file does not exist         */
/* format of component file: one line per component, places in component  */
/*  separated by whitespace                                               */
/**************************************************************************/

Component *read_components( LMPNet net, char *compfilename)
{
  FILE *f;
  Place pl;
  char buf[BUFSIZE]; /* buffer for an input line */
  char *pl_name;
  /* first_comp, last_comp: first and last components in the net */
  Component *comp,*first_comp,*last_comp;
  ComponentList *complist;
  PlaceList *placel, *first_placel,*last_placel;
  Cardinal32 Componentid=0;

  /* set 'component' pointer to NULL for each place */
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace) pl->components=NULL;
  /* net does not have any components yet */
  first_comp=last_comp=NULL;

  /* --- open component file --- */
  if (compfilename==NULL) return NULL;
  if (!(f=fopen(compfilename,"r"))) return NULL;
  fprintf(stderr,"Reading component file\n");

  /* --- read components lines --- */
  fgets(buf,BUFSIZE,f);	/* PEP       */
  fgets(buf,BUFSIZE,f); /* S_List    */
  fgets(buf,BUFSIZE,f); /* FORMAT_N2 */
  fgets(buf,BUFSIZE,f); /* LIST      */
  fgets(buf,BUFSIZE,f);
  while (!feof(f)) /* read lines until end of file */
    {
      /* Create new Component */
      comp = new Component;
      comp->id = Componentid++;
      comp->next=NULL;
      sprintf(comp->name,"comp_%d",comp->id);

      /* add component to the component list (first_comp,last_comp)  */
      if (first_comp==NULL) first_comp=comp;
      else last_comp->next=comp;
      last_comp=comp;
      
      /* Create empty place list */
      first_placel=last_placel=NULL;
      
      /* Read one place at a time */
      /* (strtok() used to read place names delimited by space,\n,\t,\f) */
      pl_name=strtok(buf,",\"\n\t\f");
      while (pl_name!=NULL)
	{
	  /* find place name in the net, get pointer to it */
	  /* (find_P: function from neunetze.c) */
	  pl=find_P(net,pl_name);  
	  if (pl!=NULL) /* place name found */
	    {
	      /* add entry to place list */
	      placel=new PlaceList;
	      placel->pl=pl;
	      placel->next=NULL;
	      if (first_placel==NULL) first_placel=placel;
	      else last_placel->next=placel;
	      last_placel=placel;
	      
	      /* add Component to Component list for this place */
	      complist=new ComponentList;
	      complist->comp=comp;
	      complist->next=pl->components;
	      pl->components=complist;
	    }
	  else /* report error */
	    fprintf(stderr,"PEP2SMV warning: place %s in component %s "\
		    "does not exist.\n",buf,comp->name);
	  /* get next place name */
	  pl_name=strtok(NULL,",\"\n\t\f");
	}
      
      /* add place list to component */
      comp->places=first_placel;
  
      /* read next line from component file */
      fgets(buf,BUFSIZE,f);
    }
  
  /* close component file, return first component in the net */
  fclose(f);
  return first_comp;
}

/**************************************************************************/
/*                 print_componentInfo                                    */
/**************************************************************************/
/* print information about components in the net:                         */
/* - components and places that belong to them                            */
/* - for all places: components they belong to                            */
/* verify if components are really one-token-P-invariants, and report     */
/*  error if they are not                                                 */
/**************************************************************************/

void print_componentInfo(LMPNet net,Component *firstcomp)
{
  Place pl;
  Transition tr;
  TSEdge ts;
  STEdge st;
  Component *comp;
  ComponentList *complist;
  PlaceList *placel;
  int i;
  
  /* --- print all components and their place list --- */
  printf("Components information:\n");
  for (comp=firstcomp;comp!=NULL;comp=comp->next)
    {
      printf("Component %s (%d): ",comp->name,comp->id);
      /* count how many places are initially marked in this component */
      i=0;
      for (placel=comp->places;placel!=NULL;placel=placel->next)
	{
	  printf(" %s",placel->pl->name);
	  if (placel->pl->marks>0) i++;
	}
      printf("\n");
      /* report error if not exactly one place is marked */
      if (option_use_pseudo && i>1)
      {
        printf("---Error: %d Places initially marked in "
		       " component %s, should be 0 or 1!\n",i,comp->name);
	exit(10);
      }
      else if (!option_use_pseudo && i!=1)
      {
	printf("---Error: %d Places initially marked in "
		       " component %s, should be 1!\n",i,comp->name);
	exit(10);
      }
    }
  
  /* --- print all places and the components they are in --- */
  printf("For places:\n");
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    {
      printf("Place %s: ",pl->name);
      for (complist=pl->components;complist!=NULL;complist=complist->next)
	printf(" %s (%d)",complist->comp->name,complist->comp->id);
      printf("\n");
    }

  /* --- verify if components are one-token-invariants: --- */
  /* for all transitions: */
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
    {
      /* set visited to 0 for all components */
      for (comp=firstcomp;comp!=NULL;comp=comp->next) comp->visited=0;

      for (ts=tr->postset;ts;ts=ts->nextplace)
	/* for every place in the preset, increase visited */
	/*  for every component it is in */
	for (complist=ts->place->components;complist!=NULL;complist=complist->next) 
	  complist->comp->visited++;
      for (st=tr->preset;st;st=st->nextplace)
	/* for every place in the postset, decrease visited */
	/* for every component it is in */
	for (complist=st->place->components;complist!=NULL;complist=complist->next) 
	  complist->comp->visited--;
    
      /* if visited is not 0 for a component, there were not as many */
      /* places in this component in the preset as there were in the */
      /* postset --> report error */
      for (comp=firstcomp;comp!=NULL;comp=comp->next) 
	if (option_use_pseudo && abs(comp->visited) > 1
	    || !option_use_pseudo && comp->visited)
	{
	  printf("---Error: %s is not a Component (transition %s)\n",
		 comp->name,tr->name);
	  exit(10);
	}
    }
}

/**************************************************************************/
/*                           write_SMV4                                   */
/**************************************************************************/
/* write SMV representation for net models 4-7 to file                    */
/*  compress: use 'unchanged' compression algorithm (with defines)        */
/*  selfloop_deadlock: add 'selfloop' formula in deadlock state           */
/*                                                                        */
/*  compress  selfloop_deadlock  model                                    */
/*     f              t            4                                      */
/*     t              t            5                                      */
/*     f              f            6                                      */
/*     t              f            7                                      */
/*                                                                        */
/* if componentfilename!=NULL, make use of components contained in        */
/* the componentfile (create component variables instead of place         */
/* variables)                                                             */
/**************************************************************************/

void write_SMV4( LMPNet net, FILE *file, bool compress, 
		 bool selfloop_deadlock, char *componentfilename)
{
 Place pl;
  Transition tr;
  TSEdge ts;
  STEdge st;
  Cardinal32 i=0;
  ChangeLog *cl;

  Component *firstcomp,*comp;
  ComponentList *complist;
  PlaceList *placel;
  
  /* --- read component file, if specified --- */
  firstcomp=read_components(net,componentfilename);
  if ((firstcomp==NULL)&&(componentfilename!=NULL)) 
    {
      fprintf(stderr,
	 "PEP2SMV warning: component file doesn't exist or wrong format!\n");
      for (pl=net->Places;pl!=NULL;pl=pl->nextplace) pl->components=NULL;
    }
  /* --- print information on components --- */
  /*     and verify if they are one-token-P-invariants */
  print_componentInfo(net,firstcomp);
  
  /* --- write header --- */
  fprintf(file,"-- generated by PEP2SMV from %s ",NNinput_file);
  if (!compress&selfloop_deadlock) fprintf(file,"(model 4)\n");
  else if (compress&selfloop_deadlock) fprintf(file,"(model 5)\n");
  else if (!compress&!selfloop_deadlock) fprintf(file,"(model 6)\n");
  else if (compress&!selfloop_deadlock) fprintf(file,"(model 7)\n");
  fprintf(file,"-- (C) 1996-1997 Guido Wimmel\n\n");
  
  /* --- MODULE main --- */
  fprintf(file,"MODULE main\nVAR\n");

  /* --- initialisation --- */
  /* determine number of components */
  i=0;
  for (comp=firstcomp;comp!=NULL;comp=comp->next) i++;

  /* assign numbers to place variables (places not in components) */
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    if (pl->components==NULL) pl->numname=i++;

  /* create ChangeLog object (i=total number of variables) */
  cl = new ChangeLog(i,compress);

  /* --- VAR declaration --- */

  /* place variables: 'Pi: boolean;' (Pi not in component) */
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    if (pl->components==NULL)
      {
	fprintf(file,"%s: boolean;\n",pl->name);
	pl->visited_pre=pl->visited_post=NULL; /* for selfloop det. */
	/* set name of place variable in ChangeLog object */
	cl->setName(pl->numname,pl->name);
      }

  /* component variables: 'comp_x: {m_Pi,...,m_Pj}; */
  for (comp=firstcomp;comp!=NULL;comp=comp->next)
    {
      fprintf(file,"%s: {",comp->name);
      for (placel=comp->places;placel!=NULL;placel=placel->next)
	fprintf(file,"m_%s,",placel->pl->name);
      if (option_use_pseudo) fprintf(file,"unmarked,");	/* SSCH */
      fseek(file,-1,SEEK_CUR);
      fprintf(file,"};\n");
      /* set name of component variable in ChangeLog object */
      cl->setName(comp->id,comp->name);
    }

  /* set visited to 0 for all components to determine initially unmarked
     pseudo-invariants */
  for (comp=firstcomp;comp!=NULL;comp=comp->next) comp->visited=0;

  /* --- INIT declaration --- */
  fprintf(file,"INIT\n(");
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    /* place variables: Pi=0 or Pi=1 (initial marking) */
    if (pl->components==NULL)
      fprintf(file,"%s=%d&",pl->name,pl->marks);
    else
      /* component variables: 'comp_x=m_Pi' if Pi is marked and in */
      /* component comp_x */
      if (pl->marks>0)
	{
	  for (complist=pl->components;complist!=NULL;complist=complist->next)
	  {
	    fprintf(file,"%s=m_%s&",complist->comp->name,pl->name);
	    complist->comp->visited++;
	  }
	}

  /* SSCH: if pseudo component is initially unmarked */
  if (option_use_pseudo)
    for (comp = firstcomp; comp; comp = comp->next)
      if (!comp->visited) fprintf(file,"%s=unmarked&",comp->name);

  fseek(file,-1,SEEK_CUR);
  fprintf(file,")\n");
  
  /* --- TRANS declaration --- */
  fprintf(file,"TRANS\n(\n");
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
    {
      /* --- one subformula TRANSt for each transition --- */

      /* 'Tr.enabled &' */
      fprintf(file,"--%s\n %s.enabled&",tr->name,tr->name);
      
      /* set visited to 0 for all components in order to detect whether
	 a component is left unmarked after the transition is fired    */
      for (comp=firstcomp;comp!=NULL;comp=comp->next) comp->visited=0;

      /* Mark places in preset of current trans. to recognize selfloops */
      for (st=tr->preset;st;st=st->nextplace)
	st->place->visited_pre=tr;
	
      /* 'next(Place_x)=1' for all places in postset not in component */
      /* 'next(Component_x)=m_Place_x' if Place_x is in a component */
      for (ts=tr->postset;ts;ts=ts->nextplace)
	{
	  ts->place->visited_post=tr; /* to recognize selfloops in preset */

	  for (complist=ts->place->components; complist; complist=complist->next)
		complist->comp->visited++;

	  if (ts->place->visited_pre!=tr) /* if place is not in selfloop */
	    {
	      if (ts->place->components==NULL)
		{
		  fprintf(file,"next(%s)=1&",ts->place->name);

		  /* call logChange, so this place variable is not */
		  /* included in the 'unchanged' expressions next(Pi)=Pi */
		  cl->logChange(ts->place->numname);
		}
	      else
		  for (complist=ts->place->components;complist!=NULL;
		       complist=complist->next)
		    {
		      fprintf(file,"next(%s)=m_%s&",
			      complist->comp->name,ts->place->name);
		      /* call logChange for component variable */
		      cl->logChange(complist->comp->id);
		    }
	    }
	}
            
      /* 'next(Place_x)=0' for all places in preset, if not in component */
      for (st=tr->preset;st;st=st->nextplace)
	if ((st->place->visited_post!=tr)&&(st->place->components==NULL))
	  {
	    fprintf(file,"next(%s)=0&",st->place->name);
	    cl->logChange(st->place->numname);
	  }
      
      /* SSCH: if firing the transition causes the pseudo component
	       to be left unmarked					*/
      if (option_use_pseudo)
        for (st=tr->preset;st;st=st->nextplace)
          for (complist=st->place->components; complist; complist=complist->next)
            if (!complist->comp->visited)
	    {
              fprintf(file,"next(%s)=unmarked&",complist->comp->name);
	      cl->logChange(complist->comp->id);
	    }

      /* generate 'unchanged' expressions next(Pi)=Pi&... */
      /* (or u_Pi_Pj&... if compression algorithm is used) */
      cl->genUnchanged(file);
      fseek(file,-1,SEEK_CUR);

      /* next subformula, joined with '|' (OR) */
      fprintf(file,"\n|\n");
    }

  /* for selfloops in deadlock state (models 6 and 7): */
  /* add 'deadlock&unchanged_(all variables)' */
  if (selfloop_deadlock)
    {
      fprintf(file,"-- selfloop for deadlocks\n");
      fprintf(file,"deadlock&");
      cl->genUnchanged(file);
      fseek(file,-1,SEEK_CUR);
      fprintf(file,"\n");
    } else {
      fseek(file,-2,SEEK_CUR);
    }
  fprintf(file,")\n");
  
  /* --- DEFINE declaration ---  */
  fprintf(file,"DEFINE\n");
  /* 'Trans_x.enabled:=Place_x&Place_y&...;' (places in preset) */
  /* 'Trans_x.enabled:=1;' if no places in preset */
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
    {
      fprintf(file,"%s.enabled:=",tr->name);
      for (i=0,st=tr->preset;st;i++,st=st->nextplace)
	fprintf(file,"%s&",st->place->name);
      if (i!=0) fseek(file,-1,SEEK_CUR); else fprintf(file,"1");
      fprintf(file,";\n");
    }

  /* defines for places in components */
  /* 'Place_x:=Component_x=m_Place_x;' */
  /* (Component_x is first component containing Place_x) */
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    if (pl->components!=NULL)
      fprintf(file,"%s:=%s=m_%s;\n",
	      pl->name,pl->components->comp->name,pl->name);

  /* Deadlock formula */
  /* 'deadlock:=!(Trans_1.enabled|...|Trans_n._enabled);' */
  fprintf(file,"deadlock:=!(");
  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
    fprintf(file,"%s.enabled|",tr->name);
  fseek(file,-1,SEEK_CUR);
  fprintf(file,");\n");

  /* generate defines for 'unchanged' expressions (if compression used) */
  cl->genDefines(file);
    
  /* --- SPEC declaration --- */
  copy_spec(file);

  /* free ChangeLog object */
  delete cl;
}

/**************************************************************************/
/*                           main                                         */
/**************************************************************************/           

int main( int argc, char *argv[] )
{
    LMPNet net; 
    int i;
    char *cp;
    char outputfilename[256]="";

    /* --- initialise array smv_options: 'smv outputfilename' --- */
    /*     (for option -c) */
    smv_options[0]="smv";
    smv_options[1]=outputfilename;
    smv_options[2]=NULL;

    /* --- parse command line --- */
    myname=*(argv++);
    argc--;

    /* print usage if no command line parameters given */   
    if (argc==0)
      {
	printf( 
	"PEP2SMV - Converter PEP->SMV\n"\
	"(C) 1996/97 G. Wimmel, University of Newcastle - TU Muenchen\n\n"\
	"Usage: %s inputfile [-f=<specification file name>] "\
	"[-o=<output file name]\n"\
	" [-m=<model>] [-s=<specification>] "\
	" [-b=<component file name>] "\
	" [-c{heck}] [-p{seudo}] [-- <smv options>]\n\n"\
	"Models:\n"\
	" 1: transitions as processes, FAIRNESS running for transition\n"\
	" 2: transitions as processes, global FAIRNESS\n"\
	" 3: places and transitions as modules\n"\
	" 4: explicit transition relation, selfloops in deadlock states\n"\
	" 5: same as 4, 'next(x)=x' expressions compressed (default)\n"\
	" 6/7: same as 4,5; no selfloops in deadlock states\n",
	myname);
	exit ( 1 );
      }

    /* read command line options */
    for (;argc>0;argc--,argv++)
      {
	if ((*argv)[0]=='-')
	  {
	    if (strncmp(*argv,"-f=",3)==0)
	      specfilename=*argv+3;
	    else if (strncmp(*argv,"-b=",3)==0)
	      componentfilename=*argv+3;
	    else if (strncmp(*argv,"-o=",3)==0)
	      strcpy(outputfilename,*argv+3);
	    else if (strncmp(*argv,"-m=",3)==0)
	      model=atoi(*argv+3);
	    else if (strncmp(*argv,"-s=",3)==0)
	      specification=*argv+3;
	    else if (strncmp(*argv,"-c",2)==0)
		option_run_smv=1;
	    else if (!strcmp(*argv,"-p"))	/* SSCH */
		option_use_pseudo=1;
	    else if (strncmp(*argv,"--",2)==0)
	      {
		argc--;argv++;
		for (i=1;(argc>0)&&(i<18);argc--)
		  smv_options[i++]=*argv++;
		smv_options[i++]=outputfilename;
		smv_options[i]=NULL;
	      }
	    else
	      {fprintf(stderr,"Illegal option: %s.\n",*argv);exit(10);}
	  }
	else
	  {
	    if (filename==0) filename=*argv;
	    else
	      {fprintf(stderr,"Command line error.\n");exit(10);}
	  }
      }
    if (filename==0) {
      fprintf(stderr,"Input file name missing.\n");exit(10);
    } 

    /* --- read Petri net file --- */
    CM_use_imarks = True; /* store initial marking in struct Place->marks */
    net = read_PEP( filename );
    if (place_number(net)+trans_number(net)==0)
      {fprintf(stderr, "PEP2SMV error: net must have at least" \
	       "one place and one transition.\n");exit(10);}
    
    /* --- generate output file name (extension .smv) if necessary */
    if (strlen(outputfilename)==0)
      {
	strcpy(outputfilename,filename);
	cp=strrchr(outputfilename,'.');
	if (cp==NULL) strcat(outputfilename,".smv");
	else strcpy(cp,".smv");
      }

    /* --- open output file --- */
    if ( ! ( outputfile = fopen(outputfilename, "w" ) ) )
      {
	fprintf(stderr,"PEP2SMV error: "\
		"Could not open file %s for writing\n",outputfilename);
	exit(10);
      }

    /* --- generate SMV representation for one of the models --- */
    switch(model)
      {
      case 1: 
	write_SMV1(net,outputfile,false);break;
      case 2: 
	write_SMV1(net,outputfile,true);break;
      case 3: 
	write_SMV3(net,outputfile);break;
      case 4: 
	write_SMV4(net,outputfile,false,true,
		   componentfilename);break;
      case 5: 
	write_SMV4(net,outputfile,true,true,
		   componentfilename);break;
      case 6: 
	write_SMV4(net,outputfile,false,false,
		   componentfilename);break;
      case 7: 
	write_SMV4(net,outputfile,true,false,
		   componentfilename);break;
      default: 
	fprintf(stderr,"PEP2SMV error: illegal model.\n");exit(10);
      }

    /* --- close output file if necessary --- */
    if (outputfile!=stdout) fclose(outputfile);
    /*
      remove_net( &net );
      */

    /* --- if option -c was given, execute smv --- */
    if (option_run_smv)
      execvp("smv",smv_options);

    exit ( 0 );
} /* main() */


/**************************************************************************/
/* END OF FILE pep2smv.c                                                  */
/**************************************************************************/

