/**************************************************************************/
/* CSC399 Individual Project, University of Newcastle upon Tyne           */
/* PEP2SMV - Converter from PEP .ll_net files to SMV specifications       */
/* (C) Oct. 1996 - Apr. 1997 Guido Wimmel                                 */
/*                                                                        */
/* transsmv.c - convert SMV output into transition firing sequence        */
/**************************************************************************/

/* changes since 'dissertation' release:                                  */
/* added _SEQUENCE: and _LOOP: output                                     */
/* lines after 'resources used:' are copied to output                     */
/* 19.07.97: marking of the net recalculated after each transition firing */
/*           algorithm for searching transition that has fired changed    */
/*           (mark places with MARKS_UNCHANGED,INCREASE,DECREASE;         */
/*            search all transitions if they fit and are enabled)         */
/*           (otherwise, transsmv cannot determine which transition fired */
/*           in cases  like (P1,P3) -> T1 -> (P2,P3)                      */
/*                          (P1,P4) -> T2 -> (P2,P4)                    ) */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "neunetze.h"
#include "convertPEP.h"

/* flags for change of the marking of a place */
typedef enum {MARKS_UNCHANGED = 0, MARKS_INCREASE, MARKS_DECREASE } placeflag;

/**************************************************************************/
/*                       translate                                        */
/**************************************************************************/
/* read SMV output from 'in', write transition firing sequence for 'net'  */
/*  to 'out'                                                              */
/**************************************************************************/

void translate(LMPNet net, FILE *in, FILE *out)
{
  char buffer[256],var_name[256];
  Place pl;
  Transition tr;
  TSEdge ts;
  STEdge st;
  Cardinal32 n,val;
  bool comma_fl,loop_started;
  bool fired=false;
  int sequence_nr=-1; /* count nr. of sequences output */

  Cardinal32 *initmark = new Cardinal32[net->NumPlaces];
  
  /* --- store the initial marking --- */
  n=0;
  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
    initmark[n++]=pl->marks;

  /* --- read input file line by line --- */
  fgets(buffer,256,in);
  while (buffer && !feof(in)) {
      /* --- Scan for 'specification' line --- */
      if (strstr(buffer,"specification")!=NULL)
	{
	  /* increase sequence nr. */
	  sequence_nr++;
	  if (sequence_nr>0) 
	    {
	      printf("\n");
	      /* restore initial marking */
	      n=0;
	      for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
		pl->marks=initmark[n++];
	    }
	  
	  /* copy "specification ... is true/false" */
	  fputs(buffer,out);           

	  /* --- if specification false, translate given --- */
	  /*     execution sequence into firing sequence     */
	  if (strstr(buffer,"false")!=NULL)
	    {
	      comma_fl=false;
	      /* copy "as demonstrated ..." */
	      fgets(buffer,256,in); fputs(buffer,out); 

	      /* header _SEQUENCE: */
	      if (sequence_nr)
		fprintf(out,"_SEQUENCE%d:\n",sequence_nr);
	      else
		fputs("_SEQUENCE:\n",out);

	      /* --- overread initial state --- */
	      fgets(buffer,256,in); /* state number */

	      /* if execution sequence begins with "-- loop starts here" */
              /* print this line */
	      if (strstr(buffer,"loop")!=NULL)
		{
		  if (sequence_nr)
		    fprintf(out,"\n_LOOP%d:\n",sequence_nr);
		  else
		    fputs("\n_LOOP:\n",out);
		  fgets(buffer,256,in);
		}

	      /* skip initial variable values, until empty line */
	      /* (next state) */
	      while (strlen(buffer)>1)              
		{fgets(buffer,256,in);}

	      /* --- get next state header; if there is a "loop" line -- */
	      /* before, set flag loop_started, output after the */
	      /* transition that caused the state change */
	      fgets(buffer,256,in);
	      if (loop_started=(strstr(buffer,"loop")!=NULL))           
		fgets(buffer,256,in);
	      
	      /* --- for every new state in execution sequence: --- */
	      /* find transition that has fired */
	      while (strstr(buffer,"state")!=NULL)
		{
		  /* set numname to MARKS_UNCHANGED for every place */
		  /* set visited_pre and visited_post (selfloop det.) */
		  for (pl=net->Places;pl!=NULL;pl=pl->nextplace)
		    {
		      pl->numname=MARKS_UNCHANGED;
		      pl->visited_pre=pl->visited_post=NULL;
		    }
		  
		  n=0; /* count number of place variables that changed */

		  fgets(buffer,256,in);

		  /* for every line corresponding to a variable change */
		  while (strlen(buffer)>1)
		    {
		      /* get variable name and new value */
		      sscanf(buffer,"%s = %d",var_name,&val); 
		      
		      /* determine if it is a place variable */
		      if ((pl=find_P(net,var_name))!=NULL)
			{
			  /* store how the marking of this place has changed */
			  if (val==0) 
			    pl->numname=MARKS_DECREASE;
			  else
			    pl->numname=MARKS_INCREASE;

			  n++;
			}
		   
		      fgets(buffer,256,in);
		    } /* while (strlen(buffer)>1) */
		  
		  /* find a transition that can have fired */
		  for (tr=net->Transitions;tr!=NULL;tr=tr->nexttrans)
		    {
		      fired=true;
		      for (st=tr->preset;(st!=NULL)&&fired;st=st->nextplace)
			{
			  st->place->visited_pre=tr; /* selfloop det. */
			  if (st->place->marks==0) fired=false;
			  /* this transition cannot have fired, because
			     it is not enabled */
			}
		      
		      for (ts=tr->postset;(ts!=NULL)&&fired;ts=ts->nextplace)
			{
			  ts->place->visited_post=tr; /* selfloop det. */
			  if (ts->place->visited_pre!=tr) /* no selfloop */
			    {
			      if (ts->place->numname!=MARKS_INCREASE)
				fired=false;
			    }
			  else if (ts->place->numname!=MARKS_UNCHANGED)
			    fired=false;
			}
		      
		      for (st=tr->preset;(st!=NULL)&&fired;st=st->nextplace)
			{
			  if ((st->place->visited_post!=tr)&&
			      (st->place->numname!=MARKS_DECREASE))
			    fired=false;
			}
		      
		      if (fired) /* found a transition ! */
			break;
		    }
		  
		  if (comma_fl) printf(","); else comma_fl=true;
		  
		  if (n==0) /* no change */
		    printf("(none)");
		  else if (tr==NULL) /* no transition found */
		    printf("???");
		  else
		    /* output transition that could have fired */
		    /* (sometimes there can be more than one, only */
		    /* one is output) */
		    {
		      printf("%s",tr->name);
		      
		      /* let transition fire, change marking */
		      for (st=tr->preset;st!=NULL;st=st->nextplace)
			st->place->marks=0;
		      for (ts=tr->postset;ts!=NULL;ts=ts->nextplace)
			ts->place->marks=1;
		    }
		  /* output '-- loop' if a loop started with the current */
		  /* state */
		  if (loop_started) {
		    if (sequence_nr)
		      fprintf(out,"\n_LOOP%d:\n",sequence_nr);
		    else
		      fputs("\n_LOOP:\n",out);
		    comma_fl=false;
		  }
		  
		  /* read next state number, overread "-- loop" line to */
		  /* output it later */
		  fgets(buffer,256,in);
		  if (loop_started=(strstr(buffer,"loop")!=NULL))
		    fgets(buffer,256,in);
		} /* while ("new state in sequence") */
	      
	      if (loop_started) 
		if (sequence_nr)
		  fprintf(out,"\n_LOOP%d:\n",sequence_nr);
		else
		  fputs("\n_LOOP:\n",out);
	      printf("\n");
	    } /* if ("specification false") */
	  else fgets(buffer,256,in);
	} /* if ("specification") */

      else if (strstr(buffer,"resources used")!=NULL)
	/* copy all lines after 'resources used' */
	{
	  fputs("\n",out);
	  while (!feof(in))
	    {fputs(buffer,out);fgets(buffer,256,in);}
	} else 
	  { 
	    /* overread lines not corresponding to a specification */
	    /* fputs(buffer,out); */
	    fgets(buffer,256,in);
	  }
    }
}

/**************************************************************************/
/*                           main                                         */
/**************************************************************************/           

int main( int argc, char *argv[] )
{
    LMPNet net; 
    char *filename=NULL;

    /* print usage if no command line parameters given */
    if (!--argc)
      {
       printf(
	"TransSMV - Translator SMV-Output->Petri net firing sequence\n"\
	"(C) 1996/97 G.Wimmel, University of Newcastle - TU Muenchen\n\n"\
	"Usage: transsmv -n=petri-net-file\n"\
        "       SMV result is read from standard input,\n"\
        "       translation is written to standard output.\n");
	exit(1);
      }

    /* read command line options */
    for (++argv;argc>0;--argc,++argv)
      {
	if ((*argv)[0]=='-')
	  {
	    if (strncmp(*argv,"-n=",3)==0)
	      filename=*argv+3;
	    else
	      {
		fprintf(stderr,"Illegal option: %s.\n",*argv);
		exit(10);
	      }
	  }
	else
	  {
	    fprintf(stderr,"Command line error.\n");
	    exit(10);
	  }
      }
    
    /* --- read Petri net file */
    CM_use_imarks = True; /* store initial marking in struct Place */
    net = read_PEP( filename );
    if (place_number(net)+trans_number(net)==0)
      {fprintf(stderr, "PEP2SMV error: net must have at least" \
	       "one place and one transition.\n");exit(10);}
    
    /* --- translate Petri net file --- */
    translate(net,stdin,stdout);

//    remove_net( &net );
        
    exit ( 0 );
} /* main() */

/**************************************************************************/
/* END OF FILE transsmv.c                                                 */
/**************************************************************************/
