# RPC-IO.tcl:
# RPC read/write routines for server and client sides
#
# $Id: RPC-IO.tcl,v 1.1 2001/07/11 12:21:49 pep Exp $
# $State: Exp $
#
# Copyright (c) 1996 Michael E. Shorter (mike@globalware.com)
#
# See the file "LICENSE.TERMS" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

global rpc_vars;
if [info exists rpc_vars(sourced,RPC_IO_tcl_)] {
    return;
}

package provide Tcl_RPC 1.3.0;

if {("[info proc init_var]" == "init_var") || ([auto_load init_var] == 1)} {
    init_var rpc_vars(sourced,RPC_IO_tcl_) 1;
    init_var rpc_vars(version,RPC_IO_tcl_) 1.3.0;
}

proc RPC_read {chan code cmd except} {
    global rpc_vars;

    upvar $code   got_code;
    upvar $cmd    got_cmd;
    upvar $except exception;

    # this is for the logging service
    set myid "<RPC_read|$chan|>";

    rpclog 0x0100 "$myid:< starting";

    set exception 1;
    set tries 0;

    if [catch {RPC_ll_read $chan $rpc_vars(MAXCODEBITS)} ret] {
	return -code error "$ret";
    }

    set got_code [string trim $ret];

    if {$got_code == {}} {
	set got_code $rpc_vars(RPC_code,EXCEPTION);
	set exception 1;
	set got_cmd "RPC read error";
	# return NAK to sender
#	puts -nonewline $chan $rpc_vars(RPC_NAK);
#	flush $chan;
	return;
    }

    if [catch {RPC_ll_read $chan $rpc_vars(MAXDATABITS)} ret] {
	return -code error "$ret";
    }
    set got_datasize [string trim $ret];

    if [catch {RPC_ll_read $chan $got_datasize} ret] {
	return -code error "$ret";
    }
    set got_cmd      $ret;

    rpclog 0x0200 "$myid:< fconfigure $chan => [fconfigure $chan]";
    rpclog 0x0200 "$myid:< code        => $got_code";
    rpclog 0x0200 "$myid:< datasize    => $got_datasize";
    rpclog 0x0400 "$myid:< string      => $got_cmd";

#    rpclog "$myid:< finshed reading, sending confirmation";
#    puts -nonewline $chan $rpc_vars(RPC_ACK);
#    flush $chan;

    set exception 0;
    rpclog 0x0100 "$myid:< done";
    return;
}

proc RPC_ll_read {sock bytes} {
    global rpc_vars;
    global rpcreadtimeouts;

    if {$bytes == 0} {
	return "";
    }

    set myid "<RPC_ll_read|$sock|>";

    if {$rpc_vars(RPC_BLOCKINGIO) == 1} {
	rpclog 0x0100 "$myid:blocking read $bytes bytes";
	return [read $sock $bytes];
    } else {
	set total_bytes_read 0;
	set bytes_to_read $bytes;
	set reads 0;
	set null_reads 0;
	set data "";

	rpclog 0x0100 "$myid:ready to read $bytes bytes";

	set mytid [hrt];
	set timeout $rpc_vars(RPC_READ_TIMEOUT);
	set rpc_vars(timeout,$mytid) 0;
	set afterret [after $timeout "set rpc_vars(timeout,$mytid) 1"];

	while {$reads <= $rpc_vars(RPC_MAX_READS)} {
	    incr reads;
	    if [eof $sock] {
		after cancel $afterret;
		return -code error "eof error";
	    }

	    if [catch {read $sock $bytes_to_read} data_read] {
		after cancel $afterret;
		return -code error "$data_read";
	    }
	    if {$rpc_vars(timeout,$mytid) == 1} {
		return -code error "read timeout";
	    }
	    set bread [string length "$data_read"];
	    if {$bread == 0} {
		incr null_reads;
		if {$null_reads == $rpc_vars(MAXNULLREADS)} {
		    after cancel $afterret;
		    return -code error "read timeout, channel stalled";
		}
	    } else {
		# read some bytes
		set total_bytes_read [expr $total_bytes_read + $bread];
		set bytes_to_read    [expr $bytes_to_read - $bread];
		set data "$data$data_read";
		rpclog 0x0400 "$myid:read $bread bytes, $total_bytes_read\
                               bytes total, \"$data\"";

		if {$total_bytes_read == $bytes} {
		    after cancel $afterret;
		    rpclog 0x0100 "$myid:returning \"$data\"";
		    return "$data";
		}
	    }
	    delay $rpc_vars(RPC_READ_TIMEOUT);
	}
	after cancel $afterret;
	return -code error "max read timeout";
    }
}

proc RPC_write {chan code result except} {
    global rpc_vars;

    upvar $except exception;

    set exception 1;

    # this is for the logging service
    set myid "<RPC_write|$chan|>";

    set datasize [string length $result];

    rpclog 0x0100 "$myid:writing...";

    rpclog 0x200 "$myid:> fconfigure $chan => [fconfigure $chan]";

    rpclog 0x0200 "$myid:> code     => $code";
    rpclog 0x0200 "$myid:> datasize => $datasize";
    rpclog 0x0400 "$myid:> string   => $result";

    if [catch {puts -nonewline $chan "[format $rpc_vars(format_codebits)\
                                      $code]"} ret] {
	rpclog 0x0002 "$myid:> write error";
	return -code error "$ret";
    }

    if [catch {puts -nonewline $chan "[format $rpc_vars(format_databits)\
                                       $datasize]$result"} ret] {
	set exception 1;
	rpclog 0x0002 "$myid:> write error";
	return -code error "$ret";
    }

    if [catch "flush $chan" ret] {
	rpclog 0x0002 "$myid:> flush error";
	set exception 1;
	return -code error "$ret";
    }

#    rpclog "$myid:finished writing, waiting for confirmation";
    # wait for acknowledgment.  block forever since we're in blocking mode.
    # think about redoing all of this using non-blocking IO...
#    if [catch {RPC_ll_read $chan 1} ack] {
#	return -code error "$ack";
#    }

    # just in case...
#    if {$ack != $RPC_ACK} {
#	if {$ack == $RPC_NAK} {
#	    rpclog "$myid: received negative acknowledgment from writer";
#	}
#	rpclog "$myid: didn't receive ACK/NAK";
#	rpclog "$myid:done prematurely";
#	set exception 1;
#	return;
#	return -code 1 "didn't receive acknowledgment of request";
#    }

    rpclog 0x0100 "$myid:done";

    set exception 0;
    return;
}

# Local Variables:
# mode:tcl
# End:
