/******************************************************************************/
/* File sdl.y                                                                 */
/* Bison source for SDL parser						      */
/* 1998 by Stefan Schwoon                                                     */
/******************************************************************************/

%{
#include "function.h"		/* memory allocation functions */
#include "main.h"		/* important definitions */
#include "symtab.h"		/* functions for handling the symbol table */
#include "node.h"
#include "sortrange.h"		/* predefined types */
#include "actions.h"
#include "ref.h"		/* the reference component */
#include "info.h"		/* for sdl_info files */
#include "property.h"		/* for sdl_prop files */
#include "strbuf.h"		/* dynamically growing string buffer */

/*****************************************************************************/

int num_pi_ids=0;	/* total number of process ids in system */
int yylex();

/* these are used by the rules for processes and procedures */
T_list create_list;
T_list called_procs;
T_list created_procs;

ref_instance tmpref;

/*****************************************************************************/

/* indicate parse error */
int yyerror(char *msg)
{
	Msg(ERR+64,msg); /* +64 prevents double incrementation of yynerrs */
	return 1;
}


/* check if name after END<keyword> matches the name after <keyword> */
void check_end(T_symbol sym, T_ident ident, char *keyword)
{
	if (!ident || ident->sym == sym) return;
	Msg(ERR,"named END%s does not match start",keyword);
}


void copy_ref_start(T_ref dst, T_ref src)
{
	dst->start_l = src->start_l;
	dst->start_c = src->start_c;
}


void copy_ref_end(T_ref dst, T_ref src)
{
	dst->end_l = src->end_l;
	dst->end_c = src->end_c;
}


/* access sdl_info database to retrieve information
   about a certain channel				*/
void find_channel_props(T_relay relay, char *name)
{
	initbuf(); addbuf(name); addbuf(".capacity");
	relay->capacity = prop_find_key(strbuf,sdl_defcap);

	initbuf(); addbuf(name); addbuf(".version");
	relay->version = prop_find_key(strbuf,sdl_chanver);
}

%}

%union {
	T_symbol symbol;	/* entry in the symbol table */
	T_ident idptr;
	T_list list;		/* generic list type */
	T_transition trans;	/* list of actions */
	T_action action;
	T_syntype type;		/* user-defined type */
	T_node node;		/* used for expressions */
	intset set;		/* set of integers */
	void* ptr;		/* generic pointer */
	ident_instance ident;
	ref_instance ref;	/* reference */
	int value;
}

/* most tokens are annotated with a reference to their
   occurence in the source code				*/
%token <ref> SYSTEM ENDSYSTEM
%token <ref> BLOCK ENDBLOCK
%token <ref> REFERENCED
%token <ref> CHANNEL ENDCHANNEL
%token <ref> NODELAY
%token <ref> FROM TO WITH
%token <ref> SIGNAL
%token <ref> CONNECT
%token <ref> PROCESS ENDPROCESS
%token <ref> START STATE INPUT
%token <ref> OUTPUT
%token <ref> NEXTSTATE STOP
%token <ref> VIA
%token <ref> SYNONYM SYNTYPE ENDSYNTYPE
%token <ref> DEFAULT CONSTANTS
%token <ref> DCL
%token <ref> TRUE FALSE NULL_PID
%token <ref> TASK ASSIGN
%token <ref> IMPLIES NE GE LE
%token <ref> OR XOR AND MOD NOT REM
%token <ref> DECISION ANY ELSE ENDDECISION
%token <ref> PROCEDURE ENDPROCEDURE CALL RETURN
%token <ref> IN OUT
%token <ref> CREATE
%token <ref> INPUT_NONE
%token <ref> JOIN

%token <value> CONST
%token <ident> IDENT
%token <ref> '(' COMMENT ';' ')' ',' ':'


/* priorities of operators in expressions */
%left IMPLIES
%left OR XOR
%left AND
%nonassoc '=' NE '>' GE '<' LE
%left '+' '-'
%left '*' '/' MOD REM
%right NOT unaryminus


%type <ref> end
%type <symbol> spec system syshead
%type <list> sysbodylist
%type <symbol> sysbodyentity

%type <symbol> block blkhead
%type <list> blkbodylist
%type <symbol> blkbodyentity

%type <symbol> channel
%type <value> delay
%type <ptr> path

%type <symbol> sigdef singlesigdef
%type <list> sigdeflist sortlist

%type <symbol> connect relayreferident relaydeclident
%type <list> relayreferlist relaydecllist

%type <idptr> optname
%type <list> siglist
%type <symbol> sigident

%type <symbol> data syndef syntypedef
%type <symbol> syndeflist singlesyndef
%type <value> optdefault const range
%type <value> negconst sigconst 
%type <list> optconstants rangelist noteq
%type <type> osyntype syntype

%type <symbol> dcl varident
%type <list> dcllist singledcl varlist

%type <symbol> process prcshead body
%type <value> numinst onum
%type <list> prcsdcllist prcsfmlparms 
%type <symbol> prcsdclentity

%type <symbol> procedure prcdhead
%type <list> prcddcllist
%type <symbol> prcddclentity
%type <list> prcdfmlparms prcdparmlist singleparmdcl
%type <value> parmtype

%type <trans> start trigger inputpart
%type <list> statespeclist statespec statelist simple_statelist
%type <symbol> stateident
%type <list> triggerlist
%type <list> term_transition transition tlist

%type <list> usiglist sigarglist
%type <ptr> singleusig singlesigarg

%type <action> stmt action
%type <action> output task
%type <action> decision call create
%type <action> terminator nextstate
%type <action> stop stoporreturn join 
%type <symbol> label

%type <symbol> toclause
%type <ptr> viaclause
%type <list> vialist
%type <symbol> viaident

%type <list> assignlist
%type <ptr> assignment
%type <value> constexpr
%type <node> expr

%type <node> question
%type <list> answerlist answer
%type <list> optrangelist

%type <list> optparms actparmlist

%start spec

%%

/*****************************************************************************/
/* rules start here							     */
/*****************************************************************************/

/* General remarks:
   Handling the symbol table entries is a good deal more complicated than
   in BPN2 since entries can be referenced before they are actually defined
   or even declared.
   The parser uses these routines (defined in symtab.c) to access information
   in the symbol table:

	define_ident: Treat this occurence of a symbol as the defining
		      occurence. A new symbol is created, carrying a given
		      identifier and type. The current level of the
		      specification may not contain another definition of 
		      the same identifier, or a declaration of this identifier
		      with another type.

	decl_ident:   Treat this occurence as a declaration. If no declaration
		      or definition of this identifier is present on the
		      current level, a new symbol is created. Otherwise,
		      the previous declaration/definition must refer to an
		      identifier of the same type as in this declaration.
		      All declarations must be resolved by a definition
		      on the same level of the specification.

	refer_ident:  Treat this occurence as a reference to a symbol that
		      must already be declared (with correct type) in either
		      the current or any enclosing structure.
*/

/**********/
/* SYSTEM */
/**********/

spec : system
	;

end : COMMENT ';' {$$=$2;}
	| ';'	{$$=$1;}
	;

system	: syshead
	  sysbodylist
	  ENDSYSTEM optname end
	  {
		T_structure this = (STree = $$ = $1)->e.struc;
		check_end($$,$4,"SYSTEM");
		revert_list($2);
		this->subtype = TYPE_SYSTEM;
		this->entities = $2;
		this->symtab = dec_blknum();
		this->num_connects = 0;
	  }
	;

syshead	: SYSTEM IDENT end
	  {
		$$ = define_ident(&$2,TYPE_STRUCTURE,alloc_structure());
		inc_blknum();

		/* create `env' as a pseudo-entry in the symbol table
		   so that channels can access it */
		s_lookup("env",&$2);

		define_ident(&$2,TYPE_STRUCTURE,alloc_structure());
		init_default_types();
		num_pi_ids = 2;		/* NULL and 0 */
 		create_list = alloc_list();;
		
		/* enable global procedures to access implicit variables
		   if local procedure copies are activated (default)	 */
		if (!(sdlflag & SDLF_NO_PRCD_COPIES)) init_implicit_vars();

		$$->ref = add_ref($$,REF_SYSTEM,&$1,&$3);
	  }
	;

sysbodylist
	: sysbodyentity sysbodylist
	  {	list_add($$ = $2, $1);	}
	| {	$$ = alloc_list();	}
	;
	
sysbodyentity
	: block
	| channel
	| sigdef
	| data
	| procedure
	;

/*********/
/* BLOCK */
/*********/

block	: blkhead
	  blkbodylist
	  ENDBLOCK optname end
	  {
		T_structure this = ($$ = $1)->e.struc;
		check_end($$,$4,"BLOCK");
		this->subtype = TYPE_BLOCK;
		revert_list(this->entities = $2);
		this->symtab = dec_blknum();
		/* number of connections inside this block */
		this->num_connects = 0;
	  }
	;

blkhead	: BLOCK IDENT end
	  {
		$$ = define_ident(&$2,TYPE_STRUCTURE,alloc_structure());
		inc_blknum();
		
		/* create `env' as a pseudo-entry in the symbol table
		   so that channels can access it */
		s_lookup("env",&$2);

		define_ident(&$2,TYPE_STRUCTURE,alloc_structure());
		$$->ref = add_ref($$,REF_BLOCK,&$1,&$3);
	  }
	;

blkbodylist
	: blkbodyentity blkbodylist
	  {	list_add($$ = $2, $1);	}
	| {	$$ = alloc_list();	}
	;
	
blkbodyentity
	: process
	| channel
	| connect
	| sigdef
	| data
	| procedure
	;

/***********/
/* CHANNEL */
/***********/

channel	: CHANNEL IDENT	delay	/* unidirectional */
	  path ENDCHANNEL optname end
	  {
		T_relay this = (T_relay) $4;
		$$ = define_ident(&$2,TYPE_RELAY,this);
		this->bidirectional = False;
		this->is_channel = $3;

		if($3)
		  find_channel_props(this,$$->name);

		check_end($$,$6,"CHANNEL");
		if($3)
		  $$->ref = add_ref($$,REF_CHANNEL,&$1,&$7);
		else
		  $$->ref = add_ref($$,REF_SIGNALROUTE,&$1,&$7);
	  }
	| CHANNEL IDENT	delay	/* bidirectional */
	  path path
	  ENDCHANNEL optname end
	  {
		T_relay this = (T_relay) $4;
		$$ = define_ident(&$2,TYPE_RELAY,this);
		this->bidirectional = True;
		this->is_channel = $3;

		if($3)
		  find_channel_props(this,$$->name);

		if (this->ep1 != ((T_relay) $5)->ep2 || 
		    this->ep2 != ((T_relay) $5)->ep1)
			Msg(ERR,"Endpoints of bidirectional channel "
				"must be the same in both ways");
		this->siglist2 = ((T_relay) $5)->siglist1;
		free($5);
		check_end($$,$7,"CHANNEL");
		$$->ref = add_ref($$,REF_CHANNEL,&$1,&$8);
		if($3)
		  add_ref($$,REF_CHANNEL,&$1,&$8);
		else
		  add_ref($$,REF_CHANNEL,&$1,&$8);
	  }
	;

delay : {$$=True;} | NODELAY {$$=False;}
	;

/**********/
/* SIGNAL */
/**********/

sigdef	: SIGNAL sigdeflist end
	  {
		/* Only one entry is stored in the entity list of the
		   surrounding structure (system, block).
		   This entry contains a placeholder name as its identifier. */
		T_ident dummy = s_dummy("$$Siglist");
		$$ = define_ident(dummy,TYPE_SIGDEF,$2);
		revert_list($2);
	  }
	;
		
sigdeflist :
	  singlesigdef
	  {	list_add($$ = alloc_list(), $1);	}
	| singlesigdef ',' sigdeflist
	  {	list_add($$ = $3,$1);			}
	;

singlesigdef : IDENT
	  {
		T_signal sig = alloc_signal();
		sig->args = NULL;
		$$ = define_ident(&$1,TYPE_SIGNAL,sig);
		$$->ref = add_ref($$,REF_SIGNAL,&$1.ref,&$1.ref);
	  }
	| IDENT '(' sortlist ')'
	  {
		T_syntype st;
		T_signal sig = alloc_signal();

		/* Are there any unbounded argument types like
		   Integer or Natural?				*/
		for_list ($3,st,T_syntype)
			if (!st->set) Msg(FAT,"cannot use unlimited sort");

		sig->args = $3;
		$$ = define_ident(&$1,TYPE_SIGNAL,sig);
		$$->ref = add_ref($$,REF_SIGNAL,&$1.ref,&$4);
	  }
	;

sortlist: syntype
	  { list_add($$ = alloc_list(), $1);	}
	| syntype ',' sortlist
	  {	list_add($$ = $3,$1);			}
	;

/***********/
/* CONNECT */
/***********/

/* Since connect instructions can occur before the definition of the
   signalroutes they reference, we need to handle signalroutes differently,
   i.e., they will be defined whereas channels are referenced.		   */

connect	: CONNECT relayreferlist AND relaydecllist end
	  {
		T_ident dummy = s_dummy("$$Connect");
		T_connect this = alloc_connect();
		$$ = define_ident(dummy,TYPE_CONNECT,this);
		this->outlist = $2;
		this->inlist = $4;
		$$->ref = add_ref($$,REF_CONNECT,&$1,&$5);
		add_ref($$,REF_CONNECT,&$1,&$5);
	  }
	;

relayreferlist : relayreferident
	  {	list_add($$ = alloc_list(), $1);	}
	| relayreferident ',' relayreferlist
	  {	list_add($$ = $3,$1);			}
	;

relayreferident : IDENT
	  {	$$ = refer_ident(&$1,TYPE_RELAY);	}
	;
		
relaydecllist : relaydeclident
	  {	list_add($$ = alloc_list(), $1);	}
	| relaydeclident ',' relaydecllist
	  {	list_add($$ = $3,$1);			}
	;

relaydeclident : IDENT
	  {	$$ = decl_ident(&$1,TYPE_RELAY);	}
	;

/***********************/
/* general definitions */
/***********************/

optname	: IDENT
	  {
		$$ = alloc_ident();
		memcpy($$,&$1,sizeof(ident_instance));
	  }
	| {	$$ = NULL;	}
	;

path	: FROM IDENT TO IDENT WITH siglist end
	  {
		if ($2.sym == $4.sym)
			Msg(ERR, "endpoints of channel must be different");
		$$ = alloc_relay();
		((T_relay) $$)->ep1 = decl_ident(&$2,TYPE_STRUCTURE);
		((T_relay) $$)->ep2 = decl_ident(&$4,TYPE_STRUCTURE);
		((T_relay) $$)->siglist1 = $6;
		copy_ref_end(&tmpref,&$7);
	  }
	;

siglist	: sigident
	  {	list_add($$ = alloc_list(), $1);	}
	| sigident ',' siglist
	  {	list_add($$ = $3,$1);			}
	;

sigident: IDENT
	  {	$$ = refer_ident(&$1,TYPE_SIGNAL);	}
	;

/********/
/* DATA */
/********/

data	: syndef
	| syntypedef
	;

syndef	: SYNONYM syndeflist end
	  {
		T_ident dummy = s_dummy("$$Synonym");
		$$ = define_ident(dummy,TYPE_CONST,(void*)1);
	  }
	;

syndeflist:
	  singlesyndef
	| singlesyndef ',' syndeflist
	;

singlesyndef: IDENT osyntype '=' constexpr
	  {
		$$ = define_ident(&$1,TYPE_CONST,(void*)$4);
		check_sort($2,$4);
	  }
	;

syntypedef
	: SYNTYPE IDENT '=' syntype
	  optdefault optconstants
	  ENDSYNTYPE optname end
	  {
		T_syntype syn = alloc_syntype();

	    	$$ = define_ident(&$2,TYPE_SYNTYPE,syn);
		syn->def_init = $4->def_init;
		syn->set = $4->set;
		syn->range = $4->range;

		/* intersect optconstants with range of parent type */
		if ($6)
		{
			syn->range = intersect_range($6,syn->range,True);
			free_list($6);
			if (range_isfinite(syn->range))
				syn->set = create_set(syn->range);
		}
			
		if ($5 != NONE) syn->def_init = $5;
		if (syn->def_init != NONE) check_sort(syn,syn->def_init);
		check_end($$,$8,"SYNTYPE");
	  }
	;

optdefault
	: DEFAULT constexpr /* oend */
	  {	$$ = $2;	}
	| {	$$ = NONE;	}
	;

optconstants
	: CONSTANTS '(' rangelist ')' end
	  {	$$ = $3;	}
	| {	$$ = NULL;	}
	;

rangelist: range
	  {	list_add($$ = alloc_list(), $1);	}
	| range ',' rangelist
	  {	list_add($$ = $3,$1);			}
	| noteq
	| noteq ',' rangelist
	  {
		int r;
		for_list ($1,r,int) list_add($3,r);
		free_list($1);
		$$ = $3;
	  }
	;

range	: constexpr
	  {	$$ = create_range($1,$1);		}
	| constexpr ':' constexpr
	  {	$$ = create_range($1,$3);		}
	| '=' constexpr
	  {	$$ = create_range($2,$2);		}
	| '>' constexpr
	  {	$$ = create_range($2+1,MAX_POS);	}
	| GE constexpr
	  {	$$ = create_range($2,MAX_POS);		}
	| '<' constexpr
	  {	$$ = create_range(MIN_NEG,$2-1);	}
	| LE constexpr
	  {	$$ = create_range(MIN_NEG,$2);		}
	;

noteq	: NE constexpr
	  {
		$$ = alloc_list();
		list_add($$,create_range(MIN_NEG,$2-1));
		list_add($$,create_range($2+1,MAX_POS));
	  }
	;

osyntype: syntype
	| {	$$ = NULL;	}
	;

/* As an extension of native SDL syntax, one could allow the
   specification of types {0,1} here.				*/

syntype	: IDENT
	  {	T_symbol tmp = refer_ident(&$1,TYPE_SYNTYPE);
		if (tmp)
		  $$ = tmp->e.syntype;
		else
		{
		  Msg(FAT,"Cannot proceed with undeclared identifier");
		  $$ = NULL;
		}
	  }
	;

const	: CONST
	| TRUE
	  {	$$ = TT;	}
	| FALSE
	  {	$$ = FF;	}
	| NULL_PID
	  {	$$ = DOT;	}
	;

negconst: '-' CONST
	  {
		if (($$ = -$2) < MIN_NEG) Msg(FAT,"constant too low - "
						  "you trying to break me?");
	  }
	;

/*******/
/* DCL */
/*******/

dcl	: DCL dcllist end
	  {
		T_ident dummy = s_dummy("$$Dcllist");
		$$ = define_ident(dummy,TYPE_VARDEF,$2);
		revert_list($2);
	  }
	;

dcllist	: singledcl
	| singledcl ',' dcllist
	  {
		T_symbol sym;
		for_list ($1,sym,T_symbol) list_add($3,sym);
		free_list($1);
		$$ = $3;
	  }
	;

singledcl
	: varlist syntype
	  {
		T_symbol sym;
		if (!$2->set) Msg(FAT,"cannot use unlimited sort");
		for_list ($1,sym,T_symbol) sym->e.syntype = $2;
		$$ = $1;
	  }
	| varlist syntype ASSIGN constexpr
	  {
		T_symbol sym;
		T_syntype syn = alloc_syntype();

		if (!$2->set) Msg(FAT,"cannot use unlimited sort");
		syn->set = $2->set;
		syn->def_init = $4;
		syn->range = $2->range;
		check_sort(syn,syn->def_init);

		for_list ($1,sym,T_symbol) sym->e.syntype = syn;
		$$ = $1;
	  }
	;

varlist	: varident
	  {	list_add($$ = alloc_list(), $1);	}
	| varident ',' varlist
	  {	list_add($$ = $3,$1);			}
	;

varident: IDENT
	  {
		$$ = define_ident(&$1,TYPE_VARIABLE,NULL);
		$$->ref = add_ref($$,REF_VARIABLE,&$1.ref,&$1.ref);
		$$->ref->end_c++;	/* for text editor */
	  }
	;

/***********/
/* PROCESS */
/***********/

/* process uses prcdfmlparms (from the procedure rules) to avoid duplicating
   the definitions. in/out-Parameters are thrown out later.		    */

process : prcshead
	  prcsfmlparms end
	  prcsdcllist
	  body
	  ENDPROCESS optname end
	  {
		T_structure this = ($$ = $1)->e.struc;
		T_prcsinf prcsinf = alloc_prcsinf();
		T_fmlparm parm;
		T_symbol sym;

		check_end($$,$7,"PROCESS");
		prcsinf->initial = lower_bound(this->num_connects);
		prcsinf->maximum = upper_bound(this->num_connects);
		prcsinf->num_states = make_states();
		prcsinf->called_procs = called_procs;
		prcsinf->fmlparms = $2;

		/* check sdl_info file for information */
		initbuf(); addbuf($$->name); addbuf(".capacity");
		prcsinf->capacity = prop_find_key(strbuf,sdl_inqcap);

		initbuf(); addbuf($$->name); addbuf(".version");
		prcsinf->version = prop_find_key(strbuf,sdl_procver);

		for_list ($2,parm,T_fmlparm) if (parm->type)
			Msg(ERR,"process cannot have in/out parameters");
		this->num_connects = 0;
		this->entities = $4;
		list_add(this->entities,$5);
		this->prcsinf = prcsinf;
		revert_list(this->entities);
		this->symtab = dec_blknum();

		/* increase total number of process ids */
		num_pi_ids += prcsinf->maximum;

		add_state_info($$);

		/* create requests are complicated because you can refer
		   to processes that have not yet been defined. created_procs
		   collects a list of (temporary) symbol table entries
		   containing one such entry for every create request being
		   made in the current procedure. At this point, this list
		   is evaluated, and all processes are re-declared on the
		   block level.						    */

		for_list (created_procs,sym,T_symbol)
		{
			T_createreq cr = alloc_createreq();
			T_ident id = alloc_ident();

			cr->parent = $$;
			cr->child = sym;
			if (sym->type == TYPE_UNDECLARED)
			{
				/* undo name change made in create rule */
				*(sym->name) &= ~0x80;

				s_lookup(sym->name,id);
				cr->child = decl_ident(id,TYPE_STRUCTURE);
				((T_call) sym->e.data)->sym = cr->child;
			}
			list_add(create_list,cr);
		}
		free_list(created_procs);
	  }
	;
	
prcshead: PROCESS IDENT numinst
	  /* oprcsfmlprms end oinpset oprcsdcllst */
	  {
		int initial = lower_bound($3);
		int maximum = upper_bound($3);

		initbuf();
		addbuf($2.sym->name);
		addbuf(".instances");
		maximum = prop_find_key(strbuf,
			(maximum == NONE)? sdl_prcsinst : maximum);
		$3 = create_range(initial,maximum);
			
		$$ = define_ident(&$2,TYPE_STRUCTURE,alloc_structure());
		$$->e.struc->num_connects = $3;

		$$->e.struc->subtype = TYPE_PROCESS;
		called_procs = alloc_list();

		/* initialize list of create requests in this process */
		created_procs = alloc_list();

		inc_blknum();

		/* create implicit variables in the symbol table
		   unless they were already created by the system rule */
		if (sdlflag & SDLF_NO_PRCD_COPIES) init_implicit_vars();

		if (initial > maximum)
			Msg(ERR,"initial number larger than maximum");
		if (initial < 0 || maximum <= 0)
			Msg(ERR,"number of instances is broken");

		$$->ref = add_ref($$,REF_PROCESS,&$1,&$2);
	  }
	;

prcsfmlparms
	: '(' prcdparmlist ')'
	  {	$$ = $2;		}
	| {	$$ = alloc_list();	}
	;

/* initial and maximum number of instances are hidden
   as upper and lower bounds of a range type		*/

numinst	: '(' onum ')'
	  {
		$$ = create_range(($2 == NONE)? 1 : $2, NONE);	
	  }
	| '(' onum ',' onum ')'
	  {
		$$ = create_range(($2 == NONE)? 1 : $2, $4);
	  }
	| {
		$$ = create_range(1,NONE);
	  }
	;

onum	: constexpr
	| {	$$ = NONE;	}
	;

prcsdcllist
	: prcsdclentity prcsdcllist
	  {	list_add($$ = $2, $1);	}
	| {	$$ = alloc_list();	}
	;
	
prcsdclentity
	: data
	| dcl
	| procedure
	;

/*************/
/* PROCEDURE */
/*************/

procedure:prcdhead
	  prcddcllist
	  body
	  ENDPROCEDURE optname end
	  {
		T_structure this = ($$ = $1)->e.struc;

		check_end($$,$5,"PROCEDURE");
		this->prcsinf->num_states = make_states();
		this->prcsinf->initial = 0;

		initbuf();
		addbuf($$->name);
		addbuf(".instances");
		this->prcsinf->maximum = prop_find_key(strbuf,sdl_prcdinst);

		this->entities = $2;
		list_add(this->entities,$3);
		revert_list(this->entities);
		this->symtab = dec_blknum();
		add_state_info($$);
	  }
	;

prcdhead: PROCEDURE IDENT
	  {
		$$ = define_ident(&$2,TYPE_PROCEDURE,alloc_structure());
		$$->ref = add_ref($$,REF_PROCEDURE,&$1,&$2);
		$<symbol>1 = $$;
		inc_blknum();
	  }
	  prcdfmlparms end
	  {
		revert_list($4);
		$$ = $<symbol>1;
		$$->e.struc->prcsinf = alloc_prcsinf();
		$$->e.struc->prcsinf->fmlparms = $4;
	  }
	;

prcddcllist
	: prcddclentity prcddcllist
	  {	list_add($$ = $2, $1);	}
	| {	$$ = alloc_list();	}
	;
	
prcddclentity
	: data
	| dcl
	| procedure
	;

prcdfmlparms
	: '(' prcdparmlist ')'
	  {	$$ = $2;		}
	| {	$$ = alloc_list();	}
	;

prcdparmlist
	: singleparmdcl
	| singleparmdcl ',' prcdparmlist
	  {
		T_fmlparm parm;
		for_list ($1,parm,T_fmlparm) list_add($3,parm);
		free_list($1);
		$$ = $3;
	  }
	;

singleparmdcl
	: parmtype varlist syntype
	  {
		T_symbol sym;
		if (!$3->set) Msg(FAT,"cannot use unlimited sort");
		$$ = alloc_list();
		for_list ($2,sym,T_symbol)
		{
			T_fmlparm parm = alloc_fmlparm();
			parm->type = $1;
			parm->sym = sym;
			sym->e.syntype = $3;
			sym->ref->type = $1? REF_IN_OUT_PARM : REF_IN_PARM;
			list_add($$,parm);
		}
		free_list($2);
	  }
	;

parmtype: IN
	  {	$$ = False;	}
	| IN '/' OUT
	  {	$$ = True;	}
	| {	$$ = False;	}
	;

/****************/
/* process body */
/****************/

body    : start statespeclist
	  {
		/* create a dummy entry in the entity list of the process
		   or procedure (annotated with the list of transitions) */
		T_ident dummy = s_dummy("$$PrcsBody");
		$$ = define_ident(dummy,TYPE_BODY,$2);
		list_add($2,$1);
		revert_list($2);
	  }
	;

start   : START end term_transition
	  {
		$$ = alloc_transition();
		$$->states = NULL;
		$$->signals = NULL;
		$$->actions = $3;
	  }
	;

statespeclist
	: statespec statespeclist
	  {
		T_transition trans;
		for_list ($1,trans,T_transition) list_add($2,trans);
		free_list($1);
		$$ = $2;
	  }
	| {	$$ = alloc_list();	}
	;

statespec :
	  STATE statelist end triggerlist
	  {
		T_transition trans;
		for_list ($4,trans,T_transition) trans->states = $2;
		$$ = $4;
	  }
	;

statelist : 
	  simple_statelist
	  { $$=$1; }
	| '*'
	  { $$ = alloc_list(); }
	;

simple_statelist : stateident
	  {	list_add($$ = alloc_list(), $1);	}
	| stateident ',' simple_statelist
	  {	list_add($$ = $3,$1);			}
	;

stateident : IDENT
	  {	$$ = decl_ident(&$1,TYPE_STATE);	}
	;

triggerlist
	: trigger triggerlist
	  {	list_add($$ = $2, $1);	}
	| {	$$ = alloc_list();	}
	;

trigger : inputpart
	;

/*********/
/* INPUT */
/*********/

inputpart: INPUT usiglist end term_transition
	  {
		$$ = alloc_transition();
		$$->signals = $2;
		$$->actions = $4;
		$$->ref = add_ref($$,REF_TRIGGER,&$1,&$3);
	  }
	;

usiglist: singleusig
	  {	list_add($$ = alloc_list(), $1);	}
	| singleusig ',' usiglist
	  {	list_add($$ = $3,$1);			}
	;

singleusig
	: IDENT		/* signal without arguments */
	  {
		T_signaluse sig = alloc_signaluse();
		sig->sym = refer_ident(&$1,TYPE_SIGNAL);
		if (!sig->sym)
			Msg(FAT,"incorrect signal");
		if (sig->sym->e.signal->args)
			Msg(ERR,"incorrect number of arguments to signal");
		sig->parm = NULL;
		$$ = sig;
	  }
	| IDENT '(' sigarglist ')'
	  {
		T_signaluse sig = alloc_signaluse();
		T_parameter parm;
		T_list args;
		int i = 0;

		/* check number of arguments */
		sig->sym = refer_ident(&$1,TYPE_SIGNAL);
		if (!(args = sig->sym->e.signal->args)
				|| args->count != $3->count)
			Msg(ERR,"incorrect number of arguments to signal");
		sig->parm = MYmalloc(args->count * sizeof(parameter_instance));

		/* check type of arguments */
		for_list ($3,parm,T_parameter)
		{
			T_syntype reqtype = list_elem(args,i,T_syntype);
			if (parm->flag == PARM_VARIABLE)
				check_subset(reqtype->set,
						parm->var->e.syntype->set);
			else if (parm->flag == PARM_CONST)
				check_sort(reqtype,parm->value);
			memcpy(sig->parm+(i++),parm,sizeof(parameter_instance));
			free(parm);
		}
		free_list($3);
		$$ = sig;
	  }
	| INPUT_NONE	/* spontaneous transition */
	  {
		$$ = NULL;
	  }
	;

sigarglist
	: singlesigarg
	  {	list_add($$ = alloc_list(), $1);	}
	| singlesigarg ',' sigarglist
	  {	list_add($$ = $3,$1);			}
	;

singlesigarg
	: sigconst	/* In order to avoid r/r conflicts sigconst is used.
			   Normally, this should be constexpr.		    */
	  {
		T_parameter parm = alloc_parameter();
		parm->flag = PARM_CONST;
		parm->value = $1;
		$$ = parm;
	  }
	| IDENT
	  {
		T_parameter parm = alloc_parameter();
		parm->flag = PARM_VARIABLE;
		parm->var = refer_ident(&$1,TYPE_VARIABLE);
		$$ = parm;
	  }
	| {
		T_parameter parm = alloc_parameter();
		parm->flag = PARM_UNUSED;
		$$ = parm;
	  }
	;

sigconst: const
	| negconst;

/**************/
/* TRANSITION */
/**************/

/* term_transition is a transition with exactly one terminating action
   at the end of its list of actions					*/

term_transition
	: transition
	  {
		T_list l = $$ = $1;
		if (!l->count || !list_elem(l,l->count-1,T_action)->terminating)
			Msg(ERR,"transition is not terminated");
	  }
	;

transition: tlist
	  {
		T_action act;
		bool term = False;

		revert_list($$ = $1);
		for_list ($$,act,T_action)
		{
			if (term) Msg(ERR,"terminating statement before "
					  "end of transition");
			term = act->terminating;
		}
	  }
	;

tlist	: stmt tlist
	  {	list_add($$ = $2, $1);		}
	| {	$$ = alloc_list();		}
	;

stmt	: label action
	  {	($$ = $2)->label = $1;		}
	| action
	  {	($$ = $1)->label = NULL;	}
	;

action	: output
	| task
	| decision
	| call
	| create
	| terminator
	;

label	: IDENT ':'
	  {
		$$ = define_ident(&$1,TYPE_LABEL,(void*)1);
		$$->ref = add_ref($$,REF_LABEL,&$1.ref,&$2);
	  }
	;

/**********/
/* OUTPUT */
/**********/

output	: OUTPUT usiglist toclause viaclause end
	  {
		T_output out;
		$$ = alloc_action();
		$$->type = ACTION_OUTPUT;
		$$->terminating = False;
		out = $$->e.output = (T_output) $4;
		out->signals = $2;
		out->dst = $3;
		$$->ref = add_ref($$,REF_ACTION,&$1,&$5);
	  } 
	;

toclause: TO IDENT
	  {
		if ($2.sym->type == TYPE_VARIABLE
				&& $2.sym->e.syntype->set == SORT_PID)
			$$ = refer_ident(&$2,TYPE_VARIABLE);
		else
			$$ = refer_offscope(&$2,TYPE_STRUCTURE);
	  }
	| {	$$ = NULL;	}
	;

viaclause
	: VIA vialist
	  {
		T_output out = alloc_output();
		out->via = $2;
		$$ = out;
	  }
	| {
		T_output out = alloc_output();
		out->via = alloc_list();
		$$ = out;
	  }
	;

vialist	: viaident
	  {	list_add($$ = alloc_list(), $1);	}
	| viaident VIA vialist
	  {	list_add($$ = $3,$1);			}
	;

viaident: IDENT
	  {	$$ = refer_offscope(&$1,TYPE_RELAY);	}
	;

/********/
/* TASK */
/********/

task	: TASK assignlist end
	  {
		$$ = alloc_action();
		$$->type = ACTION_TASK;
		$$->terminating = False;
		$$->e.list = $2;
		$$->ref = NULL;
		copy_ref_end(list_elem($2,$2->count-1,T_assignment)->ref,&$3);
		revert_list($2);
		copy_ref_start(list_elem($2,0,T_assignment)->ref,&$1);
	  }
	;

assignlist
	: assignment
	  {
		list_add($$ = alloc_list(), $1);
	  }
	| assignment ',' assignlist
	  {
		list_add($$ = $3,$1);
		copy_ref_end(((T_assignment)$1)->ref,&$2);
	  }
	;

assignment
	: IDENT ASSIGN expr
	  {
		T_assignment assign = alloc_assignment();
		intset type;

		assign->var = refer_ident(&$1,TYPE_VARIABLE);
		type = assign->var->e.syntype->set;
		assign->expr = $3;
		$$ = assign;

		/* check types of IDENT and expr */
		assign->ref = add_ref($$,REF_ACTION,&$1.ref,&$2);
		if ($3->sort == ETYPE_INTEGER
				&& type != SORT_PID && type != boolset) break;
		if ($3->sort == ETYPE_BOOLEAN && type == boolset) break;
		if ($3->sort == ETYPE_PID && type == SORT_PID) break;
		Msg(ERR,"type mismatch in assignment");
	  }
	;

/********/
/* EXPR */
/********/

constexpr: expr
	  {
		if (!$1->constant) Msg(ERR,"expression is not constant");
		$$ = eval_node($1);
		free_node($1);
	  }
	;

expr	: expr IMPLIES expr
	  { $$ = create_node(NODE_IMPLIES,$1,$3,ETYPE_BOOLEAN,ETYPE_BOOLEAN); }
	| expr OR expr
	  { $$ = create_node(NODE_OR,$1,$3,ETYPE_BOOLEAN,ETYPE_BOOLEAN); }
	| expr XOR expr
	  { $$ = create_node(NODE_XOR,$1,$3,ETYPE_BOOLEAN,ETYPE_BOOLEAN); }
	| expr AND expr
	  { $$ = create_node(NODE_AND,$1,$3,ETYPE_BOOLEAN,ETYPE_BOOLEAN); }
	| expr '=' expr
	  { $$ = create_node(NODE_EQ,$1,$3,ETYPE_ANY,ETYPE_BOOLEAN); }
	| expr NE expr
	  { $$ = create_node(NODE_NE,$1,$3,ETYPE_ANY,ETYPE_BOOLEAN); }
	| expr '>' expr
	  { $$ = create_node(NODE_GT,$1,$3,ETYPE_INTEGER,ETYPE_BOOLEAN); }
	| expr GE expr
	  { $$ = create_node(NODE_GE,$1,$3,ETYPE_INTEGER,ETYPE_BOOLEAN); }
	| expr '<' expr
	  { $$ = create_node(NODE_LT,$1,$3,ETYPE_INTEGER,ETYPE_BOOLEAN); }
	| expr LE expr
	  { $$ = create_node(NODE_LE,$1,$3,ETYPE_INTEGER,ETYPE_BOOLEAN); }
	| expr '+' expr
	  { $$ = create_node(NODE_PLUS,$1,$3,ETYPE_INTEGER,ETYPE_INTEGER); }
	| expr '-' expr
	  { $$ = create_node(NODE_MINUS,$1,$3,ETYPE_INTEGER,ETYPE_INTEGER); }
	| expr '*' expr
	  { $$ = create_node(NODE_MULT,$1,$3,ETYPE_INTEGER,ETYPE_INTEGER); }
	| expr '/' expr
	  { $$ = create_node(NODE_DIV,$1,$3,ETYPE_INTEGER,ETYPE_INTEGER); }
	| expr MOD expr
	  { $$ = create_node(NODE_MOD,$1,$3,ETYPE_INTEGER,ETYPE_INTEGER); }
	| expr REM expr
	  { $$ = create_node(NODE_REM,$1,$3,ETYPE_INTEGER,ETYPE_INTEGER); }
	| '-' expr %prec unaryminus
	  { $$ = create_node(NODE_NEG,$2,NULL,ETYPE_INTEGER,ETYPE_INTEGER); }
	| NOT expr
	  { $$ = create_node(NODE_NOT,$2,NULL,ETYPE_BOOLEAN,ETYPE_BOOLEAN); }
	| '(' expr ')'
	  { $$ = create_node(NODE_IMPLIES,$2,NULL,ETYPE_ANY,ETYPE_ANY); }
	| IDENT
	  {
		intset set;
		$$ = alloc_node();
		$$->type = NODE_VAR;
		$$->constant = False;
		$$->op1 = (T_node) refer_ident(&$1,TYPE_VARIABLE);

		set = ((T_symbol) $$->op1)->e.syntype->set;
		if (set == boolset) $$->sort = ETYPE_BOOLEAN;
		else if (set == SORT_PID) $$->sort = ETYPE_PID;
		else $$->sort = ETYPE_INTEGER;
	  }
	| const
	  {
		$$ = alloc_node();
		$$->type = NODE_CONST;
		$$->constant = True;
		$$->op1 = (T_node) $1;
		if ($1 == TT || $1 == FF) $$->sort = ETYPE_BOOLEAN;
		else if ($1 == DOT) $$->sort = ETYPE_PID;
		else $$->sort = ETYPE_INTEGER;
	  }
	;

/************/
/* DECISION */
/************/

decision: DECISION question end 
	  answerlist
	  ENDDECISION end
	  {
		T_list tr, l1;
		T_decision dec = alloc_decision();
		bool term = True, nonterm = True;
		int i = 0;

		$$ = alloc_action();
		$$->type = ACTION_DECISION;
		$$->e.decision = dec;
		dec->question = $2;
		dec->answers = alloc_list();
		dec->transitions = alloc_list();
		dec->refs = alloc_list();
		revert_list($4);

		$$->ref = add_ref($$,REF_ACTION,&$1,&$3);

		if ($4->count < 6)
			Msg(ERR,"need at least two decision branches");

		/* answerlist ($4) returns one big list containing three
		   entries for each answer. Here, these entries are separated
		   and put into three distinct lists.			      */
		for_list ($4,tr,T_list)
			switch (i++ % 3)
			{
			    case 0: list_add(dec->refs,tr); break;
			    case 1: list_add(dec->transitions,tr); break;
			    case 2: list_add(dec->answers,tr); break;
			}

		free_list($4);
		
		/* check for terminating decision */
		for_list (dec->transitions,tr,T_list)
		{
			bool thisterm = tr->count
			     && list_elem(tr,tr->count-1,T_action)->terminating;
			term &= thisterm;
			nonterm &= !thisterm;
		}

		if (!(term || nonterm))
			Msg(ERR,"all or none of the decision branches "
				"must be terminating");
		$$->terminating = term;
		
		/* check if answers are disjunct and if else is used properly */
		if (dec->question)
		{
			T_list irange, allrange = alloc_list();
			bool elsepart = False;
			int r;

			for_list (dec->answers,l1,T_list)
			{
				if (elsepart) Msg(ERR,"else branch must be "
						      "last in decision");
				if ((elsepart = !l1)) continue;
				irange = intersect_range(allrange,l1,False);
				if (irange->count) Msg(ERR,"answers must be "
							  "mutually exclusive");
				free_list(irange);
				for_list (l1,r,int) list_add(allrange,r);
			}
			free_list(allrange);
		}
		else
			for_list (dec->answers,l1,T_list)
			  if (l1)
			    Msg(ERR,"decision on ANY requires empty answers");
	  }
	;

question: expr
	| ANY
	  {	$$ = NULL;	}
	;

answerlist
	: answer
	| answer answerlist
	  {
		$$ = $2;
		list_add($$,list_elem($1,0,void*));
		list_add($$,list_elem($1,1,void*));
		list_add($$,list_elem($1,2,void*));
		free_list($1);
	  }
	;

answer	: '(' optrangelist ')' ':' transition
	  {
		T_ref ref = add_ref($2,REF_ANSWER,&$1,&$4);
		$$ = alloc_list();
		list_add($$,$2);
		list_add($$,$5);
		list_add($$,ref);
	  }
	| ELSE ':' transition
	  {
		T_ref ref = add_ref(NULL,REF_ANSWER,&$1,&$2);
		$$ = alloc_list();
		list_add($$,NULL);
		list_add($$,$3);
		list_add($$,ref);
	  }
	;

optrangelist
	: rangelist
	| { $$ = NULL; }
	;

/********/
/* CALL */
/********/

call	: CALL IDENT optparms end
	 {
		T_list plist;
		T_fmlparm parm;
		int i = 0;

		$$ = alloc_action();
		$$->type = ACTION_CALL;
		$$->terminating = False;
		$$->e.call = alloc_call();
		$$->e.call->sym = refer_ident(&$2,TYPE_PROCEDURE);
		$$->e.call->actparms = $3;
		$$->ref = add_ref($$,REF_ACTION,&$1,&$4);
		revert_list($3);
		list_add(called_procs,$$->e.call->sym);

		$$->e.call->fmlparms = plist
			= $$->e.call->sym->e.struc->prcsinf->fmlparms;

		if (plist->count != $3->count)
			Msg(ERR,"procedure call with wrong number of args");

		/* check type of arguments */
		for_list (plist,parm,T_fmlparm)
		{
			T_node node = list_elem($3,i++,T_node);
			intset type = parm->sym->e.syntype->set;

			if (parm->type && (!node || node->type != NODE_VAR))
				Msg(ERR,"need variable for in/out parameter");
			if (!node) continue;
			if ((node->sort == ETYPE_INTEGER
			    	&& type != SORT_PID && type != boolset)
			   || (node->sort == ETYPE_BOOLEAN && type == boolset)
			   || (node->sort == ETYPE_PID && type == SORT_PID))
				continue;
			Msg(ERR,"argument type mismatch");
		}
	  } 
	;

optparms: '(' actparmlist ')'
	  {	$$ = $2;		}
	| {	$$ = alloc_list();	}
	;

actparmlist
	: expr
	  {	list_add($$ = alloc_list(), $1);	}
	| expr ',' actparmlist
	  {	list_add($$ = $3,$1);			}
	| ',' actparmlist
	  {	list_add($$ = $2,NULL);			}
	| {	list_add($$ = alloc_list(), NULL);	}
	;

/**********/
/* CREATE */
/**********/

create	: CREATE IDENT optparms end
	  {
		T_symbol proc;

		$$ = alloc_action();
		$$->type = ACTION_CREATE;
		$$->terminating = False;
		$$->e.call = alloc_call();
		$$->ref = add_ref($$,REF_ACTION,&$1,&$4);

		/* Since forward references are possible, this rule creates
		   a temporary symbol list entry. This entry is appended
		   to created_procs and evaluated in the process rule.	*/

		$$->e.call->sym = proc = decl_ident(&$2,TYPE_STRUCTURE);
		if (!proc->e.data)
		{
			proc->type = TYPE_UNDECLARED;
			proc->e.data = $$->e.call;

			/* change the name information in the symbol table
			   so that other create requests of the same process
			   can get their own temporary entries in the table */
			*(proc->name) |= 0x80;
		}
		list_add(created_procs,proc);

		$$->e.call->actparms = $3;
		revert_list($3);
	  }
	;

/**************/
/* TERMINATOR */
/**************/

terminator
	: nextstate 
	| stop
	| join
	;


nextstate
	: NEXTSTATE IDENT end
	  {
		$$ = alloc_action();
		$$->type = ACTION_NEXTSTATE;
		$$->terminating = True;
		$$->e.state = decl_ident(&$2,TYPE_STATE);
		$$->ref = add_ref($$,REF_ACTION,&$1,&$3);
	  }
	| NEXTSTATE '-' end
	  {
		T_ident dummy = s_dummy("-");
		$$ = alloc_action();
		$$->type = ACTION_NEXTSTATE;
		$$->terminating = True;
		$$->e.state = decl_ident(dummy,TYPE_STATE);
		$$->ref = add_ref($$,REF_ACTION,&$1,&$3);
	  }
	;

stop	: stoporreturn
	  {
		$$ = $1;
		$$->type = ACTION_NEXTSTATE;
		$$->terminating = True;
		$$->e.state = NULL;
	  }
	;

/* More properly, one should check if this occurs inside a procedure
   and reject the wrong type of command. However, both commands are
   treated the same, so return inside a process is treated as stop,
   and stop inside a procedure is treated as return.		    */
stoporreturn
	: STOP end
	  {
		($$ = alloc_action())->ref = add_ref($$,REF_ACTION,&$1,&$2);
	  }
	| RETURN end
	  {
		($$ = alloc_action())->ref = add_ref($$,REF_ACTION,&$1,&$2);
	  }
	;

join	: JOIN IDENT end
	  {
		$$ = alloc_action();
		$$->type = ACTION_JOIN;
		$$->terminating = True;
		$$->e.goal = decl_ident(&$2,TYPE_LABEL);
		$$->ref = add_ref($$,REF_ACTION,&$1,&$3);
	  }
	;

%%
