/*****************************************************************************/
/* simplify.c                                                                */
/*****************************************************************************/
/* Eliminate dead transitions and places from a high/low level net.          */
/*                                                            Stefan Schwoon */
/*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "mnet.h"
#include "readHL.h"
#include "mtupel.h"

/*****************************************************************************/

static void usage(char *myname)
{
  fprintf(stderr,
    "\n%s -- simplify net based on a list of dead low level transitions.\n\n"

    "Usage: %s [Options] <infile1> [infile2] [FileOptions]\n\n"

    "     Options:\n"
    "      -p   to remove isolated places\n\n"

    "     FileOptions:\n"
    "      -dt <filename>     specify an alternate list of dead low level transitions\n"
    "      -hl <filename>     specify an alternate HL->LL reference file\n"
    "      -bh <filename>     specify an alternate BPN2->HL reference file\n"
    "      -bl <filename>     specify an alternate BPN2->LL reference file\n"
    "      -sh <filename>     specify an alternate SDL->HL reference file\n"
    "      -sl <filename>     specify an alternate SDL->LL reference file\n"
    "      -o <file1> [file2] specify alternate filenames for minimized nets\n"
    "      -ohl <filename>    specify an alternate HL->LL reference file for output\n"
    "      -obh <filename>    specify an alternate BPN2->HL reference file for output\n"
    "      -obl <filename>    specify an alternate BPN2->LL reference file for output\n\n"
    "      -osh <filename>    specify an alternate SDL->HL reference file for output\n"
    "      -osl <filename>    specify an alternate SDL->LL reference file for output\n\n"

    "You may give a high level net, a low level net, or both, as input files.\n"
    "New reference files will be created as required. All file names not explicitly\n"
    "stated on the command line will default to the input files' basenames plus ap-\n"
    "propriate extensions. By default, .min is appended to the name of output files.\n\n"

    "Version 0.9.4 (14.05.1998) Stefan Schwoon\n\n", myname, myname);

  exit(1);
} /* usage() */


/*****************************************************************************/

void update_phantom_refs(M_Net net)
{
	register M_Transition	tr;
	register M_Transition	ptr;
	M_Transition		*TrArray;
	MTupel			mt;
	int			i = 0;
	char			tmp[10];
	char			*dummy;

	for (tr = net->Transitions; tr; tr = tr->nexttrans)
		if (tr->numname > i)  i = tr->numname;

	TrArray = MYmalloc((i+1) * sizeof(M_Transition));

	for (tr = net->Transitions; tr; tr = tr->nexttrans)
		TrArray[tr->numname] = tr;

	for (ptr = net->PhTransitions; ptr; ptr = ptr->nexttrans)
	{
		ptr->partners = MPT_Create();
		if (ptr->corrtran)
		{
			dummy = ptr->corrtran;
			mt = read_MTupel(&dummy);
			for (i=0; i<mt->count; i++)
			{
				if ((tr = TrArray[MT_Place(mt,i)]) && !tr->dead)
					MPTT_AddValue(ptr->partners,tr);
			}
			free(mt->place); free(mt);
		}
		if (!ptr->partners->count)
		{
			ptr->dead = True;
			ptr->corrtran = NULL;
		}
		else
		{
		    strcpy(ptr->corrtran,"(");
		    for (i=0; i<ptr->partners->count; i++)
		    {
			sprintf(tmp,"%d,",MT_PlaceT(ptr->partners,i)->numname);
			strcat(ptr->corrtran,tmp);
		    }
		    ptr->corrtran[strlen(ptr->corrtran)-1] = ')';
		}
	}

	free(TrArray);

}

void simplify(M_Net hlnet, M_Net llnet)
{
	register M_Transition	tr;

	if (hlnet)
	{
	    for (tr = llnet->Transitions; tr; tr = tr->nexttrans)
		if (tr->parenttrans && !tr->dead) tr->parenttrans->childcount++;

	    for (tr = hlnet->Transitions; tr; tr = tr->nexttrans)
		if (!tr->childcount) tr->dead = True;
	}

	if (hlnet) update_phantom_refs(hlnet);
	update_phantom_refs(llnet);

} /* simplify() */

/*****************************************************************************/

enum { STAT_ALL, STAT_ALIVE, STAT_DEAD };

void print_statistics(M_Net net, char *typename, int mode)
{
	M_Place pl;
	M_Transition tr;
	M_STEdge st;
	M_TSEdge ts;

	int i;

	printf("%s: ",typename);
	for (pl = net->Places, i = 0; pl; pl = pl->nextplace)
		i += (mode == STAT_ALL)
		     || (pl->dead && mode == STAT_DEAD)
		     || (!pl->dead && mode == STAT_ALIVE);
	printf("%7u places, ",i);
	
	for (tr = net->Transitions, i = 0; tr; tr = tr->nexttrans)
		i += (mode == STAT_ALL)
		     || (tr->dead && mode == STAT_DEAD)
		     || (!tr->dead && mode == STAT_ALIVE);
	printf("%7u transitions, ",i);
	
	for (tr = net->Transitions, i = 0; tr; tr = tr->nexttrans)
	    for (ts = tr->postset; ts; ts = ts->nextplace)
		i += (mode == STAT_ALL)
		     || ((tr->dead || ts->place->dead) && mode == STAT_DEAD)
		     || ((!tr->dead && !ts->place->dead) && mode == STAT_ALIVE);
	for (tr = net->Transitions; tr; tr = tr->nexttrans)
	    for (st = tr->preset; st; st = st->nextplace)
		i += (mode == STAT_ALL)
		     || ((tr->dead || st->place->dead) && mode == STAT_DEAD)
		     || ((!tr->dead && !st->place->dead) && mode == STAT_ALIVE);
	printf("%7u arcs\n",i);
}
	

/*****************************************************************************/

void remove_places(M_Net net)
{
	M_Place  pl;
	M_STEdge st;
	M_TSEdge ts;

	for (pl = net->Places; pl; pl = pl->nextplace)
	{
	  if (!pl->dead)
	  {
	    for (ts = pl->preset; ts && ts->trans->dead; ts = ts->nexttrans)
	      skip;
	    for (st = pl->postset; st && st->trans->dead; st = st->nexttrans)
	      skip;
	    if (!(ts||st))
	      pl->dead = True;
	  }
	}
} /* remove_places */

/*****************************************************************************/
/* Main program.                                                             */
/*****************************************************************************/

int main(int argc, char **argv)
{
	int	i;
	char    *hlnet = NULL, *llnet = NULL, *deadtran,
		*hlllref, *bpnhlref, *bpnllref, *sdlhlref, *sdlllref,
		*desthlnet, *destllnet,
		*desthlllref, *destbpnhlref, *destbpnllref,
		*destsdlhlref, *destsdlllref;
	char    **dptr;
	char	tmpname[80], *idx;
	M_Net   hnet = NULL, lnet = NULL;
	FILE    *fileptr;
	clock_t time;
	bool    remove_isolated = False;
	bool    nohlllref = False,
		nobpnhlref = False,
		nobpnllref = False,
		nosdlhlref = False,
		nosdlllref = False;

	dptr = &hlnet;
	for (i=1; i<argc; i++)
		if (!strcmp(argv[i],"-p"))
			remove_isolated = True;

		else if (!strcmp(argv[i],"-dt"))
			dptr = &deadtran;
		else if (!strcmp(argv[i],"-hl"))
			dptr = &hlllref;
		else if (!strcmp(argv[i],"-bh"))
			dptr = &bpnhlref;
		else if (!strcmp(argv[i],"-bl"))
			dptr = &bpnllref;
		else if (!strcmp(argv[i],"-sh"))
			dptr = &sdlhlref;
		else if (!strcmp(argv[i],"-sl"))
			dptr = &sdlllref;

		else if (!strcmp(argv[i],"-o"))
			dptr = &desthlnet;
		else if (!strcmp(argv[i],"-ohl"))
			dptr = &desthlllref;
		else if (!strcmp(argv[i],"-obh"))
			dptr = &destbpnhlref;
		else if (!strcmp(argv[i],"-obl"))
			dptr = &destbpnllref;
		else if (!strcmp(argv[i],"-osh"))
			dptr = &destsdlhlref;
		else if (!strcmp(argv[i],"-osl"))
			dptr = &destsdlllref;

		else
		{
			if (!dptr) usage(argv[0]);
			*dptr = argv[i];

			strcpy(tmpname,argv[i]);
			idx = strrchr(tmpname,'.');
			if (!idx) strcat(tmpname,".");
			idx = strrchr(tmpname,'.');

			if (dptr == &hlnet)
			{
				strcpy(idx,".bpn2_hl_ref");
				bpnhlref = MYstrdup(tmpname);
				strcpy(idx,".sdl_hl_ref");
				sdlhlref = MYstrdup(tmpname);
				strcpy(idx,".hl_net.min");
				desthlnet = MYstrdup(tmpname);
			}
			if (dptr == &hlnet || dptr == &llnet)
			{
				strcpy(idx,".hl_ll_ref");
				hlllref = MYstrdup(tmpname);
				strcpy(idx,".dead_tran");
				deadtran = MYstrdup(tmpname);

				strcpy(idx,".ll_net.min");
				destllnet = MYstrdup(tmpname);
				strcpy(idx,".bpn2_ll_ref");
				bpnllref = MYstrdup(tmpname);
				strcpy(idx,".sdl_ll_ref");
				sdlllref = MYstrdup(tmpname);
			}
			if (dptr == &hlnet || dptr == &desthlnet)
			{
				strcpy(idx,".bpn2_hl_ref.min");
				destbpnhlref = MYstrdup(tmpname);
/*				strcpy(idx,".sdl_hl_ref.min");
*/				strcpy(idx,".sdl_hl_ref");
				destsdlhlref = MYstrdup(tmpname);
			}
			if (dptr == &llnet || dptr == &destllnet
			 || dptr == &hlnet || dptr == &desthlnet)
			{
				strcpy(idx,".hl_ll_ref.min");
				desthlllref = MYstrdup(tmpname);
				strcpy(idx,".bpn2_ll_ref.min");
				destbpnllref = MYstrdup(tmpname);
/*				strcpy(idx,".sdl_ll_ref.min");
*/				strcpy(idx,".sdl_ll_ref");
				destsdlllref = MYstrdup(tmpname);
			}
			if (dptr == &desthlnet) destllnet = desthlnet;

			if (dptr == &hlnet) dptr = &llnet;
			else if (dptr == &desthlnet) dptr = &destllnet;
			else dptr = NULL;
		}

	if (!hlnet) usage(argv[0]);

	time = clock();

	hnet = read_HLnet(hlnet);
	if (!strcmp(hnet->type,"PetriBox")||!strcmp(hnet->type,"PTNet"))
	{
		llnet = hlnet;
		lnet = hnet;
		hnet = NULL;
		if (!(fileptr = fopen(hlllref,"r")) || fclose(fileptr))
		{
		    nohlllref = True;
		    fprintf(stderr,"Warning: HL->LL references not found\n");
		}
	}
	else if (llnet)
	{
		lnet = read_HLnet(llnet);
		if (strcmp(lnet->type,"PetriBox")&&strcmp(lnet->type,"PTNet"))
		{
			fprintf(stderr,"Multiple high level nets specified; "
					"aborting\n");
			exit(1);
		}
	}

	if (!hnet || !(fileptr = fopen(bpnhlref,"r")) || fclose(fileptr))
		nobpnhlref = True;
	if (!hnet || !(fileptr = fopen(sdlhlref,"r")) || fclose(fileptr))
		nosdlhlref = True;

	if (hnet && nobpnhlref && nosdlhlref)
		    fprintf(stderr,"Warning: no BPN2->HL and no SDL->HL "
				   "references found\n");

	if (!lnet || !(fileptr = fopen(bpnllref,"r")) || fclose(fileptr))
		nobpnllref = True;
	if (!lnet || !(fileptr = fopen(sdlllref,"r")) || fclose(fileptr))
		nosdlllref = True;

	if (lnet && nobpnllref && nosdlllref)
		    fprintf(stderr,"Warning: no BPN2->LL and no SDL->LL "
				   "references found\n");

	if (!nohlllref) read_HLLLref(&hnet,&lnet,hlllref);
	if (!nobpnhlref) read_BPNref(hnet,bpnhlref);
	if (!nobpnllref) read_BPNref(lnet,bpnllref);
	if (!nosdlhlref) read_SDLref(hnet,sdlhlref);
	if (!nosdlllref) read_SDLref(lnet,sdlllref);

	read_deadlist(lnet,deadtran);

	simplify(hnet,lnet);
	if (remove_isolated)
	{
		if (hnet) remove_places(hnet);
		remove_places(lnet);
	}

	if (hnet)
	{
	  print_statistics(hnet,"HL net (before)   ",STAT_ALL);
	  print_statistics(hnet,"HL net (minimised)",STAT_ALIVE);
	  print_statistics(hnet,"HL net (removed)  ",STAT_DEAD);
	}
	if (llnet)
	{
	  print_statistics(lnet,"LL net (before)   ",STAT_ALL);
	  print_statistics(lnet,"LL net (minimised)",STAT_ALIVE);
	  print_statistics(lnet,"LL net (removed)  ",STAT_DEAD);
	}

	if (hnet) write_HLnet(hnet,desthlnet);
	if (llnet) write_HLnet(lnet,destllnet);
	if (!nohlllref) write_HLLLref(lnet,desthlllref,!!llnet);
	if (!nobpnhlref) write_BPNref(hnet,destbpnhlref);
	if (!nobpnllref) write_BPNref(lnet,destbpnllref);
	if (!nosdlhlref) write_SDLref(hnet,destsdlhlref);
	if (!nosdlllref) write_SDLref(lnet,destsdlllref);

/*	printf("Time needed: %f sec.\n\n",(float)(clock()-time)/CLOCKS_PER_SEC);
*/
	return 0;
}

/*****************************************************************************/
/* END OF FILE simplify.c                                                    */
/*****************************************************************************/

