/*****************************************************************************/
/* slist.c								     */
/*****************************************************************************/
/* Determine S-Invariants in a low level net.				     */
/*                                                            Stefan Schwoon */
/*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "slist.h"
#include "readHL.h"
#include "ref.h"
#include "list.h"
#include "mtupel.h"

/*****************************************************************************/

static void usage(char *myname)
{
  fprintf(stderr,
    "\n%s -- determine s-invariants in a low level net\n\n"

    "Usage: %s <mode> <infile> [FileOptions]\n\n"

    "     Mode:\n"
    "      -bpn2              create lists for BPN2 program\n"
    "      -sdl               create lists for SDL program\n"
    "      -pfa               create lists for PFA program (currently none)\n\n"

    "     FileOptions:\n"
    "      -r <filename>      specify an alternate reference file\n"
    "      -o <filename>      specify an alternate file to save s-lists in\n\n"

    "Invariants will be determined according to place references.\n"
    "All file names not explicitly stated on the command line will\n"
    "default to the input file's basename plus appropriate extensions.\n\n"

    "Version 0.9.2 (03.06.1998) Stefan Schwoon\n\n", myname, myname);

  exit(1);
} /* usage() */

/*****************************************************************************/

M_Block* Bl_Array;

void buildblocktree(M_Net net)
{
	int i = 0;
	M_Block cbl,bl;
	MTupel mt;
	
	for (bl = net->Blocks; bl; bl = bl->nextblock)
		if (bl->numname > i) i = bl->numname;

	Bl_Array = (MYmalloc)((i+1) * sizeof(M_Block));

	for (bl = net->Blocks; bl; bl = bl->nextblock)
		Bl_Array[bl->numname] = bl;

	for (bl = net->Blocks; bl; bl = bl->nextblock)
		if (bl->childblocks)
		{
			mt = read_MTupel(&bl->childblocks);
			for (i = 0; i < mt->count; i++)
			{
				cbl = Bl_Array[MT_Place(mt,i)];
				if (cbl) cbl->parentblock = bl;
			}
			free(mt->place);
			free(mt);
		}

}
			
void insert_pl(M_Net net, M_Place pl, int block, int bits, t_bi bi, int fv)
{
	if (bits & 24 || (bits & 4 && bi->type == PROCEDURE)
		      || (!bits && bi->type == VARIABLE))
	{
	    if (pl->lastdimen > bi->maxdimen)
			bi->maxdimen = pl->lastdimen;

	    if (pl->lastdimen > 0 && pl->lastdimen < bi->mindimen)
			bi->mindimen = pl->lastdimen;

	}

	if (bits & 24)
	    MPTT_AddValue(MT_PlaceMPT(net->blocklists,bi->index), pl);
	else if (bits & 4)
	{
	    if (bi->type == VARIABLE || bi->type == PROCEDURE)
		MPTT_AddValue(MT_PlaceMPT(net->blocklists,bi->index), pl);
	    else if (bi->type == STACK)
		MPTT_AddValue(MT_PlaceMPT(net->blocklists,
					bi->index + fv), pl);
	    else	/* CHANNEL */
		MPTT_AddValue(MT_PlaceMPT(net->blocklists,
			(bi->anz > 1)? bi->index + fv - 1 : bi->index), pl);
	}
	else if (!bits)
		MPTT_AddValue(MT_PlaceMPT(net->blocklists,bi->index), pl);
}
		
void makelists_bpn2(M_Net net)
{
	M_Block  bl;
	M_Place  pl;
	MTupel	 value;
	int	 i,firstvalue;
	int	 iblock,block,bits;
	t_bi	 bi;
	char	 *c;
	bool	 done;


	net->blocklists = MPT_Create();

	for (i = 0; i < net->numlists; i++)
		MPTT_AddValue(net->blocklists,MPT_Create());

	for (pl = net->Places; pl; pl = pl->nextplace)
	{
	    if (pl->ref)
	    {
		c = strchr(pl->ref,';');
		if (!c)
		{
			fprintf(stderr,"Parsing error in BPN2 references:"
					" ';' expected\n");
			exit(2);
		}
		*c++ = '\0';
		value = read_MTupel(&c);
		firstvalue = MT_Place(value,0);
		pl->lastdimen = MT_Place(value,value->count-1);
		free(value->place); free(value);

		iblock = 0;
		if (pl->blocklist)
		{
			value = read_MTupel(&pl->blocklist);
			for (i = 0; i < value->count; i++)
			    if (Bl_Array[MT_Place(value,i)]->parallel) break;
			iblock = i?MT_Place(value,pl->lastdimen>0? 0 : i-1) : 0;
			free(value->place); free(value);
		}
		
		c = pl->ref;
		done = False;
		while (*c && !done)
		{
		    bits = 0;
		    block = read_posint(&c);
		    if (*c == ':' && c++) bits = read_posint(&c);

		    for (i = 0; i < net->numblocks && !done; i++)
			if ((bi = MT_PlaceBI(net->blockinfo,i))
					&& bi->ident == block)
			{
				done = True;
				insert_pl(net,pl,block,bits,bi,firstvalue);
			}

		    if (!done && (bits & 7))
		    {
			for (bl = net->Blocks; bl; bl = bl->nextblock)
			    if (bl->defnr == block)
			    {
				if (pl->lastdimen > bl->maxdimen)
					bl->maxdimen = pl->lastdimen;
				if (pl->lastdimen > 0
						&& pl->lastdimen < bl->mindimen)
					bl->mindimen = pl->lastdimen;
			    }
			if (iblock) MPTT_AddValue(Bl_Array[iblock]->places,pl);
			done = !!iblock;
		    }

		    if (*c && *c != '*')
		    {
			fprintf(stderr,"Parsing error in BPN2 references:"
					" '*' expected\n");
			exit(2);
		    }
		    if (*c) c++;
		}

	    }
	}

} /* makelists_bpn2 */

/*****************************************************************************/

int insert_sdl_place(T_defitem def, M_Place pl, int bits, MTupel value)
{
	int i;

	#define add(list) { list_add(list,pl); return 1; }
	#define dim(x) (MT_Place(value,x))
	#define item(list,x) (list_elem(list,x,T_list))
	#define find_index(x) { for_list (def->idset,i,int) if (i == x) break; }
	#define index() (def->idset->run - def->idset->elems)

	switch (def->type)
	{
	    case REF_SYSTEM:
	    case REF_SIGNALROUTE:
	    case REF_CONNECT:
		add(def->list1);
	    case REF_CHANNEL:
		if (bits & (8192 | 2 | 8 | 16)) add(def->list1);
		if (bits & 4) add(item(def->array,dim(0)-1));
		break;
	    case REF_VARIABLE:
	    case REF_IN_PARM:
	    case REF_IN_OUT_PARM:
		if (bits & 4) find_index(dim(def->flag? 2 : 1));
		if (bits & (8 | 16)) find_index(dim(def->flag? 1 : 0));
		add(item(def->array,index()));
	    case REF_PROCESS:
		if (bits & (1 | 2 | 8 | 16)) add(def->list1);
		find_index(dim(value->count-1));
		if (bits & 32) add(item(def->array,index())); /*Kontrollfluss*/
		i = -1;
		if (bits & 4) i = 0;	/* Stack */
		if (bits & 128) i = 1;	/* parent */
		if (bits & 256) i = 2;	/* offspring */
		if (bits & 512) i = 3;	/* sender */
		if (bits & 1024) i = 4;	/* q */
		if (bits & 2048) i = 5; /* - */
		if (bits & 8192) i = 6; /* Fuellgrad */
		if (bits & 16384) i = 6 + dim(0);	/* input queue */
		if (i >= 0) add(item(def->more, i * def->size + index()));
		break;
	    case REF_PROCEDURE:
		if (bits & (1 | 2 | 8 | 16)) add(def->list1);
		if (bits & 32) add(def->list2);
		find_index(dim(value->count-((bits & 3072)? 1: 2)));
		i = -1;
		if (bits & 4) i = 0;	/* Stack */
		if (bits & 1024) i = 1;	/* q */
		if (bits & 2048) i = 2; /* - */
		if (i >= 0) add(item(def->more, i * def->size + index()));
		break;
	    case REF_ACTION:
	    case REF_LABEL:
		def = def->proc;
		find_index(dim(value->count -
			((def->type == REF_PROCESS)? 1 : 2)));
		add(item(def->array,index()));
	    case REF_SIGNAL:
	    case REF_ANSWER:
	    case REF_BLOCK:
	    case REF_SYNTYPE:
	    case REF_TRIGGER:
	    default:
		break;
	}

	return 0;
}

void makelists_sdl(M_Net net)
{
	M_Place	  pl;
	MTupel	  value;
	T_defitem def;
	int	  block,bits;
	char	  *c;
	int	  done;

	for (pl = net->Places; pl; pl = pl->nextplace)
	{
		if (!pl->ref) continue;

		c = strchr(pl->ref,';');
		if (!c)
		{
			fprintf(stderr,"Parsing error in BPN2 references:"
					" ';' expected\n");
			exit(2);
		}
		*c++ = '\0';
		value = read_MTupel(&c);

		c = pl->ref;
		done = False;
		while (*c && !done)
		{
		    bits = 0;
		    block = read_posint(&c);
		    if (*c == ':' && c++) bits = read_posint(&c);

		    for_list (net->defitems,def,T_defitem)
			if (def->nr == block)
				done = insert_sdl_place(def,pl,bits,value);

		    if (*c && *c != '*')
		    {
			fprintf(stderr,"Parsing error in BPN2 references:"
					" '*' expected\n");
			exit(2);
		    }
		    if (*c) c++;
		}
		free(value->place);
		free(value);
	}

} /* makelists_sdl */

/*****************************************************************************/
/* Main program.                                                             */
/*****************************************************************************/

int main(int argc, char **argv)
{

	int	i;
	char    *llnet = NULL, *ref, *sinv;
	char    **dptr;
	char	tmpname[80], *idx;
	M_Net   net = NULL;
	FILE    *fileptr;
	clock_t time;
	int	flag = 0;

	dptr = &llnet;
	for (i = 1; i < argc; i++)
		if (!strcmp(argv[i],"-r"))
			dptr = &ref;
		else if (!strcmp(argv[i],"-o"))
			dptr = &sinv;
		else if (!strcmp(argv[i],"-bpn2"))
			flag |= FLAG_BPN2;
		else if (!strcmp(argv[i],"-sdl"))
			flag |= FLAG_SDL;
		else if (!strcmp(argv[i],"-pfa"))
			flag |= FLAG_PFA;
		else
		{
			if (!dptr) usage(argv[0]);
			*dptr = argv[i];

			strcpy(tmpname,argv[i]);
			idx = strrchr(tmpname,'.');
			if (!idx) strcat(tmpname,".");
			idx = strrchr(tmpname,'.');

			if (dptr == &llnet)
			{
				strcpy(idx,(flag & FLAG_SDL)? ".sdl_ll_ref"
							    : ".bpn2_ll_ref");
				ref = MYstrdup(tmpname);
				strcpy(idx,".slist");
				sinv = MYstrdup(tmpname);
			}
			dptr = NULL;
		}

	if (!llnet) usage(argv[0]);
	if (!!(flag & FLAG_BPN2) + !!(flag & FLAG_SDL) + !!(flag & FLAG_PFA)!=1)
	{
		fprintf(stderr,"Error: need to specify exactly one mode out of "
				"-bpn2, -sdl or -pfa\n");
		exit(1);
	}

	time = clock();

	net = read_HLnet(llnet);
	if (strcmp(net->type,"PetriBox") && strcmp(net->type,"PTNet"))
	{
		fprintf(stderr,"Error: net in file %s is "
				"not of type PetriBox\n",llnet);
		exit(1);
	}

	if (!(flag & FLAG_PFA))
	{
	    if (!(fileptr = fopen(ref,"r")) || fclose(fileptr))
	    {
		fprintf(stderr,"Warning: references %s not found\n",ref);
		net->blockinfo = MPT_Create();
		net->numlists = 0;
	    }
	    else
		((flag & FLAG_BPN2)? read_BPNref : read_SDLref)(net,ref);
	}

	if (flag & FLAG_BPN2)
	{
		buildblocktree(net);
		makelists_bpn2(net);
		write_sinvlist_bpn2(net,sinv);
	}
	else if (flag & FLAG_SDL)
	{
		makelists_sdl(net);
		write_sinvlist_sdl(net,sinv);
	}
	else if (flag & FLAG_PFA)
	{
		write_sinvlist_pfa(net,sinv);
	}

/*	printf("Time needed: %f sec.\n\n",(float)(clock()-time)/CLOCKS_PER_SEC);
*/
	return 0;
}

/*****************************************************************************/
/* END OF FILE simplify.c                                                    */
/*****************************************************************************/

